Package: CoreGx
Type: Package
Title: Classes and Functions to Serve as the Basis for Other 'Gx'
        Packages
Version: 2.14.0
Date: 2024-03-11
Authors@R: c(
    person(given="Jermiah", family="Joseph", email="jermiah.joseph@uhn.ca",
        role=c("aut")),
    person(given="Petr", family="Smirnov", email = "petr.smirnov@uhnresearch.ca",
        role = c("aut")),
    person(given="Ian", family="Smith", email = "ianc.smith@mail.utoronto.ca",
        role = c("aut")),
    person(given="Christopher", family="Eeles", email = "christopher.eeles@uhnresearch.ca",
        role = c("aut")),
    person(given="Feifei", family="Li", email="ff.li@mail.utoronto.ca",
        role=c("aut")),
    person("Benjamin", "Haibe-Kains",
        email = "benjamin.haibe.kains@utoronto.ca", role = c("aut", "cre"))
    )
Description: A collection of functions and classes which serve as the foundation
    for our lab's suite of R packages, such as 'PharmacoGx' and 'RadioGx'. This
    package was created to abstract shared functionality from other lab package
    releases to increase ease of maintainability and reduce code repetition in
    current and future 'Gx' suite programs. Major features include a 'CoreSet'
    class, from which 'RadioSet' and 'PharmacoSet' are derived, along with get
    and set methods for each respective slot. Additional functions related to
    fitting and plotting dose response curves, quantifying statistical
    correlation and calculating area under the curve (AUC) or survival fraction
    (SF) are included. For more details please see the included documentation,
    as well as:
    Smirnov, P., Safikhani, Z., El-Hachem, N., Wang, D., She, A., Olsen, C.,
        Freeman, M., Selby, H., Gendoo, D., Grossman, P., Beck, A., Aerts, H.,
        Lupien, M., Goldenberg, A. (2015) <doi:10.1093/bioinformatics/btv723>.
    Manem, V., Labie, M., Smirnov, P., Kofia, V., Freeman, M., Koritzinksy, M.,
        Abazeed, M., Haibe-Kains, B., Bratman, S. (2018) <doi:10.1101/449793>.
VignetteBuilder: knitr
VignetteEngine: knitr::rmarkdown
biocViews: Software, Pharmacogenomics, Classification, Survival
Encoding: UTF-8
LazyData: TRUE
Depends: R (>= 4.1), BiocGenerics, SummarizedExperiment
Imports: Biobase, S4Vectors, MultiAssayExperiment, MatrixGenerics,
        piano, BiocParallel, parallel, BumpyMatrix, checkmate, methods,
        stats, utils, graphics, grDevices, lsa, data.table, crayon,
        glue, rlang, bench
Suggests: pander, markdown, BiocStyle, rmarkdown, knitr, formatR,
        testthat
License: GPL (>= 3)
Config/testthat/edition: 3
Roxygen: list(markdown=TRUE, r6=FALSE)
RoxygenNote: 7.3.1
RCollate: 'allGenerics.R' 'immutable-class.R' 'LongTable-class.R'
        'CoreSet-class.R' 'CoreSet-accessors.R' 'CoreSet-utils.R'
        'DataMapper-class.R' 'LongTable-accessors.R'
        'LongTable-utils.R' 'LongTableDataMapper-class.R'
        'LongTableDataMapper-accessors.R'
        'TreatmentResponseExperiment-class.R' 'TREDataMapper-class.R'
        'adaptiveMatthewCor.R' 'aggregate-methods.R'
        'callingWaterfall.R' 'connectivityScore.R' 'cosinePerm.R'
        'datasets.R' 'deprecated.R' 'endoaggregate-methods.R'
        'globals.R' 'gwc.R' 'matthewCor.R' 'mergeAssays-method.R'
        'methods-coerce.R' 'methods-dim.R' 'methods-dimnames.R'
        'methods-drugSensitivitySig.R' 'methods-guessMapping.R'
        'methods-metaConstruct.R' 'methods-subsetTo.R'
        'optimizeCoreGx.R' 'updateObject-methods.R' 'utilities.R'
        'utils-iteration.R' 'utils-messages.R' 'utils-optimization.R'
        'utils-testing.R' 'utils-updateS4.R'
Collate: 'allGenerics.R' 'immutable-class.R' 'LongTable-class.R'
        'CoreSet-class.R' 'CoreSet-accessors.R' 'CoreSet-utils.R'
        'DataMapper-class.R' 'LongTable-accessors.R'
        'LongTable-utils.R' 'LongTableDataMapper-class.R'
        'LongTableDataMapper-accessors.R'
        'TreatmentResponseExperiment-class.R' 'TREDataMapper-class.R'
        'adaptiveMatthewCor.R' 'aggregate-methods.R'
        'callingWaterfall.R' 'connectivityScore.R' 'cosinePerm.R'
        'datasets.R' 'deprecated.R' 'endoaggregate-methods.R'
        'globals.R' 'gwc.R' 'matthewCor.R' 'mergeAssays-method.R'
        'methods-coerce.R' 'methods-dim.R' 'methods-dimnames.R'
        'methods-drugSensitivitySig.R' 'methods-guessMapping.R'
        'methods-metaConstruct.R' 'methods-subsetTo.R'
        'optimizeCoreGx.R' 'updateObject-methods.R' 'utilities.R'
        'utils-iteration.R' 'utils-messages.R' 'utils-optimization.R'
        'utils-testing.R' 'utils-updateS4.R'
git_url: https://git.bioconductor.org/packages/CoreGx
git_branch: RELEASE_3_22
git_last_commit: d18e3bb
git_last_commit_date: 2025-10-29
Repository: Bioconductor 3.22
Date/Publication: 2025-10-29
NeedsCompilation: no
Packaged: 2025-10-30 03:18:39 UTC; biocbuild
Author: Jermiah Joseph [aut],
  Petr Smirnov [aut],
  Ian Smith [aut],
  Christopher Eeles [aut],
  Feifei Li [aut],
  Benjamin Haibe-Kains [aut, cre]
Maintainer: Benjamin Haibe-Kains <benjamin.haibe.kains@utoronto.ca>

# Generated by roxygen2: do not edit by hand

S3method("$",immutable)
S3method("$<-",immutable)
S3method("[",immutable)
S3method("[<-",immutable)
S3method("[[",immutable)
S3method("[[<-",immutable)
S3method("dimnames<-",immutable)
S3method("names<-",immutable)
S3method(Ops,immutable)
S3method(c,immutable)
S3method(mutable,default)
S3method(print,immutable)
export("assayMap<-")
export("cellInfo<-")
export("cellNames<-")
export("colDataMap<-")
export("colnames<-.immutable")
export("curation<-")
export("datasetType<-")
export("dateCreated<-")
export("fNames<-")
export("featureInfo<-")
export("mDataNames<-")
export("metadataMap<-")
export("molecularProfiles<-")
export("molecularProfilesSlot<-")
export("name<-")
export("pertNumber<-")
export("phenoInfo<-")
export("rawdata<-")
export("rowDataMap<-")
export("rownames<-.immutable")
export("sampleInfo<-")
export("sensNumber<-")
export("sensitivityInfo<-")
export("sensitivityMeasures<-")
export("sensitivityProfiles<-")
export("sensitivityRaw<-")
export("sensitivitySlot<-")
export("subset<-.immutable")
export("treatmentInfo<-")
export("treatmentNames<-")
export("treatmentResponse<-")
export(.)
export(.S4MethodContext)
export(.all_equals)
export(.collapse)
export(.distancePointLine)
export(.distancePointSegment)
export(.dmedncauchys)
export(.dmednnormals)
export(.edmedncauchys)
export(.edmednnormals)
export(.error)
export(.errorMsg)
export(.examineGOF)
export(.fitCurve)
export(.fitCurve2)
export(.formatMessage)
export(.funContext)
export(.getSupportVec)
export(.intersectList)
export(.length_unique)
export(.linInd)
export(.list_unique)
export(.magnitude)
export(.matInd)
export(.meshEval)
export(.meshEval2)
export(.message)
export(.multinom)
export(.parseToRoxygen)
export(.patternSearch)
export(.patternSearch2)
export(.pmedncauchys)
export(.pmednnormals)
export(.reformatData)
export(.residual)
export(.rmedncauchys)
export(.rmednnormals)
export(.sanitizeInput)
export(.stripWhiteSpace)
export(.summarizeMolecularNumbers)
export(.symSetDiffList)
export(.unionList)
export(.warnMsg)
export(.warning)
export(CoreSet)
export(CoreSet2)
export(LongTable)
export(LongTableDataMapper)
export(TREDataMapper)
export(TreatmentResponseExperiment)
export(aggregate2)
export(amcc)
export(as)
export(as.long.table)
export(assayCols)
export(assayIndex)
export(assayKeys)
export(assayMap)
export(buildLongTable)
export(callingWaterfall)
export(cardinality)
export(cellInfo)
export(cellNames)
export(checkColumnCardinality)
export(checkCsetStructure)
export(colDataMap)
export(colIDs)
export(colMeta)
export(collect_fn_params)
export(connectivityScore)
export(cosinePerm)
export(curation)
export(datasetType)
export(dateCreated)
export(drop_fn_params)
export(drugSensitivitySig)
export(endoaggregate)
export(fNames)
export(featureInfo)
export(getIntern)
export(guessMapping)
export(gwc)
export(idCols)
export(immutable)
export(is.immutable)
export(is.items)
export(is_optim_compatible)
export(mDataNames)
export(make_optim_function)
export(mcc)
export(metaConstruct)
export(metadataMap)
export(molecularProfiles)
export(molecularProfilesSlot)
export(mutable)
export(name)
export(optimizeCoreGx)
export(pertNumber)
export(phenoInfo)
export(rawdata)
export(reindex)
export(rowDataMap)
export(rowIDs)
export(rowMeta)
export(sampleInfo)
export(sensNumber)
export(sensitivityInfo)
export(sensitivityMeasures)
export(sensitivityProfiles)
export(sensitivityRaw)
export(sensitivitySlot)
export(sensitivitySlotToLongTable)
export(show.immutable)
export(showSigAnnot)
export(subset.immutable)
export(subsetByFeature)
export(subsetBySample)
export(subsetByTreatment)
export(subsetTo)
export(summarizeMolecularProfiles)
export(summarizeSensitivityProfiles)
export(treatmentInfo)
export(treatmentNames)
export(treatmentResponse)
export(updateSampleId)
export(updateTreatmentId)
exportClasses(CoreSet)
exportClasses(LongTable)
exportClasses(TreatmentResponseExperiment)
exportClasses(immutable_list)
exportMethods("$")
exportMethods("$<-")
exportMethods("[")
exportMethods("[[")
exportMethods("[[<-")
exportMethods("annotation<-")
exportMethods("assay<-")
exportMethods("assayMap<-")
exportMethods("assays<-")
exportMethods("colData<-")
exportMethods("colDataMap<-")
exportMethods("curation<-")
exportMethods("datasetType<-")
exportMethods("dateCreated<-")
exportMethods("fNames<-")
exportMethods("featureInfo<-")
exportMethods("mDataNames<-")
exportMethods("metadata<-")
exportMethods("metadataMap<-")
exportMethods("molecularProfiles<-")
exportMethods("molecularProfilesSlot<-")
exportMethods("name<-")
exportMethods("pertNumber<-")
exportMethods("phenoInfo<-")
exportMethods("rawdata<-")
exportMethods("rowData<-")
exportMethods("rowDataMap<-")
exportMethods("sampleInfo<-")
exportMethods("sampleNames<-")
exportMethods("sensNumber<-")
exportMethods("sensitivityInfo<-")
exportMethods("sensitivityProfiles<-")
exportMethods("sensitivityRaw<-")
exportMethods("treatmentInfo<-")
exportMethods("treatmentNames<-")
exportMethods("treatmentResponse<-")
exportMethods(aggregate)
exportMethods(annotation)
exportMethods(assay)
exportMethods(assayCols)
exportMethods(assayIndex)
exportMethods(assayKeys)
exportMethods(assayMap)
exportMethods(assayNames)
exportMethods(assays)
exportMethods(buildComboProfiles)
exportMethods(buildLongTable)
exportMethods(colData)
exportMethods(colDataMap)
exportMethods(colIDs)
exportMethods(colMeta)
exportMethods(colnames)
exportMethods(curation)
exportMethods(datasetType)
exportMethods(dateCreated)
exportMethods(dim)
exportMethods(dimnames)
exportMethods(endoaggregate)
exportMethods(fNames)
exportMethods(featureInfo)
exportMethods(getIntern)
exportMethods(guessMapping)
exportMethods(idCols)
exportMethods(lapply)
exportMethods(mDataNames)
exportMethods(mergeAssays)
exportMethods(metaConstruct)
exportMethods(metadata)
exportMethods(metadataMap)
exportMethods(molecularProfiles)
exportMethods(molecularProfilesSlot)
exportMethods(name)
exportMethods(names)
exportMethods(pertNumber)
exportMethods(phenoInfo)
exportMethods(rawdata)
exportMethods(reindex)
exportMethods(rowData)
exportMethods(rowDataMap)
exportMethods(rowIDs)
exportMethods(rowMeta)
exportMethods(rownames)
exportMethods(sampleInfo)
exportMethods(sampleNames)
exportMethods(sensNumber)
exportMethods(sensitivityInfo)
exportMethods(sensitivityMeasures)
exportMethods(sensitivityProfiles)
exportMethods(sensitivityRaw)
exportMethods(show)
exportMethods(subset)
exportMethods(subsetByFeature)
exportMethods(subsetBySample)
exportMethods(subsetByTreatment)
exportMethods(treatmentInfo)
exportMethods(treatmentNames)
exportMethods(treatmentResponse)
exportMethods(updateObject)
import(checkmate)
import(data.table)
import(methods)
importClassesFrom(MultiAssayExperiment,MultiAssayExperiment)
importClassesFrom(S4Vectors,Annotated)
importClassesFrom(S4Vectors,DataFrame)
importClassesFrom(S4Vectors,List)
importClassesFrom(S4Vectors,list_OR_List)
importFrom(Biobase,annotation)
importFrom(Biobase,assayDataElementNames)
importFrom(Biobase,exprs)
importFrom(Biobase,fData)
importFrom(Biobase,pData)
importFrom(Biobase,protocolData)
importFrom(BiocGenerics,"%in%")
importFrom(BiocGenerics,match)
importFrom(BiocParallel,"bpprogressbar<-")
importFrom(BiocParallel,"bpworkers<-")
importFrom(BiocParallel,bplapply)
importFrom(BiocParallel,bpparam)
importFrom(BiocParallel,bpprogressbar)
importFrom(BiocParallel,bpworkers)
importFrom(BumpyMatrix,splitAsBumpyMatrix)
importFrom(MatrixGenerics,colAlls)
importFrom(MatrixGenerics,rowAnys)
importFrom(MultiAssayExperiment,"experiments<-")
importFrom(MultiAssayExperiment,MultiAssayExperiment)
importFrom(MultiAssayExperiment,experiments)
importFrom(S4Vectors,DataFrame)
importFrom(S4Vectors,List)
importFrom(S4Vectors,SimpleList)
importFrom(S4Vectors,endoapply)
importFrom(S4Vectors,metadata)
importFrom(SummarizedExperiment,"assayNames<-")
importFrom(SummarizedExperiment,"colData<-")
importFrom(SummarizedExperiment,"rowData<-")
importFrom(SummarizedExperiment,Assays)
importFrom(SummarizedExperiment,SummarizedExperiment)
importFrom(SummarizedExperiment,`rowData<-`)
importFrom(SummarizedExperiment,assay)
importFrom(SummarizedExperiment,assayNames)
importFrom(SummarizedExperiment,assays)
importFrom(SummarizedExperiment,colData)
importFrom(SummarizedExperiment,rowData)
importFrom(bench,mark)
importFrom(bench,press)
importFrom(checkmate,assert)
importFrom(checkmate,assertCharacter)
importFrom(checkmate,assertClass)
importFrom(checkmate,assertDataFrame)
importFrom(checkmate,assertDataTable)
importFrom(checkmate,assertList)
importFrom(checkmate,assertLogical)
importFrom(checkmate,assertSubset)
importFrom(compiler,cmpfun)
importFrom(crayon,"%+%")
importFrom(crayon,blue)
importFrom(crayon,cyan)
importFrom(crayon,green)
importFrom(crayon,magenta)
importFrom(crayon,red)
importFrom(crayon,yellow)
importFrom(data.table,":=")
importFrom(data.table,`:=`)
importFrom(data.table,as.data.table)
importFrom(data.table,copy)
importFrom(data.table,data.table)
importFrom(data.table,dcast)
importFrom(data.table,fsetdiff)
importFrom(data.table,key)
importFrom(data.table,melt.data.table)
importFrom(data.table,merge.data.table)
importFrom(data.table,rbindlist)
importFrom(data.table,set)
importFrom(data.table,setDF)
importFrom(data.table,setDT)
importFrom(data.table,setDTthreads)
importFrom(data.table,setcolorder)
importFrom(data.table,setindexv)
importFrom(data.table,setkeyv)
importFrom(data.table,setnames)
importFrom(data.table,tstrsplit)
importFrom(grDevices,dev.off)
importFrom(grDevices,pdf)
importFrom(grDevices,rainbow)
importFrom(graphics,abline)
importFrom(graphics,hist)
importFrom(graphics,legend)
importFrom(graphics,lines)
importFrom(graphics,par)
importFrom(graphics,points)
importFrom(lsa,cosine)
importFrom(methods,getPackageName)
importFrom(methods,is)
importFrom(parallel,detectCores)
importFrom(piano,loadGSC)
importFrom(piano,runGSA)
importFrom(rlang,trace_back)
importFrom(stats,complete.cases)
importFrom(stats,cor.test)
importFrom(stats,cov.wt)
importFrom(stats,dcauchy)
importFrom(stats,dnorm)
importFrom(stats,integrate)
importFrom(stats,lm)
importFrom(stats,median)
importFrom(stats,na.omit)
importFrom(stats,optim)
importFrom(stats,pcauchy)
importFrom(stats,pchisq)
importFrom(stats,pnorm)
importFrom(stats,pt)
importFrom(stats,qnorm)
importFrom(stats,quantile)
importFrom(stats,rcauchy)
importFrom(stats,rnorm)
importFrom(stats,setNames)
importFrom(stats,var)
importFrom(utils,packageName)
importFrom(utils,sessionInfo)
importMethodsFrom(Biobase,"sampleNames<-")
importMethodsFrom(Biobase,dimnames)
importMethodsFrom(Biobase,sampleNames)
importMethodsFrom(BiocGenerics,"annotation<-")
importMethodsFrom(BiocGenerics,annotation)
importMethodsFrom(BiocGenerics,lapply)
importMethodsFrom(BiocGenerics,subset)
importMethodsFrom(BiocGenerics,updateObject)
importMethodsFrom(S4Vectors,"metadata<-")
importMethodsFrom(S4Vectors,aggregate)
importMethodsFrom(S4Vectors,metadata)
importMethodsFrom(SummarizedExperiment,"assay<-")
importMethodsFrom(SummarizedExperiment,"assays<-")
importMethodsFrom(SummarizedExperiment,assay)
importMethodsFrom(SummarizedExperiment,assayNames)
importMethodsFrom(SummarizedExperiment,assays)
importMethodsFrom(SummarizedExperiment,dim)

#' @include LongTable-class.R LongTable-accessors.R
#' @importFrom checkmate assertClass assertDataFrame
NULL


#### CoreGx dynamic documentation
####
#### Warning: for dynamic docs to work, you must set
#### Roxygen: list(markdown=TRUE, r6=FALSE)
#### in the DESCRIPTION file!


# ===================================
# Utility Method Documentation Object
# -----------------------------------


# ======================================
# Subset Methods
# --------------------------------------


##
## == subset


#' Subset a `LongTable` using an "assayIndex" data.frame
#'
#' @param x `LongTable`
#' @param index `data.frame` Table with columns "rowKey", "colKey" and
#'   ".\<assayName\>", where \<assayName\> is the value for each `assayNames(x)`.
#'   Warning: rownames are dropped internally in coercion to `data.table`,
#' @param reindex `logical(1)` Should index values be reset such that they
#'   are the smallest possible set of consecutive integers. Modifies the
#'   "rowKey", "colKey", and all assayKey columns. Initial benchmarks indicate
#'   `reindex=FALSE` saves ~20% of both execution time and memory allocation. The
#'   cost of reindexing decreases the smaller your subet gets.
#'
#' @return `LongTable` subset according to the provided index.
#'
#' @noRd
.subsetByIndex <- function(x, index, reindex=FALSE) {

    # -- validate input
    assertClass(x, "LongTable")
    assertDataFrame(index)
    if (!is.data.table(index)) setDT(index)
    x <- copy(x)

    # -- subset slots
    rData <- rowData(x, raw=TRUE)[sort(unique(index$rowKey)), ]
    cData <- colData(x, raw=TRUE)[sort(unique(sort(index$colKey))), ]
    assays <- assays(x, withDimnames=FALSE)
    metaKeys <- c("rowKey", "colKey")
    setkeyv(index, metaKeys)
    for (i in seq_along(assays)) {
        setkeyv(assays[[i]], metaKeys)
        aname <- paste0(".", names(assays)[i])
        # join based subsets use binary-search, O(log(n)) vs O(n) for vector-scan
        # see https://rdatatable.gitlab.io/data.table/articles/datatable-keys-fast-subset.html
        assays[[i]] <- assays[[i]][
            index[!is.na(get(aname)), c(metaKeys, aname), with=FALSE],
        ]
        setkeyv(assays[[i]], aname)
    }
    # -- update object
    # delete row-/colKeys by reference
    for (a in assays) a[, (metaKeys) := NULL]
    # ensure uniqueness for summary assays, fixes #149
    assays <- lapply(assays, FUN=unique)
    # raw=TRUE allows direct modification of slots
    setkeyv(rData, "rowKey")
    rowData(x, raw=TRUE) <- rData
    setkeyv(cData, "colKey")
    colData(x, raw=TRUE) <- cData
    assays(x, raw=TRUE) <- assays
    mutableIntern <- mutable(getIntern(x))
    setkeyv(index, paste0(".", names(assays)))
    mutableIntern$assayIndex <- index
    x@.intern <- immutable(mutableIntern)

    # -- optionally reindex the table
    if (reindex) {
        x <- reindex(x)
    }
    return(x)
}

#' Subset method for a LongTable object.
#'
#' Allows use of the colData and rowData `data.table` objects to query based on
#'  rowID and colID, which is then used to subset all assay `data.table`s stored
#'  in the `assays` slot.
#' This function is endomorphic, it always returns a LongTable object.
#'
#' @examples
#' # Character
#' subset(merckLongTable, 'ABT-888', 'CAOV3')
#' # Numeric
#' subset(merckLongTable, 1, c(1, 2))
#' # Logical
#' subset(merckLongTable, , colData(merckLongTable)$sampleid == 'A2058')
#' # Call
#' subset(merckLongTable, drug1id == 'Dasatinib' & drug2id != '5-FU',
#'     sampleid == 'A2058')
#'
#' @param x `LongTable` The object to subset.
#' @param i `character`, `numeric`, `logical` or `call`
#'  Character: pass in a character vector of rownames for the `LongTable` object
#'    or a valid regex query which will be evaluated against the rownames.
#'  Numeric or Logical: vector of indices or a logical vector to subset
#'    the rows of a `LongTable`.
#'  Call: Accepts valid query statements to the `data.table` i parameter,
#'    this can be used to make complex queries using the `data.table` API
#'    for the `rowData` data.table.
#' @param j `character`, `numeric`, `logical` or `call`
#'  Character: pass in a character vector of colnames for the `LongTable` object
#'    or a valid regex query which will be evaluated against the colnames.
#'  Numeric or Logical: vector of indices or a logical vector to subset
#'    the columns of a `LongTable`.
#'  Call: Accepts valid query statements to the `data.table` i parameter,
#'    this can be used to make complex queries using the `data.table` API
#'    for the `colData` data.table.
#' @param assays `character`, `numeric` or `logical` Optional list of assay
#'   names to subset. Can be used to subset the assays list further,
#'   returning only the selected items in the new LongTable.
#' @param reindex `logical(1)` Should index values be reset such that they
#'   are the smallest possible set of consecutive integers. Modifies the
#'   "rowKey", "colKey", and all assayKey columns. Initial benchmarks indicate
#'   `reindex=FALSE` saves ~20% of both execution time and memory allocation. The
#'   cost of reindexing decreases the smaller your subset gets.
#'
#' @return `LongTable` A new `LongTable` object subset based on the specified
#'      parameters.
#'
#' @importMethodsFrom BiocGenerics subset
#' @importFrom crayon magenta cyan
#' @importFrom MatrixGenerics rowAnys
#' @import data.table
#' @export
setMethod('subset', signature('LongTable'),
        function(x, i, j, assays=assayNames(x),
            reindex=TRUE) {

    # prevent modify by reference
    x <- copy(x)

    # local helper functions
    .rowData <- function(...) rowData(..., key=TRUE)
    .colData <- function(...) colData(..., key=TRUE)
    .tryCatchNoWarn <- function(...) suppressWarnings(tryCatch(...))
    .strSplitLength <- function(...) length(unlist(strsplit(...)))

    # subset rowData
    ## FIXME:: Can I parameterize this into a helper that works for both row
    ## and column data?
    if (!missing(i)) {
        ## TODO:: Clean up this if-else block
        if (.tryCatchNoWarn(is.call(i), error=function(e) FALSE)) {
            rowDataSubset <- .rowData(x)[eval(i), ]
        } else if (.tryCatchNoWarn(is.character(i), error=function(e) FALSE)) {
            ## TODO:: Implement diagnosis for failed regex queries
            idCols <- rowIDs(x, key=TRUE)
            if (max(unlist(lapply(i, .strSplitLength, split=':'))) > length(idCols))
                stop(cyan$bold('Attempting to select more rowID columns than
                    there are in the LongTable.\n\tPlease use query of the form ',
                    paste0(idCols, collapse=':')))
            imatch <- rownames(x) %in% i
            if (!any(imatch))
                imatch <- grepl(.preprocessRegexQuery(i), rownames(x),
                    ignore.case=TRUE)
            imatch <- str2lang(.variableToCodeString(imatch))
            rowDataSubset <- .rowData(x)[eval(imatch), ]
        } else {
            isub <- substitute(i)
            rowDataSubset <- .tryCatchNoWarn(.rowData(x)[i, ],
                error=function(e) .rowData(x)[eval(isub), ])
        }
    } else {
        rowDataSubset <- .rowData(x)
    }

    # subset colData
    if (!missing(j)) {
        ## TODO:: Clean up this if-else block
        if (.tryCatchNoWarn(is.call(j), error=function(e) FALSE, silent=TRUE)) {
            colDataSubset <- .colData(x)[eval(j), ]
        } else if (.tryCatchNoWarn(is.character(j), error=function(e) FALSE, silent=TRUE)) {
            ## TODO:: Implement diagnosis for failed regex queries
            idCols <- colIDs(x, key=TRUE)
            if (max(unlist(lapply(j, .strSplitLength, split=':'))) > length(idCols))
                stop(cyan$bold('Attempting to select more ID columns than there
                    are in the LongTable.\n\tPlease use query of the form ',
                    paste0(idCols, collapse=':')))
            jmatch <- colnames(x) %in% j
            if (!any(jmatch))
                jmatch <- grepl(.preprocessRegexQuery(j), colnames(x),
                    ignore.case=TRUE)
            jmatch <- str2lang(.variableToCodeString(jmatch))
            colDataSubset <- .colData(x)[eval(jmatch), ]
        } else {
            jsub <- substitute(j)
            colDataSubset <- .tryCatchNoWarn(.colData(x)[j, ],
                error=function(e) .colData(x)[eval(jsub), ])
        }
    } else {
        colDataSubset <- .colData(x)
    }

    # Subset assays to only keys in remaining in rowData/colData
    rows <- rowDataSubset$rowKey
    cols <- colDataSubset$colKey

    # -- find matching assays
    validAssays <- assays %in% assayNames(x)
    if (any(!validAssays))
        warning(.warnMsg(assays[!validAssays],
            " are not valid assay names, ignoring..."), call.=FALSE)
    keepAssays <- assayNames(x) %in% assays

    # -- subset index, then use index to subset x
    assayKeys <- paste0(".", assayNames(x)[keepAssays])
    idx <- mutable(getIntern(x, "assayIndex"))[
        rowKey %in% rows & colKey %in% cols,
        .SD,
        .SDcols=c("rowKey", "colKey", assayKeys)
    ]
    # -- drop rowKeys or colKeys which no longer have any assay observation
    #   after the initial subset, fixes #148
    validKeys <- idx[
        which(rowAnys(!is.na(idx[, assayKeys, with=FALSE]))),
        .(rowKey, colKey)
    ]
    idx <- idx[
        rowKey %in% unique(validKeys$rowKey) &
            colKey %in% unique(validKeys$colKey),
    ]
    assays(x, raw=TRUE)[!keepAssays] <- NULL  # delete assays being dropped

    return(.subsetByIndex(x, idx, reindex=reindex))
})



#' Convenience function for converting R code to a call
#'
#' This is used to pass through unevaluated R expressions into subset and
#'   `[`, where they will be evaluated in the correct context.
#'
#' @examples
#' .(sample_line1 == 'A2058')
#'
#' @param ... `pairlist` One or more R expressions to convert to calls.
#'
#' @return `call` An R call object containing the quoted expression.
#'
#' @export
. <- function(...) substitute(...)

# ---- subset LongTable helpers

#' Collapse vector of regex queries with | and replace * with .*
#'
#' @param queryString `character` Raw regex queries.
#'
#' @return `character` Formatted regex query.
#'
#' @keywords internal
#' @noRd
.preprocessRegexQuery <- function(queryString) {
    # Support vectors of regex queries
    query <- paste0(unique(queryString), collapse='|')
    # Swap all * with .*
    query <- gsub('\\.\\*', '*', query)
    return(gsub('\\*', '.*', query))
}


#' @keywords internal
#' @noRd
.validateRegexQuery <- function(regex, names) {
    ## TODO:: return TRUE if reqex query is valid, otherwise return error message
}

#' Convert an R object in a variable into a string of the code necessary to
#'   create that object
#'
#' @param variable `symbol` A symbol containing an R variable
#'
#' @return `character(1)` A string representation of the code necessary to
#'   reconstruct the variable.
#'
#' @keywords internal
#' @noRd
.variableToCodeString <- function(variable) {
    codeString <- paste0(capture.output(dput(variable)), collapse='')
    codeString <- gsub('\"', "'", codeString)
    return(codeString)
}

#' Filter a data.table object based on the rowID and colID columns
#'
#' @param DT `data.table` Object with the columns rowID and colID, preferably
#'  as the key columns.
#' @param indexList `list` Two integer vectors, one indicating the rowIDs and
#'  one indicating the colIDs to filter the `data.table` on.
#'
#' @return `data.table` A copy of `DT` subset on the row and column IDs specified
#'  in `indexList`.
#'
#' @import data.table
#' @keywords internal
#' @noRd
.filterLongDataTable <- function(DT, indexList) {

    # validate input
    if (length(indexList) > 2)
        stop("This object is 2D, please only pass in two ID vectors, one for
             rows and one for columns!")

    if (!all(vapply(unlist(indexList), is.numeric, FUN.VALUE=logical(1))))
        stop('Please ensure indexList only contains integer vectors!')

    # extract indices
    rowIndices <- indexList[[1]]
    colIndices <- indexList[[2]]

    # return filtered data.table
    return(copy(DT[rowKey %in% rowIndices & colKey %in% colIndices, ]))
}

##
## == [ method


#' [ LongTable Method
#'
#' Single bracket subsetting for a LongTable object. See subset for more details.
#'
#' This function is endomorphic, it always returns a LongTable object.
#'
#' @examples
#' # Character
#' merckLongTable['ABT-888', 'CAOV3']
#' # Numeric
#' merckLongTable[1, c(1, 2)]
#' # Logical
#' merckLongTable[, colData(merckLongTable)$sampleid == 'A2058']
#' # Call
#' merckLongTable[
#'      .(drug1id == 'Dasatinib' & drug2id != '5-FU'),
#'      .(sampleid == 'A2058'),
#'  ]
#'
#' @param x `LongTable` The object to subset.
#' @param i `character`, `numeric`, `logical` or `call`
#'  Character: pass in a character vector of drug names, which will subset the
#'    object on all row id columns matching the vector. This parameter also
#'    supports valid R regex query strings which will match on the colnames
#'    of `x`. For convenience, * is converted to .* automatically. Colon
#'    can be to denote a specific part of the colnames string to query.
#'  Numeric or Logical: these select based on the rowKey from the `rowData`
#'    method for the `LongTable`.
#'  Call: Accepts valid query statements to the `data.table` i parameter as
#'    a call object. We have provided the function .() to conveniently
#'    convert raw R statements into a call for use in this function.
#' @param j `character`, `numeric`, `logical` or `call`
#'  Character: pass in a character vector of drug names, which will subset the
#'      object on all drug id columns matching the vector. This parameter also
#'      supports regex queries. Colon can be to denote a specific part of the
#'      colnames string to query.
#'  Numeric or Logical: these select based on the rowID from the `rowData`
#'      method for the `LongTable`.
#'  Call: Accepts valid query statements to the `data.table` i parameter as
#'      a call object. We have provided the function .() to conveniently
#'      convert raw R statements into a call for use in this function.
#' @param assays `character` Names of assays which should be kept in the
#'   `LongTable` after subsetting.
#' @param ... Included to ensure drop can only be set by name.
#' @param drop `logical` Included for compatibility with the '[' primitive,
#'   it defaults to FALSE and changing it does nothing.
#'
#' @return A `LongTable` containing only the data specified in the function
#'   parameters.
#'
#' @export
setMethod('[', signature('LongTable'),
        function(x, i, j, assays=assayNames(x), ..., drop=FALSE) {
    subset(x, i, j, assays=assays, ...)
})


##
## == [[ method'


#' [[ Method for LongTable Class
#'
#' Select an assay from within a LongTable object.
#'
#' @describeIn LongTable Get an assay from a LongTable object. This method
#'   returns the row and column annotations by default to make assignment
#'   and aggregate operations easiers.
#'
#' @examples
#' merckLongTable[['sensitivity']]
#'
#' @param x `LongTable` object to retrieve assays from
#' @param i `character(1)` name or `integer` index of the desired assay.
#'
#' @importFrom crayon cyan magenta
#' @import data.table
#' @export
setMethod('[[', signature('LongTable'), function(x, i) {
    assay(x, i)
})


#' `[[<-` Method for LongTable Class
#'
#' Just a wrapper around assay<- for convenience. See
#' `?'assay<-,LongTable,character-method'.`
#'
#' @param x A `LongTable` to update.
#' @param i The name of the assay to update, must be in `assayNames(object)`.
#' @param value A `data.frame`
#'
#' @examples
#' merckLongTable[['sensitivity']] <- merckLongTable[['sensitivity']]
#'
#' @return A `LongTable` object with the assay `i` updated using `value`.
#'
#' @export
setReplaceMethod('[[', signature(x='LongTable'), function(x, i, value) {
    assay(x, i) <- value
    x
})


##
## == $ method


#' Select an assay from a LongTable object
#'
#' @examples
#' merckLongTable$sensitivity
#'
#' @param x A `LongTable` object to retrieve an assay from
#' @param name `character` The name of the assay to get.
#'
#' @return `data.frame` The assay object.
#'
#' @export
setMethod('$', signature('LongTable'), function(x, name) {
    # error handling is done inside `[[`
    x[[name]]
})

#' Update an assay from a LongTable object
#'
#' @examples
#' merckLongTable$sensitivity <- merckLongTable$sensitivity
#'
#' @param x A `LongTable` to update an assay for.
#' @param name `character(1)` The name of the assay to update
#' @param value A `data.frame` or `data.table` to update the assay with.
#'
#' @return Updates the assay `name` in `x` with `value`, returning an invisible
#' NULL.
#'
#' @export
setReplaceMethod('$', signature('LongTable'), function(x, name, value) {
    # error handling done inside `assay<-`
    x[[name]] <- value
    x
})


# ======================================
# Reindex Methods
# --------------------------------------

##
## == reindex

#' Redo indexing for a LongTable object to remove any gaps in integer indexes
#'
#' After subsetting a LongTable, it is possible that values of rowKey or colKey
#'   could no longer be present in the object. As a result there the indexes
#'   will no longer be contiguous integers. This method will calcualte a new
#'   set of rowKey and colKey values such that integer indexes are the smallest
#'   set of contiguous integers possible for the data.
#'
#' @param object The `LongTable` object to recalcualte indexes (rowKey and
#'     colKey values) for.
#'
#' @return A copy of the `LongTable` with all keys as the smallest set of
#'     contiguous integers possible given the current data.
#'
#' @export
setMethod('reindex', signature(object='LongTable'), function(object) {

    # -- extract the requisite data
    mutableIntern <- mutable(getIntern(object))
    index <- mutableIntern$assayIndex
    rData <- copy(rowData(object, raw=TRUE))
    cData <- copy(colData(object, raw=TRUE))
    aList <- copy(assays(object, raw=TRUE))

    # -- sort metadata tables by their id columns and update the index
    # Add row key to rData
    rData[, .rowKey := .I, by=c(rowIDs(object))]

    # Add column key to cData
    cData[, .colKey := .I, by=c(colIDs(object))]

    # -- update rowKey and colKey in the assayIndex, if they have changed
    if (rData[, any(rowKey != .rowKey)]) {
        index[rData, rowKey := .rowKey, on="rowKey"]
        rData[, rowKey := .rowKey]
        setkeyv(rData, "rowKey")
    }
    if (cData[, any(colKey != .colKey)]) {
        index[cData, colKey := .colKey, on="colKey"]
        cData[, colKey := .colKey]
        setkeyv(cData, "colKey")
    }
    rData[, .rowKey := NULL]
    cData[, .colKey := NULL]

    # -- add new indices for assayKeys to index
    setkeyv(index, c("rowKey", "colKey"))
    assays_ <- setdiff(colnames(index), c("rowKey", "colKey"))
    lapply(assays_, function(nm){
        # drop the first "." from nm
        name <- gsub("^\\.", "", nm)

        # if there are any more "." in the name, then raise an erro
        if (grepl("\\.", name))
            stop("Assay names cannot contain '.' characters!")

    })

    assayEqualKeys <- setNames(vector("logical", length(assays_)), assays_)
    # This loop iterates over each element in the 'assays_' vector.
    # For each element, it performs the following operations:
    # 1. It checks if the element is not missing in the current data.table.
    # 2. If the element is not missing, it creates a new column with the name
    #    ".<element>" and assigns a unique group identifier (.GRP) to each row
    #    that has a non-missing value for that element. The grouping is done by
    #    the element itself.
    # 3. It updates the 'assayEqualKeys' list with a logical value indicating
    #    whether all the values in the newly created column are equal to the
    #    corresponding values in the original column.
    #    This is done to ensure that summary assays, with repeated keys, have
    #    consistent values across all rows.
    #
    # The data.table syntax used in this code is as follows:
    # - The 'get' function is used to retrieve the value of a variable by name.
    # - The 'is.na' function is used to check if a value is missing.
    # - The ':=' operator is used to create or update columns by reference.
    # - The '.GRP' variable represents a unique group identifier.
    # - The 'by' parameter is used to specify the grouping variable(s).
    #
    # Note: The line width of the code has been limited to 80 characters for better readability.
    for (nm in assays_) {
        ## Added by to maintain cardinality of the each assayKey
        ## Required to fix #147 and ensure summary assays, with repeated keys
        index[!is.na(get(nm)), paste0(".", nm) := .GRP, by=c(nm)]
        assayEqualKeys[nm] <- index[!is.na(get(nm)), all(get(paste0(".", nm)) == get(nm))]
    }

    # -- check equality and update assayKeys in assays if they have changed
    for (.nm in names(which(!assayEqualKeys))) {
        nm <- gsub("\\.", "", .nm)
        setkeyv(index, .nm)
        aList[[nm]][index, (.nm) := get(paste0(".", .nm))]
        setkeyv(aList[[nm]], .nm)
        index[, (.nm) := get(paste0(".", .nm))]
    }
    index[, paste0(".", assays_) := NULL]
    setkeyv(index, paste0(".", assayNames(object)))

    # -- update the object with the reindexed tables and return
    rowData(object, raw=TRUE) <- rData
    colData(object, raw=TRUE) <- cData
    assays(object, raw=TRUE) <- aList
    mutableIntern$assayIndex <- index
    object@.intern <- immutable(mutableIntern)
    return(object)
})


#' @keywords internal
#' @noRd
.extractIDData <- function(assayDataList, idCols, keyName) {
    idDT <- data.table()
    for (assay in assayDataList) {
        idDT <- unique(rbindlist(list(idDT, assay[, ..idCols])))
    }
    rm(assayDataList)
    idDT[, eval(substitute(keyName := seq_len(.N)))]
    setkeyv(idDT, keyName)
    return(idDT)
}


#' @keywords interal
#' @noRd
.joinDropOn <- function(DT1, DT2, on) {
    DT1[DT2, on=on][, -get('on')]
}

library(testthat)
library(CoreGx)
library(data.table)

data(nci_TRE_small)
tre <- nci_TRE_small

# == subset

testthat::test_that("`subset,LongTable-method` works with call queries", {
    ntre <- subset(tre,
        treatment1id %in% unique(treatment1id)[1:5],
        sampleid %in% unique(sampleid)[1:5]
    )
    testthat::expect_s4_class(ntre, "LongTable")
    ## These tests need to be updated to use expect_true with .table_is_subset
    ## instead of expect_equal due the fact that the subset,LongTable-method
    ## will drop additional rowKey or colKey values than those in the initial
    ## subset statement if there are no assay observations using those keys.
    ## This change fixed #148, but now makes it impossible to store metadata
    ## when there are no observations, which may not be ideal?
    ## Alternative would be to rework the assayIndex to be free of NA values
    testthat::expect_true(
        CoreGx:::.table_is_subset(
            rowData(ntre),
            rowData(tre)[treatment1id %in% unique(treatment1id)[1:5]]
        )
    )
    testthat::expect_true(
        CoreGx:::.table_is_subset(
            colData(ntre),
            colData(tre)[sampleid %in% unique(sampleid)[1:5]]
        )
    )
    # check for NA values in the key column of the assay
    testthat::expect_true(
        !anyNA(assays(ntre, raw=TRUE)[["sensitivity"]]$sensitivity)
    )
    ntre2 <- tre[
        .(treatment1id %in% unique(treatment1id)[1:5]),
        .(sampleid %in% unique(sampleid)[1:5])
    ]
    testthat::expect_s4_class(ntre2, "LongTable")
    testthat::expect_true(
        CoreGx:::.table_is_subset(
            rowData(ntre2),
            rowData(tre)[treatment1id %in% unique(treatment1id)[1:5]]
        )
    )
    testthat::expect_true(
        CoreGx:::.table_is_subset(
            colData(ntre2),
            colData(tre)[sampleid %in% unique(sampleid)[1:5]]
        )
    )
    # check for NA values in the key column of the assay
    testthat::expect_true(
        !anyNA(assays(ntre2, raw=TRUE)[["sensitivity"]]$sensitivity)
    )
    testthat::expect_equal(ntre, ntre2)
})

testthat::test_that("`subset,LongTable-method` works with regex queries", {
    ntre <- subset(tre,
        c("Vinblastine", "Temozolomide"),
        c("HT*", "MOLT*")
    )
    testthat::expect_s4_class(ntre, "LongTable")
    testthat::expect_true(
        CoreGx:::.table_is_subset(
            rowData(ntre),
            rowData(tre)[grepl("Vinblastine|Temozolomide", rownames(tre)), ]
        )
    )
    testthat::expect_true(
        CoreGx:::.table_is_subset(
            colData(ntre),
            colData(tre)[grepl("HT*|MOLT-*", colnames(tre)), ]
        )
    )
})

testthat::test_that("`CoreGx:::.subsetByIndex` is equivalent to subsetting the raw data", {
    keepRows <- rowData(tre, key=TRUE)[treatment1id %in% treatment1id[1:5], ]
    fullAssay <- tre$sensitivity
    rawSubset <- fullAssay[treatment1id %in% keepRows$treatment1id, ]
    aindex <- mutable(getIntern(tre, "assayIndex"))
    subindex <- aindex[rowKey %in% keepRows$rowKey, ]
    ntre <- CoreGx:::.subsetByIndex(tre, subindex)
    testthat::expect_true(
        !anyNA(assays(ntre, raw=TRUE)[["sensitivity"]]$sensitivity)
    )
    assayByIndex <- ntre$sensitivity
    testthat::expect_true(all.equal(rawSubset, assayByIndex))
})

testthat::test_that("`subset,LongTable-method` works with row and column names", {
    ntre <- subset(tre, rownames(tre)[1:5], colnames(tre)[1:5])
    testthat::expect_equal(rownames(ntre), rownames(tre)[1:5])
    testthat::expect_true(all.equal(rowData(ntre), rowData(tre)[1:5, ]))
    testthat::expect_equal(colnames(ntre), colnames(tre)[1:5])
    testthat::expect_true(all.equal(colData(ntre), colData(tre)[1:5, ]))
})

testthat::test_that("`subset,LongTable-method` doesn't produce non-existing assay observations from joining", {
    all_assays <- assays(tre, key = FALSE, withDimnames = TRUE)
    select_row_idx <- seq.int(1, dim(tre)[1], by = 2)
    select_col_idx <- seq.int(1, dim(tre)[2], by = 2)
    sub_tre <- subset(tre, i = select_row_idx, j = select_col_idx)
    for (a in seq_along(all_assays)) {
        assay_sub <- assay(sub_tre, a, key = FALSE, withDimnames = TRUE)
        obs_exists <- dim(
            assay_sub[!all_assays[[a]], on = names(assay_sub)]
            )[1] == 0 ## anti-join to check elements not a subset
        testthat::expect_true({ obs_exists })
    }
})

testthat::test_that("`subset,LongTable-method` doesn't miss assay observations for either selected row/columns", {
    all_assays <- assays(tre, key = FALSE, withDimnames = TRUE)
    select_row_idx <- sample.int(n = dim(tre)[1], size = 1, replace = FALSE)
    sub_tre <- subset(tre, i = select_row_idx)
    select_row <- rowData(tre)[select_row_idx, rowIDs(tre), with = FALSE]
    for (a in seq_along(all_assays)) {
        assay_sub1 <- assay(sub_tre, a, key = FALSE, withDimnames = TRUE)
        assay_sub2 <- all_assays[[a]][select_row, ]
        testthat::expect_equal(assay_sub1, assay_sub2)
    }
    select_col_idx <- sample.int(n = dim(tre)[2], size = 1, replace = FALSE)
    sub_tre <- subset(tre, j = select_col_idx)
    select_col <- colData(tre)[select_col_idx, colIDs(tre), with = FALSE]
    assay_names <- assayNames(sub_tre)
    for (a in seq_along(assay_names)) {
        assay_sub1 <- assay(sub_tre, a, key = FALSE, withDimnames = TRUE)
        assay_sub2 <- all_assays[[a]][sampleid == select_col, ]
        testthat::expect_equal(assay_sub1, assay_sub2)
    }
    sub_tre <- tre[select_row_idx, select_col_idx]
    select_both <- cbind(select_row, select_col)
    for (a in seq_along(assay_names)) {
        assay_sub1 <- assay(sub_tre, a, key = FALSE, withDimnames = TRUE)
        assay_sub2 <- all_assays[[a]][select_both, ]
        testthat::expect_equal(assay_sub1, assay_sub2)
    }
})

testthat::test_that("`subset,LongTable-method` subset indexing behaves the same as data.table", {
    ## controled by .subsetByIndex
    sub_tre <- subset(tre, i = NULL) ## subset with row index by NULL
    testthat::expect_equal(dim(sub_tre), c(0, 0))
    sub_tre <- subset(tre, j = NULL) ## subset with column index by NULL
    testthat::expect_equal(dim(sub_tre), c(0, 0))
    sub_tre <- subset(tre, i = "") ## subset with row by empty rowname string
    testthat::expect_equal(sub_tre, tre)
    sub_tre <- subset(tre, j = "") ## subset with column by empty column name
    testthat::expect_equal(sub_tre, tre)
    sub_tre <- subset(tre, i = "", j = "") ## subset by empty row+column names
    testthat::expect_equal(sub_tre, tre)
    ## Get a subset with 2-Fluoro Ara-A of dose 6e-06 as second treatment in combination therapies
    sub_tre <- subset(tre, i = "*:2-Fluoro Ara-A:*:6e-06")
    regex <- "(?=.*\\:2-Fluoro Ara-A)(?=.*6e-06\\:*)^" ## rowData regex
    testthat::expect_equal(
        rowData(tre)[grepl(regex, rownames(tre), perl = TRUE), ],
        rowData(sub_tre)
    ) ## much nicer to query at the TRE level
    ## Subset containing ovarian cancer cell line
    sub_tre <- subset(tre, j = ".*OVCAR.*")
    testthat::expect_equal(
        colData(tre)[grepl(".*OVCAR.*", colnames(tre), perl = TRUE), ],
        colData(sub_tre)
    )
    ## Subset by negative index: TRE behaves the same as data.table
    i <- sample.int(n = dim(tre)[1], size = 1, replace = FALSE)
    sub_tre_1 <- tre[-i, ] ## Drop the i-th row
    sub_tre_2 <- tre[i, ] ## Extract the i-th row
    testthat::expect_equal(
        rowData(tre)[!rowData(sub_tre_1), on = names(rowData(tre))],
        rowData(sub_tre_2) # rowData(tre)\rowData(sub_tre_1)=rowData(sub_tre_2)
    )
    j <- sample.int(n = dim(tre)[2], size = 1, replace = FALSE)
    sub_tre_3 <- tre[, -j] ## Drop the j-th column
    sub_tre_4 <- tre[, j] ## Extract the j-th column
    testthat::expect_equal(
        colData(tre)[!colData(sub_tre_3), on = names(colData(tre))],
        colData(sub_tre_4) # colData(tre)\colData(sub_tre_3)=colData(sub_tre_4)
    )
    sub_tre_5 <- tre[-i, -j] ## Drop data containing i-th row OR j-th column
    sub_tre_6 <- tre[i, j] ## Extract i-th row AND j-th column
    all_assays <- assays(tre, key = FALSE, withDimnames = TRUE)
    for (a in seq_along(all_assays)) {
        testthat::expect_equal(
            all_assays[[a]][
                !assay(sub_tre_1, a, key = FALSE, withDimnames = TRUE),
            ],
            assay(sub_tre_2, i = a, key = FALSE, withDimnames = TRUE)
         )
        testthat::expect_equal(
            all_assays[[a]][
                !assay(sub_tre_3, a, key = FALSE, withDimnames = TRUE),
            ],
            assay(sub_tre_4, i = a, key = FALSE, withDimnames = TRUE)
        )
        ## tre[i, ] UNION tre[, j] + (tre[i, ] INTERSECT tre[, j])
        union_assay <- rbind(
            assay(sub_tre_2, i = a, key = FALSE, withDimnames = TRUE),
            assay(sub_tre_4, i = a, key = FALSE, withDimnames = TRUE)
        ) ## contains double count, not a union yet
        double_count_idx <- which(duplicated(union_assay))
        # Show that the double counted element is the intersect
        testthat::expect_equal(
            union_assay[double_count_idx, ],
            ## tre[i, ] INTERSECT tre[, j]
            assay(sub_tre_6, i = a, key = FALSE, withDimnames = TRUE),
            ignore_attr = TRUE
        )
        ## Show that these two produce equivalent union sets
 #       union_assay <- setorderv(union_assay[-double_count_idx, ],
 #                                cols = idCols(tre))
 #       testthat::expect_equal(
 #           union_assay,
 #           ## This indirect approach for tre[i, ] UNION tre[, j] is faster
 #           all_assays[[a]][
 #               !assay(sub_tre_5, a, key = FALSE, withDimnames = TRUE),
 #               on = idCols(tre)
 #           ], ## reordering done by internal reindexing
 #           ignore_attr = TRUE
 #       )
    }
})

# == reindex

testthat::test_that("`reindex,LongTale-method` does not mutate by reference", {
    .tre <- copy(tre)
    ntre <- reindex(tre)
    testthat::expect_true(all.equal(.tre, tre))
})

testthat::test_that("`reindex,LongTable-method` has same index as LongTable constructor", {
    ntre <- reindex(tre)
    testthat::expect_true(all.equal(getIntern(ntre, "assayIndex"), getIntern(tre, "assayIndex")))
    testthat::expect_true(all.equal(assays(ntre, raw=TRUE), assays(tre, raw=TRUE)))
})

testthat::test_that("`reindex,LongTable-method` does not corrupt data relationships", {
    ntre <- reindex(tre)
    for (i in seq_along(assayNames(tre))) {
        assay1 <- assay(tre, i, withDimnames=TRUE)
        setkeyv(assay1, idCols(tre))
        assay2 <- assay(ntre, i, withDimnames=TRUE)
        setkeyv(assay2, idCols(ntre))
        testthat::expect_true(all.equal(assay1, assay2))
    }
    assayL1 <- assays(tre)
    assayL2 <- assays(ntre)
    for (i in seq_along(assayL1)) {
        testthat::expect_true(all.equal(assayL1[[i]], assayL2[[i]]))
    }
})

testthat::test_that("`reindex,LongTable-method` removes gaps in keys in subset LongTable", {
    select_row <- seq.int(1, dim(tre)[1], by = 2)
    stre <- tre[select_row, ] ## subset data
    stre <- reindex(stre)
    ## check if rowData and colData keys have gaps
    row_keys <- rowData(stre, key = TRUE)$rowKey
    col_keys <- colData(stre, key = TRUE)$colKey
    has_no_gaps_in_row <- rle(diff(row_keys))$value == 1
    has_no_gaps_in_col <- rle(diff(col_keys))$value == 1
    testthat::expect_true(has_no_gaps_in_row)
    testthat::expect_true(has_no_gaps_in_col)
    ## check if assays' keys have gaps
    for (i in seq_along(assayNames(stre))) {
        assay_name <- assayNames(stre)[i]
        assay_keys <- assay(stre, i, key = FALSE, withDimnames = FALSE, metadata=FALSE)[[paste0(".", assay_name)]]
        has_no_gaps_in_assay <- rle(diff(assay_keys))$value == 1
        if (length(has_no_gaps_in_assay) > 1) print(i)
        testthat::expect_true(has_no_gaps_in_assay)
    } # Leave summary assay out for now
})

# == [[

#testthat::test_that("`[[,LongTable-method` returns assay metadata always with dimnames",{
#    testthat::expect_warning({ tre[[1, withDimnames = FALSE, metadata = TRUE]] },
#        regexp = ".*Unable.to.return.metadata.without.dimnames,.proceeding.as.if.withDimnames=TRUE.*"
#    )
#})
#
#testthat::test_that("`[[,LongTable-method` when keys = TRUE, ignore withDimnames and metadata",{
#    testthat::expect_warning({ tre[[1, keys = TRUE]] },
#        regexp = ".*Ignoring withDimnames and metadata arguments when keys=TRUE.*"
#    )
#})

#testthat::test_that("`[[,LongTable-method` allows only one assay selection at a time",{
#    testthat::expect_error({ tre[[1:2]] },
#        regexp = ".*Please specifying a single string assay name or integer index.*"
#    )
#    testthat::expect_error({ tre[[c("sensitivity", "profiles")]] },
#        regexp = ".*Please specifying a single string assay name or integer index.*"
#    )
#})

#' Tests of table 1 is a subset of table2, in which case there will be no rows
#' is the set difference.
#'
#' @param table1,table2 A `data.table` or a table-like object coercible to one 
#' via `as.data.table`.
#'
#' @return `logical(1)` TRUE if `table1` is a subset of `table2`, otherwise 
#' `FALSE`
#'
#' @importFrom data.table fsetdiff as.data.table
#'
#' @noRd
#' @keywords internal
.table_is_subset <- function(table1, table2) {
    if (!is(table1, "data.table")) table1 <- as.data.table(table1)
    if (!is(table2, "data.table")) table2 <- as.data.table(table2)
    nrow(fsetdiff(
        table1,
        table2,
    )) == 0
}

library(testthat)
library(CoreGx)
library(BiocParallel)
library(data.table)

data(nci_TRE_small)
tre <- copy(nci_TRE_small)
by <- c("treatment1id", "treatment2id", "sampleid")
sens <- tre$sensitivity


## -- Computing aggregations


testthat::test_that("`aggregate2` is equivalent to raw data.table aggregation", {
    ## Single threaded case
    agg_res <- sens |>
        subset(treatment1id %in% unique(treatment1id)[1:3]) |>
        aggregate2(
            mean(treatment1dose), mean(treatment2dose), mean(viability),
            by=by
        )
    ## Mutlithreaded case (via bplapply)
    agg_res_parallel <- sens |>
        subset(treatment1id %in% unique(treatment1id)[1:3]) |>
        aggregate2(
            mean(treatment1dose), mean(treatment2dose), mean(viability),
            by=by,
            nthread=2
        )
    ## data.table default
    dt_res <- sens[
        treatment1id %in% unique(treatment1id)[1:3],
        .(
            mean_treatment1dose=mean(treatment1dose),
            mean_treatment2dose=mean(treatment2dose),
            mean_viability=mean(viability)
        ),
        by=by
    ]
    expect_true(all.equal(
        agg_res,
        dt_res,
        check.attributes=FALSE  # to allow addition of aggregate call as attribute
    ))
    expect_true(all.equal(
        agg_res_parallel,
        dt_res,
        check.attributes=FALSE
    ))
})

testthat::test_that("`aggregate,LongTable-method` is equivalent to aggregating the raw assay data.table", {
    agg_tre <- tre |>
        aggregate(
            assay="sensitivity",
            mean_viability=mean(viability),
            by=by
        )
    agg_tre_parallel <- tre |>
        subset(treatment1id %in% unique(treatment1id)[1:3]) |>
        aggregate(
            assay="sensitivity",
            mean_viability=mean(viability),
            by=by,
            nthread=2
        )
    agg_dt <- tre$sensitivity[,
        .(mean_viability=mean(viability)),
        by=by
    ]
    agg_dt_small <- tre$sensitivity[
        treatment1id %in% unique(treatment1id)[1:3],
        .(mean_viability=mean(viability)),
        by=by
    ]
    expect_true(all.equal(
        agg_tre, agg_dt,
        check.attributes=FALSE
    ))
    expect_true(all.equal(
        agg_tre_parallel, agg_dt_small,
        check.attributes=FALSE
    ))
})

testthat::test_that("`aggregate2` and `aggregate,LongTable-method` automatic naming works correctly", {
    ## aggregate2
    agg2_named <- sens |>
        aggregate2(
            mean_viability=mean(viability), mean_treatment1dose=mean(treatment1dose),
                mean_treatment2dose=mean(treatment2dose),
            by=by
        )
    agg2_unnamed <- sens |>
        aggregate2(
            mean(viability), mean(treatment1dose), mean(treatment2dose),
            by=by
        )
    testthat::expect_true(all.equal(
        agg2_named, agg2_unnamed,
        check.attributes=FALSE
    ))
    ## aggregate,LongTable-method
    agg_named <- tre |>
        aggregate(
            "sensitivity",
            mean_viability=mean(viability), mean_treatment1dose=mean(treatment1dose),
                mean_treatment2dose=mean(treatment2dose),
            by=by
        )
    agg_unnamed <- tre |>
        aggregate(
            "sensitivity",
            mean(viability), mean(treatment1dose), mean(treatment2dose),
            by=by
        )
    testthat::expect_true(all.equal(
        agg_named, agg_unnamed,
        check.attributes=FALSE
    ))
})


## -- Assigning aggregated assays

testthat::test_that("`Assignment doesn't modify summarized assay data", {
    ntre <- copy(tre)
    sens_summary <- tre |>
        aggregate(
            "sensitivity",
            mean_viability=mean(viability, na.rm=TRUE),
            mean_treatment1dose=mean(treatment1dose, na.rm=TRUE),
            mean_treatment2dose=mean(treatment2dose, na.rm=TRUE),
            by=by
        )
    ntre$sens_summary <- sens_summary
    sens_summary2 <- unique(ntre$sens_summary[, .SD, .SDcols=colnames(sens_summary)])
    setkeyv(sens_summary, by)
    setkeyv(sens_summary2, by)
    testthat::expect_true(all.equal(
        sens_summary,
        sens_summary2,
        check.attributes=FALSE
    ))
})

testthat::test_that("`reindex,LongTable-method` doesn't corrupt referrential integrity of summarized assays", {
    ntre <- copy(tre)
    sens_summary <- tre |>
        aggregate(
            "sensitivity",
            mean_viability=mean(viability), mean_treatment1dose=mean(treatment1dose),
                mean_treatment2dose=mean(treatment2dose),
            by=by
        )
    ntre$sens_summary <- sens_summary
    ntre2 <- reindex(ntre)
    expect_true(all.equal(
        ntre$sens_summary,
        ntre2$sens_summary
    ))
})

testthat::test_that("`subset,LongTable-method` works correctly with summary assays", {
    ntre <- copy(tre)
    sens_summary <- tre |>
        aggregate(
            "sensitivity",
            mean_viability=mean(viability), mean_treatment1dose=mean(treatment1dose),
                mean_treatment2dose=mean(treatment2dose),
            by=by
        )
    ntre$sens_summary <- sens_summary
    stre <- subset(ntre, treatment1id %in% treatment1id[1:5])
    testthat::expect_true(CoreGx:::.table_is_subset(
        stre$sens_summary,
        ntre$sens_summary[treatment1id %in% treatment1id[1:5]]
    ))
})

#' @importClassesFrom S4Vectors Annotated list_OR_List
#' @importFrom data.table copy
NULL

#' An S4 Class For Mapping from Raw Experimental Data to a Specific S4 Object
#'
#' This object will be used as a way to abstract away data preprocessing.
#'
#' @section Slots:
#' * rawdata: A list-like object containing one or more pieces of raw data
#'   that will be processed and mapped to the slots of an `S4` object.
#' * metadata: A `List` of object level metadata.
#'
#' @md
#' @aliases DataMapper-class
.DataMapper <- setClass('DataMapper',
    contains=c('VIRTUAL', 'Annotated'),
    slots=list(rawdata='list_OR_List')
)

.local_class_2 <- 'DataMapper'

.docs_DataMapper_accessors <- function(...) .parseToRoxygen(
    "
    @title Accessing and modifying data in a `{class_}` object.

    @description
    Documentation for the various setters and getters which allow manipulation
    of data in the slots of a `{class_}` object.

    @param object A `{class_}` object to get or set data from.
    @param value A `list`-like object to assign to the rawdata slot. Should be
        a `data.frame` or `data.table` with the current implementation.

    @family DataMapper-accessors

    @return Accessors: See details
    @return Setters: An update `{class_}` object, returned invisibly.
    ",
    ...
)


# ==================================
# DataMapper Accessors Documentation
# ----------------------------------

#' @name DataMapper-accessors
#' @eval .docs_DataMapper_accessors(class_=.local_class_2)
NULL

#' @export
setGeneric('rawdata', function(object, ...) standardGeneric('rawdata'))

.docs_DataMapper_get_rawdata <- function(...) .parseToRoxygen(
    "
    @details
    __rawdata__: Get the raw data slot from a `{class_}` object. Returns
    a list-like containing one or more raw data inputs to the
    `{class_}` object.

    @md
    @aliases rawdata,{class_}-method
    @exportMethod rawdata
    ",
    ...
)

#' @rdname DataMapper-accessors
#' @aliases rawdata
#' @eval .docs_DataMapper_get_rawdata(class_=.local_class_2)
setMethod('rawdata', signature(object='DataMapper'), function(object) {
    rawdata_ <- object@rawdata
    return(data.table::copy(rawdata_))
})

#' @export
setGeneric("rawdata<-", function(object, ..., value)
    standardgeneric("rawdata<-"))

.docs_DataMapper_set_rawdata <- function(...) .parseToRoxygen(
    "
    @details
    __rawdata__: Set the raw data slot from a `{class_}` object.
    __value__: The `list`-like object to set for the rawdata slot. Note: this
        currently only supports `data.frame` or `data.table` objects.

    @md
    @aliases rawdata<-,{class_},{class1_}-method
    @exportMethod rawdata<-
    ",
    ...
)

#' @rdname DataMapper-accessors
#' @aliases rawdata<-
#' @eval .docs_DataMapper_set_rawdata(class_=.local_class_2, class1_='ANY')
setReplaceMethod('rawdata', signature(object='DataMapper'),
        function(object, value) {
    object@rawdata <- value
    object
})

# ==== LongTable Class

#' @include LongTableDataMapper-class.R
#' @include DataMapper-class.R
#' @include TreatmentResponseExperiment-class.R
NULL

#' @title LongTable to data.table conversion
#' @name as
#'
#' @examples
#' as(merckLongTable, 'data.table')
#'
#' @description Coerce a LongTable into a `data.table`.
#'
#' @param from `LongTable` Object to coerce.
#'
#' @return A `data.table` with the data from a LongTable.
#'
#' @import data.table
#' @export
setAs('LongTable', 'data.table', def=function(from) {

    # extract the assay data
    longTableData <- assays(from, withDimnames=FALSE, key=TRUE)

    # join assays into a single table
    DT <- longTableData[[1]]
    longTableData[[1]] <- NULL
    for (i in seq_along(longTableData)) {
        DT <- merge.data.table(DT, longTableData[[i]],
            suffixes=c('', paste0('._', i)), by=.EACHI, all.x=TRUE, all.y=TRUE)
    }

    # extract assay columns
    assayCols <- assayCols(from)

    # fix assayCols if there are duplicate column names between assays
    # the join will append '._n' where n is the assay index - 1
    ## TODO:: Make this a helper since it is reused in multiple functions
    .greplAny <- function(...) any(grepl(...))
    .paste0IfElse <- function(vector, suffix, isIn=c('rowKey', 'colKey'))
        ifelse(vector %in% isIn, vector, paste0(vector, suffix))
    hasSuffixes <- unlist(lapply(paste0('._', seq_along(longTableData)),
        FUN=.greplAny, x=colnames(DT)))
    if (any(hasSuffixes)) {
        whichHasSuffixes <- which(hasSuffixes) + 1
        assayCols[whichHasSuffixes] <-
            Map(FUN=.paste0IfElse,
                vector=assayCols[whichHasSuffixes],
                suffix=paste0('._', seq_along(longTableData))[hasSuffixes]
            )
    }

    # join the row and column data
    DT <- merge.data.table(DT, rowData(from, key=TRUE), by='rowKey')
    DT <- merge.data.table(DT, colData(from, key=TRUE), by='colKey')
    setkeyv(DT, c('rowKey', 'colKey'))

    # drop interal key columns
    DT[, c('rowKey', 'colKey') := NULL]

    # organize the returned columns
    colOrder <- unique(c(setdiff(
        colnames(DT), unlist(assayCols)),
        unlist(assayCols)
    ))
    setcolorder(DT, colOrder)

    aMap <- lapply(assayCols, FUN=setdiff,
        y=c(idCols(from), rowMeta(from), colMeta(from)))
    aMap <- Map(list, mutable(getIntern(from, "assayKeys")), aMap)

    DT <- cbind(DT, metadata(from)$experiment_metadata)

    metaCols <- names(metadata(from)$experiment_metadata)
    longTableMapper <- LongTableDataMapper(
        rowDataMap=list(rowIDs(from), rowMeta(from)),
        colDataMap=list(colIDs(from), colMeta(from)),
        assayMap=aMap,
        metadataMap=if (is.character(metaCols)) list(metaCols) else list()
    )
    metadata(longTableMapper) <- metadata(from)[names(metadata(from) != "experiment_metadata")]
    attr(DT, 'longTableDataMapper') <- longTableMapper

    # return the data.table
    return(DT)
})
# #' @title Coerce a LongTable into a `data.table`
# #' @name as
# #'
# #' @description S3 version of coerce method for convenience.
# #'
# #' @param from `LongTable` to coerce to a `data.table`
# #'
# #' @return A `data.table` containing the data from the LongTable, as well
# #'   as the `longTableDataMapper' attribute which contains the data needed to
# #'   reverse the coercion.
# #' @export
# as.data.table.long.table <- function(from, keep.rownames=FALSE,...) as(from, 'data.table')

#' @title Coerce a LongTable into a `data.frame`
#' @name as
#'
#' @description Currently only supports coercing to data.table or data.frame
#'
#' @param from `LongTable` Object to coerce.
#'
#' @return `data.table` containing the data from the LongTable, with the
#'   `longTableDataMapper' attribute containg the metadata needed to reverse
#'   the coercing operation.
#'
#' @importFrom data.table data.table setDF
#' @export
setAs('LongTable', 'data.frame', def=function(from) {
    DT <- as(from, 'data.table')
    setDF(DT)
    return(DT)
})

# #' @title Coerce a LongTable to a data.frame
# #' @name as
# #'
# #' @examples
# #' as(merckLongTable, 'data.frame')
# #'
# #' @description S3 version of coerce method fro convenience.
# #'
# #' @param x `LongTable` to coerce to `data.frame`.
# #' @param row.names An optional `character` vector of rownames. We do not
# #'   recommend using this parameter, it is included for S3 method consistency
# #'   with `as.data.frame`.
# #' @param optional `logical` Is it optional for row and column names to be
# #'   valid R names? If FALSE will use the make.names function to ensure the
# #'   row and column names are valid R names. Defaults to TRUE.
# #' @param ... Does nothing.
# #'
# #' @param from `LongTable` Object to coerce.
# #' 
# #' @return `data.frame` containing the data from the LongTable, with the
# #'   `longTableDataMapper' attribute containg the metadata needed to reverse
# #'   the coercion operation.
# #'
# #' @importFrom data.table data.table
# #' @export
# as.data.frame.long.table <- function(x, row.names, optional=TRUE, ...) {
#     DF <- as(x, 'data.frame')
#     if (!missing(row.names)) {
#         if (!is.character(x) || length(row.names) != nrow(DF))
#             stop(.errorMsg('[CoreGx::as.data.frame.LongTable] The row.names ',
#                 'argument must be a character vector with length equal to ',
#                 nrow(DF)))
#         if (!optional) {
#             row.names <- make.names(row.names)
#             colnames(DF) <- make.names(colnames(DF))
#         }
#         rownames(DF) <- row.names
#     }
#     DF
# }


#' @title Coerce to data.table to LongTable
#' @name as
#'
#' @examples
#' dataTable <- as(merckLongTable, 'data.table')
#' print(attr(dataTable, 'longTableDataMapper')) # Method doesn't work without this
#' as(dataTable, 'LongTable')
#'
#' @description Coerce a data.table with the proper configuration attributes
#'   back to a LongTable
#'
#' @param from A `data.table` with the 'longTableDataMapper' attribute, containing
#'   three lists named assayCols, rowDataCols and colDataCols. This attribute is
#'   automatically created when coercing from a `LongTable` to a `data.table`.
#'
#' @return `LongTable` object configured with the longTableDataMapper
#'
#' @export
setAs('data.table', 'LongTable', def=function(from) {

    if (!('longTableDataMapper' %in% names(attributes(from))))
        stop(.errorMsg('[CoreGx::as,data.table,LongTable] Coercing from ',
            'data.table to LongTable only works if the longTableMapper ',
            'attribute has been set!'))

    longTableMapper <- attr(from, 'longTableDataMapper')

    requiredConfig <- c('assayMap', 'rowDataMap', 'colDataMap')
    hasRequiredConfig <- vapply(requiredConfig,
        FUN=\(x, f) length(do.call(f, list(x)))[[1]] > 0,
        x=longTableMapper, FUN.VALUE=logical(1))
    if (!all(hasRequiredConfig))
        stop(.errorMsg('The longTableDataMapper object is missing data from',
            'the ', paste0(requiredConfig[!hasRequiredConfig], collapse=', '),
            ' slots! Check attributes(from).'))

    rawdata(longTableMapper) <- from
    return(metaConstruct(longTableMapper))
})
#' @name as.long.table
#' @title Coerce from data.table to LongTable
#'
#' @examples
#' dataTable <- as(merckLongTable, 'data.table')
#' print(attr(dataTable, 'longTableDataMapper')) # Method doesn't work without this
#' as.long.table(dataTable)
#'
#' @description Coerce a data.table with the proper configuration attributes
#'   back to a LongTable
#'
#' @param x A `data.frame` with the 'longTableDataMapper' attribute, containing
#'  three lists named assayCols, rowDataCols and colDataCols. This attribute
#'  is automatically created when coercing from a LongTable to a data.table.
#'
#' @return `LongTable` object configured with the longTableDataMapper
#' @export
as.long.table <- function(x) as(x, 'LongTable')


#' @name as
#' @title Coerce a SummarizedExperiment to a data.table
#'
#' @examples
#' SE <- molecularProfilesSlot(clevelandSmall_cSet)[[1]]
#' as(SE, 'data.table')
#'
#' @param from `SummarizedExperiment` object.
#'
#' @return `data.table` with long format of data in `from`
#'
#' @importFrom data.table as.data.table melt.data.table merge.data.table
#' @export
setAs(from='SummarizedExperiment', to='data.table', function(from) {
    # -- extract sample metadata
    colDT <- as.data.table(colData(from), keep.rownames='.sample')
    # -- extract feature metadata
    rowDT <- as.data.table(rowData(from), keep.rownames='.feature')
    # -- extract and process assays
    assayL <- assays(from)
    assayDtL <- lapply(assayL, as.data.table, keep.rownames='.feature')
    meltDtL <- lapply(assayDtL, melt, id.vars='.feature',
        variable.name='.sample', variable.factor=FALSE)
    assayDT <- meltDtL[[1]][, .(.sample, .feature)]
    for (i in seq_along(meltDtL))
        assayDT[[names(assayL)[[i]]]] <- meltDtL[[i]][['value']]
    # -- merge into a single long format table
    DT <- merge.data.table(assayDT, colDT, by='.sample')
    DT <- merge.data.table(DT, rowDT, by='.feature')
    # -- add metadata
    metadata <- metadata(from)
    notS4 <- !vapply(metadata, isS4, logical(1))
    if (!all(notS4))
        .warning('Dropped S4 metadata during coercion to data.table!')
    for (name in names(metadata)[notS4]) assayDT[[name]] <- metadata[[name]]
    return(DT)
})

#' @name as
#' @title Coerce a SummarizedExperiment to a data.frame
#'
#' @examples
#' SE <- molecularProfilesSlot(clevelandSmall_cSet)[[1]]
#' as(SE, 'data.frame')
#'
#' @param from `SummarizedExperiment` object.
#'
#' @return `data.frame` with long format of data in `from`.
#'
#' @importFrom data.table as.data.table melt.data.table merge.data.table
#' @export
setAs(from='SummarizedExperiment', to='data.frame', function(from) {
    setDF(as(from, 'data.table'))
})


#' @title Coerce a `LongTable` into a `SummarizedExperiment`
#' @name as
#'
#' @param from `LongTable` object coerce to a `SummarizedExperiment`. Assays
#'   are converted to `BumpyMatrix`es to allow treatment combination support
#'   and integration with the `gDR` package.
#'
#' @return `SummarizedExperiment` with each assay as a `BumpyMatrix`
#'
#' @seealso [`BumpyMatrix::BumpyMatrix`]
#'
#' @md
#' @importFrom SummarizedExperiment SummarizedExperiment
#' @export
setAs("LongTable", "SummarizedExperiment", def=function(from) {
    .longTableToSummarizedExperiment(from, assay_names=assayNames(from))
})


#' @title Convert a LongTable assay into a BumpyMatrix object
#'
#' @param LT `LongTable` with assay to convert into `BumpyMatrix`
#' @param assay `character(1)` A valid assay name in `LT`, as returned by
#'     `assayNames(LT)`.
#' @param rows `character()` The rownames associated with the assay rowKey
#' @param cols `character()` The names associated with the assay colKey
#' @param sparse `logical(1)` Should the `BumpyMatrix` be sparse (i.e., is the
#'   assay sparse).
#'
#' @return `BumpyMatrix` containing the data from `assay`.
#'
#' @md
#' @importFrom data.table data.table
#' @importFrom BumpyMatrix splitAsBumpyMatrix
.assayToBumpyMatrix <- function(LT, assay, rows, cols, sparse=TRUE) {
    assay_data <- assay(LT, assay, key=TRUE)
    assay_data[, rownames := rows[rowKey]]
    assay_data[, colnames := cols[colKey]]
    assay_data[, c('rowKey', 'colKey') := NULL]
    splitAsBumpyMatrix(assay_data[, -c('rownames', 'colnames')],
        row=assay_data$rownames, column=assay_data$colnames, sparse)
}

#' Convert LongTable to gDR Style SummarizedExperiment
#'
#' @param LT `LongTable` to convert to gDR `SummarizedExperiment` format.
#' @param assay_names `character()` Names to rename the assays to. These
#'   are assumed to be in the same order as `assayNames(LT)`.
#'
#' @return `SummarizedExperiment` object with all assay from `LT` as
#'   `BumpyMatrix`es.
#'
#' @md
#' @importFrom data.table setnames
#' @importFrom SummarizedExperiment SummarizedExperiment
.longTableToSummarizedExperiment <- function(LT, assay_names) {
    assay_list <- lapply(assayNames(LT), FUN=.assayToBumpyMatrix,
        LT=LT, rows=rownames(LT), cols=colnames(LT))
    if (!missing(assay_names) && length(assay_names) == length(assayNames(LT)))
        names(assay_list) <- assay_names
    SummarizedExperiment(
        assays=assay_list, rowData=rowData(LT), colData=colData(LT),
            metadata=c(metadata(LT), list(.intern=as.list(getIntern(LT))))
    )
}

#' @name as
#'
#' @title
#' Coerce a `LongTableDataMapper` to a `TREDataMapper`
#'
#' @param from A `LongTableDataMapper` to coerce.
#'
#' @return A `TREDataMapper` object.
#'
#' @md
#' @export
setAs("LongTableDataMapper", "TREDataMapper", def=function(from) {
    TREDataMapper(from)
})

#' @include LongTableDataMapper-class.R
NULL

## =====================================
## LongTableDataMapper Accessors Methods
## -------------------------------------


## ---------------------
## ---- all slot helpers


#' Method to subset the rawdata with the corresponding dimensions "dimDataMap"
#'   method.
#'
#' @param x `LongTableDataMapper` or inheriting class.
#' @param key `logical(1)` Should the returned value be keyed by the 'id_columns'
#'   item of the dimDataMap? Ignored when dim is "meta".
#' @param dim `character(1)` Which dimension should rawdata to subset for?
#'   Options are "row", "col" and "meta", corresponding to the associated
#'   slots of the `LongTableDataMapper`.
#'
#' @importFrom checkmate assertClass assertLogical
#' @importFrom methods getPackageName
#' @noRd
#' @keywords internal
.get_dimData <- function(x, key, dim=c("row", "col", "meta")) {

    # Input validation
    assertClass(x, "LongTableDataMapper")
    assertLogical(key)

    # Determine which slot and accessor function to use
    dim <- match.arg(dim)
    dimSlot <- paste0(dim, "DataMap")
    dimFun <- get(dimSlot)

    # Get method name to simplify debugging from S4 classes
    funContext <- paste0("\n[", getPackageName(), "::", dimSlot, ",",
        class(x)[1], "-method\n\t")

    .dimDataMap <- dimFun(x)

    # Ensure required data is present
    if (length(unlist(.dimDataMap)) < 1) stop(.errorMsg(funContext,
        "The ", dimSlot, " slot must contain valid data!"))

    return(.get_dimDataFromMap(x=x, key=(key && dim != "meta"),
        dataMap=.dimDataMap, funContext=funContext))
}

#' Method to subset the rawdata with the corresponding dimensions "dimDataMap"
#'   method.
#'
#' @param x `LongTableDataMapper` or inheriting class.
#' @param dataMap `list` The map of a `LongTableDataMapper` dimension, as
#'   returned by the '*DataMap' methods. Can also be used with a single assay
#'   from the `assayMap` method, but not the entire list of assays.
#' @param key `logical(1)` Should the returned value be keyed by the 'id_columns'
#'   item of the dimDataMap? Default is `TRUE`.
#' @param rename `logical(1)` Should columns be renamed from their value to
#'   their name, if the item has a name in the `dataMap`. Default is `TRUE`.
#' @param funContext `character(1)` Contextual information about the calling
#'   function, for debugging. Users don't need to worry about this.
#'
#' @importFrom checkmate assertClass assertLogical assertList
#' @importFrom methods getPackageName
#' @importFrom data.table setkeyv
#' @noRd
#' @keywords internal
.get_dimDataFromMap <- function(x, dataMap, key=TRUE, rename=TRUE, funContext) {

    if (missing(funContext))
        funContext <- paste0("\n[", getPackageName(), "::.get_dimDataMap]\n\t")

    # Input validation
    checkmate::assertClass(x, "LongTableDataMapper")
    checkmate::assertLogical(key)
    checkmate::assertList(dataMap, types=c("character", "NULL"), max.len=2)

    # Extract relevant data
    .rawdata <- rawdata(x)

    # Ensure required data is present
    if (length(.rawdata) < 1) .error(funContext,
        "The rawdata slot must contain valid data!")
    hasDimDataCols <- unlist(dataMap) %in% colnames(.rawdata)
    if (!all(hasDimDataCols)) .error(funContext, "Columns ",
        .collapse(unlist(dataMap)[!hasDimDataCols]),
        " are missing from rawdata!")

    # Subset rawdata, optionally keying table and/or renaming columns
    .dimData <- .rawdata[, .SD, .SDcols=unlist(dataMap)]
    if (key) setkeyv(.dimData, dataMap$id_columns)
    if (rename) {
        old <- unlist(unname(dataMap))
        new <- names(old)
        if (!is.null(new)) {
            names_idx <- new != "" & !is.na(new)
            data.table::setnames(.dimData, old[names_idx], new[names_idx])
        }
    }

    return(unique(.dimData))
}


## ---------------
## -- rawdata slot


#' @rdname LongTableDataMapper-accessors
#' @eval .docs_DataMapper_set_rawdata(class_=.local_class_3,
#' class1_='list')
setReplaceMethod('rawdata', signature=c(object='LongTableDataMapper',
        value='list'), function(object, value) {
    funContext <- .S4MethodContext('rawdata<-', class(object)[1],
        class(value)[1])

    rows <- unlist(rowDataMap(object))
    cols <- unlist(colDataMap(object))
    assays <- unlist(assayMap(object))
    meta <- unlist(metadataMap(object))

    ## TODO:: Improve parsing here such that it only throws warnings if meta-
    ##>data columns are missing

    mandatory <- c(rows, cols, assays, meta)
    if (!length(mandatory) || !length(value)) {
        object@rawdata <- value
    } else if (length(mandatory) && !length(rawdata(object))) {
        hasMandatory <- mandatory %in% colnames(value)
        if (!all(hasMandatory)) {
            stop(.errorMsg(funContext, "One or more map column is missing from value",
                ": ", paste0(mandatory[!hasMandatory], collapse=', '), '!'))
        }
        object@rawdata <- value
    } else {
        stop(.errorMsg(funContext, "In order to assign to the rawdata slot of ",
            "a LongTableDataMapper, either all the map slots must be ",
            "empty or the rawdata slot must be an empty list!"))
    }
    return(object)
})


## --------------------
## ---- rowDataMap slot


##
## -- rowDataMap

.docs_LongTableDataMapper_get_dimDataMap <- function(...) .parseToRoxygen(
    "
    @details
    __{dim_}DataMap__: `list` of two `character` vectors, the first are the
    columns required to uniquely identify each row of a `{class_}` and the
    second any additional {dim_}-level metadata. If the character vectors
    have names, the resulting columns are automatically renamed to the
    item name of the specified column.

    @examples
    {dim_}DataMap({data_})

    @md
    @aliases {dim_}DataMap,{class_}-method
    @exportMethod {dim_}DataMap
    ",
    ...
)

#' @export
setGeneric('rowDataMap', function(object, ...) standardGeneric('rowDataMap'))

#' @rdname LongTableDataMapper-accessors
#' @eval
#' .docs_LongTableDataMapper_get_dimDataMap(dim_='row', class_=.local_class_3,
#' data_=.local_data_3)
setMethod('rowDataMap', signature(object='LongTableDataMapper'),
        function(object) {
    object@rowDataMap
})

#' @export
setGeneric('rowDataMap<-', function(object, ..., value)
    standardGeneric('rowDataMap<-'))

.docs_LongTableDataMapper_set_dimDataMap <- function(...) .parseToRoxygen(
    "
    @details
    __{dim_}DataMap<-__: Update the `@{dim_}DataMap` slot of a `{class_}` object,
    returning an invisible NULL. Arguments:
    - value: A `list` or `List` where the first item is the names of the
    identifier columns -- columns needed to uniquely identify each row in
    {dim_}Data -- and the second item is the metadata associated with those
    the identifier columns, but not required to uniquely identify rows in
    the object rowData.

    @examples
    {dim_}DataMap({data_}) <- list(c('{id_col_}'), c())

    @md
    @aliases rowDataMap<-,{class_},list-method {dim_}DataMap<-{class_},List-method
    @exportMethod {dim_}DataMap<-
    ",
    ...
)


#' @rdname LongTableDataMapper-accessors
#' @eval
#' .docs_LongTableDataMapper_set_dimDataMap(dim_='row', class_=.local_class_3,
#' data_=.local_data_3, id_col_='treatmentid')
setReplaceMethod('rowDataMap', signature(object='LongTableDataMapper',
        value='list_OR_List'), function(object, value) {
    funContext <- '[CoreGx::`rowDataMap<-`,LongTableDataMapper-method]\n\t'
    rawdataCols <- colnames(rawdata(object))

    # -- Handle error conditions
    if (length(value) > 2) {
        stop(.errorMsg(funContext, 'Assignments to rowDataMap should be a list ',
            'of length 2, where the first item is the name of the id columns ',
            'and the second item is the name of the metadata columns which ',
            'map to those id columns.'))
    }

    hasIDcols <- value[[1]] %in% rawdataCols
    if (!all(hasIDcols) && length(rawdata(object))) {
        stop(.errorMsg(funContext, 'One or more of the id columns specified ',
            'in value[[1]] are not valid column names in the rawdata slot of ',
            'this ', class(object)[1], ' object!'))
    }

    if (length(value) > 1 && length(value[[2]]) != 0 && length(rawdata(object))) {
        hasMetaCols <- value[[2]] %in% rawdataCols
        if (!all(hasMetaCols)) {
            stop(.errorMsg(funContext, 'The follow metadata columns in value[[2]] ',
                'are not present in rawdata(object): ',
                .collapse(value[[2]][!hasMetaCols]), '!'))
        }
        hasOneToOneRelationship <-
            value[[2]] %in% cardinality(rawdata(object), group=value[[1]])
        if (!all(hasOneToOneRelationship)) {
            stop(.errorMsg(funContext, 'The columns ',
                .collapse(value[[2]][!hasOneToOneRelationship], ' do not have a ',
                '1:1 relationship with the specified ID columns!')))
        }
    }

    # -- Function body
    object@rowDataMap <- value
    return(object)
})


##
## -- rowData


#' Convenience method to subset the `rowData` out of the `rawdata` slot using
#'   the assigned `rowDataMap` metadata.
#'
#' @param x `LongTableDataMapper` object with valid data in the `rawdata` and
#'   `colDataMap` slots.
#' @param key `logical(1)` Should the table be keyed according to the
#'   `id_columns` of the `rowDataMap` slot? This will sort the table in memory.
#'   Default is TRUE.
#'
#' @return `data.table` The `rowData` as specified in the `rowDataMap` slot.
#'
#' @export
setMethod("rowData", signature("LongTableDataMapper"), function(x, key=TRUE) {
    .get_dimData(x, key, dim="row")
})


## --------------------
## ---- colDataMap slot


##
## -- colDataMap

#' @export
setGeneric('colDataMap', function(object, ...) standardGeneric('colDataMap'))

#' @rdname LongTableDataMapper-accessors
#' @eval
#' .docs_LongTableDataMapper_get_dimDataMap(dim_='col', class_=.local_class_3,
#' data_=.local_data_3)
setMethod('colDataMap', signature(object='LongTableDataMapper'),
        function(object) {
    object@colDataMap
})

#' @export
setGeneric('colDataMap<-', function(object, ..., value) standardGeneric('colDataMap<-'))

#' @rdname LongTableDataMapper-accessors
#' @eval
#' .docs_LongTableDataMapper_set_dimDataMap(dim_='col', class_=.local_class_3,
#' data_=.local_data_3, id_col_='sampleid')
setReplaceMethod('colDataMap',
        signature(object='LongTableDataMapper', value='list_OR_List'),
        function(object, value) {
    funContext <- '[CoreGx::`colDataMap<-`,LongTableDataMapper-method]\n\t'
    rawdataCols <- colnames(rawdata(object))

    # -- Handle error conditions
    if (length(value) > 2 || !is.list(value)) {
        .error(funContext, 'Assignments to colDataMap should be a list ',
            'of length 2, where the first item is the name of the id columns ',
            'and the second item is the name of the metadata columns which ',
            'map to those id columns.')
    }

    hasIDcols <- value[[1]] %in% rawdataCols
    if (!all(hasIDcols) && length(rawdata(object))) {
        .error(funContext, 'One or more of the id columns specified ',
            'in value[[1]] are not valid column names in the rawdata slot of ',
            'this ', class(object)[1], ' object!')
    }

    if (length(value) > 1 && length(value[[2]]) != 0 &&
            length(rawdata(object))) {
        hasMetaCols <- value[[2]] %in% rawdataCols
        if (!all(hasMetaCols)) {
            .error(funContext,
                'The follow metadata columns in value[[2]] ',
                'are not present in rawdata(object): ',
                .collapse(value[[2]][!hasMetaCols]), '!')
        }
        hasOneToOneRelationship <-
            value[[2]] %in% cardinality(rawdata(object), group=value[[1]])
        if (!all(hasOneToOneRelationship)) {
            .error(funContext, 'The columns ',
                .collapse(value[[2]][!hasOneToOneRelationship]),
                ' do not have a 1:1 relationship with the specified ID ',
                'columns!')
        }
    }

    # -- Function body
    object@colDataMap <- value
    return(object)
})


##
## -- colData


#' Convenience method to subset the `colData` out of the `rawdata` slot using
#'   the assigned `colDataMap` metadata.
#'
#' @param x `LongTableDataMapper` object with valid data in the `rawdata` and
#'   `colDataMap` slots.
#' @param key `logical(1)` Should the table be keyed according to the
#'   `id_columns` of the `colDataMap` slot? This will sort the table in memory.
#'   Default is TRUE.
#'
#' @return `data.table` The `colData` as specified in the `colDataMap` slot.
#'
#' @export
setMethod("colData", signature("LongTableDataMapper"), function(x, key=TRUE) {
    .get_dimData(x, key, dim="col")
})


## ----------------
## ---- assayMap slot


#' @export
setGeneric('assayMap', function(object, ...) standardGeneric('assayMap'))

.docs_LongTableDataMapper_get_assayMap <- function(...) .parseToRoxygen(
    "
    @details
    __assayMap__:  A `list` of character vectors. The name of each list item
    will be the assay in a `LongTableDataMapper` object that the columns in the
    `character` vector will be assigned to. Column renaming occurs automatically
    when the character vectors have names (from the value to the name).

    @examples
    assayMap({data_})

    @md
    @aliases assayMap,{class_},list-method assayMap,{class_},List-method
    @exportMethod assayMap
    ",
    ...
)

#' @rdname LongTableDataMapper-accessors
#' @eval .docs_LongTableDataMapper_get_assayMap(class_=.local_class_3, data_=.local_data_3)
setMethod('assayMap', signature(object='LongTableDataMapper'),
        function(object) {
    object@assayMap
})


#' @export
setGeneric('assayMap<-', function(object, ..., value) standardGeneric('assayMap<-'))

#' @noRd
.docs_LongTableDataMapper_set_assayMap <- function(...) .parseToRoxygen(
    "
    @details
    __assayMap<-__: Updates the `@assayMap` slot of a `{class_}` object,
    returning an invisible NULL. Arguments:
    - value:  A `list` of character vectors, where the name of each list
    item is the name of an assay and the values of each character vector
    specify the columns mapping to the assay in the `S4` object the
    `{class_}` constructs.

    @examples
    assayMap({data_}) <- list(sensitivity=c(viability1='viability'))

    @md
    @aliases assayMap<-,{class_},list-method assayMap<-,{class_},List-methhod
    @exportMethod assayMap<-
    ",
    ...
)

#' @rdname LongTableDataMapper-accessors
#' @eval .docs_LongTableDataMapper_set_assayMap(class_=.local_class_3, data_=.local_data_3)
setReplaceMethod('assayMap', signature(object='LongTableDataMapper',
        value='list_OR_List'), function(object, value) {
    funContext <- '[CoreGx::`assayMap<-,LongTableDataMapper-method`]\n\t'
    rawdataCols <- colnames(rawdata(object))
    if (length(names(value)) == 0) stop(.errorMsg('The value argument must
        be a named list-like of column name character vectors!'))

    for (i in seq_along(value)) {
        hasRawdataCols <- unlist(value[[i]]) %in% rawdataCols
        if (!all(hasRawdataCols) && length(rawdata(object))) {
            stop(.errorMsg(funContext, 'There are no columns named ',
                .collapse(unlist(value[[i]])[!hasRawdataCols]),
                ' in the rawdata of this ', class(object)[1],
                ' object. Please ensure item ',
                names(value)[i], ' of value has valid column names.'))
        }
    }

    object@assayMap <- value
    return(object)
})


#' Extract the data for an assay from a `LongTableDataMapper`
#'
#' @param x `LongTableDataMapper` The object to retrive assay data form according
#'   to the `assayMap` slot.
#' @param i `character(1)` Name of an assay in the `assayMap` slot of `x`.
#' @param withDimnames `logical(1)` For compatibility with
#'   `SummarizedExperiment::assay` generic. Not used.
#'
#' @return `data.table` Data for the specified assay extracted from the
#'   `rawdata` slot of `x`.
#'
#' @importFrom checkmate assertSubset assertCharacter
#' @keywords internal
setMethod("assay", signature(x="LongTableDataMapper"),
        function(x, i, withDimnames=TRUE) {

    # Input validation
    .assayMap <- assayMap(x)
    assertCharacter(i, max.len=1)
    assertSubset(i, names(.assayMap))

    # Execution context
    funContext <- paste0("\n[", getPackageName(), "::assay,", class(x)[1],
        "-method]")

    return(.get_dimDataFromMap(x, key=TRUE, .assayMap[[i]],
        funContext=funContext))
})


#' Extract the data for all assays from a `LongTableDataMapper`
#'
#' @param x `LongTableDataMapper` The object to retrive assay data form according
#'   to the `assayMap` slot.
#' @param withDimNames `logical(1)` For compatibility with
#'   `SummarizedExperiment::assay` generic. Not used.
#'
#' @return `list` Data for all assays extracted from the
#'   `rawdata` slot of `x` as a `list` of `data.tables`, where the `keys` for
#'   each table are their `id_columns`.
#'
#' @importFrom checkmate assertSubset assertCharacter
#' @keywords internal
setMethod("assays", signature(x="LongTableDataMapper"),
        function(x, withDimnames=TRUE) {
    lapply(names(assayMap(x)), FUN=assay, x=x) |>
        setNames(names(assayMap(x)))
})

# -- metadataMap


#' @export
setGeneric('metadataMap', function(object, ...) standardGeneric('metadataMap'))

#' @noRd
.docs_LongTableDataMapper_get_metadataMap <- function(...) .parseToRoxygen(
    "
    @details
    __metadataMap__:  A `list` of `character` vectors. Each item is an element
    of the constructed objects `@metadata` slot.

    @examples
    metadataMap({data_})

    @md
    @aliases metadataMap,{class_}-method
    @exportMethod metadataMap
    ",
    ...
)

#' @rdname LongTableDataMapper-accessors
#' @eval .docs_LongTableDataMapper_get_metadataMap(class_=.local_class_3, data_=.local_data_3)
setMethod('metadataMap', signature(object='LongTableDataMapper'),
        function(object) {
    object@metadataMap
})


#' @export
setGeneric('metadataMap<-', function(object, ..., value)
    standardGeneric('metadataMap<-'))

#' @noRd
.docs_LongTableDataMapper_set_metadataMap <- function(...) .parseToRoxygen(
    "
    @details
    __metadataMap<-__: Updates `{class_}` object in-place, then returns an
    `invisible(NULL)`. Arguments:
    - value:  A `list` of `character` vectors. The name of each list item
    is the name of the item in the `@metadata` slot of the `{class_}` object
    created when `metaConstruct` is called on the `DataMapper`, and a
    character vector specifies the columns of `@rawdata` to assign to each item.

    @examples
    metadataMap({data_}) <- list(object_metadata=c('metadata'))

    @md
    @aliases metadataMap<-,{class_}-method
    @exportMethod metadataMap<-
    ",
    ...
)

#' @rdname LongTableDataMapper-accessors
#' @eval .docs_LongTableDataMapper_set_metadataMap(class_=.local_class_3, data_=.local_data_3, col_='metadata')
setReplaceMethod('metadataMap', signature(object='LongTableDataMapper',
    value='list_OR_List'), function(object, value) {
    funContext <- '[CoreGx::`metadataMap<-,LongTableDataMapper-method`]\n\t'
    rawdataCols <- colnames(rawdata(object))
    if (length(names(value)) == 0) stop(.errorMsg('The value argument must
        be a named list-like of column name character vectors!'))

    for (i in seq_along(value)) {
        hasRawdataCols <- value[[i]] %in% rawdataCols
        if (!all(hasRawdataCols) && length(rawdata(object))) {
            stop(.errorMsg(funContext, 'There are no columns named ',
                .collapse(value[[i]][!hasRawdataCols]), ' in the rawdata ',
                'of this ', class(object)[1], ' object. Please ensure item ',
                names(value)[i], ' of value has valid column names.'))
        }
    }

    object@metadataMap <- value
    return(object)
})

#' Function computing connectivity scores between two signatures
#' 
#' A function for finding the connectivity between two signatures, using either
#' the GSEA method based on the KS statistic, or the gwc method based on a 
#' weighted spearman statistic. The GSEA analysis is implemented in the piano package. 
#' 
#' @references 
#'    F. Pozzi, T. Di Matteo, T. Aste, 'Exponential smoothing weighted
#'    correlations', The European Physical Journal B, Vol. 85, No 6, 2012. DOI:
#'    10.1140/epjb/e2012-20697-x
#' @references
#'    Varemo, L., Nielsen, J. and Nookaew, I. (2013) Enriching the gene set
#'    analysis of genome-wide data by incorporating directionality of gene
#'    expression and combining statistical hypotheses and methods. Nucleic
#'    Acids Research. 41 (8), 4378-4391. doi: 10.1093/nar/gkt111
#'    
#' @examples
#' xValue <- c(1,5,23,4,8,9,2,19,11,12,13)
#' xSig <- c(0.01, 0.001, .97, 0.01,0.01,0.28,0.7,0.01,0.01,0.01,0.01)
#' yValue <- c(1,5,10,4,8,19,22,19,11,12,13)
#' ySig <- c(0.01, 0.001, .97,0.01, 0.01,0.78,0.9,0.01,0.01,0.01,0.01)
#' xx <- cbind(xValue, xSig)
#' yy <- cbind(yValue, ySig)
#' rownames(xx) <- rownames(yy) <- c('1','2','3','4','5','6','7','8','9','10','11')
#' data.cor <- connectivityScore(xx,yy,method='gwc', gwc.method='spearman', nperm=300)
#' 
#' @param x A \code{matrix} with the first gene signature. In the case of GSEA the vector of
#'   values per gene for GSEA in which we are looking for an enrichment. In the 
#'   case of gwc, this should be a matrix, with the per gene responses in the 
#'   first column, and the significance values in the second.
#' @param y A \code{matrix} with the second signature. In the case of GSEA, this is the
#'   vector of up and down regulated genes we are looking for in our signature,
#'   with the direction being determined from the sign. In the case of gwc, this
#'   should be a matrix of identical size to x, once again with the per gene
#'   responses in the first column, and their significance in the second.
#' @param method \code{character} string identifying which method to use, out of 'fgsea' and 'gwc'
#' @param nperm \code{numeric}, how many permutations should be done to determine
#'   significance through permutation testing? The minimum is 100, default is
#'   1e4.
#' @param nthread \code{numeric}, how many cores to run parallel processing on.
#' @param gwc.method \code{character}, should gwc use a weighted spearman or pearson
#'   statistic?
#' @param ... Additional arguments passed down to gsea and gwc functions
#' 
#' @return \code{numeric} a numeric vector with the score and the p-value associated
#'   with it
#' 
#' @export
#' 
#' @importFrom piano runGSA loadGSC
#' @importFrom stats complete.cases
## TODO:: Implement 'gsea' method for this function
connectivityScore <- function(x, y, method = c("fgsea", "gwc"), nperm = 10000, nthread = 1, gwc.method = c("spearman", "pearson"), ...) {
    
    if (method == "gsea") {
        stop("The gsea method is implemented using fgsea, please change your
             method argument to 'fgsea'. Consult ?connectivityScore for
             more information")
    }
    
    method <- match.arg(method)
    gwc.method <- match.arg(gwc.method)
    if (!is.matrix(x)) {
        x <- as.matrix(x)
    }
    if (!is.matrix(y)) {
        y <- as.matrix(y)
    }
    if ((ncol(x) != 2 || ncol(y) != 2) && method == "gwc") {
        stop("x and y should have 2 columns: effect size and corresponding p-values")
    }
    if (method == "fgsea" && nrow(y) >= nrow(x)) {
        warning("GSEA method: query gene set (y) larger than signature (x)")
    }
    
    if (is.null(rownames(x)) || is.null(rownames(y)) || !length(intersect(rownames(x), rownames(y)))) {
        stop("Row names of x and y are either missing or have no intersection")
    }
    if (nperm < 100) {
        stop("The minimum number of permutations for permutation testing is 100")
    }
    switch(method, fgsea = {
        ## remove missing values
        y <- y[!is.na(y[, 1]), , drop = FALSE]
        x <- x[!is.na(x[, 1]), , drop = FALSE]
        ## create gene set
        gset <- cbind(gene = rownames(y), set = ifelse(as.numeric(y[, 1]) >= 0, "UP", "DOWN"))
        gset <- piano::loadGSC(gset)
        
        ## run enrichment analysis
        ##FIXME:: Update runGSA to use fgseaMultilevel to stop warnings
        suppressWarnings({ 
            nes <- piano::runGSA(geneLevelStats = x[, 1], geneSetStat = "fgsea", 
                                 gsc = gset, nPerm = nperm + (nperm%%nthread), 
                                 ncpus = nthread, 
            verbose = FALSE, adjMethod = "none", ...) 
        })
        
        ## merge p-values for negative and positive enrichment scores
        nes$pDistinctDir <- nes$pDistinctDirUp
        nes$pDistinctDir[is.na(nes$pDistinctDirUp), 1] <- nes$pDistinctDirDn[is.na(nes$pDistinctDirUp), 1]
        nes.up <- c(nes$statDistinctDir[which(names(nes$gsc) == "UP"), 1], nes$pDistinctDir[which(names(nes$gsc) == "UP"), 1])
        nes.down <- c(nes$statDistinctDir[which(names(nes$gsc) == "DOWN"), 1], nes$pDistinctDir[which(names(nes$gsc) == "DOWN"), 1])
        ## combine UP and DOWN
        if (length(nes.up) == 0) {
            score = c(es = -nes.down[1], p = nes.down[2])
        } else if (length(nes.down) == 0) {
            score = c(es = nes.up[1], p = nes.up[2])
        } else if (all(complete.cases(cbind(nes.up, nes.down))) && sign(nes.up[1]) != sign(nes.down[1])) {
            score <- c(es = (nes.up[1] - nes.down[1])/2, p = .combineTest(p = c(nes.up[2], nes.down[2]), method = "fisher", na.rm = TRUE))
        } else {
            score <- c(score = 0, p = 1)
        }
    }, gwc = {
        ## intersection between x and y
        ii <- intersect(rownames(x), rownames(y))
        if (length(ii) < 10) {
            stop("Less than 10 probes/genes in common between x and y")
        }
        score <- gwc(x1 = x[ii, 1], p1 = x[ii, 2], x2 = y[ii, 1], p2 = y[ii, 2], method.cor = gwc.method, nperm = nperm, ...)
        names(score) <- c("score", "p")
    })
    return(score)
}

#' @importFrom stats pchisq qnorm pnorm pt
.combineTest <- function(p, weight, method = c("fisher", "z.transform", "logit"), hetero = FALSE, na.rm = FALSE) {
    if (hetero) {
        stop("function to deal with heterogeneity is not implemented yet!")
    }
    method <- match.arg(method)
    na.ix <- is.na(p)
    if (any(na.ix) && !na.rm) {
        stop("missing values are present!")
    }
    if (all(na.ix)) 
        {
            return(NA)
        }  ## all p-values are missing
    p <- p[!na.ix]
    k <- length(p)
    if (k == 1) {
        return(p)
    }
    if (missing(weight)) {
        weight <- rep(1, k)
    }
    switch(method, fisher = {
        cp <- pchisq(-2 * sum(log(p)), df = 2 * k, lower.tail = FALSE)
    }, z.transform = {
        z <- qnorm(p, lower.tail = FALSE)
        cp <- pnorm(sum(weight * z)/sqrt(sum(weight^2)), lower.tail = FALSE)
    }, logit = {
        tt <- (-sum(log(p/(1 - p))))/sqrt(k * pi^2 * (5 * k + 2)/(3 * (5 * k + 4)))
        cp <- pt(tt, df = 5 * k + 4, lower.tail = FALSE)
    })
    return(cp)
}

#' @include LongTable-class.R
#' @include LongTable-accessors.R
#' @include TreatmentResponseExperiment-class.R
NULL

#' @importFrom BiocParallel bpparam bpworkers<- bpworkers bpprogressbar<-
#' bpprogressbar bplapply
#' @importFrom data.table rbindlist setDT merge.data.table
#' @importMethodsFrom S4Vectors aggregate
NULL

#' @noRd
.docs_CoreGx_aggregate <- function(...) CoreGx:::.parseToRoxygen(
    "
    @param by `character` One or more valid column names in `x` to compute
    groups using.
    @param ... `call` One or more aggregations to compute for each group by in x.
    If you name aggregation calls, that will be the column name of the value
    in the resulting `data.table` otherwise a default name will be parsed from
    the function name and its first argument, which is assumed to be the name
    of the column being aggregated over.
    @param nthread `numeric(1)` Number of threads to use for split-apply-combine
    parallelization. Uses `BiocParllel::bplapply` if nthread > 1 or you pass in
    `BPPARAM`. Does not modify data.table threads, so be sure to use
    setDTthreads for reasonable nested parallelism. See details for performance
    considerations.
    @param progress `logical(1)` Display a progress bar for parallelized
    computations? Only works if `bpprogressbar<-` is defined for the current
    BiocParallel back-end.
    @param BPPARAM `BiocParallelParam` object. Use to customized the
    the parallization back-end of bplapply. Note, nthread over-rides any
    settings from BPPARAM as long as `bpworkers<-` is defined for that class.
    @param enlist `logical(1)` Default is `TRUE`. Set to `FALSE` to evaluate
    the first call in `...` within `data.table` groups. See details for more
    information.
    @param moreArgs `list()` A named list where each item is an argument one of
    the calls in `...` which is not a column in the table being aggregated. Use
    to further parameterize you calls. Please note that these are not added
    to your aggregate calls unless you specify the names in the call.

    @details
    ## Use of Non-Standard Evaluation
    Arguments in `...` are substituted and wrapped in a list, which is passed
    through to the j argument of `[.data.table` internally. The function currently
    tries to build informative column names for unnamed arguments in `...` by
    appending the name of each function call with the name of its first argument,
    which is assumed to be the column name being aggregated over. If an argument
    to `...` is named, that will be the column name of its value in the resulting
    `data.table`.

    ## Enlisting
    The primary use case for `enlist=FALSE` is to allow computation of dependent
    aggregations, where the output from a previous aggregation is required in a
    subsequent one. For this case, wrap your call in `{curly}` and assign intermediate
    results to variables, returning the final results as a list where each list
    item will become a column in the final table with the corresponding name.
    Name inference is disabled for this case, since it is assumed you will name
    the returned list items appropriately.
    A major advantage over multiple calls to `aggregate` is that
    the overhead of parallelization is paid only once even for complex multi-step
    computations like fitting a model, capturing its paramters, and making
    predictions using it. It also allows capturing arbitrarily complex calls
    which can be recomputed later using the
    `update,TreatmentResponseExperiment-method`
    A potential disadvantage is increased RAM usage per
    thread due to storing intermediate values in variables, as well as any
    memory allocation overhead associate therewith.
    ",
    ...
)

#' Functional API for aggregation over a `LongTable` or inheriting class
#'
#' @description
#' Compute a group-by operation over a `LongTable` object or it's inhering
#' classes.
#'
#' @param x `LongTable` or inheriting class to compute aggregation on.
#' @param assay `character(1)` The assay to aggregate over.
#' @param subset `call` An R call to evaluate before perfoming an aggregate.
#' This allows you to aggregate over a subset of columns in an assay but have
#' it be assigned to the parent object. Default is TRUE, which includes all
#' rows. Passed through as the `i` argument in `[.data.table`.
#' @eval .docs_CoreGx_aggregate(curly="{")
#'
#' @return `data.table` of aggregation results.
#'
#' @seealso `data.table::[.data.table`, `BiocParallel::bplapply`
#'
#' @export
setMethod("aggregate", signature(x="LongTable"),
        function(x, assay, by, ...,  subset=TRUE, nthread=1, progress=TRUE,
        BPPARAM=NULL, enlist=TRUE, moreArgs=list()) {
    i <- substitute(subset)
    assay_ <- x[[assay]][eval(i), ]
    aggregate2(
        assay_,
        by=by,
        ...,
        nthread=nthread, progress=progress, BPPARAM=BPPARAM, enlist=enlist,
            moreArgs=moreArgs)
})


#' Functional S4 API for aggregation over a `data.table` object.
#'
#' @description
#' Compute a group-by operation over a `data.table` in a functional, pipe
#' compatible format.
#'
#' @details
#' This S4 method override the default `aggregate` method for a `data.frame`
#' and as such you need to call `aggregate.data.frame` directly to get the
#' original S3 method for a `data.table`.
#'
#' @param x `data.table` to compute aggregation over.
#' @param subset `call` An R call to evaluate before perfoming an aggregate.
#' This allows you to aggregate over a subset of columns in an assay but have
#' it be assigned to the parent object. Default is TRUE, which includes all
#' rows. Passed through as the `i` argument in `[.data.table`.
#' @eval .docs_CoreGx_aggregate(curly="{")
#'
#' @return `data.table` of aggregated results with an `aggregations` attribute
#' capturing metadata about the last aggregation performed on the table.
#'
#' @export
setMethod("aggregate", signature="data.table",
        function(x, by, ..., subset=TRUE, nthread=1, progress=TRUE,
        BPPARAM=NULL, enlist=TRUE, moreArgs=list()) {
    i <- substitute(subset)
    assay_ <- x[eval(i), ]
    aggregate2(
        x,
        by=by,
        ...,
        nthread=nthread, progress=progress, BPPARAM=BPPARAM, enlist=enlist,
            moreArgs=moreArgs)
})

#' Functional API for data.table aggregation which allows capture of associated
#' aggregate calls so they can be recomputed later.
#'
#' @param x `data.table`
#' @eval .docs_CoreGx_aggregate(curly="{")
#'
#' @return `data.table` of aggregation results.
#'
#' @seealso `data.table::[.data.table`, `BiocParallel::bplapply`
#'
#' @export
aggregate2 <- function(x, by, ..., nthread=1, progress=interactive(), BPPARAM=NULL,
        enlist=TRUE, moreArgs=list()) {
    ## TODO:: refator to checkmate
    stopifnot(is.data.table(x))
    stopifnot(is.character(by) && all(by %in% colnames(x)))
    stopifnot(is.logical(progress) && length(progress) == 1)
    stopifnot(is.logical(enlist) && length(enlist) == 1)
    stopifnot(is.list(moreArgs))
    stopifnot(length(moreArgs) == 0 || all(names(moreArgs) != ""))

    # -- assign moreArgs to the function scope, if it is able to find the values
    for (nm in names(moreArgs)) assign(nm, moreArgs[[nm]])

    # -- capture dots as a call and parse dot names, adding default names if
    # --   they are missing
    agg_call <- if (enlist) substitute(list(...)) else substitute(...)
    if (!enlist && ...length() > 1) warning(.warnMsg("Only one call can be ",
        "passed via ... when enlist=FALSE, ignoring all but first arugment!"))
    dot_names <- if (enlist) names(agg_call)[-1L] else ...names()
    if (is.null(dot_names) && enlist) dot_names <- rep("", length(agg_call) - 1)
    for (i in which(dot_names == "")) {
        dot_call <- agg_call[[i + 1]]
        # assumes the first argument in a function call is always the column name!
        dot_names[i] <- paste0(dot_call[1:max(2, length(dot_call))], collapse="_")
    }
    call_idx <- if (!enlist) 2L else seq(2L, length(agg_call))
    if (length(dot_names)) names(agg_call)[call_idx] <- dot_names

    # -- compute the aggregates, parallelizing if nthread > 1
    if (nthread == 1 && is.null(BPPARAM)) {
        res <- x[, eval(agg_call), by=c(by)]
    } else {
        x <- copy(x) # prevent modifying the source by reference
        # compute groups such that there is one table per thread
        x[, group_id := .GRP, by=by]
        ngrp <- x[, max(group_id)]
        grp_size <- ceiling(ngrp / nthread)
        x[, split_id := ceiling(group_id / grp_size)]
        x_split <- split(x, by="split_id")
        stopifnot(length(x_split) == nthread)
        if (is.null(BPPARAM)) {
            BPPARAM <- BiocParallel::bpparam()
        }
        # optionally add progressbar
        if (hasMethod("bpprogressbar<-", signature=c(class(BPPARAM), "logical"))) {
            BiocParallel::bpprogressbar(BPPARAM) <- progress
        } else if (isTRUE(progress)) {
            warning(.warnMsg(
                "Unable to set progressbar for BiocParallel backend: ",
                class(BPPARAM)[1]), .call=FALSE)
        }
        # optionally set nthread
        if (hasMethod("bpworkers<-", signature=c(class(BPPARAM), "integer"))) {
            BiocParallel::bpworkers(BPPARAM) <- nthread
        } else if (nthread > 1) {
            warning(.warnMsg("Unable to set nthread for BiocParallel backend: ",
                class(BPPARAM)[1]), .call=FALSE)
        }
        res <- BiocParallel::bplapply(
            x_split,
            function(x, agg_call, by) x[, eval(substitute(agg_call)), by=c(by)],
            agg_call=agg_call, by=by,
            BPPARAM=BPPARAM
        )
        res <- rbindlist(res)
    }
    attributes(res)[["aggregations"]] <- c(
        attributes(res)[["aggregations"]],
        list(
            agg_call=agg_call,
            by=by,
            enlist=enlist,
            moreArgs=moreArgs
        )
    )
    return(res)
}

#' @importFrom data.table setDTthreads as.data.table
#' @importFrom bench press mark
#' @importFrom parallel detectCores


#' @title A helper method to find the best multithreading configuration for your
#'   computer
#'
#' @param sample_data `TreatmentResponseExperiment`
#' @param set `logical(1)` Should the function modify your R environment
#'   with the predicted optimal settings? This changes the global state of your
#'   R session!
#' @param report `logical(1)` Should a `data.frame` of results be returned
#'   by number of threads and operation be returned? Defaults to `!set`.
#'
#' @return
#' If `set=TRUE`, modifies `data.table` threads via `setDTthreads()`, otherwise
#' displays a message indicating the optimal number of threads.
#' If `report=TRUE`, also returns a `data.frame` of the benchmark results.
#'
#' @examples
#' \donttest{
#'   data(merckLongTable)
#'   optimizeCoreGx(merckLongTable)
#' }
#'
#' @md
#' @export
optimizeCoreGx <- function(sample_data, set=FALSE, report=!set) {
    ncores <- max(1, parallel::detectCores() - 2)
    nthread_range <- if (ncores == 1) 1 else c(1, seq(2, ncores, 2))
    old_threads <- data.table::getDTthreads()
    message("Benchmarking assay(sample_data, withDimnames=TRUE)...")
    assay_report <- bench::press(nthread=nthread_range, {
        data.table::setDTthreads(nthread)
        gc()
        bench::mark({ assay(sample_data, 1, withDimnames=TRUE); NA })
    })
    message("Benchmarking assays(sample_data)...")
    assays_report <- bench::press(nthread=nthread_range, {
        data.table::setDTthreads(nthread)
        gc()
        bench::mark({ assays(sample_data); NA })
    })
    message("Benchmarking reindex(sample_data)...")
    reindex_report <- bench::press(nthread=nthread_range, {
        data.table::setDTthreads(nthread)
        gc()
        bench::mark({ reindex(sample_data); NA })
    })
    report_table <- as.data.table(
        rbind(assay_report, assays_report, reindex_report)
    )[,
        .(expression=as.character(expression), nthread,
            min_sec=as.numeric(min), median_sec=as.numeric(median),
            total_sec=as.numeric(total_time),
            mem_alloc_mb=as.numeric(mem_alloc) / 1e6, `itr/sec`,  `gc/sec`)
    ]
    best_results <- report_table[,
        .SD[which.min(median_sec), .(time=median_sec, nthread=nthread)],
        by=expression
    ]
    optimal_cores <- best_results[,
        as.integer(as.data.table(table(nthread))[which.max(N), nthread])
    ]
    message("Optimal cores for your machine are: ", optimal_cores)
    if (set) {
        message("Setting optimal cores")
        data.table::setDTthreads(optimal_cores)
    } else {
        data.table::setDTthreads(old_threads)
    }
    if (report) return(as.data.frame(report_table)) else return(invisible(NULL))
}

#' @include LongTable-class.R
#' @noRd
NULL

#' @title TreatmentResponseExperiment class definition
#'
#' @description Define a private constructor method to be used to build a
#'   `TreatmentResponseExperiment` object.
#'
#' @slot rowData See Slots section.
#' @slot colData See Slots section.
#' @slot assays See Slots section.
#' @slot metadata See Slots section.
#' @slot .intern See Slots section.
#'
#' @section Slots:
#' - *rowData*: A `data.table` containing the metadata associated with the
#'   row dimension of a `TreatmentResponseExperiment`.
#' - *colData*: A `data.table` containing the metadata associated with the
#'   column dimension of a `TreatmentResponseExperiment`.
#' - *assays*: A `list` of `data.table`s, one for each assay in a
#'   `TreatmentResponseExperiment`.
#' - *metadata*: An optional `list` of additional metadata for a
#'   `TreatmentResponseExperiment` which doesn't map to one of the dimensions.
#' - *.intern*: An `environment` that holds internal structural metadata
#'   about a `TreatmentResponseExperiment` object, such as which columns are
#'   required to key the object. An environment has been used to allow locking
#'   items, which can prevent accidental modification of a property required
#'   for the class to work.
#'
#' @return `TreatmentResponseExperiment` object containing the assay data from
#'   a treatment response experiment
#'
#' @md
#' @import data.table
#' @keywords internal
#' @rdname TreatmentResponseExperiment-class
#' @aliases .TreatmentResponseExperiment
#' @exportClass TreatmentResponseExperiment
.TreatmentResponseExperiment <- setClass("TreatmentResponseExperiment",
    contains="LongTable")


#' @title TreatmentResponseExperiment constructor method
#'
#' @rdname TreatmentResponseExperiment
#'
#' @description Builds a `TreatmentResponseExperiment` object from rectangular
#' objects. The `rowData` argument should contain row level metadata, while
#' the `colData` argument should contain column level metadata, for the
#' experimental assays
#' in the `assays` list. The `rowIDs` and `colIDs` lists are used to configure
#' the internal keys mapping rows or columns to rows in the assays. Each list
#' should contain at minimum one character vector, specifying which columns
#' in `rowData` or `colData` are required to uniquely identify each row. An
#' optional second character vector can be included, specifying any metadata
#' columns for either dimension. These should contain information about each
#' row but NOT be required to uniquely identify a row in the `colData` or
#' `rowData` objects. Additional metadata can be attached to a
#' `TreatmentResponseExperiment` by passing a list to the metadata argument.
#'
#' @details
#' For now this class is simply a wrapper around a `LongTable` class. In the
#' future we plan to refactor CoreGx such that the `LongTable` class is in a
#' separate pacakge. We can then specialize the implementation of
#' `TreatmentResponseExperiment` to better capture the biomedical nature of
#' this object.
#'
#' @param rowData `data.table`, `data.frame`, `matrix` A table like object
#'   coercible to a `data.table` containing the a unique `rowID` column which
#'   is used to key assays, as well as additional row metadata to subset on.
#' @param rowIDs `character`, `integer` A vector specifying
#'   the names or integer indexes of the row data identifier columns. These
#'   columns will be pasted together to make up the rownames of the
#'   `TreatmentResponseExperiment` object.
#' @param colData `data.table`, `data.frame`, `matrix` A table like object
#'   coercible to a `data.table` containing the a unique `colID` column which
#'   is used to key assays, as well as additional column metadata to subset on.
#' @param colIDs `character`, `integer` A vector specifying
#'   the names or integer indexes of the column data identifier columns. These
#'   columns will be pasted together to make up the colnames of the
#'   `TreatmentResponseExperiment` object.
#' @param assayIDs `list` A list of `character` vectors specifying the columns
#'   needed to uniquely identify each row in an `assay`. Names must match the
#'   `assays` list.
#' @param assays A `list` containing one or more objects coercible to a
#'   `data.table`, and keyed by rowIDs and colIDs corresponding to the rowID and
#'   colID columns in colData and rowData.
#' @param metadata A `list` of metadata associated with the
#'   `TreatmentResponseExperiment` object being constructed
#' @param keep.rownames `logical`, `character`
#'   Logical: whether rownames should be added as a column if coercing to a
#'   `data.table`, default is FALSE. If TRUE, rownames are added to the column
#'   'rn'.
#'   Character: specify a custom column name to store the rownames in.
#'
#' @return A `TreatmentResponseExperiment` object containing the data for a
#'   treatment response experiment configured according to the rowIDs and
#'   colIDs arguments.
#'
#'
#' @import data.table
#' @export
TreatmentResponseExperiment <- function(rowData, rowIDs, colData, colIDs,
        assays, assayIDs, metadata=list(), keep.rownames=FALSE) {
    if (!missing(rowData) && is(rowData, "LongTable")) {
        LT <- rowData
    } else {
        LT <- LongTable(rowData=rowData, rowIDs=rowIDs, colData=colData,
            colIDs=colIDs, assays=assays, assayIDs=assayIDs, metadata=metadata,
            keep.rownames=keep.rownames)
    }
    .TreatmentResponseExperiment(
        rowData=LT@rowData,
        colData=LT@colData,
        assays=LT@assays,
        .intern=LT@.intern,
        metadata=LT@metadata
    )
}

#' @name as
#'
#' @title
#' Coerce a `LongTable` to a `TreatmentResponseExperiment`
#'
#' @param from `LongTable` object to coerce to a `TreatmentResponseExperiment`.
#'
#' @return The data in `object`, as the child-class
#'   `TreatmentResponseExperiment`.
#'
#' @seealso [`TreatmentResponseExperiment`]
#'
#' @examples
#' data(clevelandSmall_cSet)
#' TRE <- as(treatmentResponse(clevelandSmall_cSet),
#'     "TreatmentResponseExperiment")
#' TRE
#'
#' @md
#' @export
setAs("LongTable", "TreatmentResponseExperiment", function(from) {
    TreatmentResponseExperiment(from)
})

#' Cleaveland_mut RadioSet subsetted and cast as CoreSet
#'
#' This dataset is just a dummy object derived from the Cleveland_mut RadioSet
#'   in the RadioGx R package. It's contents should not be interpreted and it
#'   is only present to test the functions in this package and provide
#'   examples
#' 
#' @references
#' Lamb et al. The Connectivity Map: using gene-expression signatures to connect 
#' small molecules, genes, and disease. Science, 2006.
#' 
#' @docType data
#' @name clevelandSmall_cSet
#' @usage data(clevelandSmall_cSet)
#' @keywords datasets
#' @format CoreSet object
NULL

#' Merck Drug Combination Data LongTable
#'
#' This is a LongTable object created from some drug combination data provided
#'   to our lab by Merck.
#'
#' @references
#' TODO:: Include a reference
#'
#' @docType data
#' @name merckLongTable
#' @usage data(merckLongTable)
#' @keywords datasets
#' @format LongTable object
NULL

#' Example LongTableDataMapper
#'
#' A dummy LongTableDataMapper object to be used in package examples.
#'
#' @docType data
#' @name exampleDataMapper
#' @usage data(exampleDataMapper)
#' @keywords datasets
#' @format LongTableDataMapper object
NULL

#' NCI-ALMANAC Drug Combination Data TreatmentResponseExperiment Subset
#'
#' This is a `TreatmentResponseExperiment` object containing a subset of
#'   NCI-ALMANAC drug combination screening data,
#'   with 2347 unique treatment combinations on 10 cancer cell lines selected.
#'
#' @references
#' Susan L. Holbeck, Richard Camalier, James A. Crowell, Jeevan Prasaad Govindharajulu, Melinda Hollingshead, Lawrence W. Anderson, Eric Polley, Larry Rubinstein, Apurva Srivastava, Deborah Wilsker, Jerry M. Collins, James H. Doroshow; The National Cancer Institute ALMANAC: A Comprehensive Screening Resource for the Detection of Anticancer Drug Pairs with Enhanced Therapeutic Activity. Cancer Res 1 July 2017; 77 (13): 3564–3576. https://doi.org/10.1158/0008-5472.CAN-17-0489
#'
#' @docType data
#' @name nci_TRE_small
#' @usage data(nci_TRE_small)
#' @keywords datasets
#' @format TreatmentResponseExperiment object
NULL

# ======================================================
# Utilty functions for message, warnings, errors and cat
# ------------------------------------------------------

#' @title .formatMessage
#'
#' @description Format one or more strings to fit nicely displayed in the
#' R console at any given width.
#'
#' @param ... One or more `character` vectors containing the strings to be
#' formatted
#' @param collapse `character(1)` A string of characters to collapse vectors
#' inside `...` with.
#'
#' @md
#' @keywords internal
#' @export
#' @noRd
.formatMessage <- function(..., collapse=', ') {
    paste0(strwrap(paste0(..., collapse=collapse)), collapse='\n')
}

#' @title .message
#'
#' @description Alternative to message which respects the local package
#' settings for verbosity in `options()`. The message is displayed if either
#' the general R option 'verbose' is TRUE, or if '<packageName()>.verbose'
#' is TRUE.
#'
#' @details
#' Defaults for package verbosity are determined in zzz.R via the .onAttach
#' function. When loading the package, if the session is interactive the
#' default verbosity is TRUE, otherwise it is FALSE.
#'
#' @md
#' @importFrom crayon blue
#' @keywords internal
#' @export
#' @noRd
.message <- function(...) {
    optionName <- paste0(packageName(), '.verbose')
    optionIsTRUE <- !is.null(getOption(optionName)) && getOption(optionName)
    verboseIsTRUE <- getOption('verbose')
    if (optionIsTRUE || verboseIsTRUE)
        message(crayon::blue$bold(.formatMessage(...)))
}

#' @title .warning
#'
#' @description Alternative to message which respects the local package
#' settings for verbosity in `options()`. The message is displayed if either
#' the general R option 'verbose' is TRUE, or if '<packageName()>.verbose'
#' is TRUE.
#'
#' @details
#' Defaults for package verbosity are determined in zzz.R via the .onAttach
#' function. When loading the package, if the session is interactive the
#' default verbosity is TRUE, otherwise it is FALSE.
#'
#' @md
#' @importFrom crayon cyan
#' @keywords internal
#' @export
#' @noRd
.warning <- function(...) {
    warning(crayon::cyan$bold(.formatMessage(...)), call.=FALSE)
}

#' @title .error
#'
#' @description Alternative to error which formats the error to fit the
#' console and prints it in magenta.
#'
#' @details
#' Defaults for package verbosity are determined in zzz.R via the .onAttach
#' function. When loading the package, if the session is interactive the
#' default verbosity is TRUE, otherwise it is FALSE.
#'
#' @md
#' @importFrom crayon magenta
#' @keywords internal
#' @export
#' @noRd
.error <- function(...) {
    stop(crayon::magenta$bold(.formatMessage(...)), call.=FALSE)
}

#' @title .funContext
#'
#' @description Build a string with the package and function name for a current
#'   function. Prepended to error message to make it easier to determine where
#'   the error or warning came from.
#'
#' @param funName `character(1)` A string with the function name, prepended
#'   with the correct connection to the package NAMESPACE. For exported functions
#'   use '::', for non-exported functions use ':::'.
#'
#' @keywords internal
#' @export
#' @noRd
.funContext <- function(funName) paste0('[', packageName(), funName, ']\n')


#' @title .parseToRoxygen
#'
#' @description Takes a string block of roxygen2 tags sepearated by new-line
#'   characteres and parses it to the appropriate format for the @eval tag,
#'   subtituting any string in { } for the argument of the same name in `...`.
#'
#' @keywords internal
#' @export
#' @noRd
.parseToRoxygen <- function(string, ...) {
    unlist(strsplit(
        with(list(...), glue::glue(string)),
    '\n'))
}

#' @keywords internal
.paste_ <- function(x, y) paste(x, y, sep='_')

#' @keywords internal
.paste_colon <- function(x, y) paste(x, y, sep=':')

#' @keywords interanl
.paste_slashes <- function(...) paste(..., sep='///')

# Navigating this file:
# - Slot section names start with ----
# - Method section names start with ==
#
# As a result, you can use Ctrl + f to find the slot or method you are looking
# for quickly, assuming you know its name.
#
# For example Ctrl + f '== molecularProfiles' would take you the molecularProfiles
# method, while Ctrl + f '---- molecularProfiles' would take you to the slot
# section.

#' @include CoreSet-class.R allGenerics.R LongTable-class.R
NULL

.local_class <- 'CoreSet'
.local_data <- 'clevelandSmall_cSet'

#### CoreGx dynamic documentation
####
#### Warning: for dynamic docs to work, you must set
#### Roxygen: list(markdown = TRUE, r6=FALSE)
#### in the DESCRPTION file!


# =======================================
# Accessor Method Documentation Object
# ---------------------------------------

#' @noRd
.docs_CoreSet_accessors <- function(...) .parseToRoxygen(
    "
    @title Accessing and modifying information in a `{class_}`

    @description
    Documentation for the various setters and getters which allow manipulation
    of data in the slots of a `{class_}` object.

    @param object A `{class_}` object.
    @param value See details.
    @param mDataType `character(1)` The name of a molecular datatype to access
    from the `molecularProfiles` of a `{class_}` object.
    @param assay `character(1)` A valid assay name in the `SummarizedExperiment`
    of `@molecularProfiles` of a {class_} object for data type `mDataType`.
    @param dimension See details.
    @param ... See details.

    @return Accessors: See details.
    @return Setters: An updated `{class_}` object, returned invisibly.
    ",
    ...
)

#' @name CoreSet-accessors
#' @eval .docs_CoreSet_accessors(class_='CoreSet')
#' @eval .parseToRoxygen("@examples data({data_})", data_=.local_data)
NULL


# ======================================
# Accessor Methods
# --------------------------------------


## ====================
## ---- annotation slot
## --------------------


##
## == annotation


#' @noRd
.docs_CoreSet_get_annotation <- function(...) .parseToRoxygen(
    "
    @details

    ## @annotation

    __annotation__: A `list` of {class_} annotations with items: 'name',
    the name of the object; 'dateCreated', date the object was created; 'sessionInfo',
    the `sessionInfo()` when the object was created; 'call', the R constructor call;
    and 'version', the object version.

    @examples

    ## @annotation

    annotation({data_})

    @md
    @importMethodsFrom BiocGenerics annotation
    @aliases annotation
    @exportMethod annotation
    ",
    ...
)

#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_get_annotation(class_=.local_class, data_=.local_data)
setMethod('annotation', signature("CoreSet"), function(object) {
    object@annotation
})


#' @noRd
.docs_CoreSet_set_annotation <- function(...) .parseToRoxygen(
    "
    @details
    __annotation<-__: Setter method for the annotation slot. Arguments:
    - value: a `list` of annotations to update the {class_} with.

    @examples
    annotation({data_}) <- annotation({data_})

    @md
    @importMethodsFrom BiocGenerics annotation<-
    @aliases annotation<-
    @exportMethod annotation<-
    ",
    ...
)

#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_set_annotation(class_=.local_class, data_=.local_data)
setReplaceMethod("annotation", signature("CoreSet", "list"),
    function(object, value)
{
    object@annotation <- value
    object
})


##
## == dateCreated


#' @export
setGeneric("dateCreated", function(object, ...) standardGeneric("dateCreated"))

.docs_CoreSet_get_dateCreated <- function(...) .parseToRoxygen(
    "
    @details
    ## @dateCreated
    __dateCreated__: `character(1)` The date the `{class_}` object was
    created, as returned by the `date()` function.
    @examples
    dateCreated({data_})

    @md
    @aliases dateCreated,{class_}-method dateCreated
    @exportMethod dateCreated
    ",
    ...
)

#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_get_dateCreated(class_=.local_class, data_=.local_data)
setMethod('dateCreated', signature("CoreSet"), function(object) {
    object@annotation$dateCreated
})


#' @export
setGeneric("dateCreated<-", function(object, ..., value)
    standardGeneric("dateCreated<-"))

.docs_CoreSet_set_dateCreated <- function(...) .parseToRoxygen(
    "
    @details
    __dateCreated<-__: Update the 'dateCreated' item in the `annotation` slot of
    a `{class_}` object. Arguments:
    - value: A `character(1)` vector, as returned by the `date()` function.
    @examples
    ## dateCreated
    dateCreated({data_}) <- date()

    @md
    @aliases dateCreated<-,{class_}-method dateCreated<-
    @exportMethod dateCreated<-
    ",
    ...
)

#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_set_dateCreated(class_=.local_class, data_=.local_data)
setReplaceMethod('dateCreated', signature(object="CoreSet", value="character"),
    function(object, value)
{
    ## TODO:: Error handling - do we ever want to allow a datetime object?
    funContext <- .funContext('dateCreated')
    if (length(value) > 1) .error(funContext, 'dateCreated must by a character
        vector of length 1, as returned by the `date()` function.')
    object@annotation$dateCreated <- value
    return(object)
})


##
## == name


#' @export
setGeneric("name", function(object, ...) standardGeneric("name"))

.docs_CoreSet_get_name <- function(...) .parseToRoxygen(
    "
    @details
    __name__: `character(1)` The name of the `{class_}`, retreived from
    the `@annotation` slot.

    @examples
    name({data_})

    @md
    @aliases name,{class_}-method name
    @exportMethod name
    ",
    ...
)

#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_get_name(class_=.local_class, data_=.local_data)
setMethod('name', signature("CoreSet"), function(object) {
    return(object@annotation$name)
})

#' @export
setGeneric("name<-", function(object, ..., value) standardGeneric("name<-"))

#' @noRd
.docs_CoreSet_set_name <- function(...) .parseToRoxygen(
    "
    @details
    __name<-__: Update the `@annotation$name` value in a `{class_}`  object.
    - value: `character(1)` The name of the `{class_}` object.

    @examples
    name({data_}) <- 'new_name'

    @md
    @aliases name<-,{class_},character-method name<-
    @exportMethod name<-
    ",
    ...
)

#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_set_name(class_=.local_class, data_=.local_data)
setReplaceMethod('name', signature("CoreSet"), function(object, value) {
    object@annotation$name <- value
    return(object)
})


## ==============
## ---- sample slot


##
## == sampleInfo


#' @export
setGeneric("sampleInfo", function(object, ...) standardGeneric("sampleInfo"))

#' @noRd
.docs_CoreSet_get_sampleInfo <- function(...) .parseToRoxygen(
    "
    ## @sample
    @details
    __{sample_}Info__: `data.frame` Metadata for all sample in a `{class_}` object.

    @md
    @aliases
    sampleInfo,{class_}-method sampleInfo
    {sample_}Info,{class_}-method
    {sample_}Info
    @exportMethod sampleInfo
    ",
    ...
)


.local_sample <- "cell"

#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_get_sampleInfo(class_=.local_class, sample_=.local_sample)
setMethod("sampleInfo", "CoreSet", function(object) {
    object@sample
})
#' @export
cellInfo <- function(...) sampleInfo(...)

#' @export
setGeneric("sampleInfo<-", function(object, value) standardGeneric("sampleInfo<-"))

#' @noRd
.docs_CoreSet_set_sampleInfo <- function(...) .parseToRoxygen(
    "
    @details
    __sampleInfo<-__: assign updated sample annotations to the `{class_}`
    object.
    Arguments:
    - value: a `data.frame` object.
    @examples
    sampleInfo({data_}) <- sampleInfo({data_})

    @md
    @aliases
    sampleInfo<-,{class_},data.frame-method
    sampleInfo<-
    {sample_}Info<-,{class_},data.frame-method
    {sample_}Info<-
    @exportMethod sampleInfo<-
    ",
    ...
)


#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_set_sampleInfo(class_=.local_class, data_=.local_data,
#'  sample_=.local_sample)
setReplaceMethod("sampleInfo", signature(object="CoreSet", value="data.frame"),
        function(object, value) {
    funContext <- .funContext('::sampleInfo')
    if (is.null(rownames(value)))
    .error(funContext, "Please provide the sampleid as rownames for the sample
        annotations")
    object@sample <- value
    object
})
#' @export
`cellInfo<-` <- function(object, value) `sampleInfo<-`(object, value=value)


##
## == sampleNames

## TODO: Implement an actual @sample slot instead of using @sample  and aliases



#' @noRd
.docs_CoreSet_get_sampleNames <- function(...) .parseToRoxygen(
    "
    @details
    __sampleNames__: `character` Retrieve the rownames of the `data.frame` in
    the `sample` slot from a {class_} object.
    @examples
    sampleNames({data_})

    @md
    @aliases
    sampleName,{class_}-method
    sampleNames
    {sample_}Name,{class_}-method
    {sample_}Names
    @exportMethod sampleNames
    ",
    ...
)

#' @importMethodsFrom Biobase sampleNames
#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_get_sampleNames(class_=.local_class, data_=.local_data,
#' sample_=.local_sample)
setMethod("sampleNames", signature("CoreSet"), function(object) {
    rownames(sampleInfo(object))
})
#' @export
cellNames <- function(object) sampleNames(object)


#' @noRd
.docs_CoreSet_set_sampleNames <- function(...) .parseToRoxygen(
    "
    @details
    __sampleNames<-__: assign new rownames to the sampleInfo `data.frame` for
    a {class_} object.
    Arguments:
    - value: `character` vector of rownames for the `sampleInfo(object)` `data.frame`.
    @examples
    sampleNames({data_}) <- sampleNames({data_})

    @md
    @aliases
    sampleNames<-,{class_},list-method
    sampleNames<-
    {sample_}Names<-,{class_},list-method
    {sample_}Names<-
    @exportMethod sampleNames<-
    ",
    ...
)


#' @importMethodsFrom Biobase sampleNames<-
#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_set_sampleNames(class_=.local_class, data_=.local_data,
#' sample_=.local_sample)
setReplaceMethod("sampleNames", signature(object="CoreSet", value="character"),
        function(object, value) {
    ## TODO: does updateSampleId also update slots other than sample?
    object <- updateSampleId(object, value)
    return(object)
})
#' @export
`cellNames<-` <- function(object, value) `sampleNames<-`(object, value=value)


## -------------------
## ---- treatment slot

## TODO: Implement an actual @treatment slot to replace @treatment and @radiation

#
# == treatmentInfo

#' @export
setGeneric('treatmentInfo', function(object, ...)
    standardGeneric('treatmentInfo'))

#' @noRd
.docs_CoreSet_get_treatmentInfo <- function(...) .parseToRoxygen(
    "
    @details
    __treatmentInfo__: `data.frame` Metadata for all treatments in a `{class_}`
    object. Arguments:
    - object: `{class_}` An object to retrieve treatment metadata from.

    @examples
    treatmentInfo({data_})

    @md
    @aliases treatmentInfo,{class_}-method treatmentInfo
    @exportMethod treatmentInfo
    ",
    ...
)

#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_get_treatmentInfo(class_=.local_class, data_=.local_data)
setMethod('treatmentInfo', signature('CoreSet'), function(object) {
    treatmentType <- switch(class(object)[1],
        'PharmacoSet'='treatment',
        'ToxicoSet'='treatment',
        'RadioSet'='radiation',
        'CoreSet'='treatment'
    )
    package <- gsub('Set', 'Gx', class(object)[1])
    if ("treatment" %in% slotNames(object)) return(object@treatment)
    treatmentInfo <- get(paste0(treatmentType, 'Info'),
        envir=asNamespace(package))
    treatmentInfo(object)
})

#' @export
setGeneric('treatmentInfo<-', function(object, ..., value)
    standardGeneric('treatmentInfo<-'))

#' @noRd
.docs_CoreSet_set_treatmentInfo <- function(...) .parseToRoxygen(
    "
    @details
    __treatmentInfo<-__: `{class_}` object with updated treatment metadata.
    object. Arguments:
    - object: `{class_}` An object to set treatment metadata for.
    - value: `data.frame` A new table of treatment metadata for `object`.

    @examples
    treatmentInfo({data_}) <- treatmentInfo({data_})

    @md
    @aliases treatmentInfo<-,{class_},data.frame-method treatmentInfo<-
    @exportMethod treatmentInfo<-
    ",
    ...
)

#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_set_treatmentInfo(class_=.local_class, data_=.local_data)
setReplaceMethod('treatmentInfo', signature(object='CoreSet',
        value='data.frame'), function(object, value) {
    object@treatment <- value
    return(invisible(object))
})

##
## == treatmentNames


#' @export
setGeneric('treatmentNames', function(object, ...)
    standardGeneric('treatmentNames'))

#' @noRd
.docs_CoreSet_get_treatmentNames <- function(...) .parseToRoxygen(
    "
    @details
    __treatmentNames__: `character` Names for all treatments in a `{class_}`
    object. Arguments:
    - object: `{class_}` An object to retrieve treatment names from.

    @examples
    treatmentNames({data_})

    @md
    @aliases treatmentNames,{class_}-method treatmentNames
    @exportMethod treatmentNames
    ",
    ...
)

#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_get_treatmentNames(class_=.local_class, data_=.local_data)
setMethod('treatmentNames', signature(object='CoreSet'), function(object) {
    rownames(treatmentInfo(object))
})


#' @export
setGeneric('treatmentNames<-', function(object, ..., value)
    standardGeneric('treatmentNames<-'))

#' @noRd
.docs_CoreSet_set_treatmentNames <- function(...) .parseToRoxygen(
    "
    @details
    __treatmentNames<-__: `{class_}` Object with updates treatment names.
    object. Arguments:
    - object: `{class_}` An object to set treatment names from.
    - value: `character` A character vector of updated treatment names.

    @examples
    treatmentNames({data_}) <- treatmentNames({data_})

    @md
    @aliases treatmentNames<-,{class_},character-method treatmentNames<-
    @exportMethod treatmentNames<-
    ",
    ...
)

#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_set_treatmentNames(class_=.local_class, data_=.local_data)
setReplaceMethod('treatmentNames',
        signature(object='CoreSet', value='character'),
        function(object, value) {
    object <- updateTreatmentId(object, new.ids=value)
    return(invisible(object))
})

## ------------------
## ---- curation slot


##
## == curation


#' @export
setGeneric("curation", function(object, ...) standardGeneric("curation"))

#' @noRd
.docs_CoreSet_get_curation <- function(...) .parseToRoxygen(
    "
    @details
    ## @curation
    __curation__: A `list` of curated mappings between identifiers in the
    {class_} object and the original data publication. {details_}
    @examples
    ## curation
    curation({data_})

    @md
    @aliases curation,{class_}-method curation
    @exportMethod curation
    ",
    ...
)

#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_get_curation(class_=.local_class, data_=.local_data,
#' details_="Contains two `data.frame`s, 'sample' with sample ids and
#' 'tissue' with tissue ids.")
setMethod('curation', signature(object="CoreSet"), function(object) {
    object@curation
})

#' @export
setGeneric("curation<-", function(object, ..., value)
    sstandardGeneric("curation<-"))

#' @noRd
.docs_CoreSet_set_curation <- function(...) .parseToRoxygen(
    "
    @details
    __curation<-__: Update the `curation` slot of a {class_} object. Arugments:
    - value: A `list` of `data.frame`s, one for each type of curated
    identifier. {details_}
    @examples
    curation({data_}) <- curation({data_})

    @md
    @aliases curation<-,{class_},list-method curation<-
    @exportMethod curation<-
    ",
    ...
)

#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_set_curation(class_=.local_class, data_=.local_data,
#' details_="For a `CoreSet` object the slot should contain tissue and
#' sample id `data.frame`s.")
setReplaceMethod("curation", signature(object="CoreSet", value="list"),
    function(object, value)
{
    object@curation <- value
    object
})



## ----------------------
## ---- datasetType slot


#
# == datasetType


#' @export
setGeneric("datasetType", function(object, ...) standardGeneric("datasetType"))

#' @noRd
.docs_CoreSet_get_datasetType <- function(...) .parseToRoxygen(
    "
    @details
    ## datasetType slot
    __datasetType__: `character(1)` The type treatment response in the
    `sensitivity` slot. Valid values are 'sensitivity', 'perturbation' or 'both'.
    @examples
    datasetType({data_})

    @md
    @aliases datasetType,{class_}-method datasetType
    @exportMethod datasetType
    ",
    ...
)

#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_get_datasetType(class_=.local_class, data_=.local_data)
setMethod("datasetType", signature("CoreSet"), function(object) {
    object@datasetType
})


#' @export
setGeneric("datasetType<-",  function(object, value)
    standardGeneric("datasetType<-"))

#' @noRd
.docs_CoreSet_set_datasetType <- function(...) .parseToRoxygen(
    "
    @details
    __datasetType<-__: Update the datasetType slot of a {class_} object.
    Arguments:
    - value: A `character(1)` vector with one of 'sensitivity', 'perturbation'
    or 'both'
    @examples
    datasetType({data_}) <- 'both'

    @md
    @aliases datasetType<-,{class_},character-method datasetType<-
    @export
    ",
    ...
)

#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_set_datasetType(class_=.local_class, data_=.local_data)
setReplaceMethod("datasetType", signature(object="CoreSet", value='character'),
    function(object, value)
{
    funContext <- .funContext('::datasetType,CoreSet,character-method')
    if (length(value) > 1) .error(funContext,
        'datasetType must be a character vector of length 1.')
    if (!is.element(value, c('sensitivity', 'perturbation', 'both')))
        .error(funContext, 'datasetType must be one of "sensitivity",
            "perturbation" or "both".')
    object@datasetType <- value
    object
})



## ---------------------------
## ---- molecularProfiles slot


##
## == molecularProfiles


#' @export
setGeneric("molecularProfiles", function(object, mDataType, assay, ...)
    standardGeneric("molecularProfiles"))

#' @noRd
.docs_CoreSet_get_molecularProfiles <- function(...) .parseToRoxygen(
    "
    @details
    ## @molecularProfiles
    __molecularProfiles__: `matrix()` Retrieve an assay in a
    `SummarizedExperiment` from the `molecularProfiles` slot of a `{class_}`
    object with the specified `mDataType`. Valid `mDataType` arguments can be
    found with `mDataNames(object)`. Exclude `mDataType` and `assay` to
    access the entire slot. Arguments:
    - assay: Optional `character(1)` vector specifying an assay in the
    `SummarizedExperiment` of the `molecularProfiles` slot of the
    `{class_}` object for the specified `mDataType`. If excluded,
    defaults to modifying the first assay in the `SummarizedExperiment` for
    the given `mDataType`.

    @md
    @aliases molecularProfiles,{class_}-method molecularProfiles
    @importClassesFrom S4Vectors DataFrame List
    @importFrom S4Vectors DataFrame
    @importFrom SummarizedExperiment colData assay assayNames
    @exportMethod molecularProfiles
    ",
    ...
)

#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_get_molecularProfiles(class_=.local_class, data_=.local_data)
setMethod(molecularProfiles, "CoreSet", function(object, mDataType, assay) {
    funContext <- .funContext(paste0('::molecularProlfiles,', class(object), '-method'))
    if (missing(mDataType) && missing(assay)) return(object@molecularProfiles)
    if (mDataType %in% names(object@molecularProfiles)) {
        if (!missing(assay)) {
            if (assay %in% assayNames(object@molecularProfiles[[mDataType]])) {
                return(SummarizedExperiment::assay(object@molecularProfiles[[mDataType]], assay))
            } else {
                .error(funContext, (paste('Assay', assay, 'not found in the SummarizedExperiment object!')))
            }
        } else {
            return(SummarizedExperiment::assay(object@molecularProfiles[[mDataType]], 1))
        }
    } else {
        stop(paste0('mDataType ', mDataType, ' not found the object!'))
    }
})

#' @export
setGeneric("molecularProfiles<-", function(object, mDataType, assay, value)
    standardGeneric("molecularProfiles<-"))

#' @noRd
.docs_CoreSet_set_molecularProfiles <- function(...) .parseToRoxygen(
    "
    @details
    __molecularProfiles<-__: Update an assay in a `SummarizedExperiment` from
    the `molecularProfiles` slot of a {class_} object with the specified
    `mDataType`. Valid `mDataType` arguments can be found with
    `mDataNames(object)`. Omit `mDataType` and `assay` to update the slot.
    - assay: Optional `character(1)` vector specifying an assay in the
    `SummarizedExperiment` of the `molecularProfiles` slot of the
    `{class_}` object for the specified `mDataType`. If excluded,
    defaults to modifying the first assay in the `SummarizedExperiment` for
    the given `mDataType`.
    - value: A `matrix` of values to assign to the `assay` slot of the
    `SummarizedExperiment` for the selected `mDataType`. The rownames and
    column names must match the associated `SummarizedExperiment`.
    @examples
    # No assay specified
    molecularProfiles({data_}, 'rna') <- molecularProfiles({data_}, 'rna')

    # Specific assay
    molecularProfiles({data_}, 'rna', 'exprs') <-
        molecularProfiles({data_}, 'rna', 'exprs')

    # Replace the whole slot
    molecularProfiles({data_}) <- molecularProfiles({data_})

    @md
    @aliases molecularProfiles<-,{class_},character,character,matrix-method
    molecularProfiles<-,{class_},character,missing,matrix-method
    molecularProfiles<-,{class_},missing,missing,list-method
    molecularProfiles<-,{class_},missing,missing,MutliAssayExperiment-method
    molecularProfiles<-
    @importFrom SummarizedExperiment assay
    @exportMethod molecularProfiles<-
    ",
    ...
)

#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_set_molecularProfiles(class_=.local_class, data_=.local_data)
setReplaceMethod("molecularProfiles", signature(object="CoreSet",
    mDataType ="character", assay="character", value="matrix"),
    function(object, mDataType, assay, value)
{
    if (mDataType %in% names(object@molecularProfiles)) {
        assay(object@molecularProfiles[[mDataType]], assay) <- value
    }
    object
})
#' @rdname CoreSet-accessors
setReplaceMethod("molecularProfiles",
    signature(object="CoreSet", mDataType ="character", assay="missing",
        value="matrix"), function(object, mDataType, assay, value)
{
    if (mDataType %in% names(object@molecularProfiles)) {
        assay(object@molecularProfiles[[mDataType]], 1) <- value
    }
    object
})
#' @rdname CoreSet-accessors
setReplaceMethod("molecularProfiles", signature(object="CoreSet",
        mDataType="missing", assay="missing", value="list_OR_MAE"),
        function(object, mDataType, assay, value) {
    object@molecularProfiles <- value
    object
})


##
## == featureInfo


#' @export
setGeneric("featureInfo", function(object, mDataType, ...)
    standardGeneric("featureInfo"))

#' @noRd
.docs_CoreSet_get_featureInfo <- function(...) .parseToRoxygen(
    "
    @details
    __featureInfo__: Retrieve a `DataFrame` of feature metadata for the specified
    `mDataType` from the `molecularProfiles` slot of a `{class_}` object. More
    specifically, retrieve the `@rowData` slot from the `SummarizedExperiment`
    from the `@molecularProfiles` of a `{class_}` object with the name
    `mDataType`.
    @examples
    featureInfo({data_}, 'rna')

    @md
    @aliases featureInfo,{class_}-method featureInfo
    @importFrom SummarizedExperiment rowData rowData<-
    @exportMethod featureInfo
    ",
    ...
)


## FIXME: Why return NULL and not throw and error instead? Or at least a warning.
#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_get_featureInfo(class_=.local_class, data_=.local_data)
setMethod(featureInfo, "CoreSet", function(object, mDataType) {
    if (mDataType %in% names(object@molecularProfiles)) {
        return(rowData(object@molecularProfiles[[mDataType]]))
    } else{
        return(NULL)
    }
})

#' @export
setGeneric("featureInfo<-", function(object, mDataType, value)
    standardGeneric("featureInfo<-"))

#' @noRd
.docs_CoreSet_set_featureInfo <- function(...) .parseToRoxygen(
    "
    @details
    __featureInfo<-__: Update the `featureInfo(object, mDataType)` `DataFrame`
    with new feature metadata. Arguments:
    - value: A `data.frame` or `DataFrame` with updated feature metadata for
    the specified molecular profile in the `molecularProfiles` slot of a
    `{class_}` object.
    @examples
    featureInfo({data_}, '{mDataType_}') <- featureInfo({data_}, '{mDataType_}')

    @aliases featureInfo<-,{class_},character,data.frame-method
    featureInfo<-,{class_},character,DataFrame-method featureInfo<-
    @importFrom SummarizedExperiment rowData rowData<-
    @importFrom S4Vectors DataFrame
    @exportMethod featureInfo<-
    ",
    ...
)

#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_set_featureInfo(class_=.local_class, data_=.local_data,
#'   mDataType_='rna')
setReplaceMethod("featureInfo", signature(object="CoreSet",
    mDataType ="character",value="data.frame"),
    function(object, mDataType, value)
{
    if (mDataType %in% names(object@molecularProfiles)) {
        rowData(object@molecularProfiles[[mDataType]]) <-
            DataFrame(value, rownames = rownames(value))
    }
    object
})
setReplaceMethod("featureInfo", signature(object="CoreSet",
    mDataType ="character",value="DataFrame"),
    function(object, mDataType, value)
{
    if (mDataType %in% names(object@molecularProfiles)) {
        rowData(object@molecularProfiles[[mDataType]]) <-
            DataFrame(value, rownames = rownames(value))
    }
    object
})


##
## == phenoInfo


#' @export
setGeneric("phenoInfo", function(object, mDataType, ...)
    standardGeneric("phenoInfo"))

#' @noRd
.docs_CoreSet_get_phenoInfo <- function(...) .parseToRoxygen(
    "
    @details
    __phenoInfo__: Return the `@colData` slot from the `SummarizedExperiment` of
    `mDataType`, containing sample-level metadata, from a `{class_}` object.

    @examples
    phenoInfo({data_}, '{mDataType_}')

    @md
    @importFrom SummarizedExperiment colData
    @aliases phenoInfo
    @exportMethod phenoInfo
    ",
    ...
)

#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_get_phenoInfo(class_=.local_class, data_=.local_data, mDataType_='rna')
setMethod(phenoInfo, signature(object='CoreSet', mDataType='character'),
    function(object, mDataType)
{
    if (mDataType %in% mDataNames(object)) { # Columns = Samples
        return(colData(object@molecularProfiles[[mDataType]]))
    }else{
        ## FIXME:: Is there a reason we throw a NULL instead of an error?
        return(NULL)
    }
})

#' @export
setGeneric("phenoInfo<-", function(object, mDataType, value)
    standardGeneric("phenoInfo<-"))

#' @noRd
.docs_CoreSet_set_phenoInfo <- function(...) .parseToRoxygen(
    "
    @details
    __phenoInfo<-__: Update the `@colData` slot of the `SummarizedExperiment`
    of `mDataType` in the `@molecularProfiles` slot of a `{class_}` object.
    This updates the sample-level metadata in-place.
    - value: A `data.frame` or `DataFrame` object where rows are samples
    and columns are sample metadata.

    @examples
    phenoInfo({data_}, '{mDataType_}') <- phenoInfo({data_}, '{mDataType_}')

    @md
    @importFrom SummarizedExperiment colData colData<-
    @importFrom S4Vectors DataFrame
    @aliases phenoInfo<-,{class_},character,data.frame-method
    phenoInfo<-,{class_},character,DataFrame-method phenoInfo<-
    @exportMethod phenoInfo<-
    ",
    ...
)

#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_set_phenoInfo(class_=.local_class, data_=.local_data, mDataType_='rna')
setReplaceMethod("phenoInfo", signature(object="CoreSet", mDataType ="character",
    value="data.frame"), function(object, mDataType, value)
{
    if(mDataType %in% mDataNames(object)) {
        colData(object@molecularProfiles[[mDataType]]) <-
            DataFrame(value, rownames = rownames(value))
    }
    object
})
setReplaceMethod("phenoInfo", signature(object="CoreSet",
    mDataType ="character", value="DataFrame"),
    function(object, mDataType, value)
{
    if (mDataType %in% mDataNames(object)) {
        colData(object@molecularProfiles[[mDataType]]) <- value
    }
    object
})


##
## == fNames

#' @export
setGeneric('fNames', function(object, mDataType, ...) standardGeneric('fNames'))

#' @noRd
.docs_CoreSet_get_fNames <- function(...) .parseToRoxygen(
    "
    @details
    __fNames__: `character()` The features names from the `rowData` slot of a
    `SummarizedExperiment` of `mDataType` within a `{class_}` object.

    @examples
    fNames({data_}, '{mDataType_}')

    @md
    @aliases fNames,{class_},character-method fNames
    @exportMethod fNames
    ",
    ...
)

#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_get_fNames(class_=.local_class, data_=.local_data,
#' mDataType_='rna')
setMethod('fNames', signature(object='CoreSet', mDataType='character'),
    function(object, mDataType)
{
    rownames(featureInfo(object, mDataType))
})


#' @export
setGeneric('fNames<-', function(object, mDataType, ..., value)
    standardGeneric('fNames<-'))

#' @noRd
.docs_CoreSet_set_fNames <- function(...) .parseToRoxygen(
    "
    @details
    __fNames__: Updates the rownames of the feature metadata (i.e., `rowData`)
    for a `SummarizedExperiment` of `mDataType` within a `{class_}` object.
    - value: `character()` A character vector of new features names for the
    `rowData` of the `SummarizedExperiment` of `mDataType` in the
    `@molecularProfiles` slot of a `{class_}` object. Must be the same
    length as `nrow(featureInfo(object, mDataType))`,
    the number of rows in the feature metadata.

    @examples
    fNames({data_}, '{mDataType_}') <- fNames({data_}, '{mDataType_}')

    @md
    @aliases fNames<-,{class_},character,character-method fNames<-
    @exportMethod fNames<-
    ",
    ...
)

#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_set_fNames(class_=.local_class, data_=.local_data,
#' mDataType_='rna')
setReplaceMethod('fNames', signature(object='CoreSet', mDataType='character',
    value='character'), function(object, mDataType, value)
{
    rownames(featureInfo(object, mDataType)) <- value
    object
})


##
## == mDataNames


#' @export
setGeneric("mDataNames", function(object, ...) standardGeneric("mDataNames"))

#' @noRd
.docs_CoreSet_get_mDataNames <- function(...) .parseToRoxygen(
    "
    @details
    __mDataNames__: `character` Retrieve the names of the molecular data types
    available in the `molecularProfiles` slot of a `{class_}` object. These
    are the options which can be used in the `mDataType` parameter of various
    `molecularProfiles` slot accessors methods.
    @examples
    mDataNames({data_})

    @md
    @aliases mDataNames,{class_}-method mDataNames
    @exportMethod mDataNames
    ",
    ...
)

#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_get_mDataNames(class_=.local_class, data_=.local_data)
setMethod("mDataNames", "CoreSet", function(object) {
    return(names(object@molecularProfiles))
})


#' @export
setGeneric("mDataNames<-", function(object, ..., value) standardGeneric("mDataNames<-"))

#' @noRd
.docs_CoreSet_set_mDataNames <- function(...) .parseToRoxygen(
    "
    @details
    __mDataNames__: Update the molecular data type names of the
    `molecularProfiles` slot of a {class_} object. Arguments:
    - value: `character` vector of molecular datatype names, with length
    equal to `length(molecularProfilesSlot(object))`.
    @examples
    mDataNames({data_}) <- mDataNames({data_})

    @md
    @aliases mDataNames<-,{class_},ANY-method mDataNames<-
    @exportMethod mDataNames<-
    ",
    ...
)

#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_set_mDataNames(class_=.local_class, data_=.local_data)
setReplaceMethod("mDataNames", "CoreSet", function(object, value) {
    names(object@molecularProfiles) <- value
    return(object)
})

##
## == molecularProfilesSlot


#' @export
setGeneric("molecularProfilesSlot", function(object, ...)
    standardGeneric("molecularProfilesSlot"))

#' @noRd
.docs_CoreSet_get_molecularProfilesSlot <- function(...) .parseToRoxygen(
    "
    @details
    __molecularProfilesSlot__: Return the contents of the `@molecularProfiles`
    slot of a `{class_}` object. This will either be a `list` or
    `MultiAssayExperiment` of `SummarizedExperiment`s.

    @examples
    molecularProfilesSlot({data_})

    @md
    @aliases moleculerProfilesSlot,{class_}-method molecularProfilesSlot
    @exportMethod molecularProfilesSlot
    ",
    ...
)

#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_get_molecularProfilesSlot(class_=.local_class, data_=.local_data)
setMethod("molecularProfilesSlot", signature("CoreSet"), function(object) {
    object@molecularProfiles
})


#' @export
setGeneric("molecularProfilesSlot<-",
    function(object, value) standardGeneric("molecularProfilesSlot<-"))

#' @noRd
.docs_CoreSet_set_molecularProfilesSlot <- function(...) .parseToRoxygen(
    "
    @details
    __molecularProfilesSlot<-__: Update the contents of the `@molecularProfiles`
    slot of a `{class_}` object. Arguemnts:
    - value: A `list` or `MultiAssayExperiment` of `SummarizedExperiment`s. The
    `list` and `assays` should be named for the molecular datatype in each
    `SummarizedExperiment`.

    @examples
    molecularProfilesSlot({data_}) <- molecularProfilesSlot({data_})

    @md
    @aliases molecularProfilesSlot<-,{class_},list-method
    molecularProfilesSlot<-{class_},MultiAssayExperiment-method
    molecularProfilesSlot<-
    @exportMethod molecularProfilesSlot<-
    ",
    ...
)

#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_set_molecularProfilesSlot(class_=.local_class, data_=.local_data)
setReplaceMethod("molecularProfilesSlot", signature("CoreSet", "list_OR_MAE"),
    function(object, value) {
    # funContext <- .S4MethodContext('molecularProfilesSlot<-', class(object),
    #     class(value))
    # if (!is(value, class(object@molecularProfiles)[1])) .error(funContext,
    #     'The class of value must be the same as the current @molecularProfiles!')
    object@molecularProfiles <- value
    object
})


## ---------------------
## ---- sensitivity slot


#
# == sensitivityInfo

#' @noRd
.docs_CoreSet_get_sensitivityInfo <- function(...) .parseToRoxygen(
    "
    @details

    ## @treatmentResponse

    ### Arguments:
    - `dimension`: Optional `character(1)` One of 'treatment', 'sample' or
    'assay' to retrieve `rowData`, `colData` or the 'assay_metadata' assay from
    the `{class_}` `@sensitvity` `LongTable` object, respectively. Ignored with
    warning if `@treatmentResponse` is not a `LongTable` object.
    -  `...`: Additional arguments to the `rowData` or `colData`.
    `LongTable` methods. Only used if the sensitivity slot contains a
    `LongTable` object instead of a `list` and the `dimension` argument is
    specified.

    ### Methods:

    __sensitivityInfo__: `DataFrame` or `data.frame` of sensitivity treatment combo
    by sample metadata for the `{class_}` object. When the `dimension`
    parameter is used, it allows retrieval of the dimension specific metadata
    from the `LongTable` object in `@treatmentResponse` of a {class_} object.

    @examples
    sensitivityInfo({data_})

    @md
    @aliases sensitivityInfo,{class_},missing-method
    sensitivityInfo,{class_},character-method
    @exportMethod sensitivityInfo
    ",
    ...
)

#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_get_sensitivityInfo(class_=.local_class,
#' data_=.local_data)
setMethod(sensitivityInfo, signature("CoreSet"),
        function(object, dimension, ...) {
    funContext <- .funContext('::sensitivityInfo')
    # case where sensitivity slot is a LongTable
    if (is(treatmentResponse(object), 'LongTable')) {
        if (!missing(dimension)) {
            switch(dimension,
                sample={ return(colData(treatmentResponse(object), ...)) },
                treatment={ return(rowData(treatmentResponse(object), ...)) },
                assay={ return(assay(treatmentResponse(object), 'assay_metadata')) },
                .error(funContext, 'Invalid value for the dimension argument.
                    Please select on of "sample", "treatment" or "assay'))
        } else {
            return(.rebuildInfo(treatmentResponse(object)))
        }
    # sensitivity is a list
    } else {
        if (!missing(dimension))
            .warning(funContext,' The dimension argument is only valid if the
                sensitivity slot contains a LongTable object. Ignoring the
                dimension and ... parameters.')
        return(treatmentResponse(object)$info)
    }
})


#' Replicate the $info slot in the old sensitivity list from the new LongTable
#'   object
#'
#' @param longTable `LongTable`
#'
#' @keywords internal
#' @importFrom MatrixGenerics colAlls
#' @importFrom data.table setkeyv merge.data.table `:=` setDF
#' @noRd
.rebuildInfo <- function(longTable) {

    # Extract the information needed to reconstruct the sensitivityInfo
    #   data.frame
    aidx <- which(assayNames(longTable) %in% "assay_metadata")
    if (!length(aidx)) aidx <- 1
    assayIndexDT <- assay(longTable, aidx, key=TRUE)
    if (aidx == 1) assayIndexDT <- assayIndexDT[, .(rowKey, colKey)]
    setkeyv(assayIndexDT, c('rowKey', 'colKey'))
    rowDataDT <- rowData(longTable, key=TRUE)
    setkeyv(rowDataDT, 'rowKey')
    colDataDT <- colData(longTable, key=TRUE)
    setkeyv(colDataDT, 'colKey')

    rowIDcols <- rowIDs(longTable)[!grepl('dose$', rowIDs(longTable))]
    colIDcols <- colIDs(longTable)
    rownameCols <- c(rowIDcols, colIDcols)

    # join the tables into the original data
    infoDT <- merge.data.table(assayIndexDT, rowDataDT, all.x=TRUE)
    setkeyv(infoDT, 'colKey')
    infoDT <- merge.data.table(infoDT, colDataDT, all.x=TRUE)[,
        -c('rowKey', 'colKey')
    ]
    infoDT <- tryCatch({
        infoDT[, .SD, .SDcols=!patterns('treatment.*dose$')]
    }, error=function(e) infoDT)

    # determine which columns map 1:1 with new identifiers and subset to those
    infoDT_first <- infoDT[, head(.SD, 1), by=rownameCols]
    infoDT_last <- infoDT[, tail(.SD, 1), by=rownameCols]
    keepCols <- colnames(infoDT_first)[
        colAlls(infoDT_first == infoDT_last, na.rm=TRUE)
        ]
    infoDT_sub <- unique(infoDT[, ..keepCols])

    # pad the dropped NA values, if they exists
    if ("sensitiivtyInfo_NA" %in% names(metadata(longTable))) {
            na_info <- copy(metadata(longTable)$sensitivityInfo_NA)
        setnames(na_info, "treatmentid", "treatment1id")
        na_info <- cbind(
            na_info,
            unique(infoDT_sub[, .SD, .SDcols=!patterns("^treatment1id$|^sampleid$|^replicate_id$|^rn$")])
        )
        na_info[, replicate_id := seq_len(.N), by=.(treatment1id, sampleid)]
        infoDT_sub <- rbind(infoDT_sub, na_info)
    }
    if ("experiment_metadata" %in% names(metadata(longTable))) {
        infoDT_sub <- cbind(
            infoDT_sub,
            as.data.table(metadata(longTable)$experiment_metadata)
        )
    }

    # rebuild the rownames
    idCols <- grep("^treatment[0-9]*id", colnames(infoDT_sub), value=TRUE)
    infoDT_sub[, treatmentid := Reduce(.paste_slashes, mget(..idCols))]
    infoDT_sub[, treatment_uid := Reduce(.paste_colon, mget(..rowIDcols))]
    infoDT_sub[, sample_uid := Reduce(.paste_colon, mget(..colIDcols))]
    infoDT_sub[, exp_id := Reduce(.paste_, .(treatment_uid, sample_uid))]

    # convert to data.frame
    setDF(infoDT_sub, rownames=infoDT_sub$exp_id)
    return(infoDT_sub)
}

#' @noRd
.docs_CoreSet_set_sensitivityInfo <- function(...) .parseToRoxygen(
    "
    @details

    __sensitivityInfo__<-: Update the `@treatmentResponse` slot metadata for a
    `{class_}` object. When used without the `dimension` argument is behaves
    similar to the old {class_} implementation, where the `@treatmentResponse` slot
    contained a list with a `$info` `data.frame` item. When the `dimension`
    arugment is used, more complicated assignments can occur where 'sample'
    modifies the `@sensitvity` `LongTable` colData, 'treatment' the rowData and
    'assay' the 'assay_metadata' assay.
    Arguments:
    - value: A `data.frame` of treatment response experiment metadata,
    documenting experiment level metadata (mapping to treatments and samples). If
    the `@treatmentResponse` slot doesn't contain a `LongTable` and `dimension` is
    not specified, you can only modify existing columns as returned by
    `sensitivityInfo(object)`.
    @examples
    sensitivityInfo({data_}) <- sensitivityInfo({data_})

    @md
    @aliases sensitivityInfo<-,{class_},missing,data.frame-method
    sensitvityInfo<-,{class_},character,data.frame-method
    @import data.table
    @exportMethod sensitivityInfo<-
    ",
    ...
)

#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_set_sensitivityInfo(class_=.local_class, data_=.local_data)
setReplaceMethod("sensitivityInfo", signature(object="CoreSet", value="data.frame"),
                function(object, dimension, ..., value) {

    funContext <- .funContext('::sensitivityInfo<-')
    if (is(treatmentResponse(object), 'LongTable')) {
        # coerce to data.table
        if (!is.data.table(value)) value <- data.table(value)
        if (missing(dimension)) {
            valueCols <- colnames(value)
            # get existing column names
            rowDataCols <- colnames(rowData(object@treatmentResponse))
            colDataCols <- colnames(colData(object@treatmentResponse))
            # drop any value columns that don't already exist
            hasValueCols <- valueCols %in% c(rowDataCols, colDataCols)
            if (!all(hasValueCols))
                .message(funContext, 'Dropping columns ',
                    .collapse(valueCols[!hasValueCols]), ' from value. Currently
                    this setter only allows modifying existing columns when
                    @treatmentResponse is a LongTable. For more fine grained updates
                    please use the dimension argument.')
            # update the object
            rowData(object@treatmentResponse, ...) <-
                unique(value[, .SD, .SDcols=valueCols %in% rowDataCols])
            colData(object@treatmentResponse, ...) <-
                unique(value[, .SD, .SDcols=valueCols %in% colDataCols])
        } else {
            switch(dimension,
                treatment={ rowData(object@treatmentResponse, ...) <- value },
                sample={ colData(object@treatmentResponse, ...) <- value },
                assay={ assay(object@treatmentResponse, 'assay_metadata') <- value },
                .error(funContext, 'Invalid argument to dimension parameter.
                    Please choose one of "sample", "treatment" or "assay"'))
        }
    } else {
        if (!missing(dimension))
            .warning(funContext, 'The dimension argument is only valid if the
                sensitivity slot contains a LongTable object. Ignoring dimension
                and ... parameters.')
        object@treatmentResponse$info <- value
    }
    return(object)
})


#
# == sensitvityMeasures


#' @noRd
.docs_CoreSet_get_sensitivityMeasures <- function(...) .parseToRoxygen(
    "
    @details
    __sensitivityMeaures__: Get the 'sensitivityMeasures' available in a `{class_}`
    object. Each measure reprents some summary of sample sensitivity to a given
    treatment, such as ic50, ec50, AUC, AAC, etc. The results are returned as a
    `character` vector with all available metrics for the PSet object.
    @examples
    sensitivityMeasures({data_}) <- sensitivityMeasures({data_})

    @md
    @exportMethod sensitivityMeasures
    ",
    ...
)

#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_get_sensitivityMeasures(class_=.local_class, data_=.local_data)
setMethod(sensitivityMeasures, "CoreSet", function(object) {
    return(colnames(sensitivityProfiles(object)))
})

#' @noRd
.docs_CoreSet_set_sensitityMeasures <- function(...) .parseToRoxygen(
    "
    @details
    __sensitivityMeaures__: Update the sensitivity meaure in a `{class_}`
    object. Thesee values are the column names of the 'profiles' assay and
    represent various compued sensitviity metrics such as ic50, ec50, AUC, AAC,
    etc.
    - value: A `character` vector of new sensitivity measure names, the
    then length of the character vector must matcht he number of columns of the
    'profiles' assay, excluding metadata and key columns.
    @examples
    sensitivityMeasures({data_}) <- sensitivityMeasures({data_})

    @md
    @exportMethod sensitivityMeasures
    ",
    ...
)

#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_set_sensitityMeasures(class_=.local_class, data_=.local_data)
setReplaceMethod('sensitivityMeasures',
    signature(object='CoreSet', value='character'),
    function(object, value)
{
    colnames(sensitivityProfiles(object)) <- value
    object
})


#
# == sensitivityProfiles


#' @noRd
.docs_CoreSet_get_sensitivityProfiles <- function(...) .parseToRoxygen(
    "
    @details
    __sensitivityProfiles__: Return the sensitivity profile summaries from the
    sensitivity slot. This data.frame cotanins vaarious sensitivity summary
    metrics, such as ic50, amax, EC50, aac, HS, etc as columns, with rows as
    treatment by sample experiments.
    @examples
    sensitivityProfiles({data_})

    @md
    @exportMethod sensitivityProfiles
    ",
    ...
)

#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_get_sensitivityProfiles(class_=.local_class, data_=.local_data)
setMethod(sensitivityProfiles, "CoreSet", function(object) {
    funContext <- .funContext('::sensitivityProfiles')
    if (is(treatmentResponse(object), 'LongTable')) {
        if (!('profiles' %in% assayNames(treatmentResponse(object)))) {
            .error(funContext, 'The LongTable object in the sensivitiy slot
                is not formatted correctly: it must contain an assay
                named "profiles"!')
        } else {
            .rebuildProfiles(treatmentResponse(object))
        }
    } else {
        return(object@treatmentResponse$profiles)
    }
})

#' @keywords internal
.rebuildProfiles <- function(object) {
    profDT <- object$profiles
    rowCols <- lapply(rowIDs(object)[
        !grepl("treatment[0-9]*dose|drug[0-9]*dose", rowIDs(object))
    ], as.name)
    colCols <- lapply(colIDs(object), as.name)
    trt <- bquote(paste(..(rowCols), sep=":"), splice=TRUE)
    smp <- bquote(paste(..(colCols), sep=":"), splice=TRUE)
    profDT[, treatment_uid := eval(trt), by=.I]
    profDT[, sample_uid := eval(smp), by=.I]
    profDT[, exp_id := paste0(treatment_uid, "_", sample_uid), by=.I]
    assayCols <- setdiff(colnames(assay(object, "profiles", raw=TRUE)), ".profiles")
    sensProf <- unique(profDT[, .SD, .SDcols=c(assayCols, "exp_id")])
    obsPerExpId <- sensProf[, .N, by="exp_id"][, max(N)]
    if (obsPerExpId > 1) warning(.warnMsg("Multiple profile values per",
        " experiment id, summarizing with mean!"), call.=FALSE)
    sensProf <- sensProf[, lapply(.SD, mean, na.rm=TRUE), by="exp_id"]
    return(sensProf)
}

#' @noRd
.docs_CoreSet_set_sensitivityProfiles <- function(...) .parseToRoxygen(
    "
    @details
    __sensitivityProfiles<-__: Update the sensitivity profile summaries the
    sensitivity slot. Arguments:
    -value: A `data.frame` the the same number of rows as as returned by
    `sensitivityProfiles(object)`, but potentially modified columns, such as the
    computation of additional summary metrics.
    @examples
    sensitivityProfiles({data_}) <- sensitivityProfiles({data_})

    @md
    @exportMethod sensitivityProfiles<-
    ",
    ...
)

#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_set_sensitivityProfiles(class_=.local_class, data_=.local_data)
setReplaceMethod("sensitivityProfiles",
    signature(object="CoreSet", value="data.frame"),
    function(object, value)
{
    if (is(treatmentResponse(object), 'LongTable'))
        warning(.warnMsg("The ", class(object)[1], " class structure has been",
            " updated! Assignment via sensitivityProfiles no long works, please",
            " see vignette('The LongTable Class', package='CoreGx') for more",
            " information."))
    else
        object@treatmentResponse$profiles <- value
    return(object)
})


#
# == sensitivityRaw


#' @noRd
.docs_CoreSet_get_sensitivityRaw <- function(...) .parseToRoxygen(
    "
    @details
    __sensitivityRaw__: Access the raw sensitiity measurents for a {class_}
    object. A 3D `array` where rows are experiment_ids, columns are doses
    and the third dimension is metric, either 'Dose' for the doses used or
    'Viability' for the sample viability at that dose.
    @examples
    head(sensitivityRaw({data_}))

    @md
    @exportMethod sensitivityRaw
    ",
    ...
)

#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_get_sensitivityRaw(class_=.local_class, data_=.local_data)
setMethod("sensitivityRaw", signature("CoreSet"), function(object) {
    if (is(treatmentResponse(object), 'LongTable'))
        return(.rebuildRaw(treatmentResponse(object)))
    else
        return(object@treatmentResponse$raw)
})

#' Replicate the $raw slot in the old @treatmentResponse list from a LongTable
#'
#' @param longTable `LongTable`
#'
#' @return A 3D `array` where rows are experiment_ids, columns are doses
#' and the third dimension is metric, either 'Dose' for the doses used or
#' 'Viability' for the sample viability at that dose.
#'
#' @keywords internal
#' @importFrom data.table merge.data.table dcast
#' @noRd
.rebuildRaw <- function(longTable) {

    ## TODO:: This function currently assumes there will only be one valid
    ## dose per treatment combination, which may not be true.

    funContext <- .funContext(':::.rebuildRaw')
    if (!('sensitivity' %in% assayNames(longTable)))
        .error(funContext, 'There is no assay named sensitivity. Not sure
            how to retrieve sensitivityRaw without a sensitivity assay. Try
            renaming your assays in the @treatmentResponse LongTable object?')

    # Extract the information needed to reconstruct the sensitivityRaw array
    viability <- longTable$sensitivity

    # Early return for single treatment sensitivity experimentss
    ## TODO:: refactor this into a helper?
    if ('assay_metadata' %in% assayNames(longTable) &&
        'old_column' %in% colnames(longTable$assay_metadata))
    {
        metadataDT <- copy(longTable$assay_metadata)
        sensitivityDT <- copy(longTable$sensitivity)
        # .NATURAL joins on all identical columns
        assayDT <- metadataDT[sensitivityDT, on=.NATURAL]
        if (length(colIDs(longTable)) > 1) {
            assayDT[, sampleid := Reduce(.paste_colon, mget(colIDs(longTable)))]
        }
        assayDT[, exp_id := paste0(treatment1id, '_', sampleid)]
        .mean <- function(x) mean(as.numeric(x), na.rm=TRUE)
        doseDT <- dcast(assayDT, exp_id ~ old_column, value.var='treatment1dose',
            fun.aggregate=.mean)
        viabDT <- dcast(assayDT, exp_id ~ old_column, value.var='viability',
            fun.aggregate=.mean)
        sensRaw <- array(dim=list(nrow(doseDT), ncol(doseDT) -1, 2),
            dimnames=list(doseDT$exp_id, colnames(doseDT)[-1],
                c('Dose', 'Viability')))
        sensRaw[, , 'Dose'] <- as.matrix(doseDT[, !'exp_id'])
        sensRaw[, , 'Viability'] <- as.matrix(viabDT[, !'exp_id'])
        return(sensRaw)
    }

    # Build the rownames
    .paste_colons <- function(...) paste(..., sep=':')
    # viability[, row_ids := Reduce(.paste_colons, mget(rowIDs(longTable)))]
    # viability[, col_ids := Reduce(.paste_colons, mget(colIDs(longTable)))]
    # viability[, rownames := paste0(row_ids, '_', col_ids)]
    # viability[, c('row_ids', 'col_ids') := NULL]

    viability[, rownames := {
        row_ids <- Reduce(.paste_colons, mget(rowIDs(longTable)))
        col_ids <- Reduce(.paste_colons, mget(colIDs(longTable)))
        paste0(row_ids, '_', col_ids)
    }]

    # Merge the doses into vectors in a list column
    viability[, dose := Reduce(.paste_slashes, mget(colnames(.SD))),
        .SDcols=patterns('^.*[d|D]ose$')]

    # Repeat the dose values if there are more viabilities
    numReplicates <- viability[, ncol(.SD), .SDcols=patterns('^[V|v]iability.*')]
    if (numReplicates > 1) {
        viability[, paste0('dose', seq_len(numReplicates)) := dose]
        viability[, dose := NULL]
    }

    # Build the array
    sensRaw <- array(dim=list(nrow(viability), numReplicates, 2),
        dimnames=list(viability$rownames, paste0('dose', seq_len(numReplicates)),
            c('Dose', 'Viability')))
    sensRaw[, , 'Dose'] <- as.matrix(viability[, .SD,
        .SDcols=patterns('^dose.*')])
    sensRaw[, , 'Viability'] <- as.matrix(viability[, .SD,
        .SDcols=patterns('^[V|v]iability.*')])

    return(sensRaw)
}

#' @noRd
.docs_CoreSet_set_sensitivityRaw <- function(...) .parseToRoxygen(
    "
    @details

    __sensitvityRaw<-__: Update the raw dose and viability data in a `{class_}`.
    - value: A 3D `array` object where rows are experiment_ids, columns are
    replicates and pages are c('Dose', 'Viability'), with the corresponding
    dose or viability measurement for that experiment_id and replicate.

    @examples
    sensitivityRaw({data_}) <- sensitivityRaw({data_})

    @md
    @importFrom data.table data.table as.data.table := merge.data.table tstrsplit
    @exportMethod sensitivityRaw<-
    "
    ,
    ...
)

#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_set_sensitivityRaw(class_=.local_class, data_=.local_data)
setReplaceMethod('sensitivityRaw', signature("CoreSet", "array"),
    function(object, value)
{
    funContext <- .funContext("::sensitivityRaw<-")
    if (is(treatmentResponse(object), 'LongTable')) {

        ## TODO:: validate value
        tre <- treatmentResponse(object)

        viabilityCols <- assayCols(tre, "sensitivity")
        # Handle the non-treatment combo case
        if (length(viabilityCols) != ncol(value)) {
            valueDT <- as.data.table(value)
            valueDT <- dcast(valueDT, V1 + V2 ~ V3, value.var='value')
            setnames(valueDT, old=c('Dose', 'Viability'),
                new=c('treatment1dose', 'viability'))
            valueDT[, V2 := NULL]  # delete the array column names
            valueDT[, (idCols(tre)) := tstrsplit(V1, ':|_', type.convert=TRUE)]
            valueDT[, V1 := NULL]
            assay(tre, i='sensitivity') <- valueDT
        } else {
            # Process into a the proper format for the sensitivity assay
            # NOTE: as.matrix deals with the case where there is only a single
            #   viability column in the sensitivityRaw array object,
            #   in which case the drop=TRUE argument causes a vector to be
            #   returned
            raw <- as.data.table(as.matrix(value[, , 'Viability']),
                keep.rownames='rn', na.rm=FALSE)
            coerceCols <- colnames(raw)[-1]
            raw[, (coerceCols) := lapply(.SD, as.numeric), .SDcols=!'rn']
            raw[, (idCols(tre)) := tstrsplit(rn, ':|_', type.convert=TRUE)]
            raw[, c('rn') := NULL]
            colnames(raw) <- gsub('^dose\\d*|^V\\d*', 'viability', colnames(raw))
            # Update the assay
            assay(tre, i='sensitivity') <- raw
        }

        # Update the object
        treatmentResponse(object) <- tre
    } else {
        object@treatmentResponse$raw <- value
        object
    }
    return(object)
})


#
# == sensitivitySlot


#' @export
setGeneric("treatmentResponse", function(object, ...) standardGeneric("treatmentResponse"))

#' @noRd
.docs_CoreSet_get_treatmentResponse <- function(...) .parseToRoxygen(
    "
    __treatmentResponse__: Retrive the contents of `@treatmentResponse` from a `{class_}`
    object.

    @examples
    treatmentResponse({data_})

    @md
    @aliases treatmentResponse,{class_}-method treatmentResponse
    @aliases sensitivitySlot
    @exportMethod treatmentResponse
    ",
    ...
)

#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_get_treatmentResponse(class_=.local_class,
#'   data_=.local_data)
setMethod("treatmentResponse", signature("CoreSet"), function(object) {
    object@treatmentResponse
})
#' @export
sensitivitySlot <- function(...) treatmentResponse(...)


#' @export
setGeneric("treatmentResponse<-", function(object, ..., value)
    standardGeneric("treatmentResponse<-"))

.docs_CoreSet_set_treatmentResponse <- function(...) .parseToRoxygen(
    "
    __treatmentResponse<-__: Assign data to the `@treatmentResponse` slot of a
    `{class_}` object.
    - value: Either a `TreatmentResponseExperiment` class object, or a list with
    an 'info' `data.frame` of experiment metadata, 'profiles' `data.frame` with
    summary statistics from the sensitivity experiment and a 'raw' 3D array
    where rows are experiments, columns are replicates and pages are 'Dose'
    or 'Viability' containing their respective values for that treatment by sample
    experiment. The type of `value` must match type of the current `@treatmentResponse`
    slot of the `{class_}` object.

    @examples
    treatmentResponse({data_}) <- treatmentResponse({data_})

    @md
    @aliases treatmentResponse<- treamentResponse<-,{class_},list-method
    treatmentResponse<-,{class_},LongTable-method
    @aliases sensitivitySlot<-
    @exportMethod treatmentResponse<-
    ",
    ...
)


#' @rdname CoreSet-accessors
#' @include LongTable-class.R
#' @eval .docs_CoreSet_set_treatmentResponse(class_=.local_class, data_=.local_data)
setReplaceMethod("treatmentResponse", signature(object="CoreSet", value="list_OR_LongTable"),
    function(object, value)
{
    # funContext <- .S4MethodContext('sensitivitySlot<-', class(object), class(value))
    # ## TODO:: Maybe try coercing the list to a LongTable and vice versa?
    if (!is(object@treatmentResponse, class(value)[1])) .error(funContext, 'The types
        of the current and @treatmentResponse slot and the value parameter must be
        the same!')
    object@treatmentResponse <- value
    return(object)
})
#' @export
`sensitivitySlot<-` <- function(..., value) `treatmentResponse<-`(..., value=value)


##
## == sensNumber


#' @export
setGeneric("sensNumber", function(object, ...) standardGeneric("sensNumber"))

#' @noRd
.docs_CoreSet_get_sensNumber <- function(...) .parseToRoxygen(
    "
    @details
    __sensNumber__: Return a count of viability observations in a `{class_}`
    object for each treatment-combo by sample combination.

    @examples
    sensNumber({data_})

    @md
    @aliases sensNumber
    @exportMethod sensNumber
    ",
    ...
)

#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_get_sensNumber(class_=.local_class, data_=.local_data)
setMethod(sensNumber, "CoreSet", function(object){
    return(
        if (is(object@treatmentResponse, 'LongTable'))
            .rebuildSensNumber(object@treatmentResponse)
        else
            object@treatmentResponse$n
    )
})

.rebuildSensNumber <- function(object) {
    sensitivityDT <- object$sensitivity
    # Melt replicates so they get counted
    sensitivityDT_melted <- melt(sensitivityDT,
        measure=patterns('^viability'), variable.name='replicate',
        value.name='viability')

    # Determine the treatment and sample combos, ignoring other identifiers
    .paste_colon <- function(x, y) paste(x, y, sep=':')
    treatmentidCols <- sensitivityDT[, colnames(.SD), .SDcols=patterns('treatment.*id')]
    sampleidCols <- sensitivityDT[, colnames(.SD), .SDcols=patterns('sample.*id')]

    # Parse the columns to dcast by to get the counts
    sensitivityDT_melted[, .treatmentCombo := Reduce(.paste_colon, mget(treatmentidCols))]
    sensitivityDT_melted[, .sampleCombo := Reduce(.paste_colon, mget(sampleidCols))]

    # Count existing sensitivity measurements
    .count_not_NA <- function(x) sum(!is.na(x))
    sensNumbDT <- dcast(sensitivityDT_melted, .treatmentCombo ~ .sampleCombo,
        value.var='viability', fun.aggregate=.count_not_NA)
    sensNumberM <- as.matrix(sensNumbDT[, !'.treatmentCombo'])
    rownames(sensNumberM) <- sensNumbDT[['.treatmentCombo']]

    return(sensNumberM)

    ## TODO:: Pad for any missing treatments or samples
    allDrugCombos <- rowData(object)[, Reduce(.paste_colon, mget(treatmentidCols))]
    allSampleCombos <- colData(object)[, Reduce(.paste_colon, mget(sampleidCols))]

}

#' @export
setGeneric("sensNumber<-", function(object, value) standardGeneric("sensNumber<-"))

#' @noRd
.docs_CoreSet_set_sensNumber <- function(...) .parseToRoxygen(
    "
    @details
    __sensNumber<-__: Update the 'n' item, which holds a matrix with a count
    of treatment by sample-line experiment counts, in the `list` in `@treatmentResponse`
    slot of a `{class_}` object. Will error when `@sensitviity` contains
    a `LongTable` object, since the counts are computed on the fly. Arguments:
    - value: A `matrix` where rows are samples and columns are treatments, with a
    count of the number of experiments for each combination as the values.

    @examples
    sensNumber({data_}) <- sensNumber({data_})

    @md
    @aliases sensNumber<-
    @exportMethod sensNumber<-
    ",
    ...
)

#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_set_sensNumber(class_=.local_class, data_=.local_data)
setReplaceMethod('sensNumber', signature(object="CoreSet", value="matrix"),
    function(object, value)
{
    if (is(treatmentResponse(object), 'LongTable')) {
        object
    } else {
        object@treatmentResponse$n <- value
        object
    }
})


## ======================
## ---- perturbation slot

##
## == pertNumber


#' @export
setGeneric("pertNumber", function(object, ...) standardGeneric("pertNumber"))

#' @noRd
.docs_CoreSet_get_pertNumber <- function(...) .parseToRoxygen(
    "
    @details
    __pertNumber__: `array` Summary of available perturbation experiments
    from in a `{class_}` object. Returns a 3D `array` with the number of
    perturbation experiments per treatment and sample, and data type.

    @examples
    pertNumber({data_})

    @md
    @aliases pertNumber
    @exportMethod pertNumber
    ",
    ...
)

#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_get_pertNumber(class_=.local_class, data_=.local_data)
setMethod(pertNumber, "CoreSet", function(object){
    return(object@perturbation$n)
})

#' @export
setGeneric("pertNumber<-", function(object, value) standardGeneric("pertNumber<-"))

.docs_CoreSet_set_pertNumber <- function(...) .parseToRoxygen(
    "
    @details
    __pertNumber<-__: Update the `@perturbation$n` value in a `{class_}` object,
    which stores a summary of the available perturbation experiments. Arguments:
    - value: A new 3D `array` with the number of perturbation experiments per
    treatment and sample, and data type

    @examples
    pertNumber({data_}) <- pertNumber({data_})

    @md
    @aliases pertNumber<-
    @exportMethod pertNumber<-
    ",
    ...
)

#' @rdname CoreSet-accessors
#' @eval .docs_CoreSet_set_pertNumber(class_=.local_class, data_=.local_data)
setReplaceMethod('pertNumber', signature(object="CoreSet", value="array"),
    function(object, value)
{
  object@perturbation$n <- value
  object
})

#' @include DataMapper-class.R
#' @include LongTableDataMapper-class.R
#' @include TreatmentResponseExperiment-class.R
#' @include LongTableDataMapper-accessors.R
NULL

#' A Class for Mapping Between Raw Data and an `TreatmentResponseExperiment`
#'   Object
#'
#' @slot rawdata See Slots section.
#' @slot rowDataMap See Slots section.
#' @slot colDataMap See Slots section.
#' @slot assayMap See Slots section.
#' @slot metadataMap See Slots section.
#'
#' @inheritSection LongTableDataMapper-class Slots
#'
#' @md
#' @aliases TREDataMapper-class
.TREDataMapper <- setClass("TREDataMapper", contains=c("LongTableDataMapper"))

#' Constructor for the `TREDataMapper` class, which maps from one or
#'   more raw experimental data files to the slots of a `LongTable` object.
#'
#' @details
#' The `guessMapping` method can be used to test hypotheses about the
#' cardinality of one or more sets of identifier columns. This is helpful
#' to determine the id columns for `rowDataMap` and `colDataMap`, as well
#' as identify columns mapping to `assays` or `metadata`.
#'
#' To attach metadata not associated with `rawdata`, please use the `metadata`
#' assignment method on your `TREDataMapper`. This metadata will be
#' merge with any metadata from `metadataMap` and added to the `LongTable`
#' which this object ultimately constructs.
#'
#' @param rawdata A `data.frame` of raw data from a treatment response
#' experiment. This will be coerced to a `data.table` internally. We recommend
#' using joins to aggregate your raw data if it is not present in a single file.
#' @param rowDataMap A list-like object containing two `character` vectors.
#' The first is column names in `rawdata` needed to uniquely identify each row,
#' the second is additional columns which map to rows, but are not required to
#' uniquely identify them. Rows should be treatments.
#' @param colDataMap A list-like object containing two `character` vectors.
#' The first is column names in `rawdata` needed to uniquely identify each
#' column, the second is additional columns which map to rows, but are not
#' required to uniquely identify them. Columns should be samples.
#' @param assayMap A list-like where each item is a `list` with two `character`
#' vectors defining an assay, the first containing the identifier columns in
#' `rawdata` needed to uniquely identify each row an assay, and the second the
#' `rawdata` columns to be mapped to that assay. The names of `assayMap`
#' will be the names of the assays in the `TreatmentResponseExperiment` that
#' is created when calling `metaConstruct` on this `DataMapper` object. If the
#' character vectors have names, the value columns will be renamed accordingly.
#' @param metadataMap A list-like where each item is a `character` vector of
#' `rawdata` column names to assign to the `@metadata` of the `LongTable`,
#' where the name of that assay is the name of the list item. If names are
#' omitted, assays will be numbered by their index in the list.
#'
#' @return A `TREDataMapper` object, with columns mapped to it's slots according
#' to the various maps in the `LongTableDataMapper` object.
#'
#' @seealso [`guessMapping`]
#'
#' @md
#' @importFrom data.table setDT
#' @export
TREDataMapper <- function(rawdata=data.frame(),
        rowDataMap=list(character(), character()),
        colDataMap=list(character(), character()),
        assayMap=list(list(character(), character())),
        metadataMap=list(character())) {

    if (is(rawdata, "LongTableDataMapper")) {
        lt_dm <- rawdata
    } else {
        lt_dm <- LongTableDataMapper(rawdata=rawdata, rowDataMap=rowDataMap,
            colDataMap=colDataMap, assayMap=assayMap, metadataMap=metadataMap)
    }

    .TREDataMapper(
        rawdata=rawdata(lt_dm),
        rowDataMap=rowDataMap(lt_dm),
        colDataMap=colDataMap(lt_dm),
        assayMap=assayMap(lt_dm),
        metadataMap=metadata(lt_dm)
    )
}

## ===========================================
## TREDataMapper Accessors Documentation
## -------------------------------------------

.local_class_5 <- "TREDataMapper"
.local_data_5 <- "exampleDataMapper"

#' @name TREDataMapper-accessors
#'
#' @eval .docs_DataMapper_accessors(class_=.local_class_5)
#' @eval .docs_DataMapper_get_rawdata(class_=.local_class_5)
#'
#' @param value See details.
NULL


## ---------------
## -- rawdata slot


#' @rdname TREDataMapper-accessors
#' @eval .docs_DataMapper_set_rawdata(class_=.local_class_5,
#' class1_='list')
setReplaceMethod("rawdata", signature=c(object="TREDataMapper",
        value="list"), function(object, value) {
    callNextMethod(object=object, value=value)
})


## --------------------
## ---- rowDataMap slot


##
## -- rowDataMap

#' @rdname TREDataMapper-accessors
#' @eval
#' .docs_LongTableDataMapper_get_dimDataMap(dim_='row', class_=.local_class_5,
#' data_=.local_data_5)
#' @aliases rowDataMap
setMethod('rowDataMap', signature(object='TREDataMapper'), function(object) {
    callNextMethod()
})


#' @rdname TREDataMapper-accessors
#' @eval
#' .docs_LongTableDataMapper_set_dimDataMap(dim_='row', class_=.local_class_5,
#' data_=.local_data_5, id_col_='treatmentid')
#' @aliases rowDataMap<-
setReplaceMethod('rowDataMap', signature(object='TREDataMapper',
        value='list_OR_List'), function(object, value) {
    callNextMethod()
})


##
## -- rowData


#' Convenience method to subset the `rowData` out of the `rawdata` slot using
#'   the assigned `rowDataMap` metadata.
#'
#' @param x `TREDataMapper` object with valid data in the `rawdata` and
#'   `colDataMap` slots.
#' @param key `logical(1)` Should the table be keyed according to the
#'   `id_columns` of the `rowDataMap` slot? This will sort the table in memory.
#'   Default is TRUE.
#'
#' @return `data.table` The `rowData` as specified in the `rowDataMap` slot.
#'
#' @export
setMethod("rowData", signature(x="TREDataMapper"), function(x, key=TRUE) {
    callNextMethod()
})


## --------------------
## ---- colDataMap slot


##
## -- colDataMap

#' @rdname TREDataMapper-accessors
#' @eval
#' .docs_LongTableDataMapper_get_dimDataMap(dim_='col', class_=.local_class_5,
#' data_=.local_data_5)
#' @aliases colDataMap
setMethod('colDataMap', signature(object='TREDataMapper'),
        function(object) {
    callNextMethod()
})

#' @rdname TREDataMapper-accessors
#' @eval
#' .docs_LongTableDataMapper_set_dimDataMap(dim_='col', class_=.local_class_5,
#' data_=.local_data_5, id_col_='sampleid')
#' @aliases colDataMap<-
setReplaceMethod('colDataMap',
        signature(object="TREDataMapper", value="list_OR_List"),
        function(object, value) {
    callNextMethod()
})


##
## -- colData


#' Convenience method to subset the `colData` out of the `rawdata` slot using
#'   the assigned `colDataMap` metadata.
#'
#' @param x `TREDataMapper` object with valid data in the `rawdata` and
#'   `colDataMap` slots.
#' @param key `logical(1)` Should the table be keyed according to the
#'   `id_columns` of the `colDataMap` slot? This will sort the table in memory.
#'   Default is TRUE.
#'
#' @return `data.table` The `colData` as specified in the `colDataMap` slot.
#'
#' @export
setMethod("colData", signature(x="TREDataMapper"), function(x, key=TRUE) {
    callNextMethod()
})


## ----------------
## ---- assayMap slot

#' @rdname TREDataMapper-accessors
#' @eval .docs_LongTableDataMapper_get_assayMap(class_=.local_class_5, data_=.local_data_5)
#' @aliases assayMap
setMethod('assayMap', signature(object='TREDataMapper'),
        function(object) {
    callNextMethod()
})


#' @rdname TREDataMapper-accessors
#' @eval .docs_LongTableDataMapper_set_assayMap(class_=.local_class_5, data_=.local_data_5)
#' @aliases assayMap<-
setReplaceMethod('assayMap', signature(object='TREDataMapper',
        value='list_OR_List'), function(object, value) {
    callNextMethod()
})


#' Extract the data for an assay from a `TREDataMapper`
#'
#' @param x `TREDataMapper` The object to retrive assay data form according
#'   to the `assayMap` slot.
#' @param i `character(1)` Name of an assay in the `assayMap` slot of `x`.
#' @param withDimnames `logical(1)` For compatibility with
#'   `SummarizedExperiment::assay` generic. Not used.
#'
#' @return `data.table` Data for the specified assay extracted from the
#'   `rawdata` slot of `x`.
#'
#' @importFrom checkmate assertSubset assertCharacter
#' @keywords internal
setMethod("assay", signature(x="TREDataMapper"),
        function(x, i, withDimnames=TRUE) {
    callNextMethod()
})


#' Extract the data for all assays from a `TREDataMapper`
#'
#' @param x `TREDataMapper` The object to retrive assay data form according
#'   to the `assayMap` slot.
#' @param withDimnames `logical(1)` For compatibility with
#'   `SummarizedExperiment::assay` generic. Not used.
#'
#' @return `list` Data for all assays extracted from the
#'   `rawdata` slot of `x` as a `list` of `data.tables`, where the `keys` for
#'   each table are their `id_columns`.
#'
#' @importFrom checkmate assertSubset assertCharacter
#' @keywords internal
setMethod("assays", signature(x="TREDataMapper"),
        function(x, withDimnames=TRUE) {
    callNextMethod()
})

# -- metadataMap

#' @rdname TREDataMapper-accessors
#' @eval .docs_LongTableDataMapper_get_metadataMap(class_=.local_class_5, data_=.local_data_5)
#' @aliases metadataMap
setMethod('metadataMap', signature(object='TREDataMapper'),
        function(object) {
    callNextMethod()
})


#' @rdname TREDataMapper-accessors
#' @eval .docs_LongTableDataMapper_set_metadataMap(class_=.local_class_5, data_=.local_data_5, col_='metadata')
#' @aliases metadataMap<-
setReplaceMethod('metadataMap', signature(object='TREDataMapper',
        value='list_OR_List'), function(object, value) {
    callNextMethod()
})

#' Compute the correlation between a molecular feature and treatment response
#'
#' @param object An object inheriting form the `CoreGx::CoreSet` class
#' @param ... Allow definition of new arguments to this generic
#'
#' @return A 3D array of genes x drugs x metric
#'
#' @export
#' @keywords internal
setGeneric("drugSensitivitySig", function(object, ...)
    standardGeneric("drugSensitivitySig"))

#' @include CoreSet-class.R LongTable-class.R
#' @importClassesFrom MultiAssayExperiment MultiAssayExperiment
#' @importFrom MultiAssayExperiment MultiAssayExperiment
#' @import checkmate
NULL

setClassUnion('list_OR_MAE', c('list', 'MultiAssayExperiment'))

.local_class <- 'CoreSet'
.local_data <- "clevelandSmall_cSet"

#' @title
#' CoreSet - A generic data container for molecular profiles and
#'   treatment response data
#'
#' @slot annotation See Slots section.
#' @slot molecularProfiles See Slots section.
#' @slot sample See Slots section.
#' @slot treatment See Slots section.
#' @slot treatmentResponse See Slots section.
#' @slot perturbation See Slots section.
#' @slot curation See Slots section.
#' @slot datasetType See Slots section.
#'
#' @details
#' The CoreSet (cSet) class was developed as a superclass for pSets in the
#' PharmacoGx and RadioGx packages to contain the data generated in screens
#' of cancer sample lines for their genetic profile and sensitivities to therapy
#' (Pharmacological or Radiation). This class is meant to be a superclass which
#' is contained within the PharmacoSet (pSet) and RadioSet (rSet) objects
#' exported by PharmacoGx and RadioGx. The format of the data is similar for
#' both pSets and rSets, allowing much of the code to be abstracted into
#' the CoreSet super-class. However, the models involved with quantifying
#' sampleular response to Pharmacological and Radiation therapy are widely
#' different, and extension of the cSet class allows the
#' packages to apply the correct model for the given data.
#'
#' @section Slots:
#' * annotation: A `list` of annotation data about the ``r .local_class``,
#'   including the `$name` and the session information for how the object
#'   was created, detailing the exact versions of R and all the packages used.
#' * molecularProfiles: A `list` or `MultiAssayExperiment` containing
#    a set of `SummarizedExperiment`s with molecular profile data for a given
#'   ``r .local_class`` object.
#' * sample: A `data.frame` containg the annotations for all the samples
#'   profiled in the data set, across all molecular data types and
#'   treatment response experiments.
#' * treatment: A `data.frame` containing the annotations for all treatments
#'   in the dataset, including the mandatory 'treatmentid' column to uniquely
#'   identify each treatment.
#' * treatmentResponse: A `list` or `LongTable` containing all the data for the
#'   treatment response experiment, including `$info`, a `data.frame`
#'   containing the experimental info, `$raw` a 3D `array` containing raw data,
#'   `$profiles`, a `data.frame` containing sensitivity profiles
#'   statistics, and `$n`, a `data.frame` detailing the number of
#'   experiments for each sample-drug/radiationInfo pair
#' * perturbation: `list` containing `$n`, a `data.frame`
#'   summarizing the available perturbation data. This slot is currently
#'   being deprecated.
#' * curation: `list` containing mappings for `treatment`,
#'   `sample` and `tissue` names used in the data set to universal
#'   identifiers used between different ``r .local_class`` objects.
#' * datasetType: `character` string of 'sensitivity',
#'   'perturbation', or both detailing what type of data can be found in the
#'   ``r .local_class``, for proper processing of the data
#'
#' @seealso [`CoreSet-accessors`]
#'
#' @md
#' @aliases CoreSet-class
#' @exportClass CoreSet
.CoreSet <- setClass("CoreSet",
    slots=list(
        treatmentResponse="list_OR_LongTable",
        annotation="list",
        molecularProfiles="list_OR_MAE",
        sample="data.frame",
        treatment="data.frame",
        datasetType="character",
        perturbation="list",
        curation="list"
    )
)

# The default constructor above does a poor job of explaining the required structure of a CoreSet.
# The constructor function defined below guides the user into providing the required components of the curation and senstivity lists
# and hides the annotation slot which the user does not need to manually fill.
# This also follows the design of the Expression Set class.

## ==========================
## CONSTRUCTOR
## --------------------------

#' CoreSet constructor
#'
#' A constructor that simplifies the process of creating CoreSets, as well
#' as creates empty objects for data not provided to the constructor. Only
#' objects returned by this constructor are expected to work with the CoreSet
#' methods.
#'
#' ## __WARNING__:
#' Parameters to this function have been renamed!
#' * cell is now sample
#' * drug is now treatment
#'
#' @param name A \code{character} string detailing the name of the dataset
#' @param molecularProfiles A \code{list} of SummarizedExperiment objects containing
#'   molecular profiles for each molecular data type.
#' @param sample A \code{data.frame} containing the annotations for all the sample
#'   profiled in the data set, across all data types. Must contain the mandatory
#'   `sampleid` column which uniquely identifies each sample in the object.
#' @param treatment A `data.frame` containing annotations for all treatments
#'   profiled in the dataset. Must contain the mandatory `treatmentid` column
#'   which uniquely identifies each treatment in the object.
#' @param sensitivityInfo A \code{data.frame} containing the information for the
#'   sensitivity experiments. Must contain a 'sampleid' column with unique
#'   identifiers to each sample, matching the `sample` object and a 'treatmentid'
#'   columns with unique indenifiers for each treatment, matching the `treatment`
#'   object.
#' @param sensitivityRaw A 3 Dimensional \code{array} contaning the raw drug
#'   dose response data for the sensitivity experiments
#' @param sensitivityProfiles \code{data.frame} containing drug sensitivity profile
#'   statistics such as IC50 and AUC
#' @param sensitivityN,perturbationN A \code{data.frame} summarizing the
#'   available sensitivity/perturbation data
#' @param curationSample,curationTissue,curationTreatment A \code{data.frame} mapping
#'   the names for samples, tissues and treatments used in the data set to
#'   universal identifiers used between different CoreSet objects
#' @param datasetType A `character(1)` string of 'sensitivity',
#'   'preturbation', or 'both' detailing what type of data can be found in the
#'   `CoreSet`, for proper processing of the data
#' @param verify `logical(1)`Should the function verify the CoreSet and
#'   print out any errors it finds after construction?
#' @param ... Catch and parse any renamed constructor arguments.
#'
#' @return An object of class `CoreSet`
#'
#' @examples
#' data(clevelandSmall_cSet)
#' clevelandSmall_cSet
#'
#' @export
#'
#' @include LongTable-class.R
#' @import methods
#' @importFrom utils sessionInfo
#' @importFrom stats na.omit
#' @importFrom SummarizedExperiment rowData colData assays
CoreSet <- function(name, molecularProfiles=list(), sample=data.frame(),
    sensitivityInfo=data.frame(), sensitivityRaw=array(dim=c(0,0,0)),
    sensitivityProfiles=matrix(), sensitivityN=matrix(nrow=0, ncol=0),
    perturbationN=array(NA, dim=c(0,0,0)), curationSample=data.frame(),
    curationTissue=data.frame(), curationTreatment=data.frame(),
    treatment=data.frame(), datasetType=c("sensitivity", "perturbation", "both"),
    verify=TRUE, ...
) {

    # .Deprecated("CoreSet2", package=packageName(), msg="The CoreSet class is
    #     being redesigned. Please use the new constructor to ensure forwards
    #     compatibility with future releases! Old objects can be updated with
    #     the updateObject method.", old="CoreSet")

    # parse deprecated parameters to ensure changes don't break old code
    dotnames <- ...names()
    if ("cell" %in% dotnames) {
        .warning("The cell parameter is deprecated, assigning to sample...")
        sample <- cell
    }
    if ("drug" %in% dotnames) {
        .warning("The drug paramter is deprecated, assigning to treatment...")
        treatment <- drug
    }

    # ensure new sampleid and treatmentid identifiers are honoured
    sample <- .checkForSampleId(sample)
    treatment <- .checkForTreatmentId(treatment)
    sensitivityInfo <- .checkForSampleId(sensitivityInfo)
    sensitivityInfo <- .checkForTreatmentId(sensitivityInfo)
    curationSample <- .checkForIdColumn(curationSample, c("sampleid", "unique.sampleid"), "cellid")
    curationTreatment <- .checkForIdColumn(curationTreatment, c("treatmentid", "unique.treatmentid"), "drugid")
    for (nm in names(molecularProfiles)) {
        colData(molecularProfiles[[nm]]) <- .checkForSampleId(
            colData(molecularProfiles[[nm]]))
        # handle perturbation case
        colData(molecularProfiles[[nm]]) <- .checkForIdColumn(
            colData(molecularProfiles[[nm]]), "treatmentid", "drugid",
            error=FALSE)
    }

    datasetType <- match.arg(datasetType)

    annotation <- list()
    annotation$name <- as.character(name)
    annotation$dateCreated <- date()
    annotation$sessionInfo <- sessionInfo()
    annotation$call <- match.call()

    for (i in seq_len(length(molecularProfiles))){
        if (!is(molecularProfiles[[i]], "SummarizedExperiment")) {
            stop(sprintf("Please provide the %s data as a SummarizedExperiment",
                names(molecularProfiles[i])))
        } else {
            rowData(molecularProfiles[[i]]) <-
                rowData(molecularProfiles[[i]])[
                    rownames(assays(molecularProfiles[[i]])[[1]]), , drop=FALSE
            ]
            colData(molecularProfiles[[i]]) <- colData(molecularProfiles[[i]])[
                colnames(assays(molecularProfiles[[i]])[[1]]), , drop=FALSE
            ]
        }
    }

    sensitivity <- list()

    if (!all(rownames(sensitivityInfo) == rownames(sensitivityProfiles) &
        rownames(sensitivityInfo) == dimnames(sensitivityRaw)[[1]])) {
        stop("Please ensure all the row names match between the sensitivity data.")
    }

    sensitivity$info <- as.data.frame(sensitivityInfo, stringsAsFactors=FALSE)
    sensitivity$raw <- sensitivityRaw
    sensitivity$profiles <- as.data.frame(sensitivityProfiles,
        stringsAsFactors=FALSE)
    sensitivity$n <- sensitivityN

    curation <- list()
    curation$sample <- as.data.frame(curationSample, stringsAsFactors=FALSE)
    curation$tissue <- as.data.frame(curationTissue, stringsAsFactors=FALSE)

    perturbation <- list()
    perturbation$n <- perturbationN
    if (datasetType == "perturbation" || datasetType == "both") {
        perturbation$info <- "The metadata for the perturbation experiments is
            available for each molecular type by calling the appropriate info
            function. \n For example, for RNA transcriptome perturbations, the
            metadata can be accessed using rnaInfo(cSet)."
    } else {
        perturbation$info <- "Not a perturbation dataset."
    }

    object  <- .CoreSet(annotation=annotation,
        molecularProfiles=molecularProfiles,
        sample=as.data.frame(sample), datasetType=datasetType,
        treatmentResponse=sensitivity, perturbation=perturbation,
        curation=curation, treatment=treatment)
    if (verify) { checkCsetStructure(object)}

    ## TODO:: Are these functions identitical in inheriting packages?
    if(length(sensitivityN) == 0 &&
            datasetType %in% c("sensitivity", "both")) {
        sensNumber(object) <- .summarizeSensitivityNumbers(object)
    }
    if(length(perturbationN) == 0  &&
            datasetType %in% c("perturbation", "both")) {
        pertNumber(object) <- .summarizePerturbationNumbers(object)
    }
    return(object)
}


#' Utility to help identify and fix deprecated identifiers
#'
#' @param new_col `character(1)` The new identifier.
#' @param old_col `character(1)` A regex matching any old identifers to
#' replace.
#'
#' @return `rectangular` object, with old_col updated to new_col if it exists.
#'
#' @noRd
.checkForIdColumn <- function(df, new_col, old_col, error=TRUE) {
    if (nrow(df) == 0 || ncol(df) == 0) return(df)
    name <- as.character(substitute(df))
    if (!any(colnames(df) %in% new_col)) {
        if (old_col %in% colnames(df)) {
            .warning("The ", old_col, "identifier is deprecated, updating to",
                new_col, " in ", name, "!")
            colnames(df) <- gsub(old_col, new_col[1], colnames(df))
        } else {
            if (error)
                .error("The ", new_col[1], " identifier is mandatory in ", name, "!")
        }
        return(df)
    }
    return(df)
}

#' @noRd
.checkForTreatmentId <- function(df)
    .checkForIdColumn(df, new_col="treatmentid", old_col="drugid")

#' @noRd
.checkForSampleId <- function(df)
    .checkForIdColumn(df, new_col="sampleid", old_col="cellid")


#' @noRd
.docs_CoreSet2_constructor <- function(...) .parseToRoxygen(
    "
    @title Make a CoreSet with the updated class structure

    @description
    New implementation of the CoreSet constructor to support MAE and TRE. This
    constructor will be swapped with the original `CoreSet` constructor as
    part of an overhaul of the CoreSet class structure.

    @param name A `character(1)` vector with the `{class_}` objects name.
    @param treatment A `data.frame` with treatment level metadata. {tx_}
    @param sample A `data.frame` with sample level metadata for the union
        of samples in `treatmentResponse` and `molecularProfiles`. {sx_}
    @param molecularProfiles A `MultiAssayExperiment` containing one
        `SummarizedExperiment` object for each molecular data type.
    @param treatmentResponse A `LongTable` or `LongTableDataMapper` object
        containing all treatment response data associated with the `{class_}`
        object.
    @param curation {cx_}
    @param perturbation A deprecated slot in a `{class_}` object included
    for backwards compatibility. This may be removed in future releases.
    @param datasetType A deprecated slot in a `{class_}` object included
    for backwards compatibility. This may be removed in future releases.

    @examples
    data({data_})
    {data_}

    @return A `CoreSet` object storing standardized and curated treatment
        response and multiomic profile data associated with a given publication.

    @importFrom MultiAssayExperiment MultiAssayExperiment
    @importFrom checkmate assertCharacter assertDataFrame assertClass assert
    assertList assertSubset
    ",
    ...
)

#' @eval .docs_CoreSet2_constructor(class_=.local_class,
#' tx_="",
#' sx_="",
#' cx_="A `list(2)` object with two items named `treatment` and `sample` with
#' mappings from publication identifiers to standardized identifiers for
#' both annotations, respectively.",
#' data_=.local_data)
#' @md
#' @export
CoreSet2 <- function(name="emptySet", treatment=data.frame(),
    sample=data.frame(), molecularProfiles=MultiAssayExperiment(),
    treatmentResponse=LongTable(), datasetType="sensitivity",
    perturbation=list(n=array(dim=3), info="No perturbation data!"),
    curation=list(sample=data.frame(), treatment=data.frame())
) {

    # -- update old curation names
    names(curation) <- gsub("drug|radiation", "treatment", names(curation))
    names(curation) <- gsub("cell", "sample", names(curation))

    ## -- input validation
    assertCharacter(name, len=1)
    assertDataFrame(treatment)
    assertDataFrame(sample)
    assertClass(molecularProfiles, "MultiAssayExperiment")
    assert(
        checkClass(treatmentResponse, "LongTable"),
        checkClass(treatmentResponse, "LongTableDataMapper")
    )
    assertList(curation, min.len=2)
    assertSubset(c("sample", "treatment"), choices=names(curation))

    ## -- capture object creation environment
    annotation <- list(name=name, dateCreated=date(),
        sessionInfo=sessionInfo(), call=match.call())

    ## -- conditionally materialize DataMapper
    if (is(treatmentResponse, 'LongTableDataMapper'))
        treatmentResponse <- metaConstruct(treatmentResponse)


    ## -- handle missing rownames for sample
    if (!all(sample$sampleid == rownames(sample)))
        rownames(sample) <- sample$sampleid

    object <- .CoreSet(
        annotation=annotation,
        sample=sample,
        treatment=treatment,
        molecularProfiles=molecularProfiles,
        treatmentResponse=treatmentResponse,
        datasetType=datasetType,
        curation=curation,
        perturbation=perturbation
    )

    ## -- data integrity checks
    # molecularProfiles
    validProfiles <- .checkMolecularProfiles(object)

    # treatmentResponse
    validTreatments <- .checkTreatmentResponse(object)

    diagnosis <- c(!isTRUE(validProfiles), !isTRUE(validTreatments))
    if (any(diagnosis)) {
        .error(paste0(list(validProfiles, validTreatments)[diagnosis],
            collapse="\n", sep="\n"))
    }
    return(object)
}

#' Show a CoreSet
#'
#' @param object `CoreSet` object to show via `cat`.
#'
#' @seealso [`cat`]
#'
#' @examples
#' show(clevelandSmall_cSet)
#'
#' @return Prints the CoreSet object to the output stream, and returns
#'   invisible NULL.
#' 
#' @importFrom crayon %+% yellow red green blue cyan magenta
#' 
#' @md
#' @export
setMethod("show", signature=signature(object="CoreSet"), function(object) {

    if (!.hasSlot(object, "sample") || !.hasSlot(object, "treatment"))
        stop(.errorMsg("This ", class(object)[1], " object appears to be out",
            "of date! Please run object <- updateObject(object) to update ",
            "the object for compatibility with the current release."),
            call.=FALSE)

    cat(yellow$bold$italic(paste0("<", class(object)[1], ">\n")))
    space <- "  "
    cat(yellow$bold$italic("Name: ") %+% green(name(object)), "\n")

    cat(yellow$bold$italic("Date Created: ") %+% green(dateCreated(object)), "\n")

    # cat("Number of samples: ", nrow(sampleInfo(object)), "\n")
    cat(yellow$bold$italic("Number of samples: "), green(nrow(sampleInfo(object))), "\n")

    mProfiles <- molecularProfilesSlot(object)
    mProfileNames <- names(mProfiles)
    if (is(mProfiles, "MultiAssayExperiment")) {
        cat(yellow$bold$italic("Molecular profiles: "))
        cat(yellow$bold$italic(paste0("<", class(mProfiles)[1], ">"), '\n'))

        # changing this to just experiments() as its cleaner
        # showMAE <- capture.output(show(mProfiles))
        # dropAfter <- which(grepl("Functionality", showMAE)) - 1
        # showCompactMAE <- showMAE[1:dropAfter]
        # cat(space, paste0(showCompactMAE, collapse="\n  "), "\n")

        showExp <- capture.output(experiments(mProfiles))
        cat(space, yellow$bold(showExp[1], '\n'))

        # iterate through rest of experiments 
        # split by ":" and print the first element in yellow, the rest in cyan
        for (i in 2:length(showExp)) {
            splitExp <- strsplit(showExp[i], ":")[[1]]
            
            # print the first 5 characters in splitExp[1] in yellow
            cat(space, yellow$bold(substr(splitExp[1], 1, 5)))
            # print after the first 5 characters in spllitExp[1] in cyan
            cat(
                red(substr(splitExp[1], 6, nchar(splitExp[1]))),
                green(paste0(":", splitExp[2:length(splitExp)], collapse=":"),'\n'))
        }

        # samplenames <- sort(sampleNames(object))
        # samplenames <- 
        #     if (length(samplenames) > 6) {
        #         paste0(.collapse(head(samplenames, 3)), ' ... ', .collapse(tail(samplenames, 3)))
        #     } else {
        #         .collapse(samplenames)
        #     }
        # x <- yellow$bold(paste0("colnames(", length(sampleNames(object)), "):"))
        # cat(space, x, green(samplenames), "\n")
    } else {
        cat(yellow$bold$italic("Molecular profiles: \n"))
        if (!length(mProfileNames)) cat(space, "None\n")
        for (item in mProfileNames) {
            title <- switch(item,
                "dna"="DNA",
                "rna"="RNA",
                "rnaseq"="RNAseq",
                "snp"="SNP",
                "cnv"="CNV",
                item
            )
            cat(title, ":\n")
            cat(paste0(space, "Dim: ",
                paste0(dim(molecularProfiles(object, mDataType=item)), collapse=", ")),
                "\n"
            )
        }
    }
    cat(yellow$bold$italic("Treatment response: "))
    if (is(treatmentResponse(object), "LongTable")) {
        show(treatmentResponse(object))
        # showLT <- capture.output(show(treatmentResponse(object)))
        # cat(space, paste0(showLT, collapse="\n  "), "\n")
    } else {
        cat("Drug pertubation:\n")
        cat(space,
            "Please look at pertNumber(cSet) to determine number of experiments",
            " for each drug-sample combination.\n")
        cat("Drug sensitivity:\n")
        cat(space, "Number of Experiments: ", nrow(sensitivityInfo(object)), "\n")
        cat(space, "Please look at sensNumber(cSet) to determine number of ",
            "experiments for each drug-sample combination.\n")
    }
})


#' Update the sample ids in a cSet object
#'
#' @examples
#' updateSampleId(clevelandSmall_cSet, sampleNames(clevelandSmall_cSet))
#'
#' @param object The object for which the sample ids will be updated
#' @param new.ids The new ids to assign to the object
#'
#' @return \code{CoreSet} The modified CoreSet object
#'
#' @keywords internal
#' @importFrom S4Vectors endoapply
#' @importFrom SummarizedExperiment colData rowData
#' @export
updateSampleId <- function(object, new.ids=vector("character")) {

    if (length(new.ids) != nrow(sampleInfo(object))){
        stop("Wrong number of sample identifiers")
    }

    if (datasetType(object) == "sensitivity" || datasetType(object) == "both") {
        myx <- match(sensitivityInfo(object)[, "sampleid"],
            rownames(sampleInfo(object)))
        if (is(treatmentResponse(object), 'LongTable')) {
            LT <- treatmentResponse(object)
            whichSampleIds <- which(colData(LT)$sampleid %in% sampleNames(object))
            colData(LT)$sampleid <- new.ids[whichSampleIds]
            treatmentResponse(object) <- LT
        } else {
            sensitivityInfo(object)[, "sampleid"] <- new.ids[myx]
        }
    }

    molecularProfilesSlot(object) <- lapply(molecularProfilesSlot(object), function(SE) {
        myx <- match(colData(SE)[["sampleid"]],
            rownames(sampleInfo(object)))
        colData(SE)[["sampleid"]]  <- new.ids[myx]
        return(SE)
    })

    if (any(duplicated(new.ids))) {
        warning("Duplicated ids passed to updateSampleId. Merging old ids into",
            " the same identifier")

        if(ncol(sensNumber(object)) > 0) {
            sensMatch <- match(rownames(sensNumber(object)),
                rownames(sampleInfo(object)))
        }
        if(dim(pertNumber(object))[[2]] > 0) {
            pertMatch <- match(dimnames(pertNumber(object))[[1]],
                rownames(sampleInfo(object)))
        }

        curMatch <- match(rownames(curation(object)$sample),
            rownames(sampleInfo(object)))
        duplId <- unique(new.ids[duplicated(new.ids)])

        for(id in duplId){
            if (ncol(sensNumber(object)) > 0) {
                myx <- which(new.ids[sensMatch] == id)
                sensNumber(object)[myx[1],] <- apply(sensNumber(object)[myx, ],
                    2, sum)
                sensNumber(object) <- sensNumber(object)[-myx[-1], ]
                # sensMatch <- sensMatch[-myx[-1]]
        }
        if (dim(pertNumber(object))[[1]] > 0) {
            myx <- which(new.ids[pertMatch] == id)
            pertNumber(object)[myx[1], , ] <- apply(pertNumber(object)[myx, , ],
                c(1,3), sum)
            pertNumber(object) <- pertNumber(object)[-myx[-1], , ]
        }

        myx <- which(new.ids[curMatch] == id)
        curation(object)$sample[myx[1],] <- apply(curation(object)$sample[myx, ], 2,
            FUN=paste, collapse="///")
        curation(object)$sample <- curation(object)$sample[-myx[-1], ]
        curation(object)$tissue[myx[1],] <- apply(curation(object)$tissue[myx, ],
            2, FUN=paste, collapse="///")
        curation(object)$tissue <- curation(object)$tissue[-myx[-1], ]

        myx <- which(new.ids == id)
        sampleInfo(object)[myx[1],] <- apply(sampleInfo(object)[myx,], 2,
            FUN=paste, collapse="///")
        sampleInfo(object) <- sampleInfo(object)[-myx[-1], ]
        new.ids <- new.ids[-myx[-1]]
        if(ncol(sensNumber(object)) > 0){
            sensMatch <- match(rownames(sensNumber(object)),
                rownames(sampleInfo(object)))
        }
        if(dim(pertNumber(object))[[1]] > 0){
            pertMatch <- match(dimnames(pertNumber(object))[[1]],
                rownames(sampleInfo(object)))
        }
        curMatch <- match(rownames(curation(object)$sample),
            rownames(sampleInfo(object)))
        }
    } else {
        if (dim(pertNumber(object))[[1]] > 0) {
            pertMatch <- match(dimnames(pertNumber(object))[[1]],
                rownames(sampleInfo(object)))
        }
        if (ncol(sensNumber(object)) > 0) {
            sensMatch <- match(rownames(sensNumber(object)),
                rownames(sampleInfo(object)))
        }
        curMatch <- match(rownames(curation(object)$sample),
            rownames(sampleInfo(object)))
    }
    if (dim(pertNumber(object))[[1]] > 0) {
        dimnames(pertNumber(object))[[1]] <- new.ids[pertMatch]
    }
    if (ncol(sensNumber(object)) > 0) {
        rownames(sensNumber(object)) <- new.ids[sensMatch]
    }
    rownames(curation(object)$sample) <- new.ids[curMatch]
    rownames(curation(object)$tissue) <- new.ids[curMatch]
    rownames(sampleInfo(object)) <- new.ids
    return(object)
}

# updateFeatureNames <- function(object, new.ids=vector("character")){
#
#   if (length(new.ids)!=nrow(sampleInfo(object))){
#     stop("Wrong number of sample identifiers")
#   }
#
#   if(datasetType(object)=="sensitivity"|datasetType(object)=="both"){
#     myx <- match(sensitivityInfo(object)[,"sampleid"],rownames(sampleInfo(object)))
#     sensitivityInfo(object)[,"sampleid"] <- new.ids[myx]
#
#   }
#
#   molecularProfilesSlot(object) <- lapply(molecularProfilesSlot(object), function(eset){
#
#     myx <- match(colData(eset)[["sampleid"]],rownames(sampleInfo(object)))
#     colData(eset)[["sampleid"]]  <- new.ids[myx]
#     return(eset)
#       })
#   myx <- match(rownames(curation(object)$sample),rownames(sampleInfo(object)))
#   rownames(curation(object)$sample) <- new.ids[myx]
#   rownames(curation(object)$tissue) <- new.ids[myx]
#   if (dim(pertNumber(object))[[1]]>0){
#     myx <- match(dimnames(pertNumber(object))[[1]], rownames(sampleInfo(object)))
#     dimnames(pertNumber(object))[[1]] <- new.ids[myx]
#   }
#   if (nrow(sensNumber(object))>0){
#     myx <- match(rownames(sensNumber(object)), rownames(sampleInfo(object)))
#     rownames(sensNumber(object)) <- new.ids[myx]
#   }
#   rownames(sampleInfo(object)) <- new.ids
#   return(object)
#
# }


### TODO:: Add updating of sensitivity Number tables
#' Update the treatment ids in a cSet object
#'
#' @examples
#' updateTreatmentId(clevelandSmall_cSet, treatmentNames(clevelandSmall_cSet))
#'
#' @param object The object for which the treatment ids will be updated
#' @param new.ids The new ids to assign to the object
#'
#' @return `CoreSet` The modified CoreSet object
#'
#' @keywords internal
#' @importFrom S4Vectors endoapply
#' @importFrom SummarizedExperiment colData rowData
#' @export
updateTreatmentId <- function(object, new.ids = vector('character')){

    if (nrow(treatmentInfo(object)) < 1) {
        message("No treatments in this object! Returning without modification.")
        return(object)
    }

    if (length(new.ids) != nrow(treatmentInfo(object))) {
        stop('Wrong number of drug identifiers')
    }
    if (datasetType(object) == 'sensitivity' || datasetType(object) == 'both') {
        myx <- match(sensitivityInfo(object)[, "treatmentid"], rownames(treatmentInfo(object)))
        sensitivityInfo(object)[, "treatmentid"] <- new.ids[myx]
    }
    if (datasetType(object) == 'perturbation' || datasetType(object) == 'both') {
        molecularProfilesSlot(object) <- lapply(molecularProfilesSlot(object),
                function(SE) {
            myx <- match(
                SummarizedExperiment::colData(SE)[["treatmentid"]],
                rownames(treatmentInfo(object))
            )
            SummarizedExperiment::colData(SE)[["treatmentid"]] <- new.ids[myx]
            return(SE)
        })
    }
    if (any(duplicated(new.ids))) {
        warning('Duplicated ids passed to updateTreatmentId. Merging old ids ',
            'into the same identifier')
        if (ncol(sensNumber(object)) > 0){
            sensMatch <- match(colnames(sensNumber(object)),
                rownames(treatmentInfo(object)))
        }
        if (dim(pertNumber(object))[[2]] > 0) {
            pertMatch <- match(dimnames(pertNumber(object))[[2]],
                rownames(treatmentInfo(object)))
        }
        if ("treatment" %in% names(curation(object))) {
            curMatch <- match(rownames(curation(object)$treatment),
                rownames(treatmentInfo(object)))
        }
        duplId <- unique(new.ids[duplicated(new.ids)])
        for(id in duplId) {
            if (ncol(sensNumber(object))>0){
                myx <- which(new.ids[sensMatch] == id)
                sensNumber(object)[, myx[1]] <- apply(sensNumber(object)[, myx], 1, sum)
                sensNumber(object) <- sensNumber(object)[, -myx[-1]]
                # sensMatch <- sensMatch[-myx[-1]]
            }
            if (dim(pertNumber(object))[[2]] > 0) {
                myx <- which(new.ids[pertMatch] == id)
                pertNumber(object)[,myx[1],] <- apply(pertNumber(object)[,myx,],
                    c(1,3), sum)
                pertNumber(object) <- pertNumber(object)[,-myx[-1], ]
                # pertMatch <- pertMatch[-myx[-1]]
            }
            if ("treatment" %in% names(curation(object))) {
                myx <- which(new.ids[curMatch] == id)
                curation(object)$treatment[myx[1], ] <-
                    apply(curation(object)$treatment[myx, ], 2, paste,
                        collapse='///')
                curation(object)$treatment <- curation(object)$treatment[-myx[-1], ]
                # curMatch <- curMatch[-myx[-1]]
            }

            myx <- which(new.ids == id)
            treatmentInfo(object)[myx[1],] <- apply(treatmentInfo(object)[myx,],
                2, paste, collapse='///')
            treatmentInfo(object) <- treatmentInfo(object)[-myx[-1], ]
            new.ids <- new.ids[-myx[-1]]
            if (ncol(sensNumber(object)) > 0) {
                sensMatch <- match(colnames(sensNumber(object)),
                    rownames(treatmentInfo(object)))
            }
            if (dim(pertNumber(object))[[2]] > 0) {
                pertMatch <- match(dimnames(pertNumber(object))[[2]],
                    rownames(treatmentInfo(object)))
            }
            if ("treatment" %in% names(curation(object))) {
                curMatch <- match(rownames(curation(object)$treatment),
                    rownames(treatmentInfo(object)))
            }
        }
    } else {
        if (dim(pertNumber(object))[[2]]>0){
            pertMatch <- match(dimnames(pertNumber(object))[[2]],
                rownames(treatmentInfo(object)))
        }
        if (ncol(sensNumber(object))>0){
            sensMatch <- match(colnames(sensNumber(object)),
                rownames(treatmentInfo(object)))
        }
        if ("treatment" %in% names(curation(object))) {
            curMatch <- match(rownames(curation(object)$treatment),
                rownames(treatmentInfo(object)))
        }
    }
    if (dim(pertNumber(object))[[2]]>0){
        dimnames(pertNumber(object))[[2]] <- new.ids[pertMatch]
    }
    if (ncol(sensNumber(object))>0){
        colnames(sensNumber(object)) <- new.ids[sensMatch]
    }
    if ("treatment" %in% names(curation(object))) {
        rownames(curation(object)$treatment) <- new.ids[curMatch]
    }
    rownames(treatmentInfo(object)) <- new.ids
    return(object)
}


.summarizeSensitivityNumbers <- function(object) {

    if (datasetType(object) != "sensitivity" && datasetType(object) != "both") {
        stop ("Data type must be either sensitivity or both")
    }

    ## unique drug identifiers
    # drugn <- sort(unique(treatmentResponse(object)$info[ , "treatmentid"]))

    ## consider all drugs
    drugn <- rownames(treatmentInfo(object))

    ## unique drug identifiers
    # samplen <- sort(unique(treatmentResponse(object)$info[ , "sampleid"]))

    ## consider all sample
    samplen <- rownames(sampleInfo(object))

    sensitivity.info <- matrix(0, nrow=length(samplen), ncol=length(drugn),
        dimnames=list(samplen, drugn))
    drugids <- sensitivityInfo(object)[, "treatmentid"]
    sampleids <- sensitivityInfo(object)[, "sampleid"]
    sampleids <- sampleids[grep("///", drugids, invert=TRUE)]
    drugids <- drugids[grep("///", drugids, invert=TRUE)]

    tt <- table(sampleids, drugids)
    sensitivity.info[rownames(tt), colnames(tt)] <- tt

    return(sensitivity.info)
}

#' @export
#' @keywords internal
.summarizeMolecularNumbers <- function(object) {

    ## consider all molecular types
    mDT <- mDataNames(object)

    ## consider all sample lines
    samplen <- rownames(sampleInfo(object))

    molecular.info <- matrix(0, nrow=length(samplen), ncol=length(mDT),
        dimnames=list(samplen, mDT))

    for(mDataType in mDT) {
        tt <- table(phenoInfo(object, mDataType)$sampleid)
        molecular.info[names(tt), mDataType] <- tt
    }
    return(molecular.info)
}

#' @importFrom SummarizedExperiment colData rowData
.summarizePerturbationNumbers <- function(object) {

    if (datasetType(object) != "perturbation" && datasetType(object) != "both") {
        stop ("Data type must be either perturbation or both")
    }

    ## consider all drugs
    drugn <- rownames(treatmentInfo(object))

    ## consider all sample lines
    samplen <- rownames(sampleInfo(object))

    perturbation.info <- array(0, dim=c(length(samplen), length(drugn),
        length(molecularProfilesSlot(object))),
        dimnames=list(samplen, drugn, names((molecularProfilesSlot(object)))))

    for (i in seq_len(length(molecularProfilesSlot(object)))) {
        if (nrow(colData(molecularProfilesSlot(object)[[i]])) > 0 &&
                all(is.element(c("sampleid", "treatmentid"),
                    colnames(colData(molecularProfilesSlot(object)[[i]]))))) {
            tt <- table(colData(molecularProfilesSlot(object)[[i]])[ , "sampleid"],
                colData(molecularProfilesSlot(object)[[i]])[ , "treatmentid"])
            perturbation.info[rownames(tt), colnames(tt),
                names(molecularProfilesSlot(object))[i]] <- tt
        }
    }

    return(perturbation.info)
}

#' A function to verify the structure of a CoreSet
#'
#' This function checks the structure of a PharamcoSet, ensuring that the
#' correct annotations are in place and all the required slots are filled so
#' that matching of samples and drugs can be properly done across different types
#' of data and with other studies.
#'
#' @examples
#' checkCsetStructure(clevelandSmall_cSet)
#'
#' @param object A `CoreSet` to be verified
#' @param plotDist Should the function also plot the distribution of molecular
#'   data?
#' @param result.dir The path to the directory for saving the plots as a string.
#'   Defaults to this R sessions `tempdir()`.
#'
#' @return Prints out messages whenever describing the errors found in the
#'   structure of the cSet object passed in.
#'
#' @export
#'
#' @md
#' @importFrom graphics hist
#' @importFrom grDevices dev.off pdf
#' @importFrom SummarizedExperiment assay rowData colData
#' @importFrom S4Vectors metadata
checkCsetStructure <- function(object, plotDist=FALSE, result.dir=tempdir()) {

    msg <- c()

    # Make directory to store results if it doesn't exist
    if (!file.exists(result.dir) && plotDist) {
        dir.create(result.dir, showWarnings=FALSE, recursive=TRUE)
    }

    ####
    ## Checking molecularProfiles
    ####
    for (i in seq_along(molecularProfilesSlot(object))) {
        profile <- molecularProfilesSlot(object)[[i]]
        nn <- names(molecularProfilesSlot(object))[i]

        # Testing plot rendering for rna and rnaseq
        if ((metadata(profile)$annotation == "rna" ||
                metadata(profile)$annotation == "rnaseq") && plotDist) {
            pdf(file=file.path(result.dir, sprintf("%s.pdf", nn)))
            hist(assay(profile, 'exprs'), breaks=100)
            dev.off()
        }

        ## Test if sample and feature annotations dimensions match the assay
        if (nrow(rowData(profile)) != nrow(assays(profile)$exprs)) {
            msg <- c(msg, paste0(nn, " number of features in rowData is ",
                "different from SummarizedExperiment slots"))
        }
        if (nrow(colData(profile)) != ncol(assays(profile)$exprs)) {
            msg <- c(msg, paste0(nn, "number of samples in colData is ",
                "different from expression slots", nn))
        }

        # Checking sample metadata for required columns
        if (!("sampleid" %in% colnames(colData(profile)))) {
            msg <- c(msg, paste0(nn, " sampleid does not exist in colData ",
                "(samples) columns"))
        }
        if (!("batchid" %in% colnames(colData(profile)))) {
            msg <- c(msg, sprintf(nn, " batchid does not exist in colData ",
                "(samples) columns"))
        }

        # Checking mDataType of the SummarizedExperiment for required columns
        if (metadata(profile)$annotation == "rna" ||
                metadata(profile)$annotation == "rnaseq") {
            if (!("BEST" %in% colnames(rowData(profile)))) {
                msg <- c(msg, paste0(nn, " BEST does not exist in rowData ",
                    "(features) columns"))
            }
            if (!("Symbol" %in% colnames(rowData(profile)))) {
                msg <- c(msg, paste0(nn, " Symbol does not exist in rowData ",
                    "(features) columns"))
            }
        }

        # Check that all sampleids from the cSet are included in molecularProfiles
        if ("sampleid" %in% colnames(rowData(profile))) {
            if (!all(colData(profile)[, "sampleid"] %in% rownames(sampleInfo(object)))) {
                msg <- c(msg, paste0(nn, " not all the sample lines in this ",
                    "profile are in sample lines slot"))
            }
        } else {
            msg <- c(msg, paste0(nn, " sampleid does not exist in colData ",
                "(samples)"))
        }
    }

    #####
    # Checking sample
    #####
    if ("tissueid" %in% colnames(sampleInfo(object))) {
        if ("unique.tissueid" %in% colnames(curation(object)$tissue)) {
            if (length(intersect(rownames(curation(object)$tissue),
                    rownames(sampleInfo(object)))) != nrow(sampleInfo(object))) {
                msg <- c(msg, paste0("rownames of curation tissue slot should",
                    " be the same as sample slot (curated sample ids)"))
            } else {
                if (length(intersect(sampleInfo(object)$tissueid,
                        curation(object)$tissue$unique.tissueid)) !=
                            length(table(sampleInfo(object)$tissueid))) {
                    msg <- c(msg, paste0("tissueid should be the same as unique",
                        " tissue id from tissue curation slot"))
                }
            }
        } else {
            msg <- c(msg, paste0("unique.tissueid which is curated tissue id",
                " across data set should be a column of tissue curation slot"))
        }
        if (any(is.na(sampleInfo(object)[,"tissueid"]) |
                sampleInfo(object)[, "tissueid"] == "", na.rm=TRUE)) {
            msg <- c(msg, paste0(
                    "There is no tissue type for these samples",
                    paste(
                        rownames(sampleInfo(object))[
                            which(is.na(sampleInfo(object)[,"tissueid"]) |
                                sampleInfo(object)[,"tissueid"] == "")
                            ],
                        collapse=" ")))
        }
    } else {
        msg <- c(msg, "tissueid does not exist in sample slot")
    }

    if("unique.sampleid" %in% colnames(curation(object)$sample)) {
        if (length(intersect(curation(object)$sample$unique.sampleid,
                rownames(sampleInfo(object)))) != nrow(sampleInfo(object))) {
            msg <- c(msg, "rownames of sample slot should be curated sample ids")
        }
    } else {
        msg <- c(msg, paste0("unique.sampleid which is curated sample id across",
            " data set should be a column of sample curation slot"))
    }

    if (length(intersect(rownames(curation(object)$sample),
            rownames(sampleInfo(object)))) != nrow(sampleInfo(object))) {
        msg <- c(msg, paste0("rownames of curation sample slot should be the",
            " same as sample slot (curated sample ids)"))
    }

    if (!is(sampleInfo(object), "data.frame")) {
        msg <- c(msg, "sample slot class type should be dataframe")
    }
    if (length(msg)) return(paste0(msg, collapse="\n")) else TRUE
}

#' @importFrom MultiAssayExperiment MultiAssayExperiment experiments
#' @importFrom S4Vectors List
#' @importFrom BiocGenerics %in% match
.checkMolecularProfiles <- function(object) {
    msg <- character()
    # ---- Make a MutliAssayExperiment, if it isn't one already
    molecProf <- molecularProfilesSlot(object)
    isSummarizedExperiment <- all(as(lapply(experiments(molecProf), is,
        'SummarizedExperiment'), 'List'))
    if (!all(isSummarizedExperiment)) {
        nmsg <- .formatMessage('All molecular profiles must be stored as
            SummarizedExperiment objects. The following are not ',
            paste(names(which(!isSummarizedExperiment)), collapse=', '))
        msg <- c(msg, nmsg)
    }
    tryCatch({
        MAE <- if (is(molecProf, 'MultiAssayExperiment')) molecProf else
            MultiAssayExperiment(molecProf)
    }, error=function(e) msg <- c(msg, paste0('Failed coercing to
        MultiAssayExperiment: ', as.character(e))))

    # ---- Check for correct metadata columns
    # -- sample identifiers
    colDataL <- lapply(experiments(MAE), FUN=colData)
    colColNameL <- as(lapply(colDataL, FUN=colnames), 'List')
    hasSampleId <- any(colColNameL %in%  'sampleid')
    if (!all(hasSampleId)) {
        nmsg <- .formatMessage('All SummarizedExperiments must have a sampleid
            column. This is not the case for ',
            paste(names(which(!hasSampleId)), collapse=', '), '!')
        msg <- c(msg, nmsg)
    }
    hasBatchId <- any(colColNameL %in% 'batchid')
    if (!all(hasBatchId)) {
        nmsg <- .formatMessage('All SummarizedExpeirments must have a batchid
            column. This is not the case for ',
            paste(names(which(!hasBatchId)), collapse=', '), '!')
        msg <- c(msg, nmsg)
    }
    # -- feature identifiers
    rowDataL <- lapply(experiments(MAE), FUN=rowData)
    rowColNameL <- as(lapply(rowDataL, colnames), 'List')
    # hasGeneId <- rowColNameL %in% 'geneid'
    hasSymbol <- rowColNameL %in% 'Symbol'
    hasBEST <- rowColNameL %in% 'BEST'
    # hasEnsemblId <- rowColNamesL %in% 'ensemblid'

    # ---- Check all samples are in the @sample slot
    samples <- sampleNames(object)

    
    sampleIdL <- as(lapply(colDataL, `[[`, i='sampleid'), 'List')
    hasValidSamples <- sampleIdL %in% samples
    if (!all(all(hasValidSamples))) {
        nmsg <- .formatMessage('All sampleids in the @molecularProfiles slot
            must also be in the @sample slot. This is not the case
            for ', paste(names(which(all(hasValidSamples))), collapse=', '))
        msg <- c(msg, nmsg)
    }

    # ---- Return messages if something is wrong, or TRUE if everything is good
    return(if (length(msg)) msg else TRUE)
}

.checkTreatmentResponse <- function(object) {
    msg <- character()
    # ---- Extract sensitivity data
    samples <- sampleNames(object)
    sensSlot <- treatmentResponse(object)
    if (!is(sensSlot, "TreatmentResponseExperiment")) {
        nmsg <- "The treatmentReponse parameter must be a
            TreatmentResponseExperiment!"
        msg <- c(msg, nmsg)
        return(msg)
    }
    return(if (length(msg)) msg else TRUE)
}

## ----setup, include=FALSE-----------------------------------------------------
library(knitr)
knitr::opts_chunk$set(echo = TRUE)

## ----coregx_load, hide=TRUE, message=FALSE------------------------------------
library(CoreGx)
library(Biobase)
library(SummarizedExperiment)

## ----CoreSet_class------------------------------------------------------------
getClass("CoreSet")

## ----CoreSet_accessors--------------------------------------------------------
methods(class="CoreSet")

## ----the_cSet_object----------------------------------------------------------
data(clevelandSmall_cSet)
clevelandSmall_cSet

## ----cSet_accessor_demo-------------------------------------------------------
mProf <- molecularProfiles(clevelandSmall_cSet, "rna")
mProf[seq_len(5), seq_len(5)]

## ----cSet_accessor_demo1------------------------------------------------------
cInfo <- sampleInfo(clevelandSmall_cSet)
cInfo[seq_len(5), seq_len(5)]

## ----cSet_accessor_demo2------------------------------------------------------
sensProf <- sensitivityProfiles(clevelandSmall_cSet)
sensProf[seq_len(5), seq_len(5)]

## ----defining_a_new_class-----------------------------------------------------
DemoSet <- setClass("DemoSet",
                    representation(demoSlot="character"),
                    contains="CoreSet")
getClass("DemoSet")

## ----inherit_methods----------------------------------------------------------
methods(class="DemoSet")

## ----defining_methods_for_a_new_class-----------------------------------------
clevelandSmall_dSet <- DemoSet(clevelandSmall_cSet)
class(clevelandSmall_dSet@molecularProfiles[['rna']])

expressionSets <- lapply(molecularProfilesSlot(clevelandSmall_dSet), FUN=as,
  'ExpressionSet')
molecularProfilesSlot(clevelandSmall_dSet) <- expressionSets

# Now this will error
tryCatch({molecularProfiles(clevelandSmall_dSet, 'rna')},
         error=function(e)
             print(paste("Error: ", e$message)))

## ----redefine_the_method------------------------------------------------------
setMethod(molecularProfiles,
          signature("DemoSet"),
          function(object, mDataType) {
            pData(object@molecularProfiles[[mDataType]])
          })

## ----testing_new_method-------------------------------------------------------
# Now we test our new method
mProf <- molecularProfiles(clevelandSmall_dSet, 'rna')
head(mProf)[seq_len(5), seq_len(5)]

## ----defining_setter_methods--------------------------------------------------
# Define generic for setter method
setGeneric('demoSlot<-', function(object, value) standardGeneric('demoSlot<-'))

# Define a setter method
setReplaceMethod('demoSlot',
                 signature(object='DemoSet', value="character"),
                 function(object, value) {
                   object@demoSlot <- value
                   return(object)
                 })

# Lets add something to our demoSlot
demoSlot(clevelandSmall_dSet) <- c("This", "is", "the", "demoSlot")

## ----defining_getter_methods--------------------------------------------------
# Define generic for getter method
setGeneric('demoSlot', function(object, ...) standardGeneric("demoSlot"))

# Define a getter method
setMethod("demoSlot",
          signature("DemoSet"),
          function(object) {
            paste(object@demoSlot, collapse=" ")
          })

# Test our getter method
demoSlot(clevelandSmall_dSet)

## ----sessionInfo--------------------------------------------------------------
sessionInfo()

#' @include LongTable-class.R
#'
#' @importFrom data.table key merge.data.table
#' @import checkmate
NULL

#' Merge assays with an `S4` object.
#'
#' @param object `S4` An S4 object a list-like slot containing assays for the
#'   object.
#' @param ... Allow new arguments to be defined for this generic.
#'
#' @return A modified version of `object`.
#'
#' @examples
#' "This is a generic method!"
#'
#' @exportMethod mergeAssays
setGeneric("mergeAssays", function(object, ...) standardGeneric("mergeAssays"))


#' Endomorphically merge assays within a `LongTable` or inheriting class
#'
#' @param object A `LongTable` or inheriting class.
#' @param x `character(1)` A valid assay name in `object`.
#' @param y `character(1)` A valid assay name in `object`.
#' @param target `character(1)` Name of the assay to assign the result to.
#' Can be a new or existing assay. Defaults to `x`.
#' @param ... Fallthrough arguments to merge.data.table to specify the join
#'   type. Use this to specify which columns to merge on. If excluded, defaults
#'   to by=assayKeys(objecty, y).
#' @param metadata `logical` A logical vector indicating whether to attach
#' metadata to either assay before the merge occurs. If only one value is
#' passed that value is used for both assays. Defaults to `FALSE`.
#'
#' @return A copy of `object` with assays `x` and `y` merged and assigned to
#' `target`.
#'
#' @seealso [`merge.data.table`]
#'
#' @author
#' Christopher Eeles
#'
#' @export
setMethod("mergeAssays", signature("LongTable"),
        function(object, x, y, target=x, ..., metadata=FALSE) {
    checkmate::qassert(target, "S1")
    z <- .merge_longtable_assays(object, x=x, y=y, ...,
        metadata=metadata)
    object[[target]] <- z
    object
})

#' @noRd
.merge_longtable_assays <- function(object, x, y, ..., metadata=FALSE) {
    # -- input validation
    checkmate::qassert(x, "S1")
    checkmate::qassert(y, "S1")
    checkmate::assertChoice(x, assayNames(object))
    checkmate::assertChoice(x, assayNames(object))
    checkmate::qassert(metadata, c("B1", "B2"))
    if (length(metadata) == 1) metadata <- c(metadata, metadata)

    # -- extract assays to merge
    x_ <- assay(object, x, summarize=TRUE, metadata=metadata[1])
    y_ <- if (x == y) x_ else
        assay(object, y, summarize=TRUE, metadata=metadata[2])

    # -- handle by argument
    by <- NA
    by_args <- c("by", "by.x", "by.y")
    if (!(any(by_args %in% ...names())))  {
        by <- assayKeys(object, y)
    }

    if (!is.character(by))
        merge.data.table(x=x_, y=y_, ...)
    else
        merge.data.table(x=x_, y=y_, by=by, ...)
}

#' @title List of `deprecated` or `defunct` methods in the `CoreGx` R package.
#'
#' @details
#' ## deprecated
#' `CoreSet`: The `CoreSet` constructor is being updated to have a new API. This
#' API is currently available via the `CoreSet2` constructor. In Bioconductor
#' 3.16, the old constructor will be renamed `CoreSet2` and the new constructor
#' will be renamed `CoreSet`.
#'
#' ## defunct
#' `buildLongTable`: This function no longer works as building a `LongTable` or
#' `TreatmentResponseExperiment` now uses a `DataMapper` and the `metaConstruct`
#' method. See `vignette("LongTable")` for a detailed description of how
#' to create a LongTable object.
#'
#' @name CoreGx-deprecated
#' @aliases
#' CoreGx-defunct
NULL


# ===== Deprecated


# ===== Defunct

# -- buildLongTableClass

# ==== LongTable Class

#' LongTable build method
#'
#' @description Create a LongTable object from a single data.table or
#'   data.frame object.
#'
#' @param from `character` Path to the .csv file containing the data and
#'   metadata from which to build the `LongTable`.
#' @param colDataCols `list` List with two `character` vectors, the first
#'   specifying one or more columns to be used as column identifiers (e.g.,
#'   drug name columns) and the second containing any additional metadata
#'   columns related to the column identifiers. If you wish to rename any of
#'   these columns, assign the new names to their respective character vectors.
#' @param rowDataCols `list` List with two `character` vectors, the first
#'   specifying one or more columns to be used as cell identifiers (e.g.,
#'   cell-line name columns) and the second containing any additional metadata
#'   columns related to the cell identifiers. If you wish to rename any of
#'   these columns, assign the new names to their respective character vectors.
#' @param assayCols `list` A named list of character vectors specifying how to
#'   parse assay columns into a list of `data.table`s. Each list data.table
#'   will be named for the name of corresponding list item and contain the columns
#'   specified in the character vector of column names in each list item. If
#'   there are no names for assayCols, the assays will be numbered by instead.
#'
#' @return A `LongTable` object containing one or more assays, indexed by
#'   rowID and colID.
#'
#' @import data.table
#' @export
setMethod('buildLongTable', signature(from='data.frame'),
        function(from, rowDataCols, colDataCols, assayCols) {

    .Defunct("metaConstruct", msg="This method has been deprecated
        in favour of use of the LongTableDataMapper metadata object
        with the metaConstruct method to build a LongTable object!")

    # -- local helpers
    .unlist <- function(x) unlist(x, recursive=TRUE, use.names=FALSE)

    # -- handle missing params
    missingParams <- c(missing(rowDataCols), missing(colDataCols), missing(assayCols))
    if (any(missingParams))
        stop(.errorMsg('[CoreGx::buildLongTable,data.frame] The following',
            ' parameters are required:',
            .collapse(c('rowDataCols', 'colDataCols', 'assayCols')[missingParams])))

    # -- validate input and return useful messages if invalid
    ## TODO:: Check input parameters are valid

    # -- convert to data.table by reference
    if (!is.data.table(from))
        from <- data.table(from)

    # -- build drug and cell metadata tables and index by the appropriate ID
    colData <- unique(from[, .unlist(colDataCols), with=FALSE])
    setorderv(colData, colDataCols[[1]])  # order by id columns
    colData[, colKey := seq_len(.N)]
    rowData <- unique(from[, .unlist(rowDataCols), with=FALSE])
    setorderv(rowData, rowDataCols[[1]])  # order by id columns
    rowData[, rowKey := seq_len(.N)]

    # -- add the row and column ids to the value data
    assayData <- from[rowData, on=.unlist(rowDataCols)][colData, on=as.character(unlist(colDataCols))]
    rm(from)
    assayData[, as.character(unique(c(.unlist(rowDataCols), .unlist(colDataCols)))) := NULL]
    # row reason to prevent sort in join because key sorts
    setkey(assayData, rowKey, colKey)

    setkey(rowData, rowKey)
    setkey(colData, colKey)

    # -- rename columns, if necessary
    rowDataColnames <- lapply(rowDataCols, names)
    notNullRownames <- !vapply(rowDataColnames, FUN=is.null, FUN.VALUE=logical(1))
    if (any(notNullRownames))
        for (i in which(notNullRownames)) {
            setnames(rowData, rowDataCols[[i]], names(rowDataCols[[i]]))
            rowDataCols[[i]] <- names(rowDataCols[[i]])
        }

    colDataColnames <- lapply(colDataCols, names)
    notNullColnames <- !vapply(colDataColnames, FUN=is.null, FUN.VALUE=logical(1))
    if (any(notNullColnames))
        for (i in which(notNullColnames)) {
            setnames(colData, colDataCols[[i]], names(colDataCols[[i]]))
            colDataCols[[i]] <- names(colDataCols[[i]])
        }

    # -- drop colKey or rowKey from assayCols, since we are adding it back in the
    # next step
    ## TODO:: Add a check to see if the keys are there to avoid dropping/re-adding
    .drop.in <- function(x, y) x[!(x %in% y)]
    assayCols <- lapply(assayCols, .drop.in, y=c('colKey', 'rowKey'))

    # -- add the index columns to the different assay column vectors
    # this allows the .selectDataTable helper to be more general
    .prependToVector <- function(vector, values) c(values, vector)
    assayCols <- lapply(assayCols, FUN=.prependToVector, values=c('rowKey', 'colKey'))
    if (is.null(names(assayCols))) names(assayCols) <- paste0('assay', seq_along(assayCols))
    assays <- lapply(assayCols, .selectDataTable, DT=assayData)

    # -- remove the colname suffixes by reference from assays which had the same
    # colnames prior to joining into a single DT
    for (assay in assays) {
        setnames(assay, colnames(assay), gsub('\\._\\d+$', '', colnames(assay)))
    }

    return(LongTable(rowData=rowData, rowIDs=rowDataCols[[1]],
                     colData=colData, colIDs=colDataCols[[1]],
                     assays=assays))
})

#' LongTable build method from character
#'
#' @description LongTable Create a LongTable object from a single .csv file
#'
#' @param from `character` Path to the .csv file containing the data and
#'   metadata from which to build the `LongTable`.
#' @param colDataCols `list` List with two `character` vectors, the first
#'   specifying one or more columns to be used as column identifiers (e.g.,
#'   drug name columns) and the second containing any additional metadata
#'   columns related to the column identifiers.
#' @param rowDataCols `list` List with two `character` vectors, the first
#'   specifying one or more columns to be used as cell identifiers (e.g.,
#'   cell-line name columns) and the second containing any additional metadata
#'   columns related to the cell identifiers.
#' @param assayCols `list` A named list of character vectors specifying how to
#'   parse assay columns into a list of `data.table`s. Each list data.table
#'   will be named for the name of corresponding list item and contain the columns
#'   specified in the character vector of column names in each list item.
#'
#' @return A `LongTable` object containing one or more assays, indexed by
#'   rowID and colID.
#'
#' @import data.table
#' @importFrom crayon magenta
#' @export
setMethod('buildLongTable', signature(from='character'),
          function(from, rowDataCols, colDataCols, assayCols)
{
    if (length(from) > 1)  # Call list subsetting method
        buildLongTable(as.list(from), rowDataCols, colDataCols, assayCols)
    if (!file.exists(from))
        stop(magenta$bold("The is no file at path: ", from, '. Please double
            check the location of the source file!'))

    # read in data
    tableData <- .freadNA(from)

    return(buildLongTable(from=tableData, rowDataCols, colDataCols, assayCols))
})

#' LongTable build method from list
#'
#' @description Create a LongTable object from a list containing file paths,
#'   data.frames and data.tables.
#'
#' @examples
#' \dontrun{
#' assayList <- assays(merckLongTable, withDimnames=TRUE)
#' rowDataCols <- list(rowIDs(merckLongTable), rowMeta(merckLongTable))
#' colDataCols <- list(colIDs(merckLongTable), colMeta(merckLongTable))
#' assayCols <- assayCols(merckLongTable)
#' longTable <- buildLongTable(from=assayList, rowDataCols, colDataCols, assayCols)
#' }
#'
#' @param from `list` A list containing any combination of character file paths,
#'  data.tables and data.frames which will be used to construct the LongTable.
#' @param colDataCols `list` List with two `character` vectors, the first
#'   specifying one or more columns to be used as column identifiers (e.g.,
#'   drug name columns) and the second containing any additional metadata
#'   columns related to the column identifiers.
#' @param rowDataCols `list` List with two `character` vectors, the first
#'   specifying one or more columns to be used as cell identifiers (e.g.,
#'   cell-line name columns) and the second containing any additional metadata
#'   columns related to the cell identifiers.
#' @param assayCols `list` A named list of character vectors specifying how to
#'   parse assay columns into a list of `data.table`s. Each list data.table
#'   will be named for the name of corresponding list item and contain the columns
#'   specified in the character vector of column names in each list item.
#'
#' @return A `LongTable` object constructed with the data in `from`.
#'
#' @import data.table
#' @importFrom crayon magenta cyan
#' @export
setMethod('buildLongTable', signature(from='list'),
        function(from, rowDataCols, colDataCols, assayCols) {

    # Prevent modify by reference for data.tables in list
    from <- copy(from)

    # local helpers
    ##FIXME:: This is exactly what the Map function is (an alias for mapply with
    ##   SIMPLIFY=FALSE)
    .mapply <- function(...) mapply(..., SIMPLIFY=FALSE)

    # preprocess from list
    isChar <- is.items(from, 'character')
    isDT <- is.items(from, FUN=is.data.table)
    isDF <- is.items(from, FUN=is.data.frame) & !isDT

    if (!all(isChar | isDT | isDF))
        stop(.errorMsg('\n[CoreGx::buildLongTable,list-method] List items at',
            ' indexes ', .collapse(which(!(isChar | isDT | isDF ))),
            ' are not character, data.table or data.frame.', collapse=', '))

    if (any(isChar)) from <- c(from[!isChar], lapply(from[isChar], FUN=.freadNA))
    if (any(isDF)) for (i in which(isDF)) from[[i]] <- data.table(from[[i]])

    # validate mappings
    ## TODO:: Ensure there is no case where joining on rowMeta or colMeta gives
    #  different results than just ids
    joinCols <- unique(unlist(c(rowDataCols, colDataCols)))
    dataColNames <- lapply(from, FUN=colnames)
    joinColsIn <- lapply(dataColNames, `%in%`, x=joinCols)
    hasAllIdCols <- unlist(lapply(joinColsIn, FUN=all))
    if (!all(hasAllIdCols)) {
        missingCols <- unique(unlist(.mapply(`[`, x=joinCols, i=joinColsIn)))
        stop(.errorMsg('[CoreGx::buildLongTable,list] Assay(s) ',
            .collapse(which(hasAllIdCols)), ' are missing one or more id ',
            'columns: ', .collapse(missingCols), collapse=', '))
    }

    # Set keys for faster joins
    for (i in seq_along(from)) setkeyv(from[[i]], cols=joinCols)

    # join assays into a single table
    DT <- from[[1]]
    from[[1]] <- NULL
    for (i in seq_along(from))
        DT <- merge.data.table(DT, from[[i]], suffixes=c('', paste0('._', i)))

    # fix assayCols if there are duplicate column names between assays
    # the join will append '._n' where n is the assay index - 1
    nonDataCols <- setdiff(colnames(DT), unique(c(unlist(rowDataCols), unlist(colDataCols))))
    assaySuffixCols <- lapply(paste0('\\._', seq_along(from)), grep, x=nonDataCols, value=TRUE)
    .length.gt.0 <- function(x) length(x) > 0
    hasSuffixes <- unlist(lapply(assaySuffixCols, FUN=.length.gt.0))
    duplicatedCols <- lapply(assaySuffixCols[hasSuffixes], gsub,
        pattern='\\._\\d+', replacement='')

    .which.in <- function(x, y) which(x %in% y)
    whichHasSuffixes <- which(hasSuffixes) + 1
    whichDuplicated <- .mapply(.which.in,
        x=assayCols[whichHasSuffixes], y=duplicatedCols)
    assayCols[whichHasSuffixes] <-
        .mapply(replace, x=assayCols[whichHasSuffixes],
            list=whichDuplicated, values=assaySuffixCols[hasSuffixes])

    # construct new LongTable
    buildLongTable(from=DT, rowDataCols, colDataCols, assayCols)
})


# ---- Helper Methods

#' fread with more default na.strings
#'
#' @keywords internal
#' @noRd
.freadNA <- function(...) {
    as.na <- unique(c(getOption('datatable.na.string'),
        c('NA', 'NULL', 'NaN', 'missing', 'None',
            'none', 'na', 'null', 'Null', 'Na')))
    fread(..., na.strings=as.na)
}


#' Select a set of column names from a data.table, returning a copy of the
#'   data.table with duplicate rows removed
#'
#' @param colNames `character` The column names to select from the data.table
#' @param DT `data.table`, `data.frame`, `matrix` An object coercible to a `data.table`.
#'   Please note rownames will be dropped by default.
#' @param keep.rownames `logical` or `character` Passed through to the data.table coercing if DT is not a
#'   `data.table`. If TRUE, rownames will be caputured in the `rn` column; if FALSE (default) rownames will
#'   be dropped; if `character`, rownames will be captured in a column with the same name.
#'
#' @return `data.table` Copy of `DT` containing only the specified columns, with duplicate rows removed.
#'
#' @import data.table
#' @keywords internal
#' @noRd
.selectDataTable <- function(colNames, DT, keep.rownames=FALSE) {
    # validate input
    if (!is.data.table(DT)) {
        tryCatch({
            DT <- data.table(DT, keep.rownames=keep.rownames)
        }, warning=function(w) {
            warning(w)
        }, error=function(e) {
            message(e)
            stop("Argument to DT parameter must be coercible to a data.table!")
        })
    }
    if (!is.character(colnames(DT))) stop("Currently only character column ids are supported!")
    missingColumns <- setdiff(colNames, colnames(DT))
    if (length(missingColumns) > 0)
        warning(paste0("There are no columns named ", paste0(missingColumns, collapse=", "), 'in DT.
            Continuing subset without these columns.'))

    # perform subset and copy to prevent modify by refence issues
    selectedDT <- copy(unique(DT[, .SD, .SDcols=colnames(DT) %in% colNames]))

    return(selectedDT)
}

#' Update the `CoreSet` class after changes in it struture or API
#'
#' @param object A `CoreSet` object to update the class structure for.
#' @param verify A `logical(1)` indicating is `validObject` should be called
#' after updating the object. Defaults to `TRUE`, only set `FALSE` for debugging.
#' @param verbose TRUE or FALSE, indicating whether information about the update
#' should be reported
#' @return `CoreSet` with update class structure.
#'
#' @md
#'
#' @importFrom MultiAssayExperiment MultiAssayExperiment
#' @importMethodsFrom BiocGenerics updateObject
#' @export
setMethod('updateObject', signature(object="CoreSet"),
        function(object, verify=FALSE, verbose = FALSE) {

    if (verbose) {
        message("updateObject object = 'CoreSet'")
    }
    
    if (!.hasSlot(object, "sample")) {
        cell <- object@cell
        sample_ <- cell
    } else {
        sample_ <- object@sample
    }
    colnames(sample_) <- gsub("cellid", "sampleid", colnames(sample_))

    if (!.hasSlot(object, "treatment")) {
        if (.hasSlot(object, "drug")) {
            treatment <- object@drug
        } else if (.hasSlot(object, "radiation")) {
            treatment <- object@radiation
        } else  {
            treatment <- data.frame()
        }
    } else {
        treatment <- object@treatment
    }
    colnames(treatment) <- gsub("drugid", "treatmentid", colnames(treatment))

    if (!.hasSlot(object, "treatmentResponse")) {
        treatmentResponse <- object@sensitivity
    } else {
        treatmentResponse <- object@treatmentResponse
    }

    if (is(treatmentResponse, "LongTable")) {
        treatmentResponse <- updateObject(treatmentResponse)
        mutableIntern <- mutable(getIntern(treatmentResponse))
    } else {
        colnames(treatmentResponse$info) <- gsub("cellid", "sampleid",
            colnames(treatmentResponse$info))
        colnames(treatmentResponse$info) <- gsub("drugid",
            "treatmentid", colnames(treatmentResponse$info))
    }

    mProf <- object@molecularProfiles
    for (i in seq_along(mProf)) {
        colnames(colData(mProf[[i]])) <- gsub("cellid", "sampleid",
            colnames(colData(mProf[[i]])))
        colnames(colData(mProf[[i]])) <- gsub("drugid", "treatmentid",
            colnames(colData(mProf[[i]])))
    }
    curation_ <- object@curation
    names(curation_) <- gsub("cell", "sample", names(curation_))
    names(curation_) <- gsub("drug", "treatment", names(curation_))
    colnames(curation_$sample) <- gsub("cellid", "sampleid",
        colnames(curation_$sample))
    if ("treatment" %in% names(curation_)) {
        colnames(curation_$treatment) <- gsub("drugid",
            "treatmentid", colnames(curation_$treatment))
    }

    cSet <- .CoreSet(
        sample=sample_,
        treatment=treatment,
        treatmentResponse=treatmentResponse,
        molecularProfiles=mProf,
        annotation=object@annotation,
        curation=curation_,
        perturbation=object@perturbation,
        datasetType=object@datasetType
    )

    if (verify) isValid(cSet)

    return(cSet)
})

#' Update the `LongTable` class after changes in it struture or API
#'
#' @param object A `LongTable` object to update the class structure for.
#' @param verify A `logical(1)` indicating is `validObject` should be called
#' after updating the object. Defaults to `TRUE`, only set `FALSE` for debugging.
#' @param verbose TRUE or FALSE, indicating whether information about the update
#' should be reported
#' @return `LongTable` with update class structure.
#'
#' @md
#'
#' @importMethodsFrom BiocGenerics updateObject
#' @export
setMethod("updateObject", signature(object="LongTable"),
        function(object, verify=FALSE, verbose = FALSE) {

     if (verbose) {
        message("updateObject object = 'CoreSet'")
    }
    if (is.environment(getIntern(object))) {
        rData <- rowData(object, key=TRUE)
        rIDs <- rowIDs(object)
        cData <- colData(object, key=TRUE)
        cIDs <- colIDs(object)
        id_cols <- c(rIDs, cIDs)
        assays_ <- assays(object, raw=TRUE)
        setkeyv(rData, "rowKey")
        assays_ <- lapply(assays_, merge.data.table,
            y=rData[, c("rowKey", rIDs), with=FALSE],
            by="rowKey"
        )
        setkeyv(cData, "colKey")
        assays_ <- lapply(assays_, merge.data.table,
            y=cData[, c("colKey", cIDs), with=FALSE],
            by="colKey"
        )
        rData[, rowKey := NULL]
        cData[, colKey := NULL]
        for (a_ in assays_) a_[, c("rowKey", "colKey") := NULL]
        mdata <- metadata(object)
        assayMap <- lapply(assays_, function(x, y) y, y=id_cols)
        oclass <- class(object)[1]
        object <- LongTable(
            rowData=rData, rowIDs=rIDs,
            colData=cData, colIDs=cIDs,
            assays=assays_, assayIDs=assayMap,
            metadata=mdata
        )
        object <- as(object, oclass)  # Coerce to inherting class if needed
    }
    if (verify) isValid(object)
    return(object)
})

# Navigating this file:
# - Slot section names start with ----
# - Method section names start with ==
#
# As a result, you can use Ctrl + f to find the slot or method you are looking
# for quickly, assuming you know its name.
#
# For example Ctrl + f '== molecularProfiles' would take you the molecularProfiles
# method, while Ctrl + f '---- molecularProfiles' would take you to the slot
# section.

#' @include LongTable-class.R allGenerics.R
NULL

#### CoreGx dynamic documentation
####
#### Warning: for dynamic docs to work, you must set
#### Roxygen: list(markdown = TRUE, r6=FALSE)
#### in the DESCRPTION file!

.local_class_4 <- 'LongTable'
.local_data_4 <- 'merckLongTable'


# =======================================
# Accessor Method Documentation Object
# ---------------------------------------


#' @noRd
.docs_LongTable_accessors <- function(...) .parseToRoxygen(
    "
    @title Accessing and modifying information in a `{class_}`

    @description
    Documentation for the various setters and getters which allow manipulation
    of data in the slots of a `{class_}` object.

    @return Accessors: See details.
    @return Setters: An updated `{class_}` object, returned invisibly.
    ",
    ...
)

#' @name LongTable-accessors
#' @eval .docs_LongTable_accessors(class_=.local_class_4)
#' @eval .parseToRoxygen("@examples data({data_})", data_=.local_data_4)
NULL


# ======================================
# Accessor Methods
# --------------------------------------


## ==================
## ---- .intern Slot
## ------------------


##
## == getIntern


#' Get the symbol(s) x from the object@.intern slot of a LongTable
#'
#' This is used as an alternative to R attributes for storing structural
#' metadata of an S4 objects.
#'
#' @examples
#' getIntern(merckLongTable, 'rowIDs')
#' getIntern(merckLongTable, c('colIDs', 'colMeta'))
#'
#' @describeIn LongTable Access structural metadata present within a
#'   LongTable object. This is mostly for developmer use.
#'
#' @param object `LongTable`
#' @param x `character` One or more symbol name strings to retrieve from
#'     the object@.intern environment.
#'
#' @return `immutable` value of x if length(x) == 1 else named list of values
#'     for all symbols in x.
#'
#' @include LongTable-class.R
#' @export
setMethod('getIntern', signature(object='LongTable', x='character'),
        function(object, x) {
    return(if (length(x) == 1) object@.intern[[x]] else object@.intern[x])
})
#' @describeIn LongTable Access all structural metadata present within a
#'   LongTable object. This is primarily for developmer use.
#'
#' @param object `LongTable`
#' @param x `missing` This argument is excluded from from the function call.
#'
#' @return An `immutable` list.
#'
#' @examples
#' getIntern(merckLongTable)
#'
#' @aliases getIntern,LongTable,missing-method
#' @export
setMethod('getIntern', signature(object='LongTable', x='missing'),
    function(object, x) object@.intern
)

#' Set the .intern slot of a LongTable
#'
#' @param object `LongTable`
#' @param value An `immutable_list` object, being a class union between `list`
#'   and `immutable` S3 classes.
#'
#' @return Updates the object and returns invisibly.
#'
#' @keywords internal
setReplaceMethod("getIntern", signature(object="LongTable",
    value="immutable_list"), function(object, value) {
        object@.intern <- value
        return(object)
})

## ==================
## ---- rowData Slot
## ------------------

#' Retrieve the row metadata table from a LongTable object
#'
#' @examples
#' rowData(merckLongTable)
#'
#' @describeIn LongTable Get the row level annotations for a `LongTable` object.
#'
#' @param x A `LongTable` object to retrieve the row metadata from.
#' @param key `logical` Should the rowKey column also be returned? Defaults
#'     to FALSE.
#' @param use.names `logical` This parameter is just here to stop matching
#'     the positional argument to use.names from the rowData generic. It
#'     doesn't do anything at this time and can be ignored.
#' @param ... For developer use only! Pass raw=TRUE to modify the slot
#'   directly. This will corrupt your data if you don't know what you are
#'   doing!
#'
#' @return A `data.table` containing rowID, row identifiers, and row metadata.
#'
#' @importFrom data.table data.table copy
#' @export
setMethod('rowData', signature(x='LongTable'),
        function(x, key=FALSE, use.names=FALSE, ...) {
    if (any(...names() == "raw") && isTRUE(...elt(which(...names() == "raw")))) {
        return(x@rowData)
    } else {
        return(if (key) copy(x@rowData[, -'.rownames']) else
            copy(x@rowData[, -c('.rownames', 'rowKey')]))
    }
})

#' Helper method to share functionality between rowData and colData replace methods
#'
#' @param x `LongTable` or inheriting class to update dimData for.
#' @param dim `character(1)` One of "row" or "col" indicating with dimension
#'   to updated metadata for.
#' @param value #' @param value A `data.table` or `data.frame` to update the
#'   `rowData` or `colData` of `x` with.
#'
#' @return An updated version of `value` which meets all the requirements for
#'   assignment to a `LongTable` or inheriting class.
#'
#' @noRd
#' @keywords internal
.update_dimData <- function(x, dim, value) {

    titleDim <- paste0(toupper(substr(dim, 1, 1)), substr(dim, 2, nchar(dim)))
    dimIDs <- get(paste0(dim, "IDs"))
    dimKey <- paste(dim, "Key")
    dimData <- paste0(dim, "Data")

    # type check input
    if (is(value, 'data.frame')) setDT(value)
    if (!is(value, 'data.table'))
        stop(.errorMsg('\n[CoreGx::', dim, 'Data<-] Please pass a data.frame or ',
            'data.table to update the ', dim, 'Data slot. We recommend modifying the',
            ' object returned by ', dim, 'Data(x) then reassigning it with ',
            dim, 'Data(x)',
            ' <- new', titleDim, 'Data'),
            call.=FALSE
        )

    # remove key column
    if (dimKey %in% colnames(value)) {
        value[, (dimKey) := NULL]
        .message('\n[CoreGx::', dim, ,'Data<-] Dropping ', dim, 'Key from replacement',
            ' value, this function will deal with mapping the ', dim, 'Key',
            ' automatically.')
    }

    # assemble information to select proper update method
    dimIDCols <- dimIDs(x)
    sharedDimIDCols <- intersect(dimIDCols, colnames(value))

    # error if all the row/colID columns are not present in the new row/colData
    equalDimIDs <- dimIDCols %in% sharedDimIDCols
    if (!all(equalDimIDs)) warning(.warnMsg('\n[CoreGx::', dim,
        'Data<-] The ID columns ', dimIDCols[!equalDimIDs],
        ' are not present in value. The function ',
        'will attempt to join with existing ', dim, 'IDs, but this may fail!',
        collapse=', '), call.=FALSE)

    dimIDs_ <- dimIDs(x, data=TRUE, key=TRUE)

    ## TODO:: Throw error if user tries to modify ID columns

    duplicatedIDcols <- value[, .N, by=c(sharedDimIDCols)][, N > 1]
    if (any(duplicatedIDcols))
        warning(.warnMsg("\n[CoreGx::", dim, "Data<-,", class(x)[1], "-method] The ",
            "ID columns are duplicated for rows ",
            .collapse(which(duplicatedIDcols)),
            "! These rows will be dropped before assignment."),
        call.=FALSE)

    dimData <- dimIDs_[unique(value), on=.NATURAL, allow.cartesian=FALSE]
    dimData[, (dimKey) := .I]
    dimData <- dimData[!duplicated(get(dimKey)), ]
    setkeyv(dimData, dimKey)
    dimData[, .rownames := Reduce(.paste_colon, mget(dimIDCols))]

    ## TODO:: Add some sanity checks before returing

    return(dimData)
}


#' Updates the `rowData` slot as long as the ID columns are not changed.
#'
#' @examples
#' rowData(merckLongTable) <- rowData(merckLongTable)
#'
#' @describeIn LongTable Update the row annotations for a `LongTable` object.
#'   Currently requires that all columns in rowIDs(longTable) be present in
#'   value.
#'
#' @param x A `LongTable` object to modify.
#' @param value A `data.table` or `data.frame` to update the `rowData` of
#'   `x` with.
#' @param ... For developer use only! Pass raw=TRUE to modify the slot
#'   directly. This will corrupt your data if you don't know what you are
#'   doing!
#'
#' @return A copy of the `LongTable` object with the `rowData`
#'   slot updated.
#'
#' @md
#' @importFrom crayon cyan magenta
#' @importFrom SummarizedExperiment `rowData<-`
#' @importFrom data.table setDT
#' @export
setReplaceMethod('rowData', signature(x='LongTable'), function(x, ..., value) {

    if (any(...names() == "raw") && isTRUE(...elt(which(...names() == "raw")))) {
        x@rowData <- value
        return(invisible(x))
    }

    x@rowData <- .update_dimData(x=x, dim="row", value=value)
    return(invisible(x))
})


## ==================
## ---- colData Slot
## ------------------


#' Retrieve the column metadata table from a LongTable object
#'
#' @examples
#' colData(merckLongTable)
#'
#' # Get the keys as well, mostly for internal use
#' colData(merckLongTable, key=TRUE)
#'
#' @describeIn LongTable Get the column level annotations for a LongTable
#'   object.
#'
#' @param x A `LongTable` to retrieve column metadata from.
#' @param key `logical` Should the colKey column also be returned? Defaults to
#'     FALSE.
#' @param ... For developer use only! Pass raw=TRUE to return the slot for
#'   modification by reference.
#'
#' @return A `data.table` containing row identifiers and metadata.
#'
#' @import data.table
#' @export
setMethod('colData', signature(x='LongTable'),
        function(x, key=FALSE, dimnames=FALSE, ...) {
    if (any(...names() == "raw") && isTRUE(...elt(which(...names() == "raw")))) {
        return(x@colData)
    }
    return(if (key) copy(x@colData[, -'.colnames']) else
        copy(x@colData[, -c('.colnames', 'colKey')]))
})

#' Updates the `colData` slot as long as the ID columns are not changed.
#'
#' @examples
#' colData(merckLongTable) <- colData(merckLongTable)
#'
#' @describeIn LongTable Update the colData of a LongTable object. Currently
#'   requires that all of the colIDs(longTable) be in the value object.
#'
#' @param x A `LongTable` object to modify.
#' @param value A `data.table` or `data.frame` to update with. Must have
#'   all of the colIDs currently in the `LongTable` object in order to ensure
#'   assay key mappings are consistent.
#' @param ... For developer use only! Pass raw=TRUE to modify the slot
#'   directly. This will corrupt your data if you don't know what you are
#'   doing!
#'
#' @return A copy of the `LongTable` object with the `colData`
#'   slot updated.
#'
#' @importFrom crayon cyan magenta
#' @importFrom SummarizedExperiment colData<-
#' @importFrom data.table data.table setDT
#' @export
setReplaceMethod('colData', signature(x='LongTable'),
        function(x, ..., value) {
    if (any(...names() == "raw") && isTRUE(...elt(which(...names() == "raw")))) {
        x@colData <- value
        return(invisible(x))
    }
    x@colData <- .update_dimData(x=x, dim="col", value=value)
    return(invisible(x))
})

## ==================
## ---- assaySlot
## ------------------


##
## == assays


#' Return a list of `data.table` objects with the assay measurements from a
#'  `LongTable` object.
#'
#' @examples
#' assays(merckLongTable)
#'
#' @describeIn LongTable Get a list containing all the assays in a `LongTable`.
#'
#' @param x `LongTable` What to extract the assay data from.
#' @param withDimnames `logical` Should the returned assays be joined to
#'   the row and column identifiers (i.e., the pseudo-dimnames of the object).
#' @param metadata `logical` Should row and column metadata also be joined
#'   to the returned assays. This is useful for modifying assays before
#'   reconstructing a new LongTable.
#' @param key `logical` Should the key columns also be returned? Defaults
#'   to !`withDimnames`.
#' @param ... For developer use only! Pass raw=TRUE to return the slot for
#'   modification by reference.
#'
#' @return A `list` of `data.table` objects, one per assay in the object.
#'
#' @importMethodsFrom SummarizedExperiment assays
#' @import data.table
#' @export
setMethod('assays', signature(x='LongTable'), function(x, withDimnames=TRUE,
        metadata=withDimnames, key=!withDimnames, ...) {
    # secret arguments for internal use
    if (any(...names() == "raw") && isTRUE(...elt(which(...names() == "raw")))) {
        return(x@assays)
    }

    # input validation
    if (!withDimnames && metadata)
        warning(.warnMsg('[CoreGx::assays] Cannot use metadata=TRUE when',
            ' withDimnames=FALSE. Ignoring the metadata argument.'),
            call.=FALSE)

    # optionally join with rowData and colData
    assayIndex <- assayIndex(x)
    if (metadata) {
        rData <- rowData(x, key=TRUE)
        cData <- colData(x, key=TRUE)
    } else {
        rData <- rowIDs(x, data=TRUE, key=TRUE)
        cData <- colIDs(x, data=TRUE, key=TRUE)
    }
    if (withDimnames) {
        setkeyv(assayIndex, "rowKey")
        assayIndex <- rData[assayIndex, , on="rowKey"]
        setkeyv(assayIndex, "colKey")
        assayIndex <- cData[assayIndex, , on="colKey"]
    }

    # honor row and column ordering guarantees from CoreGx design documentation
    aList <- copy(x@assays)
    aNames <- names(aList)
    # prepend with . to match assay index naming convention
    names(aList) <- paste0(".", aNames)
    corder <- c(
        if (withDimnames) idCols(x),
        if (key) c("rowKey", "colKey"),
        if (withDimnames && metadata) c(sort(rowMeta(x)), sort(colMeta(x)))
    )
    for (i in seq_along(aList)) {
        setkeyv(assayIndex, names(aList)[i])
        aList[[i]] <- assayIndex[aList[[i]], ]
        deleteCols <- setdiff(names(aList), names(aList)[i])
        if (length(deleteCols) > 0) {
            aList[[i]][, (deleteCols) := NULL]
        }
        if (withDimnames || key) aList[[i]][, (names(aList)[i]) := NULL]
        if (!key) {
            aList[[i]][, c("rowKey", "colKey") := NULL]
        }
        if (withDimnames) setkeyv(aList[[i]], idCols(x))
        else if (key) setkeyv(aList[[i]], c("rowKey", "colKey"))
        else setkeyv(aList[[i]], names(aList)[[i]])
        if (!is.null(corder)) setcolorder(aList[[i]], corder) else
            setcolorder(aList[[i]])
    }
    # reset names to non dot version
    names(aList) <- aNames
    return(aList)
})


#' Setter method for the assays slot in a LongTable object
#'
#' @examples
#' assays(merckLongTable) <- assays(merckLongTable, withDimnames=TRUE)
#'
#' @describeIn LongTable Update the assays in a LongTable object. The rowIDs
#'   and colIDs must be present in all assays to allow successfully remapping
#'   the keys. We recommend modifying the list returned by
#'   assays(longTable, withDimnames=TRUE) and the reassigning to the
#'   `LongTable`.
#'
#' @param x A `LongTable` to modify the assays in.
#' @param value A `list` of `data.frame` or `data.table` objects, all of which
#'   contain the row and column identifiers and metadata.
#' @param ... For developer use only! Pass raw=TRUE to modify the slot
#'   directly. This will corrupt your data if you don't know what you are
#'   doing!
#'
#' @return A copy of the `LongTable` with the assays modified.
#'
#' @importMethodsFrom SummarizedExperiment assays<-
#' @import data.table
#' @export
setReplaceMethod('assays', signature(x='LongTable', value='list'),
        function(x, ..., value) {
    if (any(...names() == "raw") && isTRUE(...elt(which(...names() == "raw")))) {
        x@assays <- value
    } else {
        assay_names <- names(value)
        for (name in assay_names) {
            x[[name]] <- value[[name]]
        }
    }
    return(x)
})


##
## == assay


#' Get an assay from a LongTable object
#'
#' @describeIn LongTable Retrieve an assay `data.table` object from the
#'   `assays` slot of a `LongTable` object.
#'
#' @examples
#' # Default annotations, just the key columns
#' assay(merckLongTable, 'sensitivity')
#' assay(merckLongTable, 1)
#'
#' # With identifiers joined
#' assay(merckLongTable, 'sensitivity', withDimnames=TRUE)
#'
#' # With identifiers and metadata
#' assay(merckLongTable, 'profiles', withDimnames=TRUE, metadata=TRUE)
#'
#' @param x `LongTable` The `LongTable` object to get the assay from.
#' @param i `integer` or `character` vector containing the index or name
#'   of the assay, respectively.
#' @param withDimnames `logical(1)` Should the dimension names be returned
#'   joined to the assay. This retrieves both the row and column identifiers
#'   and returns them joined to the assay. For
#' @param summarize `logical(1)` If the assays is a summary where some of
#'   `idCols(x)` are not in `assayKeys(x, i)`, then those missing columns
#'   are dropped. Defaults to `FALSE`. When `metadata` is `TRUE`, only
#'   metadata columns with 1:1 cardinality with the assay keys for `i`.
#' @param metadata `logical(1)` Should all of the metadata also be joined to
#'   the assay. This is useful when modifying assays as the resulting list
#'   has all the information needed to recreated the LongTable object. Defaults
#'   to `withDimnames`.
#' @param key `logical` Should the key columns also be returned? Defaults to
#'   !withDimnames. This is incompatible with `summarize=TRUE`, which will
#'   drop the key columns regardless of the value of this argument.
#' @param ... For developer use only! Pass raw=TRUE to return the slot for
#'   modification by reference.
#'
#' @importMethodsFrom SummarizedExperiment assay
#' @importFrom crayon magenta cyan
#' @import data.table
#' @export
setMethod('assay', signature(x='LongTable'), function(x, i, withDimnames=TRUE,
        summarize=withDimnames, metadata=!summarize,
        key=!(summarize || withDimnames), ...) {
    # secret arguments for internal use
    if (any(...names() == "raw") && isTRUE(...elt(which(...names() == "raw")))) {
        return(x@assays[[i]])
    }

    ## TODO:: Update input validation to use checkmate where possible
    # validate input
    if (length(i) > 1)
        .error('\n[CoreGx::assay] Please specifying a single string ',
            'assay name or integer index. See assayNames(x) for available ',
            'assays.')

    keepAssay <- if (is.character(i)) which(assayNames(x) == i) else i
    assayName <- assayNames(x)[keepAssay]
    .assayName <- paste0(".", assayName)
    if (length(keepAssay) < 1)
        stop(.errorMsg('\n[CoreGx::assay] There is no assay ', i,
            ' in this LongTable. Use assayNames(longTable) for a list',
            'of valid assay names.'),
            call.=FALSE)

    if (!withDimnames && metadata)
        warning(.warnMsg('\n[CoreGx::assay] Cannot use metadata=TRUE when',
            ' withDimnames=FALSE. Ignoring the metadata argument.'),
            call.=FALSE)

    if (summarize && key)
        warning(.warnMsg('\n[CoreGx::assay] Cannot use key=TRUE when',
            ' summarize=TRUE. Ignoring the key argument.'),
            call.=FALSE)

    # extract the specified assay
    assayData <- copy(x@assays[[keepAssay]])

    # extract the assay index
    assayIndex <- na.omit(unique(assayIndex(x)[,
        c("rowKey", "colKey", .assayName),
        with=FALSE
    ]))

    # handle summarized assays
    aKeys <- assayKeys(x, assayName)
    # only compute summaries for assays that are actually summarized
    summarize <- summarize && !all(idCols(x) %in% aKeys)
    if (summarize) {
        assayIndex <- assayIndex[, first(.SD), by=.assayName]
    }

    setkeyv(assayIndex, .assayName)
    assayData <- assayData[assayIndex, on=.assayName]
    # # optinally join with row and column metadata
    setkeyv(assayData, "rowKey")
    if (withDimnames && !metadata) {
        assayData <- rowIDs(x, data=TRUE, key=TRUE)[assayData, ]
        setkeyv(assayData, "colKey")
        assayData <- colIDs(x, data=TRUE, key=TRUE)[assayData, ]
    } else if (withDimnames && metadata) {
        assayData <- rowData(x, key=TRUE)[assayData, ]
        setkeyv(assayData, "colKey")
        assayData <- colData(x, key=TRUE)[assayData, ]
    }
    # honour row and column ordering guarantees
    ## See: https://github.com/bhklab/CoreGx/wiki/CoreGx-Design-Documentation
    if (withDimnames || key) {
        assayData[, (.assayName) := NULL]
        if (withDimnames) setkeyv(assayData, idCols(x)) else
            setkeyv(assayData, c("rowKey", "colKey"))
    } else {
        setkeyv(assayData, .assayName)
    }
    if (!key) assayData[, c("rowKey", "colKey") := NULL]
    corder <- c(
        if (withDimnames) idCols(x),
        if (key) c("rowKey", "colKey"),
        if (withDimnames && metadata) c(sort(rowMeta(x)), sort(colMeta(x)))
    )
    if (!is.null(corder)) setcolorder(assayData, corder) else setcolorder(assayData)

    # Drop columns with wrong cardinality from summary assays
    if (summarize) {
        summaryCols <- assayCols(x, assayName)
        if (metadata) {
            rCols <- colnames(rowData(x))
            cCols <- colnames(colData(x))
            rBy <- intersect(rCols, aKeys)
            cBy <- intersect(cCols, aKeys)
            rKeep <- rCols[
                rowData(x)[, lapply(.SD, uniqueN), by=c(rBy)][, lapply(.SD, max)] == 1
            ]
            cKeep <- cCols[
                colData(x)[, lapply(.SD, uniqueN), by=c(cBy)][, lapply(.SD, max)] == 1
            ]
            summaryCols <- c(summaryCols, rKeep, cKeep)
        }
        # use of %in% should maintain column ordering gaurantees
        assayData <- assayData[, colnames(assayData) %in% summaryCols, with=FALSE]
        setkeyv(assayData, aKeys)
    }

    ## Add [] to ensure assay always prints, even after modify by reference
    ## See: https://stackoverflow.com/questions/33195362/data-table-is-not-displayed-on-first-call-after-being-modified-in-a-function
    return(assayData[])
})


#' Add or replace an assay in a LongTable by name or index
#'
#' @description Add or replace an assay in a LongTable by name. Currently
#'    this function only works when the assay has all columns in row and column
#'    data tables (i.e., when assays is retured withDimnames=TRUE).
#'
#' @examples
#' assay(merckLongTable, 'sensitivity') <-
#'      assay(merckLongTable, 'sensitivity', withDimnames=TRUE)
#' assay(merckLongTable, 'sensitivity') <- merckLongTable$sensitivity
#'
#' @param x A `LongTable` to update.
#' @param i `integer` or `character` vector containing the index or name
#'   of the assay to update.
#' @param value
#' A `data.frame` or `data.table` to update the assay data
#'   with. This must at minumum contain the row and column data identifier
#'   columns to allow correctly mapping the assay keys. We recommend modifying
#'   the results returned by assay(longTable, 'assayName', withDimnames=TRUE).
#'   For convenience, both the `[[` and `$` LongTable accessors return an assay
#'   with the dimnames.
#'
#' @return `LongTable` With updated assays slot.
#'
#' @describeIn LongTable
#'
#' @md
#' @importMethodsFrom SummarizedExperiment assay<-
#' @importFrom data.table data.table fsetdiff setcolorder set setDT
#' @export
setReplaceMethod('assay', signature(x='LongTable'), function(x, i, value) {
    stopifnot(is.character(i) || is.numeric(i))

    funContext <- CoreGx:::.S4MethodContext('assay', class(x))
    if (length(i) > 1) .error(funContext, ' Only a single assay ',
        'name can be assiged with assay(x, i) <- value.')

    # -- determine if the assay already exists
    if (is.numeric(i)) i <- assayNames(x)[i]
    .i <- paste0(".", i)
    assayExists <- i %in% assayNames(x)

    if (is.null(value)) {
        keepAssays <- setdiff(assayNames(x), i)
        return(x[, , keepAssays])
    }

    if (!is.data.frame(value)) .error(funContext, ' Only a data.frame or',
        ' data.table can be assiged to the assay slot!')
    value <- copy(value)  # prevent modify by reference
    if (!is.data.table(value)) setDT(value)

    # -- extract strucutral metadata from .intern slot
    mutable_intern <- mutable(getIntern(x))
    aIndex <- mutable_intern$assayIndex
    aKeys <- mutable_intern$assayKeys

    # -- determine the id columns if the assay doesn't already exits
    if (!any(assayExists)) {
        assayKey <- intersect(idCols(x), colnames(value))
    } else {
        if (sum(assayExists) > 1)
            .error(funContext, "Only one assay can be modified at a time.",
                " Please set i to be a character(1) vector.")
        assayKey <- aKeys[[i]]
    }
    # -- add assayKey column to the value
    setkeyv(value, assayKey)
    value[, (.i) := .I]

    # -- join assay with existing metadata
    rKeys <- intersect(rowIDs(x), assayKey)
    cKeys <- intersect(colIDs(x), assayKey)
    rIndex <- rowIDs(x, data=TRUE, key=TRUE)[, .SD, .SDcols=c("rowKey", rKeys)]
    setkeyv(rIndex, rKeys)
    cIndex <- colIDs(x, data=TRUE, key=TRUE)[, .SD, .SDcols=c("colKey", cKeys)]
    setkeyv(cIndex, cKeys)

    # -- add an index to the assay
    setkeyv(rIndex, "rowKey")
    setkeyv(cIndex, "colKey")
    setkeyv(aIndex, "rowKey")
    annotatedIndex <- merge.data.table(aIndex, rIndex, all=TRUE)
    setkeyv(annotatedIndex, "colKey")
    annotatedIndex <- merge.data.table(annotatedIndex, cIndex, all=TRUE)

    # -- update assayIndex with the new assay
    setkeyv(annotatedIndex, assayKey)
    if (.i %in% colnames(annotatedIndex)) annotatedIndex[, (.i) := NULL]
    # FIXME:: This is really slow with by=.EACHI when the cardinality is high
    annotatedIndex[value, (.i) := get(.i), on=assayKey, by=.EACHI]
    annotatedIndex[, (assayKey) := NULL]
    setkeyv(annotatedIndex, unique(c(paste0(".", assayNames(x)), .i)))

    # -- detect and warn users if they have modified id columns
    # rowIDs
    presentRowIDs <- intersect(rowIDs(x), colnames(value))  # allow summary over some keys
    if (!(length(presentRowIDs) > 0)) stop(.errorMsg("No rowIDs(x) present in",
        "value! Cannot summarize over an entire dimension."), call.=FALSE)
    ## set check.attributes=FALSE to allow unequal table keys
    equalRowIDs <- .table_is_subset(
        unique(value[, presentRowIDs, with=FALSE])[order(mget(presentRowIDs))],
        unique(rowIDs(x, data=TRUE)[order(mget(presentRowIDs)), presentRowIDs,
            with=FALSE])
    )
    if (!isTRUE(equalRowIDs))
        stop(.errorMsg("One or more rowIDs(x) columns have been modified.",
                " Identifier columns cannot be modified via assay assignment!"),
            call.=FALSE
        )
    # colIDs
    presentColIDs <- intersect(colIDs(x), colnames(value))  # allow summary over some keys
    if (!(length(presentColIDs) > 0)) stop(.errorMsg("No colIDs(x) present in",
        "value! Cannot summarize over an entire dimension."), call.=FALSE)
    equalColIDs <- .table_is_subset(
        unique(value[, presentColIDs, with=FALSE])[order(mget(presentColIDs))],
        unique(colIDs(x, data=TRUE)[order(mget(presentColIDs)), presentColIDs,
            with=FALSE])
    )
    if (!isTRUE(equalColIDs))
        stop(.errorMsg("One or more colIDs(x) column have been modified.",
                " Identifier columns cannot be modified via assay assignment!"),
            call.=FALSE
        )

    # -- remove metadata columns for the assay
    throwAwayCols <- c(idCols(x), rowMeta(x), colMeta(x))
    keepCols <- setdiff(colnames(value), throwAwayCols)
    assayValue <- unique(value[, keepCols, with=FALSE])
    setkeyv(assayValue, .i)

    # -- update the object
    setcolorder(annotatedIndex, c("rowKey", "colKey"))
    mutable_intern$assayIndex <- annotatedIndex
    mutable_intern$assayKeys[[i]] <- assayKey
    x@.intern <- immutable(mutable_intern)
    x@assays[[i]] <- assayValue

    return(x)
})

##
## == assayNames


#' Retrieve the assay names from a `LongTable` object.
#'
#' @examples
#' assayNames(merckLongTable)
#' names(merckLongTable)
#'
#' @describeIn LongTable Return the names of the assays contained in a
#'   `LongTable`
#'
#' @param x A `LongTable` object to retrieve the assay names from.
#'
#' @return `character` Names of the assays contained in the `LongTable`.
#'
#' @importMethodsFrom SummarizedExperiment assayNames
#' @aliases names,LongTable-method names
#' @export
setMethod('assayNames', signature(x='LongTable'), function(x) {
    return(names(x@assays))
})
#' @export
setMethod("names", signature(x="LongTable"), function(x) {
    return(assayNames(x))
})


## ==================
## ---- metadata Slot
## ------------------


#' Getter method for the metadata slot of a `LongTable` object
#'
#' @param x The `LongTable` object from which to retrieve the metadata list.
#'
#' @return `list` The contents of the `metadata` slot of the `LongTable`
#'   object.
#'
#' @importMethodsFrom S4Vectors metadata
#' @export
setMethod('metadata', signature(x='LongTable'), function(x) {
    return(x@metadata)
})


#' Setter method for the metadata slot of a `LongTable` object
#'
#' @param x `LongTable` The LongTable to update
#' @param value `list` A list of new metadata associated with a `LongTable`
#'   object.
#'
#' @return `LongTable` A copy of the `LongTable` object with the `value` in
#'   the metadata slot.
#'
#' @importMethodsFrom S4Vectors metadata<-
#' @importFrom crayon cyan magenta
#' @export
setReplaceMethod('metadata', signature(x='LongTable'), function(x, value) {
    if (!is(value, 'list'))
        stop(magenta$bold('The `metadata` slot must be a list!'))
    x@metadata <- value
    return(x)
})

#' Convert the old sensitivity slot format into a `LongTable` and update the
#' `CoreSet` object.
#'
#' @param object Inheriting from `CoreSet`.
#' @param mapper Should the `LongTableDataMapper` object be early returned,
#' instead of the `LongTable` object. This can be useful if the conversion
#' fails or corrupts your data. You can then modify the `DataMapper` as
#' necessary to fix the sensititivity data.
#'
#' @return A `LongTable` constructed from `object@treatmentResponse`, or a
#' `LongTableDataMapper` if `mapper`=TRUE.
#'
#' @keywords internal
#' @noRd
#' @importFrom data.table data.table as.data.table merge.data.table
#' melt.data.table
.sensitivityToTRE <- function(object, mapper=FALSE) {

    # -- validate input
    funContext <- .funContext(':::.sensitivitySlotToLongTable')
    if (!is(object, 'CoreSet')) .error(funContext, ' object must inherit from
        the CoreSet class.')
    oldSensitivity <- treatmentResponse(object)

    if (!is(oldSensitivity, 'list')) .error(funContext, ' @sensitivty slot
        is not a `list`?')

    # -- extract the old data as data.tables

    # sensitivityInfo
    infoDT <- as.data.table(oldSensitivity$info, keep.rownames=TRUE)
    rowCols <- c(treatment1id="treatmentid", treatment1dose='dose')
    colCols <- c(sampleid="sampleid")

    # sensitivityProfiles
    profDT <- as.data.table(oldSensitivity$profiles, keep.rownames=TRUE)

    # sensitivityRaw
    doseDT <- as.data.table(oldSensitivity$raw[, , 1], keep.rownames=TRUE)
    meltedDoseDT <- na.omit(melt.data.table(doseDT, id.vars='rn',
        variable.name='old_column', value.name='dose'))
    meltedDoseDT[, dose := as.numeric(dose)]
    viabDT <- as.data.table(oldSensitivity$raw[, , 2], keep.rownames=TRUE)
    meltedViabDT <- na.omit(melt.data.table(viabDT, id.vars='rn',
        variable.name='old_column', value.name='viability'))
    meltedViabDT[, viability := as.numeric(viability)]

    # -- merge into a single long format data.table
    assayDT <- merge.data.table(meltedDoseDT, meltedViabDT,
        by=c('rn', 'old_column'))
    assayMap <- list(sensitivity=c('viability'),
        profiles=setdiff(colnames(profDT), 'rn'))

    rawdataDT <- merge.data.table(assayDT, profDT, by='rn')
    rawdataDT <- merge.data.table(rawdataDT, infoDT, by='rn')
    rawdataDT[, replicate_id := seq_len(.N), by=c(rowCols, colCols)]

    if (max(rawdataDT$replicate_id) > 1) {
        # Handle case where there is only 1 drug (i.e., radiation in RadioGx)
        if (length(unique(rawdataDT[[rowCols[1]]])) == 1) {
            rowCols <- c(rowCols, 'replicate_id')
        } else {
            colCols <- c(colCols, 'replicate_id')
        }
    } else {
        rawdataDT[, replicate_id := NULL]
    }

    groups <- list(
        rowDataMap=rowCols,
        colDataMap=colCols,
        assayMap=c(rowCols, colCols)
    )

    # -- capute the na rownames to make recreation easier in .rebuildInfo
    missing_rows <- setdiff(infoDT$rn, rawdataDT$rn)
    na_index <- infoDT[rn %in% missing_rows, .(rn, treatmentid, sampleid)]

    # -- build a LongTableDataMapper object
    TREdataMapper <- TREDataMapper(rawdata=rawdataDT)
    guess <- guessMapping(TREdataMapper, groups, subset=TRUE)

    assayCols <- unlist(assayMap)

    # do not steal any assay columns for the row or column data
    guess$rowDataMap[[2]] <- setdiff(guess$rowDataMap[[2]], assayCols)
    guess$colDataMap[[2]] <- setdiff(guess$colDataMap[[2]], assayCols)
    guess$metadata[[2]] <- setdiff(guess$metadata[[2]],
        c(assayCols, guess$rowDataMap[[2]], guess$colDataMap[[2]]))
    assayMap$assay_metadata <- setdiff(guess$assayMap$mapped_columns, assayCols)
    assayMap <- lapply(assayMap, FUN=function(x, y) list(y, x),  # add id columns
        y=guess$assayMap[[1]])

    # update the data mapper
    rowDataMap(TREdataMapper) <- guess$rowDataMap
    colDataMap(TREdataMapper) <- guess$colDataMap
    assayMap(TREdataMapper) <- assayMap
    metadataMap(TREdataMapper) <-
        list(experiment_metadata=guess$metadata$mapped_columns)
    metadata(TREdataMapper) <- list(sensitivityInfo_NA=na_index)

    # build the object
    return(if (!mapper) metaConstruct(TREdataMapper) else TREdataMapper)
}


#' Compare the values of sensitivityInfo before and after use of
#' .sensitivityToTRE
#'
#' @param object `CoreSet` to be updated to the new
#' `TreatmentResponseExperiment` sensitivity format.
#' @param FUN `function` The function to compare results from.
#'
#' @return None, displays results of `all.equal` on the sensitivityInfo for
#'   the columns which should be conserved.
#'
#' @keywords internal
#' @noRd
#' @importFrom data.table data.table as.data.table merge.data.table
#' melt.data.table
.compareTreatmentResponse <- function(object, FUN) {

    new_object <- copy(object)
    tre <- CoreGx:::.sensitivityToTRE(object)
    new_object@treatmentResponse <- tre

    old_res <- copy(FUN(object))
    new_res <- copy(FUN(new_object))

    setDT(old_res, keep.rownames="rownames")
    setDT(new_res, keep.rownames="rownames")

    equal_columns <- colnames(old_res)
    all.equal(
        old_res[order(rownames), .SD, .SDcols=equal_columns],
        new_res[order(rownames), .SD, .SDcols=equal_columns]
    )
}

## Utility function for the Gx suite packages ------------------------------
## FIXME:: Put these functions in sections based on similar purposes

# ExpressionSet to SummarizedExperiment -----------------------------------

#' CSet molecularProfiles from ESets to SEs
#'
#' Converts all ExpressionSet objects within the molecularProfiles slot of a
#'   CoreSet to SummarizedExperiments
#'
#' @param cSet \code{S4} A CoreSet containing molecular data in ExpressionSets
#'
#' @return \code{S4} A CoreSet containing molecular data in a
#'   SummarizedExperiments
#'
#' @importFrom BiocParallel bplapply
#' @importFrom SummarizedExperiment SummarizedExperiment Assays assay
#'   assayNames assayNames<-
#' @importFrom Biobase exprs fData pData annotation protocolData
#'   assayDataElementNames
#' @importFrom S4Vectors SimpleList DataFrame
#' @importFrom stats setNames
#' @keywords internal
.convertCSetMolecularProfilesToSE <- function(cSet) {

    eSets <- molecularProfilesSlot(cSet)  # Extract eSet data

    molecularProfilesSlot(cSet) <- lapply(eSets, function(eSet) {

        # Change rownames from probes to EnsemblGeneId for rna data type
        if (grepl("^rna$", Biobase::annotation(eSet))) {
            rownames(eSet) <- Biobase::fData(eSet)$EnsemblGeneId
        }

        # Build summarized experiment from eSet TODO:: Do we want to pass an environment for better memory efficiency?
        SE <- SummarizedExperiment::SummarizedExperiment(assays = SimpleList(as.list(Biobase::assayData(eSet))), rowData = S4Vectors::DataFrame(Biobase::fData(eSet),
            rownames = rownames(Biobase::fData(eSet))), colData = S4Vectors::DataFrame(Biobase::pData(eSet), rownames = rownames(Biobase::pData(eSet))),
            metadata = list(experimentData = eSet@experimentData, annotation = Biobase::annotation(eSet), protocolData = Biobase::protocolData(eSet)))
        ## TODO:: Determine if this can be done in the SE constructor?  Extract names from expression set
        assayNames(SE) <- assayDataElementNames(eSet)
        # Assign SE to cSet
        mDataType <- Biobase::annotation(eSet)
        molecularProfilesSlot(cSet)[[mDataType]] <- SE
    })
    setNames(cSet@molecularProfiles, names(eSets))
    cSet
}

# sanityCheck -------------------------------------------------------------

## TODO:: Add documentation!
#' @export
#' @noRd
.sanitizeInput <- function(x, y, lower, upper, pars, x_as_log, y_as_log,
        y_as_pct, trunc, verbose = FALSE) {
    # Set to 2 to see debug printouts

    if (!is.logical(x_as_log)) {
        if (verbose == 2) {
            message("x_as_log:")
            message(x_as_log)
        }
        stop("'x_as_log' is not a logical.")
    }

    if (!is.logical(y_as_log)) {
        if (verbose == 2) {
            message("y_as_log:")
            message(y_as_log)
        }
        stop("'y_as_log' is not a logical.")
    }

    if (!is.logical(y_as_pct)) {
        if (verbose == 2) {
            message("y_as_pct:")
            message(y_as_pct)
        }
        stop("'y_as_pct' is not a logical.")
    }

    if (!is.logical(trunc)) {
        if (verbose == 2) {
            message("trunc:")
            message(trunc)
        }
        stop("'trunc' is not a logical.")
    }

    if (y_as_pct && y_as_log) {
        if (verbose == 2) {
            message("y_as_pct:")
            message(y_as_pct)
            message("y_as_log:")
            message(y_as_log)
        }
        warning("y_as_pct and y_as_log flags should almost certainly not both be TRUE.")
    }

    if (!(verbose %in% c(0, 1, 2))) {
        message("verbose:")  #can't have the if(verbose == 2) statement here since verbose itself is the problem!
        message(verbose)
        stop("'verbose' flag is not set correctly.")
    }

    if (!missing(x)) {
        if (!all(is.finite(x) | is.na(x)) || (x_as_log && any(x == -Inf))) {
            if (verbose == 2) {
                message("x:")
                message(x)
            }
            stop("x must contain only real numbers, NA-values, and/or -Inf (if x_as_log flag is set to TRUE).")
        }

        if (x_as_log == FALSE && min(x) < 0) {
            if (verbose == 2) {
                message("x:")
                message(x)
                message("x_as_log:")
                message(x_as_log)
            }
            stop("Negative x-values encountered. Data may be inappropriate, or 'x_as_log' flag may be set incorrectly.")
        }
        if (length(unique(x)) < 3) {
            stop("Please pass in at least 3 unique dose points.")
        }
    }

    if (missing(y)) {
        if (missing(pars)) {
            stop("Both 'pars' and 'y' missing, please pass in some data!")
        } else {

            if (pars[[1]] < 0 || pars[[2]] < 0) {
                # HS or alpha
                if (verbose == 2) {
                  message("pars:")
                  message(pars)
                }
                warning("Curve parameters may be inappropriately set to negative values.")
            }

            if (length(pars) == 3) {
                # and thus we are in PharmacoGx
                if (x_as_log == FALSE && pars[[3]] < 0) {
                  message("pars:")
                  message(pars[[3]])
                  message("x_as_log:")
                  message(x_as_log)
                  stop("'x_as_log' flag may be set incorrectly, as the EC50 is negative when a positive value is expected.")
                }

                if (y_as_pct == FALSE) {
                  if (pars[[2]] > 1) {
                    if (verbose == 2) {
                      message("pars:")
                      message(pars[[2]])
                      message("y_as_pct:")
                      message(y_as_pct)
                    }
                    warning("Warning: 'y_as_pct' flag may be set incorrectly.")
                  }
                }
            } else if (length(pars) == 2) {
                if (pars[[1]] < pars[[2]]) {
                  if (verbose) {
                    warning("Alpha is greater than beta.")
                    if (verbose == 2) {
                      message("pars:")
                      message(pars)
                    }
                  }
                }
            } else {
                stop("Pars does not have the correct length.")
            }
        }

    } else {

        if (!all(is.finite(y) | is.na(y)) || (y_as_log && any(y == -Inf))) {
            if (verbose == 2) {
                message("y:")
                message(y)
            }
            stop("y must contain only real numbers, NA-values, and/or -Inf (if y_as_log is set to TRUE).")
        }

        if (min(y, na.rm=TRUE) < 0) {
            if (verbose) {
                warning("Warning: Negative y data.")
                if (verbose == 2) {
                  message("y:")
                  message(y)
                }
            }
        }

        if (max(y, na.rm=TRUE) > (1 + 99 * y_as_pct)) {
            if (verbose) {
                warning("Warning: y data exceeds negative control.")
                if (verbose == 2) {
                  message("y:")
                  message(y)
                }
            }
        }

        if (missing(pars)) {

            if (y_as_log == FALSE && min(y, na.rm=TRUE) < 0) {
                if (verbose) {
                  warning("Negative y-values encountered. y data may be inappropriate, or 'y_as_log' flag may be set incorrectly.")
                  if (verbose == 2) {
                    message("y:")
                    message(y)
                    message("y_as_log:")
                    message(y_as_log)
                  }
                }
            }

            if (y_as_pct == TRUE && max(y, na.rm=TRUE) < 5) {
                if (verbose) {
                  warning("Warning: 'y_as_pct' flag may be set incorrectly.")
                  if (verbose == 2) {
                    message("y:")
                    message(y)
                    message("y_as_pct:")
                    message(y_as_pct)
                  }
                }
            }

            if (y_as_pct == FALSE && max(y, na.rm=TRUE) > 5) {
                if (verbose) {
                  warning("Warning: 'y_as_pct' flag may be set incorrectly.")
                  if (verbose == 2) {
                    message("y:")
                    message(y)
                    message("y_as_pct:")
                    message(y_as_pct)
                  }
                }
            }

            if (!missing(x) && length(x) != length(y)) {
                if (verbose == 2) {
                  message("x:")
                  message(x)
                  message("y:")
                  message(y)
                }
                stop("Vector of x-values is not of same length as vector of y-values.")
            }

        } else {
            stop("Please pass in only one of 'pars' and 'y', as it is unclear which to use in the computation.")
        }
    }

    if (!missing(lower) && !missing(upper)) {
        if (!(is.double(lower))) {
            if (verbose == 2) {
                message("lower:")
                message(lower)
            }
            stop("The lower bound must be a positive real number.")
        }

        if (!(is.double(lower))) {
            if (verbose == 2) {
                message("upper:")
                message(upper)
            }
            stop("The upper bound must be a positive real number.")
        }

        if (lower >= upper) {
            if (verbose == 2) {
                message("lower:")
                message(lower)
                message("upper:")
                message(upper)
            }
            stop("The lower bound of the range of allowed x-values must be less than the upper bound.")
        }

        if (lower < 0) {
            if (verbose == 2) {
                message("lower:")
                message(lower)
            }
            stop("The lower bound of the range of allowed x-values must be nonnegative.")
        }

        if (upper < 0) {
            if (verbose == 2) {
                message("upper:")
                message(upper)
            }
            stop("The upper bound of the range of allowed x-values must be nonnegative.")
        }
    }
}


# getSupportVec -----------------------------------------------------------

## get vector of interpolated concentrations for graphing purposes
#' .getSupportVec
#'
#' @param x An input vector of dosages
#' @param output_length The length of the returned support vector
#'
#' @return \code{numeric} A numeric vector of interpolated concentrations
#'
#' @export
#' @noRd
.getSupportVec <- function(x, output_length = 1001) {
    return(seq(from = min(x), to = max(x), length.out = output_length))
}

#### reformatData ------------------------------------------------------------

#' @export
#' @noRd
.reformatData <- function(x, y, pars, x_to_log, y_to_log, y_to_frac, trunc) {
    if (!(is.logical(x_to_log))) {
        stop("x_to_log must be a logical.")
    }

    if (!(is.logical(y_to_log))) {
        stop("y_to_log must be a logical.")
    }

    if (!(is.logical(y_to_frac))) {
        stop("y_to_frac must be a logical.")
    }

    if(y_to_log && y_to_frac){
        warning("Both y_to_log and y_as_frac set. Will first convert to fraction and then take the logarithm.\n
            If this is not what is intended, please reformat data prior to curve fitting. ")
    }

    if (x_to_log) {
        x <- log10(x)
    }
    ### Note, if Y is passed in, it is sorted in this same order below. This is not obvious from the code right away.
    if (is.unsorted(x, na.rm=TRUE)) {
        InputUnsorted <- TRUE
        warning("x-values passed in unsorted. Sorting x-values and corresponding y-values (if passed in).")
        xOrder <- order(x)
        x <- x[xOrder]
    } else {
        InputUnsorted <- FALSE
    }

    if (!missing(y)) {

        if (InputUnsorted) {
            y <- y[xOrder]
        }

        if (any(is.na(x) & (!is.na(y)))) {
            warning("Missing x-values with non-missing y-values encountered. Removed y-values correspoding to those x-values.")
            myx <- !is.na(x)
            x <- as.numeric(x[myx])
            y <- as.numeric(y[myx])
        }

        if (any((!is.na(x)) & is.na(y))) {
            warning("Missing y-values with non-missing x-values encountered. Removed x values correspoding to those y-values.")
            myy <- !is.na(y)
            x <- as.numeric(x[myy])
            y <- as.numeric(y[myy])
        }

        myxy <- complete.cases(x,y)
        x <- x[myxy]
        y <- y[myxy]

        if (y_to_frac) {
            y <- y/100
        }

        if (trunc) {
            y = pmin(as.numeric(y), 1)
            y = pmax(as.numeric(y), 0)
        }

        if (x_to_log) {
            x0s <- which(x == -Inf)
            if (length(x0s) > 0) {
                x <- x[-x0s]
                y <- y[-x0s]
            }
        }

        if (y_to_log) {
            if (any(y <= 0)) {
                warning("Transforming y to log with non-positive y values present, therefore removing.")
                x <- x[y > 0]
                y <- y[y > 0]
                if (!length(x)) {
                  stop("No valid positive y values encountered, please pass in some data.")
                }
            }
            y <- log(y)
        }



         if (length(unique(x)) < 3) {
            stop("Less than 3 unique dose points left after cleaning data, please pass in enough valid measurements.")

        }

        return(list(x = x, y = y))
    }

    if (!missing(pars)) {

        if (x_to_log && length(pars) == 3) {
            pars[[3]] <- log10(pars[[3]])
        }

        if (y_to_frac && length(pars) == 3) {
            pars[[2]] <- pars[[2]]/100
        }

        return(list(x = x, pars = pars))
    }
}




# multinom ----------------------------------------------------------------

#' @export
#' @noRd
.multinom <- function(x, y) {
    coeff <- 1
    for (i in seq_len(length(y))) {
        coeff <- coeff * choose(x, y[i])
        x <- x - y[i]
    }
    return(coeff)
}

# medncauchys -------------------------------------------------------------

## TODO:: Add documentation to these functions

#' A random sample distributed as the median of N Cauchy distributed variables
#'
#' Naming follows R conventions.
#'
#' @param N How many samples to sample
#' @param n The number of Cauchy distributions to take the median of
#' @param scale the scale of the Cauchy distribution.
#'
#' @importFrom stats rcauchy
#' @export
#' @keywords internal
#' @noRd
.rmedncauchys <- function(N, n, scale) {
    x <- matrix(NA, nrow = 1, ncol = N)
    for (i in seq_len(N)) {
        x[i] <- median(rcauchy(n, scale = scale))
    }
    return(x)
}

#' PDF of the median of N Cauchy distributed variables
#'
#' This function calculates the PDF/density for a variable distributed as the median value of n IID Cauchy variables. Naming follows R conventions.
#'
#' @param x Where to evaluate the density function
#' @param n The number of Cauchy distributions to take the median of
#' @param scale the scale of the Cauchy distribution.
#' @param divisions How many maximum divisions to use in numerical integration
#'
#' @importFrom stats dcauchy pcauchy integrate
#' @export
#' @keywords internal
#' @noRd
.dmedncauchys = function(x, n, scale, divisions = 100) {
    n <- rep(n, times = length(x)/length(n))
    scale <- rep(scale, times = length(x)/length(scale))
    y <- matrix(NA, nrow = 1, ncol = length(x))
    for (g in seq_along(x)) {
        if (n[g]%%2 == 0) {
            y[g] <- 2 * .multinom(n[g], c(n[g]/2 - 1, n[g]/2 - 1)) * tryCatch(integrate(f = function(j) {
                (pcauchy(x[g] - j/2, scale = scale[g]))^(n[g]/2 - 1) * (1 - pcauchy(x[g] + j/2, scale = scale[g]))^(n[g]/2 - 1) * dcauchy(x[g] -
                  j/2, scale = scale[g]) * dcauchy(x[g] + j/2, scale = scale[g])
            }, lower = 0, upper = Inf, subdivisions = divisions)[[1]], error = function(e) {
                if (divisions == 1) {
                  wseq <- c(1, 4, 1)
                } else {
                  wseq <- c(1, 4, rep(c(2, 4), times = divisions - 1), 1)
                }
                aseq <- seq(from = 0, to = pi/2, length.out = 2 * divisions + 1)
                tseq <- tan(aseq)/2
                return(sum((pcauchy(x[g] + tseq, scale = scale[g]))^(n[g]/2 - 1) * (pcauchy(x[g] - tseq, scale = scale[g]))^(n[g]/2 - 1) *
                  dcauchy(x[g] + tseq, scale = scale[g]) * dcauchy(x[g] - tseq, scale = scale[g])/(cos(aseq))^2 * wseq) * (aseq[2] - aseq[1])/6)
            })
        } else {
            y[g] <- .multinom(n[g], c((n[g] - 1)/2, (n[g] - 1)/2)) * (pcauchy(x[g], scale = scale[g]))^((n[g] - 1)/2) * (1 - pcauchy(x[g],
                scale = scale[g]))^((n[g] - 1)/2) * dcauchy(x[g], scale = scale[g])
        }
    }
    return(y)
}


#' CDF of the median of N Cauchy distributed variables
#'
#' This function calculates the CDF/distribution for a variable distributed as the median value of n IID Cauchy variables. Naming follows R conventions.
#'
#' @param x Where to evaluate the distribution function
#' @param n The number of Cauchy distributions to take the median of
#' @param scale the scale of the Cauchy distribution.
#' @param divisions How many maximum divisions to use in numerical integration
#'
#' @importFrom stats pcauchy integrate
#' @export
#' @keywords internal
#' @noRd
.pmedncauchys = function(x, n, scale, divisions = 100) {
    n <- rep(n, times = length(x)/length(n))
    scale <- rep(scale, times = length(x)/length(scale))
    y <- integer(length(x))
    for (g in seq_along(x)) {
        if (n[g] %% 2 == 0) {
            y[g] <- tryCatch(integrate(f = function(k) {
                .dmedncauchys(k, n[g], scale[g])
            }, lower = -Inf, upper = x[g], subdivisions = divisions)[[1]], error = function(e) {
                wseq <- c(1, 4, rep(c(2, 4), times = divisions - 1), 1)
                aseq <- seq(from = -pi/2, to = atan(x[g]), length.out = 2 * divisions + 1)
                return(sum(.dmedncauchys(tan(aseq), n[g], scale[g]) * wseq/(cos(aseq))^2) * (aseq[3] - aseq[1])/6)
            })
        } else {
            y[g] <- 0
            Fx <- pcauchy(x[g], scale = scale[g])
            for (i in 0:((n[g] - 1)/2)) {
                y[g] <- y[g] + choose((n[g] - 1)/2, i) * (-1)^i * Fx^((n[g] + 1)/2 + i)/((n[g] + 1)/2 + i)
            }
            y[g] <- y[g] * .multinom(n[g], c((n[g] - 1)/2, (n[g] - 1)/2))
        }
    }
    return(y)
}

#' Expectation of the likelihood of the median of N Cauchy distributions for
#' truncated data
#'
#' This function calculates the expected value of the PDF of a median of N
#' Cauchy with given scale parameter, calculated over the region from x to
#' infinity if x>0, and -infinity to x otherwise.
#'
#' This is used in curve fitting when data has been truncated. Since for
#' truncated data, we don't know what the "real" value was, the reasoning is
#' we take the expected value.
#'
#' This increases robustness to extreme outliers while not completely ignoring
#' the fact that points are truncated, and seems to work well in practice. The
#' name of the function follows:
#' e(xpectation)d(istribution)med(ian)ncauchys - following R conventions.
#'
#' @param x Where the truncation occurred
#' @param n The number of Cauchy distributions to take the median of
#' @param scale the scale of the Cauchy distribution.
#' @param divisions How many maximum divisions to use in numerical integration
#'
#' @importFrom stats integrate
#' @keywords internal
#' @export
#' @noRd
.edmedncauchys = function(x, n, scale, divisions = 100) {
    n <- rep(n, times = length(x)/length(n))
    scale <- rep(scale, times = length(x)/length(scale))
    y <- numeric(length(x))
    for (g in seq_along(y)) {
        if (x[g] > 0) {
            upper <- Inf
            lower <- x[g]
        } else {
            upper <- x[g]
            lower <- -Inf
        }
        y[g] <- tryCatch(integrate(f = function(k) {
            (.dmedncauchys(k, n[g], scale[g]))^2
        }, lower = lower, upper = upper, subdivisions = divisions)[[1]], error = function(e) {
            wseq <- c(1, 4, rep(c(2, 4), times = divisions - 1), 1)
            aseq <- seq(from = atan(lower), to = atan(upper), length.out = 2 * divisions + 1)
            return(sum((.dmedncauchys(tan(aseq), n[g], scale[g]))^2 * wseq/(cos(aseq))^2) * (aseq[3] - aseq[1])/6)
        })
    }
    return(y)
}

#### mednnormals -------------------------------------------------------------

#' A random sample distributed as the median of N Normally distributed variables
#'
#' Naming follows R conventions.
#'
#' @param N How many samples to sample
#' @param n The number of normal distributions to take the median of
#' @param scale the SD of the normal distribution.
#'
#' @export
#' @keywords internal
#' @noRd
.rmednnormals = function(N, n, scale) {
    x <- matrix(NA, nrow = 1, ncol = N)
    for (i in seq_len(N)) {
        x[i] <- median(rnorm(n, sd = scale))
    }
    return(x)
}

#' PDF of the median of N Normally distributed variables
#'
#' This function calculates the PDF/density for a variable distributed as the median value of n IID Normal variables. Naming follows R conventions.
#'
#' @param x where to evaluate the density
#' @param n The number of normal distributions to take the median of
#' @param scale the SD of the normal distribution.
#' @param divisions How many maximum divisions to use in numerical integration
#'
#' @importFrom stats rnorm  dnorm
#' @export
#' @keywords internal
#' @noRd
.dmednnormals = function(x, n, scale, divisions = 100) {
    n <- rep(n, times = length(x)/length(n))
    scale <- rep(scale, times = length(x)/length(scale))
    y <- matrix(NA, nrow = 1, ncol = length(x))
    for (g in seq_along(x)) {
        if (n[g]%%2 == 0) {
            y[g] <- 2 * .multinom(n[g], c(n[g]/2 - 1, n[g]/2 - 1)) * tryCatch(integrate(f = function(j) {
                (pnorm(x[g] - j/2, sd = scale[g]))^(n[g]/2 - 1) * (1 - pnorm(x[g] + j/2, sd = scale[g]))^(n[g]/2 - 1) * dnorm(x[g] - j/2,
                  sd = scale[g]) * dnorm(x[g] + j/2, sd = scale[g])
            }, lower = 0, upper = Inf, subdivisions = divisions)[[1]], error = function(e) {
                if (divisions == 1) {
                  wseq <- c(1, 4, 1)
                } else {
                  wseq <- c(1, 4, rep(c(2, 4), times = divisions - 1), 1)
                }
                aseq <- seq(from = 0, to = pi/2, length.out = 2 * divisions + 1)
                tseq <- tan(aseq)/2
                return(sum((pnorm(x[g] + tseq, sd = scale[g]))^(n[g]/2 - 1) * (pnorm(x[g] - tseq, sd = scale[g]))^(n[g]/2 - 1) * dnorm(x[g] +
                  tseq, sd = scale[g]) * dnorm(x[g] - tseq, sd = scale[g])/(cos(aseq))^2 * wseq) * (aseq[2] - aseq[1])/6)
            })
        } else {
            if(n[g]==1){
                y[g] <- dnorm(x[g], sd = scale[g]) ## This reduces to the simple case for n equals 1 below, but we can save many calls to pnorm, which just get raised to a power of 0.
            } else {
                y[g] <- .multinom(n[g], c((n[g] - 1)/2, (n[g] - 1)/2)) * (pnorm(x[g], sd = scale[g]))^((n[g] - 1)/2) * (1 - pnorm(x[g], sd = scale[g]))^((n[g] -
                                                                                                                                                              1)/2) * dnorm(x[g], sd = scale[g])
            }

        }
    }
    return(y)
}


#' CDF of the median of N Normally distributed variables
#'
#' This function calculates the CDF/distribution for a variable distributed as the median value of n IID Normal variables. Naming follows R conventions.
#'
#' @param x Where to evaluate the Distribution
#' @param n The number of normal distributions to take the median of
#' @param scale the SD of the normal distribution.
#' @param divisions How many maximum divisions to use in numerical integration
#'
#' @importFrom stats integrate
#' @export
#' @keywords internal
#' @noRd
.pmednnormals = function(x, n, scale, divisions = 100) {
    n <- rep(n, times = length(x)/length(n))
    scale <- rep(scale, times = length(x)/length(scale))
    y <- numeric(length(x))
    for (g in seq_along(x)) {
        if (n[g]%%2 == 0) {
            y[g] <- tryCatch(integrate(f = function(k) {
                .dmednnormals(k, n[g], scale[g])
            }, lower = -Inf, upper = x[g], subdivisions = divisions)[[1]], error = function(e) {
                wseq <- c(1, 4, rep(c(2, 4), times = divisions - 1), 1)
                aseq <- seq(from = -pi/2, to = atan(x[g]), length.out = 2 * divisions + 1)
                return(sum(.dmednnormals(tan(aseq), n[g], scale[g]) * wseq/(cos(aseq))^2) * (aseq[3] - aseq[1])/6)
            })
        } else {
            y[g] <- 0
            Fx <- pnorm(x[g], sd = scale[g])
            for (i in 0:((n[g] - 1)/2)) {
                y[g] <- y[g] + choose((n[g] - 1)/2, i) * (-1)^i * Fx^((n[g] + 1)/2 + i)/((n[g] + 1)/2 + i)
            }
            y[g] <- y[g] * .multinom(n[g], c((n[g] - 1)/2, (n[g] - 1)/2))
        }
    }
    return(y)
}

#' Expectation of the likelihood of the median of N normal distributions for truncated data
#'
#' This function calculates the expected value of the PDF of a median of N normals with SD=scale, calculated over the region from x to infinity if x>0, and -infinity to x otherwise.
#' This is used in curve fitting when data has been truncated. Since for truncated data, we don't know what the "real" value was, the reasoning is we take the expected value.
#' This increases robustness to extreme outliers while not completely ignoring the fact that points are truncated, and seems to work well in practice. The name of the function follows:
#' e(xpectation)d(istribution)med(ian)nnormals - following R conventions.
#'
#' @param x Where the truncation occurred
#' @param n The number of normal distributions to take the median of
#' @param scale the SD of the normal distribution.
#' @param divisions How many maximum divisions to use in numerical integration
#'
#' @importFrom stats integrate
#' @export
#' @keywords internal
#' @noRd
.edmednnormals = function(x, n, scale, divisions = 100) {
    n <- rep(n, times = length(x)/length(n))
    scale <- rep(scale, times = length(x)/length(scale))
    y <- numeric(length(x))
    for (g in seq_along(y)) {
        if(n[g]==1){ ## The n=1 case is called very often, and there are significant savings (20x) to not calling numerical integration.
            if(x[g]>0){
                pnorm(x[g], sd=scale[g]/sqrt(2), lower.tail = FALSE)/(scale*2*sqrt(pi))
            } else {
                pnorm(x[g], sd=scale[g]/sqrt(2), lower.tail = TRUE)/(scale*2*sqrt(pi))
            }
        } else {
            if (x[g] > 0) {
                upper <- Inf
                lower <- x[g]
            } else {
                upper <- x[g]
                lower <- -Inf
            }
            y[g] <- tryCatch(integrate(f = function(k) {
                (.dmednnormals(k, n[g], scale[g]))^2
            }, lower = lower, upper = upper, subdivisions = divisions)[[1]], error = function(e) {
                wseq <- c(1, 4, rep(c(2, 4), times = divisions - 1), 1)
                aseq <- seq(from = atan(lower), to = atan(upper), length.out = 2 * divisions + 1)
                return(sum((.dmednnormals(tan(aseq), n[g], scale[g]))^2 * wseq/(cos(aseq))^2) * (aseq[3] - aseq[1])/6)
            })
        }

    }
    return(y)
}




# Set Operations ----------------------------------------------------------

## TODO:: Can we implement this as an extension of the BiocGenerics::setdiff?
#' Utility to find the symmetric set difference of a list of two or more
#' vectors or lists
#'
#' The function finds the symmetric set differnces between all the arguments,
#' defined as Union(args)-Intersection(args)
#'
#' @examples
#' list1 <- list('a', 'b', 'c')
#' list2 <- list('a', 'c')
#' list3 <- list('a', 'c', 'd')
#' listAll <- .symSetDiffList(list1, list2, list3)
#' listAll
#'
#' @param ... A list of or any number of vector like objects of the same mode,
#'   which could also be operated on by the native R set operations
#'
#' @return A vector like object of the same mode as the first argument,
#'   containing only the symmetric set difference
#'
#' @export
#' @keywords internal
.symSetDiffList <- function(...) {
    return(setdiff(.unionList(...), .intersectList(...)))
}

## FIXME:: This should be implemented as an extension of the intersect generic provided in the BiocGenerics package!
#' Intersect A List of More Than Two Vectors
#'
#' Utility to find the intersection between a list of more than two vectors or
#'   lists This function extends the native intersect function to work on two
#'   or more arguments.
#'
#' @examples
#' list1 <- list('a', 'b', 'c')
#' list2 <- list('a', 'c')
#' list3 <- list('a', 'c', 'd')
#' listAll <- .intersectList(list1, list2, list3)
#' listAll
#'
#' @param ... A list of or any number of vector like objects of the same mode,
#'   which could also be operated on by the native R set operations
#'
#' @return A vector like object of the same mode as the first argument,
#'   containing only the intersection common to all arguments to the function
#'
#' @export
#' @keywords internal
.intersectList <- function(...) {
    args <- list(...)
    nargs <- length(args)
    if (nargs == 0) {
        return(args)
    }
    if (nargs == 1) {
        if (nargs == 1 && is.list(args[[1]])) {
            do.call(".intersectList", args[[1]])
        } else {
            return(args[[1]])
        }
    } else if (nargs == 2) {
        return(intersect(args[[1]], args[[2]]))
    } else {
        return(intersect(args[[1]], .intersectList(args[-1])))
    }
}


## FIXME:: This should be implemented as an extension of the union generic from BiocGenerics
#' Utility to find the union between a list of more than two vectors or
#' lists
#'
#' This function extends the native union function to work on two or more
#' arguments.
#'
#' @examples
#' list1 <- list('a', 'b')
#' list2 <- list('a', 'c')
#' list3 <- list('c', 'd')
#' listAll <- .unionList(list1, list2, list3)
#' listAll
#'
#' @param ... A list of or any number of vector like objects of the same mode,
#'   which could also be operated on by the native R set operations
#'
#' @return A vector like object of the same mode as the first argument,
#'   containing all the elements of all arguments passed to the function
#'
#' @export
#' @keywords internal
.unionList <- function(...) {
    args <- list(...)
    nargs <- length(args)
    return(unique(unlist(do.call(c, args))))
}


#' @param matInd array indices
#' @param dimsizes array containing size of array of interest in each dimension
#'
#' @export
#' @keywords internal
#' @noRd
.linInd <- function(matInd, dimsizes) {
    y <- matInd[1]
    if (length(dimsizes) > 1) {
        for (i in seq(2, length(dimsizes))) {
            y <- y + (matInd[i] - 1) * prod(dimsizes[seq_len(i - 1)])
        }
    }
    return(y)
}

#' @export
#' @keywords internal
#' @noRd
# @param linInd linear index @param dimsizes array containing size of array of interest in each dimension
.matInd <- function(linInd, dimsizes) {
    y <- matrix(0, nrow = length(dimsizes), ncol = 1)
    if (NROW(y) > 1) {
        for (i in seq(2, length(dimsizes))) {
            y[i] <- ceiling(linInd/prod(dimsizes[seq_len(i - 1)]))
            linInd <- linInd - (y[i] - 1) * prod(dimsizes[seq_len(i - 1)])
        }
    }
    y[1] <- linInd
    return(y)
}


# Not Used? ------------------------------------------------------------------

### TODO:: Determine type of objects intended for this function
#' Getter for attributes of an object
#'
#' @param pars The object for which attributes are to be returned
#' @return A named vector where index `Rsquare` contains the attributes of the object
#' @export
#' @keywords internal
#' @noRd
.examineGOF <- function(pars) {
    return(c(Rsquare = attr(pars, "Rsquare")))
}


# Different in PharmacoGx? ------------------------------------------------


## FIXME:: This function already exists as base::trimws? Is there any reason we need to reimplement it?
#' @export
#' @keywords internal
#' @noRd
.stripWhiteSpace <- function(str, method = c("both", "head", "tail")) {
    method <- match.arg(method)
    str2 <- NULL
    if (length(str) == 1) {
        switch(method, both = {
            str2 <- gsub("^[ \t]+", "", str)
            str2 <- gsub("[ \t]+$", "", str2)
        }, head = {
            str2 <- gsub("^[ \t]+", "", str)
        }, tail = {
            str2 <- gsub("[ \t]+$", "", str)
        })
        return(str2)
    } else {
        str2 <- vapply(str, .stripWhiteSpace, method = method, FUN.VALUE = character(1))
        return(str2)
    }
}


# ==== LongTable

#' Convenience function for collapsing a character vector
#'
#' @examples
#' .collapse(c("Vector", "of", "words")
#'
#' @param ... `pairlist` One or more character vectors
#' @param collapse `character` Argument to collapse of paste0, default is ' '.
#'
#' @return `character` A single character vector.
#'
#' @keywords internal
#' @export
#' @noRd
.collapse <- function(..., collapse=' ')
    paste0(..., collapse=collapse)

#' Returns a colorized error message (magenta)
#'
#' @examples
#' cat(.errorMsg('This ', 'is ', 'an ', 'error ', 'message', time = TRUE))
#'
#' @param ... `pairlist` One or more strings or character vectors, also
#'   accepts any params to paste0.
#' @param time `logical` Indicates whether to include timestamp or not. Default is FALSE.
#'
#' @return `character` Colorized string with results from paste0(...)
#'
#' @keywords internal
#' @export
#' @noRd
.errorMsg <- function(..., time = FALSE, collapse=', ') {
    msg <- paste0(..., collapse=collapse)
    if (time) {
        timestamp <- format(Sys.time(), "%Y-%m-%d %H:%M:%S")
        msg <- paste0(timestamp, " ", msg)
    }
    magenta$bold(msg)
}

#' Returns a colorized warning message (cyan)
#'
#' @examples
#' cat(.warnMsg('This ', 'is ', 'a ', 'warning ', 'message', time = TRUE))
#'
#' @param ... `pairlist` One or more strings or character vectors, also
#'   accepts any params to paste0.
#' @param time `logical` Indicates whether to include timestamp or not. Default is FALSE.
#'
#' @return `character` Colorized string with results from paste0(...)
#'
#' @keywords internal
#' @export
#' @noRd
.warnMsg <- function(..., time = FALSE, collapse=', ') {
    msg <- paste0(..., collapse=collapse)
    if (time) {
        timestamp <- format(Sys.time(), "%Y-%m-%d %H:%M:%S")
        msg <- paste0(timestamp, " ", msg)
    }
    cyan$bold(msg)
}

#' Returns a colorized info message (green)
#' @examples
#' cat(.infoMsg('This ', 'is ', 'an ', 'info ', 'message', time = TRUE))
#' @param ... `pairlist` One or more strings or character vectors, also
#'  accepts any params to paste0.
#' @param time `logical` Indicates whether to include timestamp or not. Default is FALSE.
#' @return `character` Colorized string with results from paste0(...)
#' @keywords internal
#' @export
#' @noRd
#' @aliases .info
.infoMsg <- function(..., time = FALSE, collapse=', ') {
    msg <- paste0(..., collapse=collapse)
    if (time) {
        timestamp <- format(Sys.time(), "%Y-%m-%d %H:%M:%S")
        msg <- paste0(timestamp, " ", msg)
    }
    green$bold(msg)
}

#' Get the types of all items in a list
#'
#' @examples
#' list <- list(c(1,2,3), c('a','b','c'))
#' is.items(list, 'character')
#'
#' @param list A `list` to get the types from
#' @param ... `pairlist` Additional arguments to FUN
#' @param FUN `function` or `character` Either a function, or the name
#'   of a function which returns a single logical value. The default function
#'   uses `is`, specify the desired type in `...`. You can also use other
#'   type checking functions such as is.character, is.numeric, or is.data.frame.
#'
#' @return `logical` A vector indicating if the list item is the specified
#'   type.
#'
#' @export
is.items <- function(list, ..., FUN=is)
    vapply(list, FUN=FUN, FUN.VALUE=logical(1), ...)

#' @export
.length_unique <- function(x) length(unique(x))

#' @export
.list_unique <- function(x) list(unique(x))

#' @export
.all_equals <- function(x, y) all(x == y)


#' Return the name of the function and the name of the package that function
#'   is in when called within an R function.
#'
#' For providing context in user messages, warnings and errors
#'
#' @param n `integer` How far up the call stack to look for context. Defaults to
#'   2 since it is assumed this function will be used inside of `message`,
#'   `warning` or `stop`.
#'
#' @return A `character` vector with the name of the function
#'   `.getExecutionContext` was called from, as well as the package name,
#'   if applicable.
#'
#' @md
#' @keywords internal
#' @importFrom rlang trace_back
#' @importFrom utils packageName
#' @noRd
#' @aliases .context
.getExecutionContext <- function(n=2) {

    # name of function which called this function
    callStack <- rlang::trace_back()$calls
    context <- deparse(callStack[[length(callStack) - n]][1])

    # remove function arguments
    context <- gsub('\\(.*\\)', '', context)

    # deal with getting function names from inside an lapply statement
    ## TODO:: clean this up
    if (grepl('.*lapply.*', context)) {
        context <- deparse(callStack[[length(callStack) - (n + 1)]][3])
        context <- gsub('\\(.*\\)', '', context)
        # deal with S4 lapply calls (e.g., endoapply)
        if (grepl('.*match.fun.*', context)) {
            context <- deparse(callStack[[length(callStack) - (n + 5)]][3])
            context <- gsub('\\(.*\\)', '', context)
        }
    } else if (grepl('.*mapply.*', context)) {
        context <- deparse(callStack[[length(callStack) - (n + 1)]][1])
        context <- gsub('\\(.*\\)', '', context)
        if (grepl('.*match.fun.*', context)) {
            context <- deparse(callStack[[length(callStack) - (n + 5)]][1])
            context <- gsub('\\(.*\\)', '', context)
        }
    } else if (grepl('.*FUN.*', context)) {
        context <- deparse(callStack[[length(callStack) - (n + 2)]][3])
        context <- gsub('\\(.*\\)', '', context)
        # deal with S4 lapply calls (e.g., endoapply)
        if (grepl('.*match.fun.*', context)) {
            context <- tryCatch({
                deparse(callStack[[length(callStack) - (n + 6)]][3])
            }, error=function(e) 'context_failed')
            context <- gsub('\\(.*\\)', '', context)
        }
    }
    if (!grepl('::', context)) context <- paste0(packageName(), '::', context)

    return(paste0('\n[', context, '] '))
}
#' @noRd
.context <- .getExecutionContext


#'
#'
#'
#'
#' @md
#' @export
.S4MethodContext <- function(generic, ...) {
    dots <- list(...)
    formals <- selectMethod(generic, signature=dots)
    context <- paste0(
        formals@target@package[1], '::`', # what package is the method from
        formals@generic, ',',  # what is the name of the generic
        paste0(formals@target@.Data, collapse=','), '-method`') # what is the method signature
    return(context)
}

# Let it live here for now...

#' Build an assay table with an `S4` object.
#'
#' @param object `S4` An S4 object a list-like slot containing assays for the
#'   object.
#' @param ... Allow new arguments to be defined for this generic.
#'
#' @return `data.table`.
#'
#' @examples
#' "This is a generic method!"
#'
#' @exportMethod buildComboProfiles
setGeneric("buildComboProfiles", function(object, ...) standardGeneric("buildComboProfiles"))

#' Build an assay table with selected assay profiles for drug combinations
#'
#' @examples
#' \dontrun{
#' combo_profile_1 <- buildComboProfiles(tre, c("auc", "SCORE"))
#' combo_profile_2 <- buildComboProfiles(tre, c("HS", "EC50", "E_inf", "ZIP"))
#' }
#'
#' @param object `LongTable` or inheriting class containing curated drug combination data.
#' @param profiles `character` a vector of profile names, i.e., column names of assays.
#'
#' @return A `data.table` containing fields
#'   `treatment1id`, `treatment1dose`, `treatment2id`, `treatment2dose`, `sampleid`,
#'   which are used as keys to keep track of profiles,
#'   along with columns of selected profiles from their assays.
#'   Each `*_1` is the monothearpy profile of treatment 1 in the combination,
#'   and the same rule applies to treatment 2.
#'
#' @import data.table
#' @importFrom methods is
#' @export
#' @docType methods
setMethod("buildComboProfiles", signature(object = "LongTable"),
          function(object, profiles) {
    if (!is.character(profiles)) {
        stop("argument `profiles` must be `character`")
    } else if (length(profiles) == 0) {
        stop("argument `profiles` must not be empty")
    }

    if (is.null(object[["sensitivity"]])) {
        stop("Assay sensitivity is missing", call. = FALSE)
    } else if (!"treatment2id" %in% idCols(object)) {
        stop("This `TreatmentResponseExperiment` does not contain drug combination data.",
             call. = FALSE)
    }

    get_combo_viability <- ("combo_viability" %in% profiles)
    #if (get_combo_viability) {
    #    profiles <- profiles[!profiles %in% "combo_viability"]
    #    # and enable option for including combo viability
    #}

    combo_keys <- c("treatment1id", "treatment2id",
                    "treatment1dose", "treatment2dose", "sampleid")
    if (any(combo_keys %in% profiles)) {
        profiles <- profiles[!profiles %in% combo_keys]
        # and enable option for including dose here?
    }

    ## stop if none of the assays contain user selected profiles
    which_profiles <- lapply(assayCols(object), function(x) {
        if (any(x %in% profiles))
            return(x %in% profiles)
    })
    which_profiles[sapply(which_profiles, is.null)] <- NULL
    if (length(which_profiles) == 0)
        stop("No profiles found in any assay!")

    ## check whether there are profiles not present in assays
    profiles_exist <- vapply(profiles, function(x){
        any(sapply(assayCols(object), function(y) x %in% y))
    }, logical(1))
    profiles_not_exist <- names(profiles_exist)[which(profiles_exist == FALSE)]
    if (length(profiles_not_exist) > 0)
        warning(
            'No profiles named ',
            paste(profiles_not_exist, collapse = ", "),
            ' in any of the assays, thus will not be included in the returned table.',
            call. = FALSE
        )

    if (!is.null(object[["combo_viability"]])) {
        #if (get_combo_viability) {
        #    combo_profiles <- object[["combo_viability"]][,
        #        c(combo_keys, "combo_viability"), with = FALSE
        #    ]
        #} else {
        #    combo_profiles <- object[["combo_viability"]][,
        #        combo_keys, with = FALSE
        #    ]
        #}
        combo_profiles <- object[["combo_viability"]][,
            combo_keys, with = FALSE
        ]
        ## we know replicates have been averaged in combo_viability
    } else {
        #if (get_combo_viability) {
        #    object |>
        #        subset(!is.na(treatment2dose)) |>
        #        aggregate(
        #            assay = "sensitivity",
        #            combo_viability = (mean(viability) / 100),
        #            by = combo_keys
        #        ) -> combo_profiles
        #} else {
        #    combo_profiles <- unique(
        #        object$sensitivity[
        #            !is.na(treatment2dose),
        #            combo_keys,
        #            with = FALSE
        #        ],
        #        by = combo_keys
        #    )
        #}
        combo_profiles <- unique(
            object$sensitivity[
                !is.na(treatment2dose),
                combo_keys,
                with = FALSE
            ],
            by = combo_keys
        )
    }
    setkeyv(combo_profiles, combo_keys)

    assay_to_query <- names(which_profiles)

    ## how do we handle replicate rows?
    for (i in seq_along(which_profiles)) {
        assay_cols <- assayCols(object)[[assay_to_query[i]]]
        query_profiles <- assay_cols[which_profiles[[i]]]
        assay_ <- object[[assay_to_query[i]]]
        if (!("treatment2id" %in% assay_cols)) {
            ## Assays for monotherapy data
            ## Here I assume monotherapy assay tables have fewer keys
            ## and less cardinality than treatment combo tables
            ## might need extra condition check?
            monotherapy_keys <- c("treatment1id", "sampleid")
            if ("treatment1dose" %in% assay_cols)
                monotherapy_keys <- c(monotherapy_keys, "treatment1dose")
            assay_ <- assay_[, c(monotherapy_keys, query_profiles), with = FALSE]
            combo_profiles <- combo_profiles[assay_, ,
                on = c(treatment1id = "treatment1id", sampleid = "sampleid")
            ]
            ## remove single agents not tested in drug combination screening
            combo_profiles <- combo_profiles[!is.na(treatment2dose)]
            combo_profiles <- merge(
                combo_profiles,
                assay_,
                by.x = c("treatment2id", "sampleid"),
                by.y = c("treatment1id", "sampleid"),
                suffixes = c("_1", "_2")
            )
            ## remove single agents not tested in drug combination screening
            combo_profiles <- combo_profiles[!is.na(treatment1dose)]
            ## Edge case: adding profiles in sensitivity first
            ## then add monotherapy profiles
            replicates <- c("tech_rep", "bio_rep")
            has_reps <- replicates %in% colnames(combo_profiles)
            if (any(has_reps)) {
                setkeyv(combo_profiles, c(combo_keys, replicates[has_reps]))
            } else {
                setkeyv(combo_profiles, combo_keys)
            }
        } else {
            assay_keys <- key(assay_)
            common_keys <- intersect(key(combo_profiles), assay_keys)
            if (dim(combo_profiles)[1] >= dim(assay_)[1]) {
                ## Here assume less cardinality implies no replicates
                assay_ <- assay_[, c(common_keys, query_profiles), with = FALSE]
                setkeyv(assay_, common_keys)
                ## Left-join on combo_profiles
                combo_profiles <- combo_profiles[assay_, ,
                    on = common_keys
                ]
            } else { ## might have replicates
                ## likely to happen for profiles in sensitivity assay
                replicates <- c("tech_rep", "bio_rep")
                reps_in_common_keys <- replicates %in% common_keys
                reps_in_assay_keys <- replicates %in% assay_keys
                if (!all(reps_in_common_keys) & any(reps_in_assay_keys))
                    miss_rep_key <- xor(reps_in_common_keys, reps_in_assay_keys)
                if (any(miss_rep_key)) {
                    ## treat it as a profile to add first
                    add_rep_key <- replicates[which(reps_in_assay_keys)]
                    query_profiles <- c(query_profiles, add_rep_key)
                    setkeyv(assay_, c(common_keys, add_rep_key))
                }
                assay_ <- assay_[, c(common_keys, query_profiles), with = FALSE]
                ## Left-join on assay_, replicate has become a key if present in assay_
                combo_profiles <- assay_[combo_profiles, ,
                    on = common_keys
                ]
            }
        }
    }
    return(combo_profiles)
})

#' @include DataMapper-class.R
NULL

#' A Class for Mapping Between Raw Data and an `LongTable` Object
#'
#' @slot rawdata See Slots section.
#' @slot rowDataMap See Slots section.
#' @slot colDataMap See Slots section.
#' @slot assayMap See Slots section.
#' @slot metadataMap See Slots section.
#'
#' @inheritSection DataMapper-class Slots
#'
#' @section Slots:
#' * rowDataMap: A list-like object containing two `character` vectors.
#' The first is column names in `rawdata` needed to uniquely identify each row,
#' the second is additional columns which map to rows, but are not required to
#' uniquely identify them. Rows should be drugs.
#' * colDataMap: A list-like object containing two `character` vectors.
#' The first is column names in `rawdata` needed to uniquely identify each
#' column, the second is additional columns which map to rows, but are not
#' required to uniquely identify them. Columns should be samples.
#' * assayMap A list-like where each item is a `list` with two elements
#' specifying an assay, the first being the identifier columns in `rawdata`
#' needed to uniquely identify each row an assay, and the second a list of
#' `rawdata` columns to be mapped to that assay. The names of `assayMap`
#' will be the names of the assays in the `LongTable` that is created when
#' calling `metaConstruct` on this `DataMapper` object.
#' * metadataMap: A list-like where each item is a `character` vector of
#' `rawdata` column names to assign to the `@metadata` of the `LongTable`,
#' where the name of that assay is the name of the list item. If names are
#' omitted, assays will be numbered by their index in the list.
#'
#' @md
#' @aliases LongTableDataMapper-class
.LongTableDataMapper <- setClass('LongTableDataMapper',
    contains=c('DataMapper'),
    slots=list(
        rowDataMap='list_OR_List',
        colDataMap='list_OR_List',
        assayMap='list_OR_List',
        metadataMap='list_OR_List'
    )
)


#' Constructor for the `LongTableDataMapper` class, which maps from one or
#'   more raw experimental data files to the slots of a `LongTable` object.
#'
#' @details
#' The `guessMapping` method can be used to test hypotheses about the
#' cardinality of one or more sets of identifier columns. This is helpful
#' to determine the id columns for `rowDataMap` and `colDataMap`, as well
#' as identify columns mapping to `assays` or `metadata`.
#'
#' To attach metadata not associated with `rawdata`, please use the `metadata`
#' assignment method on your `LongTableDataMapper`. This metadata will be
#' merged with any metadata from `metadataMap` and added to the `LongTable`
#' which this object ultimately constructs.
#'
#' @param rawdata A `data.frame` of raw data from a treatment response
#' experiment. This will be coerced to a `data.table` internally. We recommend
#' using joins to aggregate your raw data if it is not present in a single file.
#' @param rowDataMap A list-like object containing two `character` vectors.
#' The first is column names in `rawdata` needed to uniquely identify each row,
#' the second is additional columns which map to rows, but are not required to
#' uniquely identify them. Rows should be drugs.
#' @param colDataMap A list-like object containing two `character` vectors.
#' The first is column names in `rawdata` needed to uniquely identify each
#' column, the second is additional columns which map to rows, but are not
#' required to uniquely identify them. Columns should be samples.
#' @param assayMap A list-like where each item is a `list` with two `character`
#' vectors defining an assay, the first containing the identifier columns in
#' `rawdata` needed to uniquely identify each row an assay, and the second the
#' `rawdata` columns to be mapped to that assay. The names of `assayMap`
#' will be the names of the assays in the `LongTable` that is created when
#' calling `metaConstruct` on this `DataMapper` object. If the character vectors
#' have names, the value columns will be renamed accordingly.
#' @param metadataMap A list-like where each item is a `character` vector of
#' `rawdata` column names to assign to the `@metadata` of the `LongTable`,
#' where the name of that assay is the name of the list item. If names are
#' omitted, assays will be numbered by their index in the list.
#'
#' @return A `LongTable` object, with columns mapped to it's slots according
#' to the various maps in the `LongTableDataMapper` object.
#'
#' @seealso [`guessMapping`]
#'
#' @examples
#' data(exampleDataMapper)
#' exampleDataMapper
#'
#' @md
#' @importFrom checkmate assertList assertDataTable
#' @importFrom data.table setDT copy
#' @export
LongTableDataMapper <- function(rawdata=data.frame(),
        rowDataMap=list(character(), character()),
        colDataMap=list(character(), character()),
        assayMap=list(list(character(), character())),
        metadataMap=list(character())) {

    if (is(rawdata, 'data.frame') && !is(rawdata, 'data.table')) setDT(rawdata)
    assertDataTable(rawdata)
    assertList(rowDataMap, types="character", len=2)
    assertList(colDataMap, types="character", len=2)
    assertList(assayMap, types="list", min.len=1)
    for (i in seq_along(assayMap)) assertList(assayMap[[i]], types="character",
        len=2)
    assertList(metadataMap)

    .LongTableDataMapper(rawdata=copy(rawdata), rowDataMap=rowDataMap,
        colDataMap=colDataMap, assayMap=assayMap, metadataMap=metadataMap)
}

## FIXME:: Modify rawdata setter to check that columns exist in maps for case
##>when maps are assigned first, then rawdata

# ======================
# DataMapper Show Method
# ----------------------


#' @describeIn LongTableDataMapper-class Show method for LongTableDataMapper.
#' Determines how the object is displayed in the console.
#'
#' @param object A `LongTableDataMapper` to display in the console.
#'
#' @examples
#' show(exampleDataMapper)
#'
#' @return `invisible` Prints to console.
#'
#' @importFrom crayon %+% yellow red green blue cyan magenta
#' @exportMethod show
setMethod('show', signature(object='LongTableDataMapper'), function(object) {

    ## -- class
    cat(yellow$bold$italic(paste0('<', class(object)[1], '>'), '\n'))

    missingVal <- ' NA'

    ## -- rawdata
    cat(yellow$bold('rawdata:'))
    if (length(rawdata(object))) {
        cat(paste0(' dim(', paste0(dim(rawdata(object)), collapse=', '), ')\n'))
        table_data <- capture.output(
            print(head(rawdata(object), 3), trunc.cols=TRUE, class=TRUE)
        )
        table_data[1] <- paste0('  ', table_data[1])
        rawdata_head <- paste0(
            paste0(table_data[-length(table_data)], collapse='\n  '),
            paste0(
                strwrap(table_data[length(table_data)], initial='\n  ', exdent=4),
                collapse='\n'
            ))
        cat(rawdata_head, '\n\r')  # print the snapshot
    } else {
        red(cat(missingVal, '\n'))
    }

    ## -- rowDataMap
    cat(yellow$bold('rowDataMap:'))
    rows <- rowDataMap(object)
    if (length(rows)) {
        cat('\n ', green('rowIDs:'), paste0(rows[[1]], collapse=', '))
    } else {
        cat(green(missingVal))
    }
    if (length(rows) > 1) {
        cat('\n ', green('rowMeta:'), paste0(rows[[2]], collapse=', '))
        cat('\n')
    } else {
        cat('\n')
    }

    ## -- colDataMap
    cat(yellow$bold('colDataMap:'))
    cols <- colDataMap(object)
    if (length(cols)) {
        cat('\n ', green('colIDs:'), paste0(cols[[1]], collapse=', '))
    } else {
        cat(green(missingVal))
    }
    if (length(cols) > 1) {
        cat('\n ', green('colMeta:'), paste0(cols[[2]], collapse=', '))
    }

    ## -- assayMap
    cat(yellow$bold('\nassayMap:'))
    assayM <- assayMap(object)
    if (length(assayM)) {
        for (aName in names(assayM)) {
            cat('\n ', red(paste0(aName, ':')))
            cat('\n    keys:',
                paste0(strwrap(
                    paste0(assayM[[aName]][[1]], collapse=', '), exdent=2),
                collapse="\n    ")
            )
            cat('\n    values:',
                paste0(strwrap(
                    paste0(assayM[[aName]][[2]], collapse=', '), exdent=2),
                collapse="\n    ")
            )
        }
    } else {
        cat(green(missingVal))
    }

    ## -- metadataMap
    cat(yellow$bold('\nmetadataMap:'))
    metadataM <- metadataMap(object)
    if (length(metadataM)) {
        for (mName in names(metadataM))
            cat('\n ', green(paste0(mName, ':')),
                paste0(metadataM[[mName]], collapse=', '))
    } else {
        cat(green(missingVal))
    }
    cat('\n')
})

## ===========================================
## LongTableDataMapper Accessors Documentation
## -------------------------------------------

.local_class_3 <- 'LongTableDataMapper'
.local_data_3 <- 'exampleDataMapper'

#' @name LongTableDataMapper-accessors
#'
#' @eval .docs_DataMapper_accessors(class_=.local_class_3)
#' @eval .docs_DataMapper_get_rawdata(class_=.local_class_3)
#'
#' @param value See details.
NULL

#' Drug sensitivity calling using waterfall plots
#'
#' 1. Sensitivity calls were made using one of IC50, ActArea or Amax
#'
#' 2. Sort log IC50s (or ActArea or Amax) of the samples to generate a
#'   “waterfall distribution”
#'
#' 3. Identify cutoff:
#'
#'  3.1 If the waterfall distribution is non-linear (pearson cc to the linear
#'    fit <=0.95), estimate the major inflection point of the log IC50 curve as
#'    the point on the curve with the maximal distance to a line drawn between
#'    the start and end points of the distribution.
#'
#'  3.2 If the waterfall distribution appears linear (pearson cc to the linear
#'    fit > 0.95), then use the median IC50 instead.
#'
#' 4. Samples within a 4-fold IC50 (or within a 1.2-fold ActArea or 20% Amax
#'   difference) difference centered around this inflection point are classified
#'   as being “intermediate”,  samples with lower IC50s (or ActArea/Amax
#'   values) than this range are defined as sensitive, and those with IC50s (or
#'   ActArea/Amax) higher than this range are called “insensitive”.
#'
#' 5. Require at least x sensitive and x insensitive samples after applying
#'   these criteria (x=5 in our case).
#'
## FIXME:: Write a real example
#' @examples
#' # Dummy example
#' 1 + 1
#'
## FIXME:: Clarify the parameters of this function
#' @param x What type of object does this take in?
#' @param type
#'   ic50: IC50 values in micro molar (positive values)
#'   actarea: Activity Area, that is area under the drug activity curve (positive values)
#'   amax: Activity at max concentration (positive values)
#'
#' @param intermediate.fold vector of fold changes used to define the intermediate sensitivities for ic50, actarea and amax respectively
#' @param cor.min.linear \code{numeric} The minimum linear correlation to
#'   require?
#' @param name \code{character} The name of the output to use in plot
#' @param plot \code{boolean} Whether to plot the results
#'
#' @return \code{factor} Containing the drug sensitivity status of each
#'   sample.
#'
#' @importFrom stats complete.cases  cor.test lm median
#' @importFrom graphics par points abline lines legend
#' @importFrom grDevices rainbow
#'
#' @export
#' @keywords internal
callingWaterfall <- function(x, type = c("IC50", "AUC", "AMAX"), intermediate.fold = c(4, 1.2, 1.2), cor.min.linear = 0.95, name = "Drug",
    plot = FALSE) {

    type <- match.arg(type)
    if (any(!is.na(intermediate.fold) & intermediate.fold < 0)) {
        intermediate.fold <- intermediate.fold[!is.na(intermediate.fold) & intermediate.fold < 0] <- 0
    }
    if (is.null(names(x))) {
        names(x) <- paste("X", seq_along(x), sep = ".")
    }

    xx <- x[complete.cases(x)]
    switch(type, IC50 = {
        xx <- -log10(xx)
        ylabel <- "-log10(IC50)"
        ## 4 fold difference around IC50 cutoff
        if (length(intermediate.fold) == 3) {
            intermediate.fold <- intermediate.fold[1]
        }
        if (intermediate.fold != 0) {
            interfold <- log10(intermediate.fold)
        } else {
            interfold <- 0
        }
    }, AUC = {
        ylabel <- "AUC"
        ## 1.2 fold difference around Activity Area cutoff
        if (length(intermediate.fold) == 3) {
            intermediate.fold <- intermediate.fold[2]
        }
        interfold <- intermediate.fold
    }, AMAX = {
        ylabel <- "Amax"
        ## 1.2 fold difference around Amax
        if (length(intermediate.fold) == 3) {
            intermediate.fold <- intermediate.fold[3]
        }
        interfold <- intermediate.fold
    })

    if (length(xx) < 3) {
        tt <- array(NA, dim = length(x), dimnames = list(names(x)))
        if (interfold == 0) {
            tt <- factor(tt, levels = c("resistant", "sensitive"))
        } else {
            tt <- factor(tt, levels = c("resistant", "intermediate", "sensitive"))
        }
        return(tt)
    }

    oo <- order(xx, decreasing = TRUE)
    ## test linearity with Pearson correlation
    cc <- stats::cor.test(-xx[oo], seq_along(oo), method = "pearson")
    ## line between the two extreme sensitivity values
    dd <- cbind(y = xx[oo][c(1, length(oo))], x = c(1, length(oo)))
    rr <- lm(y ~ x, data = data.frame(dd))
    ## compute distance from sensitivity values and the line between the two extreme sensitivity values
    ddi <- apply(cbind(seq_along(oo), xx[oo]), 1, function(x, slope, intercept) {
        return(.distancePointLine(x = x[1], y = x[2], a = slope, b = intercept))
    }, slope = rr$coefficients[2], intercept = rr$coefficients[1])
    if (cc$estimate > cor.min.linear) {
        ## approximately linear waterfall
        cutoff <- which.min(abs(xx[oo] - median(xx[oo])))
        cutoffn <- names(cutoff)[1]
    } else {
        ## non linear waterfall identify cutoff as the maximum distance
        cutoff <- which.max(abs(ddi))
        cutoffn <- names(ddi)[cutoff]
    }
    ## identify intermediate sensitivities
    switch(type, IC50 = {
        if (interfold == 0) {
            rang <- c(xx[oo][cutoff], xx[oo][cutoff])
        } else {
            rang <- c(xx[oo][cutoff] - interfold, xx[oo][cutoff] + interfold)
        }
    }, AUC = {
        if (interfold == 0) {
            rang <- c(xx[oo][cutoff], xx[oo][cutoff])
        } else {
            rang <- c(xx[oo][cutoff]/interfold, xx[oo][cutoff] * interfold)
        }
    }, AMAX = {
        if (interfold == 0) {
            rang <- c(xx[oo][cutoff], xx[oo][cutoff])
        } else {
            rang <- c(xx[oo][cutoff]/interfold, xx[oo][cutoff] * interfold)
        }
    })


    ## check whether range is either min or max
    if (rang[2] >= max(xx)) {
        rang[2] <- sort(unique(xx), decreasing = TRUE)[2]
    }
    if (rang[2] <= min(xx)) {
        rang[2] <- sort(unique(xx), decreasing = FALSE)[2]
    }
    if (rang[1] <= min(xx)) {
        rang[1] <- sort(unique(xx), decreasing = FALSE)[2]
    }
    if (rang[1] >= max(xx)) {
        rang[1] <- sort(unique(xx), decreasing = TRUE)[2]
    }

    ## compute calls
    calls <- rep(NA, length(xx))
    names(calls) <- names(xx)
    calls[xx < rang[1]] <- "resistant"
    calls[xx >= rang[2]] <- "sensitive"
    calls[xx >= rang[1] & xx < rang[2]] <- "intermediate"

    if (plot) {
        par(mfrow = c(2, 1))
        ccols <- rainbow(4)
        mycol <- rep("grey", length(xx))
        names(mycol) <- names(xx)
        mycol[calls == "sensitive"] <- ccols[2]
        mycol[calls == "intermediate"] <- ccols[3]
        mycol[calls == "resistant"] <- ccols[4]
        mycol[cutoffn] <- ccols[1]
        mypch <- rep(16, length(xx))
        names(mypch) <- names(xx)
        mypch[cutoffn] <- 19
        plot(xx[oo], col = mycol[oo], pch = mypch[oo], ylab = ylabel, main = sprintf("%s\nWaterfall", name))
        points(x = cutoff, y = xx[cutoffn], pch = mypch[cutoffn], col = mycol[cutoffn])
        graphics::abline(a = rr$coefficients[1], b = rr$coefficients[2], lwd = 2, col = "darkgrey")
        lines(x = c(cutoff, cutoff), y = c(par("usr")[3], xx[cutoffn]), col = "red")
        lines(x = c(par("usr")[1], cutoff), y = c(xx[cutoffn], xx[cutoffn]), col = "red")
        legend("topright", legend = c(sprintf("resistant (n=%i)", sum(!is.na(calls) & calls == "resistant")), sprintf("intermediate (n=%i)",
            sum(!is.na(calls) & calls == "intermediate")), sprintf("sensitive (n=%i)", sum(!is.na(calls) & calls == "sensitive")), "cutoff",
            sprintf("R=%.3g", cc$estimate)), col = c(rev(ccols), NA), pch = c(16, 16, 16, 19, NA), bty = "n")

        plot(ddi, pch = mypch[oo], col = mycol[oo], ylab = "Distance", main = sprintf("%s\n%s", name, "Distance from min--max line"))
        points(x = cutoff, y = ddi[cutoffn], pch = mypch[cutoffn], col = mycol[cutoffn])
        legend("topright", legend = c("resistant", "intermediate", "sensitive", "cutoff"), col = rev(ccols), pch = c(16, 16, 16, 19), bty = "n")
    }

    tt <- rep(NA, length(x))
    names(tt) <- names(x)
    tt[names(calls)] <- calls
    if (interfold == 0) {
        tt <- factor(tt, levels = c("resistant", "sensitive"))
    } else {
        tt <- factor(tt, levels = c("resistant", "intermediate", "sensitive"))
    }
    return(tt)
}


# Helper Functions --------------------------------------------------------

#' Calculate shortest distance between point and line
#'
#' @examples .distancePointLine(0, 0, 1, -1, 1)
#'
#' @description This function calculates the shortest distance between a point
#'   and a line in 2D space.
#'
#' @param x x-coordinate of point
#' @param y y-coordinate of point
#' @param a `numeric(1)` The coefficient in line equation a * x + b * y + c = 0.
#'   Defaults to 1.
#' @param b `numeric(1)` The coefficient in line equation a * x + b * y + c = 0.
#'   Defaults to 1.
#' @param c `numeric(1)` The intercept in line equation a * x + b * y + c = 0.
#'   Defaults to 0.
#'
#' @return `numeric` The shortest distance between a point and a line.
#'
#' @export
#' @keywords internal
.distancePointLine <- function(x, y, a=1, b=1, c=0) {

    if (!(all(is.finite(c(x, y, a, b, c))))) {
        stop("All inputs to .distancePointLine must be real numbers.")
    }

    return(abs(a * x + b * y + c) / sqrt(a^2 + b^2))
}

#' @export
#' @keywords internal
.magnitude <- function(p1, p2) {
    return(sqrt(sum((p2 - p1)^2)))
}

#' Calculate shortest distance between point and line segment
#'
#' @description This function calculates the shortest distance between a point
#'   and a line segment in 2D space.
#'
#' @param x x-coordinate of point
#' @param y y-coordinate of point
#' @param x1 x-coordinate of one endpoint of the line segment
#' @param y1 y-coordinate of line segment endpoint with x-coordinate x1
#' @param x2 x-coordinate of other endpoint of line segment
#' @param y2 y-coordinate of line segment endpoint with x-coordinate x2
#'
#' @return \code{numeric} The shortest distance between a point and a line
#'   segment
#'
#' @examples .distancePointSegment(0, 0, -1, 1, 1, -1)
#'
#' @export
#' @keywords internal
.distancePointSegment <- function(x, y, x1, y1, x2, y2) {
    if (!(all(is.finite(c(x, y, x1, x2, y1, y2))))) {
        stop("All inputs to linePtDist must be real numbers.")
    }

    bestEndpointDistance <- min(c(.magnitude(c(x, y), c(x1, y1)), .magnitude(c(x, y), c(x2, y2))))

    if (.magnitude(c(x, y), c((x1 + x2)/2, (y1 + y2)/2)) < bestEndpointDistance) {
        # length to point has only one local minimum which is the global minimum; iff this condition is true then the shortest distance is to a
        # point on the line segment vertical line segment
        if (x1 == x2) {
            a <- 1
            b <- 0
            c <- -x1
        } else {
            a <- (y2 - y1)/(x2 - x1)
            b <- -1
            c <- y1 - a * x1
        }
        return(.distancePointLine(x, y, a, b, c))
    } else {
        return(bestEndpointDistance)
    }
}

#' @title Constructor for "immutable" S3-class property
#'
#' @description
#' This method should allow any S3 object in R to become immutable by
#' intercepting `[<-`, `[[<-`, `$<-` and `c` during S3-method dispatch and
#' returning an error.
#'
#' Reverse with call to the `mutable` function.
#'
#' @details
#' The motivation for this class was to create pseudo-private slots in an R
#' S4 object by preventing mutation of those slots outside of the accessors
#' written for the class. It should behave as expected for R object which
#' operate with 'copy-on-modify' semantics, including most base R functions and
#' S3 objects.
#'
#' An environment was not suitable for this case due
#' to the 'copy-by-reference' semantics, such that normal R assignment, which
#' users assume makes a copy of the object, actually references the same
#' environment in both the original and copy of the object.
#'
#' WARNING: This implementation is unable to intercept modifications to a
#' `data.table` via the `set*` group of methods. This is because these methods
#' are not S3 generics and therefore no mechanism exists for hooking into them
#' to extend their functionality. In general, this helper class will only work
#' for objects with an S3 interface.
#'
#' @param object,x Any R object which uses S3 method dispatch
#'
#' @return The `object` with "immutable" prepended to its class attribute.
#'
#' @examples
#' immutable_list <- immutable(as.list(1:5))
#' class(immutable_list)
#' # errors during assignment operations
#' tryCatch({ immutable_list$new <- 1 }, error=print)
#'
#' @seealso
#' [`assignment-immutable`], [`setOps-immutable`]
#'
#' @md
#' @rdname immutable
#' @name immutable
#' @export
immutable <- function(object) {
    if (isS4(object)) stop("Can only set immutability for base and S3 classes!")
    # call mutable to prevent assigning immutable class twice
    structure(mutable(object), class=c("immutable", attributes(object)$class))
}

# register the new S3 class, so it can be used in S4 method dispatch
#' @rdname immutable
#' @name immutable
#' @export
setOldClass("immutable")

#' @rdname immutable
#' @export
setClassUnion("immutable_list", c("immutable", "list"))


#' @title Check if an R object inherits from the immutable S3-class.
#'
#' @return `logical(1)` Does the object inherit from the "immutable" S3-class?
#'
#' @examples
#' immutable_list <- immutable(as.list(1:5))
#' is.immutable(immutable_list)
#'
#' @rdname immutable
#' @export
is.immutable <- function(object) {
    is(object, "immutable")
}


#' @title Print method for objects inheriting from the "immutable" S3-class
#'
#' @param ... Fallthrough arguments to `print.default`.
#'
#' @return None, `invisible(NULL)`
#'
#' @rdname immutable
#' @md
#' @export
print.immutable <- function(x, ...) {
    other_cls <- setdiff(attributes(x)$class, "immutable")
    class(x) <- other_cls
    cat("immutable", class(x), "\n")
    print(x, ...)
}

#' @rdname immutable
#' @export
show.immutable <- function(x) print(x)


# -- Intercept subset and concatentate operations to return another "immutable"

#' @title Intercept concatenation for "immutable" class objects to return another "immutable" class object.
#'
#' @description
#' Ensures that `c` and `append` to an "immutable" class object return an
#' immutable class object.
#'
#' @param x An R object inheriting from the "immutable" S3-clas
#' @param ... Objects to concatenate to `x`.
#'
#' @return x with one or more values appended to it.
#'
#' @md
#' @export
c.immutable <- function(x, ...) {
    new_obj <- NextMethod()
    immutable(new_obj)
}

.immutable_emsg <- "Object is immutable! Use `mutable(object)` to return a mutable copy."

#' @name setOps-immutable
#' @rdname setOps-immutable
#'
#' @title Subset an immutable object, returning another immutable object.
#'
#' @param x An R object inheriting from the "immutable" S3-class.
#' @param ... Catch any additional parameters. Lets objects with arbitrary
#' dimensions be made immutable.
#'
#' @return An immutable subset of `x`.
#'
#' @examples
#' immut_mat <- immutable(matrix(1:100, 10, 10))
#' immut_mat[1:5, 1:5]
#'
#' @md
#' @aliases subset, [.immutable, [[.immutable, $.immutable
#' @export
subset.immutable <- function(x, ...) {
    sub_obj <- NextMethod()
    immutable(sub_obj)
}
#' @name [
#' @rdname setOps-immutable
#' @export
`[.immutable` <- function(x, ...) {
    if (is.data.table(x)) {
        dots <- substitute(alist(...))
        j_expr <- dots[["j"]]
        if (is.null(j_expr) && length(dots) > 2)
            j_expr <- dots[[2 + 1]]  # index plus one due to alist call
        if (!is.null(j_expr))
            j_txt <- deparse(j_expr)
            if (grepl(":=|let[ ]*\\(|set[ ]*\\(", j_txt))
                stop("This data.table is immutable! No assignment by reference ",
                "allowed. Use `mutable(x)` to return a mutable copy.",
                    call.=FALSE)
    }
    sub_obj <- NextMethod()
    immutable(sub_obj)
}
#' @name [[
#' @rdname setOps-immutable
#' @export
`[[.immutable` <- function(x, ...) {
    sub_obj <- NextMethod()
    immutable(sub_obj)
}
#' @name $
#' @rdname setOps-immutable
#' @export
`$.immutable` <- function(x, ...) {
    sub_obj <- NextMethod()
    immutable(sub_obj)
}

# -- Intercept assignment to prevent modification

#' @name assignment-immutable
#' @rdname assignment-immutable
#'
#' @title Intercept assignment operations for "immutable" S3 objects.
#'
#' @description
#' Prevents modification of objects labelled with the "immutable" S3-class by
#' intercepting assignment during S3-method dispatch and returning an error.
#'
#' @param object,x An R object inherting from the "immutable" S3-class.
#' @param ... Catch subset arguments for various dimensions.
#' @param value Not used.
#'
#' @return None, throws an error.
#'
#' @examples
#' immutable_df <- immutable(data.frame(a=1:5, b=letters[1:5]))
#' # return immutable data.frame
#' immutable_df[1:4, ]
#' # return immutable vector
#' immutable_df$a
#'
#' @md
#' @usage \\method{subset}{immutable}(object, ...) <- value
#' @aliases subset<-.immutable, [<-.immutable, [[<-.immutable, $<-.immubtale,
#' colnames<-.immutable, rownames<-.immutable, dimnames<-.immutable,
#' names<-.immutable
#' @export
`subset<-.immutable` <- function(object, ..., value) {
    stop(.immutable_emsg, call.=FALSE)
}
#' @name [<-
#' @rdname assignment-immutable
#' @export
`[<-.immutable` <- function(object, ..., value) {
    stop(.immutable_emsg, call.=FALSE)
}
#' @name [[<-
#' @rdname assignment-immutable
#' @export
`[[<-.immutable` <- function(object, ..., value) {
    stop(.immutable_emsg, call.=FALSE)
}
#' @name $<-
#' @rdname assignment-immutable
#' @export
`$<-.immutable` <- function(object, ..., value) {
    stop(.immutable_emsg, call.=FALSE)
}
#' @name names<-
#' @rdname assignment-immutable
#' @export
`names<-.immutable` <- function(x, value)
    stop(.immutable_emsg, call.=FALSE)
#' @name dimnames<-
#' @rdname assignment-immutable
#' @export
`dimnames<-.immutable` <- function(x, value)
    stop(.immutable_emsg, call.=FALSE)
#' @name colnames<-
#' @rdname assignment-immutable
#' @usage \\method{colnames}{immutable}(x) <- value
#' @export
`colnames<-.immutable` <- function(x, value)
    stop(.immutable_emsg, call.=FALSE)
#' @name rownames<-
#' @rdname assignment-immutable
#' @usage \\method{rownames}{immutable}(x) <- value
#' @export
`rownames<-.immutable` <- function(x, value)
    stop(.immutable_emsg, call.=FALSE)


# -- Remove immutability from an R object


#' @title Remove the "immutable" S3-class from an R object, allowing it to be
#' modified normally again.
#'
#' @param object An R object inheriting from the "immutable" class.
#'
#' @return The `object` with the "immutable" class stripped from it.
#'
#' @examples
#' immut_list <- immutable(list())
#' mutable(immut_list)
#'
#' @md
#' @export
mutable <- function(object) UseMethod("mutable", object)
#'
#' @md
#' @importFrom data.table copy
#' @export
mutable.default <- function(object) {
    new_class <- setdiff(attributes(object)$class, "immutable")
    structure(copy(object), class=new_class)
}


# -- Make comparisons work for immutable objects

#' @export
Ops.immutable <- function(e1, e2) {
    get(.Generic)(mutable(e1), mutable(e2))
}

#' @include allGenerics.R
#' @include aggregate-methods.R
NULL

#' Functional API for endomorphic aggregation over a `LongTable` or
#' inheriting class
#'
#' @description
#' Compute a group-by operation over a `LongTable` object or its inhering
#' classes.
#'
#' @param x `LongTable` or inheriting class to compute aggregation on.
#' @param assay `character(1)` The assay to aggregate over.
#' @param target `character(1)` The assay to assign the results to. Defaults
#' to `assay`.
#' @param subset `call` An R call to evaluate before perfoming an aggregate.
#' This allows you to aggregate over a subset of columns in an assay but have
#' it be assigned to the parent object. Default is TRUE, which includes all
#' rows. Passed through as the `i` argument in `[.data.table`.
#' @eval .docs_CoreGx_aggregate(curly="{")
#'
#' @return Object with the same class as `x`, with the aggregation results
#' assigned to `target`, using `strategy` if `target` is an existing assay in
#' `x`.
#'
#' @seealso `data.table::[.data.table`, `BiocParallel::bplapply`
#'
#' @export
setMethod("endoaggregate", signature(x="LongTable"),
        function(x, ..., assay, target=assay, by, subset=TRUE, nthread=1,
        progress=TRUE, BPPARAM=NULL, enlist=TRUE, moreArgs=list()) {
    i <- substitute(subset)
    assay_ <- x[[assay]][eval(i), ]
    res <- aggregate2(
        assay_,
        by=by,
        ...,
        nthread=nthread, progress=progress, BPPARAM=BPPARAM, enlist=enlist,
        moreArgs=moreArgs
    )
    if (target %in% assayNames(x)) {
        res <- merge.data.table(x[[assay]], res, by=by)
    }
    x[[target]] <- res
    x
})

#' Generic for Guessing the Mapping Between Some Raw Data and an S4 Object
#'
#' @param object An `S4` object containing so raw data to guess data to
#'   object slot mappings for.
#' @param ... Allow new arguments to be defined for this generic.
#'
#' @return A `list` with mapping guesses as items.
#'
#' @examples
#' "Generics shouldn't need examples!"
#'
#' @md
#' @export
setGeneric('guessMapping', function(object, ...) standardGeneric('guessMapping'))

#' Guess which columns in raw experiment data map to which dimensions.
#'
#' Checks for columns which are uniquely identified by a group of identifiers.
#' This should be used to help identify the columns required to uniquely
#' identify the rows, columns, assays and metadata of a `DataMapper` class
#' object.
#'
#' @details
#' Any unmapped columns will be added to the end of the returned `list` in an
#' item called unmapped.
#'
#' The function automatically guesses metadata by checking if any columns have
#' only a single value. This is returned as an additional item in the list.
#'
#' @param object A `LongTableDataMapper` object.
#' @param groups A `list` containing one or more vector of column names
#'   to group-by. The function uses these to determine 1:1 mappings between
#'   the combination of columns in each vector and unique values in the raw
#'   data columns.
#' @param subset A `logical` vector indicating whether to to subset out mapped
#'   columns after each grouping. Must be a single `TRUE` or `FALSE` or have
#'   the same length as groups, indicating whether to subset out mapped columns
#'   after each grouping. This will prevent mapping a column to two different
#'   groups.
#' @param data A `logical` vector indicating whether you would like the data
#'   for mapped columns to be returned instead of their column names. Defaults
#'   to `FALSE` for easy use assigning mapped columns to a `DataMapper` object.
#'
#' @return A `list`, where each item is named for the associated `groups` item
#' the guess is for. The character vector in each item are columns which are
#' uniquely identified by the identifiers from that group.
#'
#' @examples
#' guessMapping(exampleDataMapper, groups=list(rows='treatmentid', cols='sampleid'),
#' subset=FALSE)
#'
#' @md
#' @export
setMethod('guessMapping', signature(object='LongTableDataMapper'),
        function(object, groups, subset, data=FALSE) {
    funContext <- '[CoreGx::guessMapping,LongTableDataMapper-method]\n\t'

    # Extract the raw data
    mapData <- copy(rawdata(object))
    if (!is.data.table(mapData)) setDT(mapData)

    # Error handle for subset parameter
    if (!(length(subset) == length(groups) || length(subset) == 1))
        stop(.errorMsg(funContext, ' The subset parameter must be
            either length 1 or length equal to the groups parameter!'))

    if (length(subset) == 1) subset <- rep(subset, length(groups))

    # Get the id columns to prevent subsetting them out
    idCols <- unique(unlist(groups))

    # Map unique columns in the data to the metadata slot
    metadataColumns <- names(which(vapply(mapData, FUN=.length_unique,
        numeric(1)) == 1))
    metadataColumns <- setdiff(metadataColumns, idCols)
    metadata <- mapData[, .SD, .SDcols=metadataColumns]
    DT <- mapData[, .SD, .SDcols=!metadataColumns]

    # Check the mappings for each group in groups
    for (i in seq_along(groups)) {
        message(funContext, paste0('Mapping for group ', names(groups)[i],
            ': ', paste0(groups[[i]], collapse=', ')))
        mappedCols <- checkColumnCardinality(DT, groups[[i]])
        mappedCols <- setdiff(mappedCols, idCols)
        assign(names(groups)[i], DT[, .SD, .SDcols=mappedCols])
        if (subset[i]) DT <- DT[, .SD, .SDcols=!mappedCols]
    }

    # Merge the results
    groups <- c(list(metadata=NA), groups)
    mappings <- mget(names(groups))
    unmapped <- setdiff(colnames(mapData),
        unique(c(unlist(groups), unlist(lapply(mappings, colnames)))))
    if (!data) mappings <- lapply(mappings, colnames)
    mappings <- Map(f=list, groups, mappings)
    mappings <- lapply(mappings, FUN=setNames,
        nm=c('id_columns', 'mapped_columns'))

    mappings[['unmapped']] <- unmapped

    return(mappings)
})

#' Search a data.frame for 1:`cardinality` relationships between a group
#'   of columns (your identifiers) and all other columns.
#'
#' @param df A `data.frame` to search for 1:`cardinality` mappings with
#'   the columns in `group`.
#' @param group A `character` vector of one or more column names to
#'   check the cardinality of other columns against.
#' @param cardinality The cardinality of to search for (i.e., 1:`cardinality`)
#'   relationships with the combination of columns in group. Defaults to 1
#'   (i.e., 1:1 mappings).
#' @param ... Fall through arguments to data.table::`[`. For developer use.
#'   One use case is setting verbose=TRUE to diagnose slow data.table
#'   operations.
#'
#' @return A `character` vector with the names of the columns with
#'    cardinality of 1:`cardinality` with the columns listed in `group`.
#'
#' @examples
#' df <- rawdata(exampleDataMapper)
#' checkColumnCardinality(df, group='treatmentid')
#'
#' @aliases cardinality
#'
#' @md
#' @importFrom data.table setindexv
#' @importFrom MatrixGenerics colAlls
#' @export
checkColumnCardinality <- function(df, group, cardinality=1, ...) {

    funContext <- '\n[CoreGx::checkColumnCardinality]\n\t'

    # Copy to prevent accidental modify by references
    df <- copy(df)
    if (!is.data.table(df)) setDT(df)

    # Intercept slow data.table group by when nrow == .NGRP
    setkeyv(df, cols=group)
    nrowEqualsNGroup <- df[, .N, by=group, ...][, max(N), ...] == 1
    if (nrowEqualsNGroup) {
        if (cardinality != 1) stop(.errorMsg(funContext, 'The group argument
            uniquely identifies each row, so the cardinality is 1:1!'))
        columnsHaveCardinality <- setdiff(colnames(df), group)
    } else {
        dimDT <- df[, lapply(.SD, FUN=uniqueN), by=group, ...]
        columnsHaveCardinality <- colnames(dimDT)[colAlls(dimDT == cardinality)]
    }

    return(columnsHaveCardinality)
}
#' @export
cardinality <- checkColumnCardinality

## Matthews correlatipon coefficient
## TODO:: Give this function a more descriptive name
#' Compute a Mathews Correlation Coefficient
#'
#' The function computes a Matthews correlation coefficient for two factors
#' provided to the function. It assumes each factor is a factor of class labels,
#' and the enteries are paired in order of the vectors.
#'
#' Please note: we recommend you call set.seed() before using this function to
#' ensure the reproducibility of your results. Write down the seed number or
#' save it in a script if you intend to use the results in a publication.
#'
#' @examples
#' x <- factor(c(1,2,1,2,3,1))
#' y <- factor(c(2,1,1,1,2,2))
#' mcc(x,y)
#'
#' @param x,y \code{factor} of the same length with the same number of levels
#' @param nperm \code{numeric} number of permutations for significance
#' estimation. If 0, no permutation testing is done
#' @param nthread \code{numeric} can parallelize permutation texting using
#'   BiocParallels bplapply
#' @param alternative indicates the alternative hypothesis and must be one of
#' ‘"two.sided"’, ‘"greater"’ or ‘"less"’.  You can specify just
#' the initial letter.  ‘"greater"’ corresponds to positive
#' association, ‘"less"’ to negative association.
#' @param ... \code{list} Additional arguments
#'
#' @return A list with the MCC as the $estimate, and p value as $p.value
#' @export
mcc <- function(x, y, nperm = 1000, nthread = 1, alternative=c("two.sided", "less", "greater"), ...) {
    # PARAMETER CHANGE WARNING
    if (!missing(...)) {
        if ("setseed" %in% names(...)) {
            warning("The setseed parameter has been removed in this release to conform
              to Bioconductor coding standards. Please call set.seed in your
              script before running this function.")
        }
    }
    alternative <- match.arg(alternative)
    if ((length(x) != length(y)) || (!is.factor(x) || length(levels(x)) < 2) || (!is.factor(y) || length(levels(y)) < 2)) {
        stop("x and y must be factors of the same length with at least two levels")
    }
    if (length(levels(x)) != length(levels(y))) {

        warning("The number of levels x and y was different. Taking the union of all class labels.")
        levels(x) <- union(levels(x), levels(y))
        levels(y) <- union(levels(x), levels(y))

    }
    res <- list(estimate = NA, p.value = NA)
    ## compute MCC
    res[["estimate"]] <- .mcc(ct = table(x, y))
    ## compute significance of MCC using a permutation test
    if (nperm > 0) {
        splitix <- parallel::splitIndices(nx = nperm, ncl = nthread)
        splitix <- splitix[vapply(splitix, length, FUN.VALUE = numeric(1)) > 0]
        mcres <- BiocParallel::bplapply(splitix, function(x, xx, yy) {
            res <- vapply(x, function(x, xx, yy) {
                xx <- sample(xx)
                yy <- sample(yy)
                return(.mcc(ct = table(xx, yy)))
            }, xx = xx, yy = yy, FUN.VALUE = numeric(1))
            return(res)
        }, xx = x, yy = y)
        mcres <- unlist(mcres)
        #res[["p.value"]] <- sum(mcres > res["estimate"])/sum(!is.na(mcres))
        switch(alternative, "two.sided" = {
            if(res[["estimate"]] > 0) {
                res[["p.value"]] <- 2*sum(res[["estimate"]] <=  mcres )/sum(!is.na(mcres))
            } else {
                res[["p.value"]] <- 2*sum(res[["estimate"]] >=  mcres )/sum(!is.na(mcres))
            }
        }, "less" = {
            res[["p.value"]] <- sum(res[["estimate"]] >=  mcres )/sum(!is.na(mcres)) 
        }, "greater" = {
            res[["p.value"]] <- sum(res[["estimate"]] <=  mcres )/sum(!is.na(mcres))
        })
        if (res["p.value"] == 0) {
            res["p.value"] <- 1/(nperm + 1)
        }
    }
    return(res)
}

## Helper functions


## Just a lot of math, multiclass MCC
## https://en.wikipedia.org/wiki/Matthews_correlation_coefficient#Multiclass_case
.mcc <-
  function(ct, nbcat=nrow(ct)) {
    if(nrow(ct) != ncol(ct)) { stop("the confusion table should be square!") }

    ## The following code checks if there would be a division by 0 error in the computation on the Matthew's
    ## correlation coefficient. In practice, this occurs when there are multiple categories but all predictions end up
    ## in only one category - in this case, the denominator is 0. This is dealt with by adding a psuedocount to each.
    ## This is chosen because the mcc of a matrix of 1s is 0, and such should not affect the value in expectation.
    ## Note this is not necessary when alll entries lie on the diagonal.
    if( !(sum(ct)==sum(diag(ct))) && ## Check if all entries are on the diagonal. If they are, no adjustment necessary.
        ((sum(rowSums(ct) == 0) == (nbcat-1)) || ## Otherwise, check if there is entries only in one column or only in one row.
        (sum(colSums(ct) == 0) == (nbcat-1)))) {
        ct <- ct + matrix(1, ncol=nbcat, nrow=nbcat)
      } ### add element to categories if nbcat-1 predictive categories do not contain elements. Not in case where all are correct!

    if (sum(ct, na.rm = TRUE) <= 0) {
        return(NA)
    }

    myx <- matrix(TRUE, nrow = nrow(ct), ncol = ncol(ct))
    diag(myx) <- FALSE
    if (sum(ct[myx]) == 0) {
        return(1)
    }
    myperf <- 0
    for (k in seq_len(nbcat)) {
        for (m in seq_len(nbcat)) {
            for (l in seq_len(nbcat)) {
                myperf <- myperf + ((ct[k, k] * ct[m, l]) - (ct[l, k] * ct[k, m]))
            }
        }
    }
    aa <- 0
    for (k in seq_len(nbcat)) {
        cc <- 0
        for (l in seq_len(nbcat)) {
            cc <- cc + ct[l, k]
        }
        dd <- 0
        for (f in seq_len(nbcat)) {
            for (g in seq_len(nbcat)) {
                if (f != k) {
                  dd <- dd + ct[g, f]
                }
            }
        }
        aa <- aa + (cc * dd)
    }
    bb <- 0
    for (k in seq_len(nbcat)) {
        cc <- 0
        for (l in seq_len(nbcat)) {
            cc <- cc + ct[k, l]
        }
        dd <- 0
        for (f in seq_len(nbcat)) {
            for (g in seq_len(nbcat)) {
                if (f != k) {
                  dd <- dd + ct[f, g]
                }
            }
        }
        bb <- bb + (cc * dd)
    }

    myperf <- myperf/(sqrt(aa) * sqrt(bb))
    return(myperf)
}

#' Cosine Permutations
#'
#' Computes the cosine similarity and significance using permutation test. This
#'   function uses random numbers, to ensure reproducibility please call 
#'   \code{set.seed()} before running the function.
#' 
#' @examples
#' x <- factor(c(1,2,1,2,1))
#' y <- factor(c(2,2,1,1,1))
#' cosinePerm(x, y)
#' 
#' @param x \code{factor} is the factors for the first variable
#' @param y \code{factor} is the factors for the second variable
#' @param nperm \code{integer} is the number of permutations to compute the null
#'   distribution of MCC estimates
#' @param alternative \code{string} indicates the alternative hypothesis and 
#'   must be one of ‘'two.sided'’, ‘'greater'’ or ‘'less'’.  You can specify 
#'   just the initial letter.  ‘'greater'’ corresponds to positive association, 
#'   ‘'less'’ to negative association. Options are 'two.sided', 'less', or 
#'   'greater'
#' @param include.perm \code{boolean} indicates whether the estimates for the 
#'   null distribution should be returned. Default set to 'FALSE'
#' @param nthread \code{integer} is the number of threads to be used to perform 
#'   the permutations in parallel
#' @param ... A \code{list} of fallthrough parameters 
#' 
#' @return A \code{list} estimate of the cosine similarity, p-value and 
#'   estimates after random permutations (null distribution) in include.perm is 
#'   set to 'TRUE'
#' 
#' @importFrom lsa cosine
#' @importFrom BiocParallel bplapply
#' 
#' @export
cosinePerm <- function(x, y, nperm = 1000, alternative = c("two.sided", "less", "greater"), include.perm = FALSE, nthread = 1, 
    ...) {
    # PARAMETER CHANGE WARNING
    if (!missing(...)) {
        if ("setseed" %in% names(...)) {
            warning("The setseed parameter has been removed in this release to conform
              to Bioconductor coding standards. Please call set.seed in your
              script before running this function.")
        }
    }
    
    alternative <- match.arg(alternative)
    if ((length(x) != length(y))) {
        stop("x and y must be vectors of the same length")
    }
    res <- c(estimate = NA, p.value = NA)
    x <- as.numeric(x)
    y <- as.numeric(y)
    ## compute cosine
    res["estimate"] <- drop(lsa::cosine(x = x, y = y))
    ## compute significance of cosine using a permutation test
    if (nperm > 0) {
        splitix <- parallel::splitIndices(nx = nperm, ncl = nthread)
        splitix <- splitix[vapply(splitix, length, FUN.VALUE = numeric(1)) > 0]
        mcres <- BiocParallel::bplapply(splitix, function(x, xx, yy) {
            res <- vapply(x, function(x, xx, yy) {
                xx <- sample(xx)
                yy <- sample(yy)
                return(drop(lsa::cosine(x = xx, y = yy)))
            }, xx = xx, yy = yy, FUN.VALUE = numeric(1))
            return(res)
        }, xx = x, yy = y)
        mcres <- unlist(mcres)
        switch(alternative, two.sided = {
            res["p.value"] <- 2 * (min(sum(mcres < res["estimate"]), sum(mcres > res["estimate"]))/sum(!is.na(mcres)))
        }, less = {
            res["p.value"] <- sum(mcres < res["estimate"])/sum(!is.na(mcres))
        }, greater = {
            res["p.value"] <- sum(mcres > res["estimate"])/sum(!is.na(mcres))
        })
        if (res["p.value"] == 0) {
            res["p.value"] <- 1/(nperm + 1)
        }
    }
    res <- as.list(res)
    if (include.perm) {
        res <- c(res, list(estimate.random = mcres))
    }
    return(res)
}

#' Curve fitting via `stats::optim` L-BFGS-B with fall-back grid/pattern search
#'   if convergence is not achieved.
#'
#' @description Function to fit curve via stats::optim
#'
#' @param par `numeric` Vector of intial guesses for the parameters. For each
#'    index `i` of `par`, par\[i\] must be within the range (`lower\[i\]`,
#'   `upper\[i\]`). If only a single `upper` or `lower` value is present,
#'   that range is used for all parameters in `par`.
#' @param x `numeric` Values to evaluate `fn` for.
#' @param y `numeric` Target output values to optimze `fn` against.
#' @param fn `function` A function to optimize. Any `fn` arguments passed via
#'   `...` will be treated as constant and removed from the optimization. It
#'   is assumed that the first argument is the x value to optimize over and
#'   any subsequent arguments are free parameters to be optimized. Transformed
#'   to be optim compatible via `make_optim_function` is the first arguement
#'   isn't already `par`.
#' @param loss `character(1)` or `function` Either the name of one of the bundled
#'   loss functions (see details) or a custom loss function to compute for
#'   the output of `fn` over `x`.
#' @param lower `numeric(1)` Lower bound for parameters. Parallel to `par`.
#' @param upper `numeric(1)` Upper bound for paramteres. Parallel to `par`.
#' @param density `numeric` how many points in the dimension of each parameter
#'   should be evaluated (density of the grid)
#' @param step initial step size for pattern search.
#' @param precision `numeric` smallest step size used in pattern search, once
#'   step size drops below this value, the search terminates.
#' @param span `numeric` Can be safely kept at 1, multiplicative ratio for
#'   initial step size in pattern search. Must be larger than precision.
#' @param ... `pairlist` Fall through arguments to `fn`.
#' @param loss_args `list` Additional argument to the `loss` function.
#'   These get passed to losss via `do.call` analagously to using `...`.
#' @param optim_only `logical(1)` Should the fall back methods when optim fails
#'   be skipped? Default is `FALSE`.
#' @param control `list` List of control parameters to pass to `optim`. See
#'   `?optim` for details.
#'
#' @examples
#' \dontrun{
#'   # Four parameter hill curve equation
#'   hillEqn <- function(x, Emin, Emax, EC50, lambda) {
#'       (Emin + Emax * (x / EC50)^lambda) / (1 + (x / EC50)^lambda)
#'   }
#'   # Make some dummy data
#'   doses <- rev(1000 / (2^(1:20)))
#'   lambda <- 1
#'   Emin <- 1
#'   Emax <- 0.1
#'   EC50 <- median(doses)
#'   response <- hillEqn(doses, Emin=Emin, lambda=lambda, Emax=Emax, EC50=EC50)
#'   nresponse <- response + rnorm(length(response), sd=sd(response)*0.1) # add noise
#'   # 3-parameter optimization
#'   3par <- .fitCurve2(
#'       par=c(Emax, EC50, lambda),
#'       x=doses,
#'       y=nresponse,
#'       fn=hillEqn,
#'       Emin=Emin, # set this as constant in the function being optimized (via ...)
#'       loss=.normal_loss,
#'       loss_args=list(trunc=FALSE, n=1, scale=0.07),
#'       upper=c(1, max(doses), 6),
#'       lower=c(0, min(doses), 0)
#'   )
#'   # 2-parameter optimization
#'   2par <- .fitCurve2(
#'       par=c(Emax, EC50),
#'       x=doses,
#'       y=nresponse,
#'       fn=hillEqn,
#'       Emin=Emin, # set this as constant in the function being optimized (via ...)
#'       lambda=1,
#'       loss=.normal_loss,
#'       loss_args=list(trunc=FALSE, n=1, scale=0.07),
#'       upper=c(1, max(doses)),
#'       lower=c(0, min(doses))
#'   )
#' }
#'
#' @details
#' TODO
#'
#' @return `numeric` Vector of optimal parameters for `fn` fit against `y`
#'   on the values of `x`.
#'
#' @importFrom stats optim
#' @export
.fitCurve2 <- function(par, x, y, fn, loss, lower=-Inf, upper=Inf,
        precision=1e-4, density=c(2, 10, 5), step= 0.5 / density, ...,
        loss_args=list(), span=1, optim_only=FALSE,
        control=list(factr=1e-08, ndeps=rep(1e-4, times=length(par)), trace = 0)
        ) {
    stopifnot(is.function(fn))
    stopifnot(is.function(loss) || is.character(loss))
    stopifnot(c("par", "x", "y", "fn") %in% formalArgs(loss))
    stopifnot(
        is.null(names(loss_args)) || all(names(loss_args) %in% formalArgs(loss))
    )
    stopifnot(
        (length(par) == length(upper) && length(par) == length(lower)) ||
        (length(upper) == 1 && length(lower) == 1)
    )
    stopifnot(span > precision)
    optim_fn <- if (is_optim_compatible(fn)) {
        fn
    } else {
        make_optim_function(fn, ...)
    }
    guess <- optim(
        par=par,
        fn=function(p)
            do.call(loss,
                args=c(
                    list(par=p, x=x, y=y, fn=optim_fn), # mandatory loss args
                    loss_args # additional args to loss
                )
            ),
        upper=upper,
        lower=lower,
        control=control,
        method="L-BFGS-B"
    )

    failed <- guess[["convergence"]] != 0
    if (failed) guess <- list(par=par, convergence=(-1))
    guess <- guess[["par"]]

    guess_residual <- do.call(loss,
        args=c(list(par=guess, x=x, y=y, fn=optim_fn), loss_args))
    gritty_guess_residual <- do.call(loss,
        args=c(list(par=par, x=x, y=y, fn=optim_fn), loss_args))

    if (
        (failed || any(is.na(guess)) || guess_residual >= gritty_guess_residual)
        && !optim_only
    ) {
        guess <- .meshEval2(density=density, par=guess, x=x, y=y, fn=optim_fn,
            lower=lower, upper=upper, ..., loss=loss,loss_args=loss_args)
        guess_residual <- do.call(loss,
            args=c(list(par=guess, x=x, y=y, fn=optim_fn), loss_args))

        guess <- .patternSearch2(span=span, precision=precision, step=step,
            par=guess, par_residual=guess_residual, x=x, y=y, fn=optim_fn,
            loss=loss, lower=lower, upper=upper, ..., loss_args=loss_args)
    }

    y_hat <- optim_fn(par=guess, x=x)

    Rsqr <- 1 - (var(y - y_hat) / var(y))
    attr(guess, "Rsquare") <- Rsqr

    return(guess)
}


#' Compute the loss using the expectation of the likelihood of the median
#'   for N samples from a probability density function.
#'
#' @param .pdf `function` Probability density function to use for computing loss.
#' @param .edf `function` Expected likelihood function for the median of `n`
#'   random samples from `.pdf`.
#' @inheritParams .fitCurve2
#' @param n `numeric(1)`
#' @param scale `numeric(1)`
#' @param trunc `logical(1)`
#'
#' @return `numeric(1)` Loss of `fn` on `x` relative to `y`.
#'
#' @keywords interal
#' @noRd
.sampling_loss <- function(.pdf, .edf, par, x, y, fn, ..., n=1, scale=0.07,
        trunc=FALSE) {
    diffs <- fn(par=par, x=x, ...) - y
    if (trunc == FALSE) {
        if (n == 1 && grepl("normals", deparse(substitute(.pdf))))
            return(sum(diffs^2))
        return(sum(-log(.pdf(diffs, n, scale))))
    } else {
        down_truncated <- abs(y) >= 1
        up_truncated <- abs(y) <= 0
        return(
            sum(-log(.pdf(diffs[!(down_truncated | up_truncated)], n, scale))) +
            sum(-log(.edf(-diffs[up_truncated | down_truncated], n, scale)))
        )
    }
}


#' See docs for `.sampling_loss`
#' @keywords interal
#' @noRd
.normal_loss <- function(par, x, y, fn, ..., n=1, scale=0.07,
        trunc=FALSE) {
    .sampling_loss(.pdf=.dmednnormals, .edf=.edmednnormals, par=par, x=x, y=y,
        fn=fn, ..., n=n, scale=scale, trunc=trunc)
}



#' See docs for `.sampling_loss`
#' @keywords internal
#' @noRd
.cauchy_loss <- function(par, x, y, fn, ..., n=1, scale=0.07,
        trunc=FALSE) {
    .sampling_loss(.pdf=.dmedncauchys, .edf=.edmedncauchys, par=par, x=x, y=y,
        fn=fn, ..., n=n, scale=scale, trunc=trunc)
}

#' @export
#' @keywords internal
#' @noRd
.meshEval2 <- function(density, par, x, y, fn,
        loss=.normal_loss, lower, upper, ..., loss_args=list()) {
    # input validation
    stopifnot(is.function(fn))
    stopifnot(is.function(loss) || is.character(loss))
    stopifnot(c("par", "x", "y", "fn") %in% formalArgs(loss))
    stopifnot(
        is.null(names(loss_args)) || all(names(loss_args) %in% formalArgs(loss))
    )
    # make function amenable to use via optim
    optim_fn <- if (is_optim_compatible(fn)) {
        fn
    } else {
        make_optim_function(fn, ...)
    }
    par_loss <- do.call(loss,
        args=c(list(par=par, x=x, y=y, fn=optim_fn), loss_args))

    periods <- matrix(NA, nrow = length(par), ncol = 1)
    names(periods) <- names(par)
    periods[1] <- 1

    if (length(par) > 1) {
        for (p in seq(2, length(par))) {
            ## the par-1 is because we want 1 increment of par variable once
            ##   all previous variables have their values tested once.
            periods[p] <- periods[p - 1] * (density[p - 1] *
                (upper[p - 1] - lower[p - 1]) + 1)
        }
    }

    currentPars <- lower

    ## The plus one is because we include endpoints.
    for (point in seq_len(prod((upper - lower) * density + 1))) {

        test_par_loss <- do.call(loss,
            c(list(par=currentPars, x=x, y=y, fn=optim_fn), loss_args))

        ## Check for something catastrophic going wrong
        if (
            !length(test_par_loss) ||
            (!is.finite(test_par_loss) && test_par_loss != Inf)
        ) {
            stop(paste0(" Test Guess Loss is: ", test_par_loss, "\n",
                "Other Pars:\n", "x: ", paste(x, collapse = ", "), "\n",
                "y: ", paste(y, collapse = ", "), "\n", "n: ", n, "\n",
                "pars: ", currentPars, "\n", "scale: ", scale, "\n",
                "family : ", family, "\n", "Trunc ", trunc))
        }
        ## save the guess if its an improvement
        if (test_par_loss < par_loss) {
            par <- currentPars
            par_loss <- test_par_loss
        }
        ## increment the variable(s) that should be incremented this loop
        for (p in seq_along(par)) {
            if (point %% periods[p] == 0) {
                currentPars[p] <- currentPars[p] + 1 / density[p]
                if (currentPars[p] > upper[p]) {
                    currentPars[p] <- lower[p]
                }
            }
        }
    }

    return(par)
}


#' @export
#' @keywords internal
#' @noRd
.patternSearch2 <- function(span, precision, step, par, par_residual, x, y,
        lower, upper, fn, loss, ..., loss_args=list()) {
    # input validation
    stopifnot(is.function(fn))
    stopifnot(is.function(loss) || is.character(loss))
    stopifnot(c("par", "x", "y", "fn") %in% formalArgs(loss))
    stopifnot(
        is.null(names(loss_args)) || all(names(loss_args) %in% formalArgs(loss))
    )
    # make function amenable to use via optim
    optim_fn <- if (is_optim_compatible(fn)) {
        fn
    } else {
        make_optim_function(fn, ...)
    }
    # setup matrix for searching the parameter space
    neighbours <- matrix(nrow = 2 * length(par), ncol = length(par))
    neighbour_loss <- matrix(NA, nrow = 1, ncol = nrow(neighbours))

    while (span > precision) {
        for (neighbour in seq_len(nrow(neighbours))) {
            neighbours[neighbour, ] <- par
            dimension <- ceiling(neighbour / 2)
            if (neighbour %% 2 == 1) {
                neighbours[neighbour, dimension] <- pmin(
                    par[dimension] + span * step[dimension],
                    upper[dimension]
                )
            } else {
                neighbours[neighbour, dimension] <- pmax(
                    par[dimension] - span * step[dimension],
                    lower[dimension]
                )
            }

            neighbour_loss[neighbour] <- do.call(loss, args=c(
                list(par=neighbours[neighbour, ], x=x, y=y, fn=optim_fn),
                    loss_args))
        }

        if (min(neighbour_loss) < par_residual) {
            par <- neighbours[which.min(neighbour_loss)[1], ]
            par_residual <- min(neighbour_loss)
        } else {
            span <- span / 2
        }
    }
    return(par)
}



#' Drop parameters from a function and replace them with constants
#'   inside the function body.
#'
#' @param fn `function` A non-primitive function to remove parameters from
#'   (via `base::formals(fn)`).
#' @param args `list` A list where names are the function arguments (parameters)
#'   to remove and the values are the appopriate value to replace the parameter
#'   with in the function body.
#'
#' @return `function` A new non-primitize function with the parameters named in
#'   `args` deleted and their values fixed with the values from `args` in the
#'   function body.
#'
#' @importFrom compiler cmpfun
#' @export
drop_fn_params <- function(fn, args) {
    stopifnot(is.function(fn) && !is.primitive(fn))
    if (length(args) == 0) return(fn)
    stopifnot(all(names(args) %in% formalArgs(fn)))
    stopifnot(is.list(args))
    # Delete the arguments we are deparamterizing
    formals(fn)[formalArgs(fn) %in% names(args)] <- NULL
    # Replace the symbols with the new fixed value inside the fuction body
    deparse_body <- deparse(body(fn))
    for (i in seq_along(args)) {
        deparse_body <- gsub(names(args)[i], args[[i]], deparse_body)
    }
    # Parse the new function body back to a call
    body(fn, envir=parent.frame()) <- str2lang(paste0(deparse_body, collapse="\n"))
    fn <- compiler::cmpfun(fn)
    return(fn)
}

#' Collects all function arguments other than the first into a single list
#'   parameter.
#'
#' @description
#' Useful for converting a regular function into a function amenable to
#' optimization via `stats::optim`, which requires all free parameters be
#' passed as a single vector `par`.
#'
#' @details
#' Takes a function of the form f(x, ...), where ... is any number of additional
#'   function parameters (bot not literal `...`!) and parses it to a function of
#'   the form f(par, x) where `par` is a vector of values for ... in
#'   the same order as the arguments appear in `fn`.
#'
#' @param fn `function` A non-primitive function to refactor such that the first
#'   argument becomes the second argument and all other parameters must be
#'   passed as a vector to the first argument of the new function via the `par`
#'   parameter.
#'
#' @return `function` A new non-primitive function where the first argument is
#'   `par`, which takes a vector of parameters being optimized, and the
#'   second argument is the old first argument to `fn` (usually `x` since this
#'   is the independent variable to optimize the function over).
#'
#' @export
collect_fn_params <- function(fn) {
    stopifnot(is.function(fn) && !is.primitive(fn))
    # Capture the current formal args
    formal_args <- formalArgs(fn)
    if ("..." %in% formalArgs(fn))
        stop("No support for fn with ... in signature!")
    # Replace args other than the first with a list
    args <- paste0("par[[", seq_along(formal_args[-1]), "]]") |>
        as.list() |>
        setNames(formal_args[-1])
    fn <- drop_fn_params(fn, args=args)
    # Add the `par` list to the formal args of the function
    formals(fn) <- c(alist(par=), formals(fn))
    return(fn)
}


#' Takes a non-primitive R function and refactors it to be compatible with
#'   optimization via `stats::optim`.
#'
#' @description Takes a non-primitive R function and refactors it to be compatible with optimization via `stats::optim`.
#'
#'
#' @param fn `function` A non-primitive function
#' @param ... Arguments to `fn` to fix for before building the
#'   function to be optimized. Useful for reducing the number of free parameters
#'   in an optimization if there are insufficient degrees of freedom.
#'
#' @seealso [`drop_fn_params`], [`collect_fn_params`]
#'
#' @export
make_optim_function <- function(fn, ...) {
    # NOTE: error handling done inside helper methods!
    fn1 <- drop_fn_params(fn, args=list(...))
    fn2 <- collect_fn_params(fn1)
    return(fn2)
}


#' Check whether a function signature is amenable to optimization via `stats::optim`.
#'
#' @description
#' Functions compatible with `optim` have the parameter named `par` as their
#' first formal argument where each value is a respective free parameter to
#' be optimized.
#'
#' @param fn `function` A non-primitive function.
#'
#' @return `logical(1)` `TRUE` if the first value of `formalArg(fn)` is "par",
#'   otherwise `FALSE`.
#'
#' @export
is_optim_compatible <- function(fn) formalArgs(fn)[1] == "par"






# ======================
# ==== Deprecating =====

#' .fitCurve
#'
#' Curve optimization from 1 variable to 1 variable, using L-BFSG-B from optim,
#' with fallback to pattern search if optimization fails to converge.
#'
#' @param x `numeric` input/x values for function
#' @param y `numeric` output/y values for function
#' @param f `function` function f, parameterized by parameters to optimize
#' @param density `numeric` how many points in the dimension of each parameter
#'   should be evaluated (density of the grid)
#' @param step initial step size for pattern search.
#' @param precision `numeric` smallest step size used in pattern search, once
#'   step size drops below this value, the search terminates.
#' @param lower_bounds `numeric` lower bounds for the paramater search space
#' @param upper_bounds `numeric` upper bounds for the parameter search space
#' @param median_n `integer` number of technical replicates per measured point
#'   in x. Used to evaluate the proper median distribution for the normal and
#'   cauchy error models
#' @param scale `numeric` scale on which to measure probability for the error
#'   model (roughly SD of error)
#' @param family `character` which error family to use. Currently, "normal"
#'   and "Cauchy" are implemented
#' @param trunc `logical` Whether or not to truncate the values at 100% (1.0)
#' @param verbose `logical` should diagnostic messages be printed?
#' @param gritty_guess `numeric` intitial, uninformed guess on parameter
#'   values (usually heuristic)
#' @param span ['numeric'] can be safely kept at 1, multiplicative ratio for
#'   initial step size in pattern search. Must be larger than precision.
#'
#' @keywords internal
#' @noRd
#'
#' @importFrom stats optim var
#' @export
.fitCurve <- function(x, y, f, density, step, precision, lower_bounds,
    upper_bounds, scale, family, median_n, trunc, verbose, gritty_guess,
    span = 1) {

    guess <- tryCatch(optim(par = gritty_guess, fn = function(t) {
        .residual(
            x = x, y = y, n = median_n, pars = t, f = f,
            scale = scale, family = family, trunc = trunc
        )
    }, lower = lower_bounds, upper = upper_bounds, control = list(
        factr = 1e-08,
        ndeps = rep(1e-4, times = length(gritty_guess)),
        trace = 0
    ), method = "L-BFGS-B"), error = function(e) {
        list(par = gritty_guess, convergence = -1)
    })

    failed <- guess[["convergence"]] != 0
    guess <- guess[["par"]]

    guess_residual <- .residual(x = x, y = y, n = median_n, pars = guess, f = f, scale = scale, family = family, trunc = trunc)
    gritty_guess_residual <- .residual(x = x, y = y, n = median_n, pars = gritty_guess, f = f, scale = scale, family = family, trunc = trunc)

    if (failed || any(is.na(guess)) || guess_residual >= gritty_guess_residual) {
        guess <- .meshEval(x = x, y = y, f = f, guess = gritty_guess, lower_bounds = lower_bounds, upper_bounds = upper_bounds, density = density,
            n = median_n, scale = scale, family = family, trunc = trunc)
        guess_residual <- .residual(x = x, y = y, n = median_n, pars = guess, f = f, scale = scale, family = family, trunc = trunc)

        guess <- .patternSearch(x = x, y = y, f = f, guess = guess, n = median_n, guess_residual = guess_residual, lower_bounds = lower_bounds,
            upper_bounds = upper_bounds, span = span, precision = precision, step = step, scale = scale, family = family, trunc = trunc)
    }

    y_hat <- do.call(f, list(x, par=guess))

    Rsqr <- 1 - (var(y - y_hat) / var(y))
    attr(guess, "Rsquare") <- Rsqr

    return(guess)
}



# meshEval ----------------------------------------------------------------
#' meshEval
#'
#' Generate an initial guess for dose-response curve parameters by evaluating
#' the residuals at different lattice points of the search space
#'
#' @export
#' @keywords internal
#' @noRd
# ##FIXME:: Why is this different in PharmacoGx?
.meshEval <- function(x, y, f, guess, lower_bounds, upper_bounds, density, n,
        scale, family, trunc) {
    pars <- NULL
    guess_residual <- .residual(x = x, y = y, n = n, pars = guess, f = f,
        scale = scale, family = family, trunc = trunc)

    periods <- matrix(NA, nrow = length(guess), ncol = 1)
    names(periods) <- names(guess)
    periods[1] <- 1

    if (length(guess) > 1) {
        for (par in 2:length(guess)) {
            ## the par-1 is because we want 1 increment of par variable once all previous variables have their values tested once.
            periods[par] <- periods[par - 1] * (density[par - 1] * (upper_bounds[par - 1] - lower_bounds[par - 1]) + 1)
        }
    }

    currentPars <- lower_bounds

    ## The plus one is because we include endpoints.
    for (point in seq_len(prod((upper_bounds - lower_bounds) * density + 1))) {

        test_guess_residual <- .residual(x = x, y = y, n = n, pars = currentPars, f = f, scale = scale, family = family, trunc = trunc)

        ## Check for something catastrophic going wrong
        if (!length(test_guess_residual) || (!is.finite(test_guess_residual) && test_guess_residual != Inf)) {
            stop(paste0(" Test Guess Residual is: ", test_guess_residual, "\n", "Other Pars:\n", "x: ", paste(x, collapse = ", "), "\n",
                "y: ", paste(y, collapse = ", "), "\n", "n: ", n, "\n", "pars: ", pars, "\n", "scale: ", scale, "\n", "family : ", family,
                "\n", "Trunc ", trunc))
        }
        ## save the guess if its an improvement
        if (test_guess_residual < guess_residual) {
            guess <- currentPars
            guess_residual <- test_guess_residual
        }
        ## increment the variable(s) that should be incremented this loop
        for (par in seq_along(guess)) {
            if (point%%periods[par] == 0) {
                currentPars[par] <- currentPars[par] + 1/density[par]

                if (currentPars[par] > upper_bounds[par]) {
                    currentPars[par] <- lower_bounds[par]
                }
            }
        }
    }

    return(guess)
}

#' @export
#' @keywords internal
#' @noRd
.patternSearch <- function(x, y, f, guess, n, guess_residual, lower_bounds,
        upper_bounds, span, precision, step, scale, family, trunc) {
    neighbours <- matrix(nrow = 2 * length(guess), ncol = length(guess))
    neighbour_residuals <- matrix(NA, nrow = 1, ncol = nrow(neighbours))

    while (span > precision) {
        for (neighbour in seq_len(nrow(neighbours))) {
            neighbours[neighbour, ] <- guess
            dimension <- ceiling(neighbour / 2)
            if (neighbour %% 2 == 1) {
                neighbours[neighbour, dimension] <- pmin(
                    guess[dimension] + span * step[dimension],
                    upper_bounds[dimension]
                )
            } else {
                neighbours[neighbour, dimension] <- pmax(
                    guess[dimension] - span * step[dimension],
                    lower_bounds[dimension]
                )
            }

            neighbour_residuals[neighbour] <- .residual(x = x, y = y,
                f = f, pars = neighbours[neighbour, ], n = n, scale = scale,
                family = family,
                trunc = trunc)
        }

        if (min(neighbour_residuals) < guess_residual) {
            guess <- neighbours[which.min(neighbour_residuals)[1], ]
            guess_residual <- min(neighbour_residuals)
        } else {
            span <- span / 2
        }
    }
    return(guess)
}

## TODO:: Write documentation
## FIXME:: Why is this different from PharmacoGx?
#' @title Residual calculation
#'
#' @return A \code{numeric} containing the estimated residuals for the model
#'   fit
#'
#' @export
#' @keywords internal
#' @noRd
.residual <- function(x, y, n, pars, f, scale = 0.07, family = c("normal", "Cauchy"), trunc = FALSE) {
    family <- match.arg(family)
    diffs <- f(x, par=pars) - y

    if (family != "Cauchy") {
        if (trunc == FALSE) {
            if (n == 1) {
                return(sum(diffs^2))
            }
            return(sum(-log(.dmednnormals(diffs, n, scale))))
        } else {
            down_truncated <- abs(y) >= 1
            up_truncated <- abs(y) <= 0
            return(sum(-log(.dmednnormals(diffs[!(down_truncated | up_truncated)], n, scale))) + sum(-log(.edmednnormals(-diffs[up_truncated |
                down_truncated], n, scale))))
        }
    } else {
        if (trunc == FALSE) {
            return(sum(-log(.dmedncauchys(diffs, n, scale))))
        } else {
            down_truncated <- abs(y) >= 1
            up_truncated <- abs(y) <= 0
            return(sum(-log(.dmedncauchys(diffs[!(down_truncated | up_truncated)], n, scale))) + sum(-log(.edmedncauchys(-diffs[up_truncated |
                down_truncated], n, scale))))
        }
    }
}

library(testthat)
testthat::local_edition(3)

# Create some dummy data
hillEqn <- function(x, Emin, Emax, EC50, lambda) {
    (Emin + Emax * (x / EC50)^lambda) / (1 + (x / EC50)^lambda)
}
# Set parameters for function testing
doses <- rev(1000 / (5^(1:10)))
lambda <- 0.6
Emin <- 1
Emax <- 0.5
EC50 <- median(doses)
# Helper to combine
fx <- if (is_optim_compatible(hillEqn)) hillEqn else
    make_optim_function(hillEqn, lambda=lambda, Emin=Emin)
response <- hillEqn(doses, Emin=Emin, lambda=lambda, Emax=Emax, EC50=EC50)
nresponse <- response + rnorm(length(response), sd=sd(response)*0.1)

# -- Loss function tests
lmsg <- c("LOSS FUNCTION: ")
testthat::test_that(paste0(lmsg, ".residual and .normal_loss produce equal results."), {
    trunc_vals <- c(FALSE, TRUE, FALSE, TRUE)
    nvals <- c(1, 1, 3, 3)
    for (i in seq_along(trunc_vals)) {
        n1 <- .residual(par=c(Emax=0.2, EC50=10), x=doses, y=nresponse, f=fx,
            family="normal", n=nvals[i], trunc=trunc_vals[i], scale=0.07)
        n2 <- CoreGx:::.normal_loss(par=c(Emax=0.2, EC50=10), x=doses, y=nresponse, fn=fx,
            n=nvals[i], trunc=trunc_vals[i], scale=0.07)
        testthat::expect_equal(n1, n2,
            info=paste0("trunc: ", trunc_vals[i], ", n: ", nvals[i])
        )
    }
})

testthat::test_that(paste0(lmsg, ".residual and .cauchy_loss produce equal results."), {
    trunc_vals <- c(FALSE, TRUE, FALSE, TRUE)
    nvals <- c(1, 1, 3, 3)
    for (i in seq_along(trunc_vals)) {
        c1 <- .residual(par=c(Emax=0.2, EC50=10), x=doses, y=nresponse, f=fx,
            family="Cauchy", n=nvals[i], trunc=trunc_vals[i], scale=0.07)
        c2 <- CoreGx:::.cauchy_loss(par=c(Emax=0.2, EC50=10), x=doses, y=nresponse, fn=fx,
            n=nvals[i], trunc=trunc_vals[i], scale=0.07)
        testthat::expect_equal(c1, c2,
            info=paste0("trunc: ", trunc_vals[i], ", n: ", nvals[i])
        )
    }
})

# -- Curve fitting
cmsg <- "CURVE FITTING: "
testthat::test_that(paste0(cmsg, ".fitCurve and .fitCurve2 produce equal results for 2-parameter Hill curve."), {
    par_list <- list(c(Emax=0.1, EC50=0.1), c(Emax=0.5, EC5O=500), c(Emax=0.9, EC50=100))
    for (i in seq_along(par_list)) {
        pars <- par_list[[i]]
        normal_par1 <- .fitCurve(
            gritty_guess=pars,
            x=doses,
            y=nresponse,
            f=fx,
            family="normal",
            trunc=FALSE,
            median_n=1,
            scale=0.07,
            upper_bound=c(2, max(doses)),
            lower_bound=c(0, min(doses)),
            density=c(2, 10),
            precision=1e-4,
            step=0.5 / c(2, 10)
        )
        normal_par2 <-.fitCurve2(
            par=pars,
            x=doses,
            y=nresponse,
            fn=hillEqn,
            loss=CoreGx:::.normal_loss,
            loss_args=list(trunc=FALSE, n=1, scale=0.07),
            Emin=Emin,
            lambda=lambda,
            upper=c(2, max(doses)),
            lower=c(0, min(doses)),
            density=c(2, 10),
            precision=1e-4,
            step=0.5 / c(2, 10)
        )
        testthat::expect_equal(normal_par1, normal_par2,
            info=paste0("Emax: ", pars[1], ", EC50: ", pars[2]))
        cauchy_par1 <- .fitCurve(
            gritty_guess=pars,
            x=doses,
            y=nresponse,
            f=fx,
            family="Cauchy",
            trunc=FALSE,
            median_n=1,
            scale=0.07,
            upper_bound=c(2, max(doses)),
            lower_bound=c(0, min(doses)),
            density=c(2, 10),
            precision=1e-4,
            step=0.5 / c(2, 10)
        )
        cauchy_par2 <-.fitCurve2(
            par=pars,
            x=doses,
            y=nresponse,
            fn=hillEqn,
            loss=CoreGx:::.cauchy_loss,
            loss_args=list(trunc=FALSE, n=1, scale=0.07),
            Emin=Emin,
            lambda=lambda,
            upper=c(2, max(doses)),
            lower=c(0, min(doses)),
            density=c(2, 10),
            precision=1e-4,
            step=0.5 / c(2, 10)
        )
        testthat::expect_equal(cauchy_par1, cauchy_par2,
            info=paste0("Emax: ", pars[1], ", EC50: ", pars[2]))
    }
})


testthat::test_that(
    paste0(cmsg, ".fitCurve and .fitCurve2 produce equal results for 3-parameter Hill curve."), {
    par_list <- list(
        c(Emax=0.1, EC50=0.1, lambda=1),
        c(Emax=0.5, EC5O=500, lambda=0.75),
        c(Emax=0.9, EC50=100, lambda=2)
    )
    fx <- make_optim_function(hillEqn, Emin=Emin)
    for (i in seq_along(par_list)) {
        pars <- par_list[[i]]
        normal_par1 <- .fitCurve(
            gritty_guess=pars,
            x=doses,
            y=nresponse,
            f=fx,
            family="normal",
            trunc=FALSE,
            median_n=1,
            scale=0.07,
            upper_bound=c(2, max(doses), 6),
            lower_bound=c(0, min(doses), 0),
            density=c(2, 10, 5),
            precision=1e-4,
            step=0.5 / c(2, 10, 5)
        )
        normal_par2 <-.fitCurve2(
            par=pars,
            x=doses,
            y=nresponse,
            fn=hillEqn,
            loss=CoreGx:::.normal_loss,
            loss_args=list(trunc=FALSE, n=1, scale=0.07),
            Emin=Emin,
            upper=c(2, max(doses), 6),
            lower=c(0, min(doses), 0),
            density=c(2, 10, 5),
            precision=1e-4,
            step=0.5 / c(2, 10, 5)
        )
        testthat::expect_equal(normal_par1, normal_par2,
            info=paste0("Emax: ", pars[1], ", EC50: ", pars[2]))
        cauchy_par1 <- .fitCurve(
            gritty_guess=pars,
            x=doses,
            y=nresponse,
            f=fx,
            family="Cauchy",
            trunc=FALSE,
            median_n=1,
            scale=0.07,
            upper_bound=c(2, max(doses), 6),
            lower_bound=c(0, min(doses), 0),
            density=c(2, 10, 5),
            precision=1e-4,
            step=0.5 / c(2, 10, 5)
        )
        cauchy_par2 <-.fitCurve2(
            par=pars,
            x=doses,
            y=nresponse,
            fn=hillEqn,
            loss=CoreGx:::.cauchy_loss,
            loss_args=list(trunc=FALSE, n=1, scale=0.07),
            Emin=Emin,
            upper=c(2, max(doses), 6),
            lower=c(0, min(doses), 0),
            density=c(2, 10, 5),
            precision=1e-4,
            step=0.5 / c(2, 10, 5)
        )
        testthat::expect_equal(cauchy_par1, cauchy_par2,
            info=paste0("Emax: ", pars[1], ", EC50: ", pars[2]))
    }
})

#' Subset a CoreSet object based on various parameters, such as cell lines, molecular features
#'
#' @param object An object inheriting from the `CoreGx::CoreSet` class
#' @param ... Allow definition of new arguments to this generic
#'
#' @return A subsetted version of the original `object`
#'
#' @examples
#' "Generics shouldn't need examples!"
#'
#' @export
#' @keywords internal
setGeneric("subsetTo", function(object, ...) standardGeneric("subsetTo"))

#' @include immutable-class.R
#' @include allGenerics.R
NULL

#' @title LongTable class definition
#'
#' @description Define a private constructor method to be used to build a
#'   `LongTable` object.
#'
#' @slot rowData See Slots section.
#' @slot colData See Slots section.
#' @slot assays See Slots section.
#' @slot metadata See Slots section.
#' @slot .intern See Slots section.
#'
#' @section Slots:
#' - *rowData*: A `data.table` containing the metadata associated with the
#'   row dimension of a `LongTable`.
#' - *colData*: A `data.table` containing the metadata associated with the
#'   column dimension of a `LongTable`.
#' - *assays*: A `list` of `data.table`s, one for each assay in a
#'   `LongTable`.
#' - *metadata*: An optional `list` of additional metadata for a `LongTable`
#'   which doesn't map to one of the dimensions.
#' - *.intern*: An `immutable` `list` that holds internal structural metadata
#'   about a LongTable object, such as which columns are required to key
#'   the object.
#'
#' @return `LongTable` object containing the assay data from a treatment
#'   response experiment
#'
#' @md
#' @import data.table
#' @keywords internal
#' @rdname LongTable-class
#' @aliases .LongTable
#' @exportClass LongTable
.LongTable <- setClass("LongTable",
    slots=list(
        rowData='data.table',
        colData='data.table',
        assays='list',
        metadata='list',
        .intern='immutable_list')
)


#' @title LongTable constructor method
#'
#' @rdname LongTable
#'
#' @param rowData `data.frame` A rectangular object coercible to a `data.table`.
#' @param rowIDs `character` A vector of `rowData` column names needed to
#'   uniquely identify each row in a `LongTable`.
#' @param colData `data.frame` A rectangular object coercible to a `data.table.`
#' @param colIDs `chacter` A vector of `colData` column names needed to uniquely
#'   identify each column in a `LongTable`.
#' @param assays `list` A list of rectangular objects, each coercible to
#'   a `data.table`. Must be named and item names must match the `assayIDs`
#'   list.
#' @param assayIDs `list` A list of `character` vectors specifying the columns
#'   needed to uniquely identify each row in an `assay`. Names must match the
#'   `assays` list.
#' @param metadata `list` A list of one or more metadata items associated with
#'   a LongTable experiment.
#' @param keep.rownames `logical(1)` or `character(1)` Should rownames be
#'   retained when coercing to `data.table` inside the constructor. Default
#'   is FALSE. If TRUE, adds a 'rn' column to each rectangular object that
#'   gets coerced from `data.frame` to `data.table`. If a string, that becomes
#'   the name of the rownames column.
#'
#' @return A `LongTable` object containing the data for a treatment response
#'   experiment and configured according to the rowIDs and colIDs arguments.
#'
#' @examples
#' "See vignette('The LongTable Class', package='CoreGx')"
#'
#' @importFrom data.table key setkeyv
#' @export
LongTable <- function(rowData, rowIDs, colData, colIDs, assays, assayIDs,
        metadata=list(), keep.rownames=FALSE) {

    # handle missing parameters
    isMissing <- c(rowData=missing(rowData), rowIDs=missing(rowIDs),
        colIDs=missing(colIDs), colData=missing(colData), assays=missing(assays),
        assayIDs=missing(assayIDs))

    if (any(isMissing)) stop(.errorMsg('\nRequired parameter(s) missing: ',
        names(isMissing)[isMissing], collapse='\n\t'))

    # check parameter types and coerce or error
    if (!is(colData, "data.table")) {
        tryCatch({
            colData <- data.table(colData, keep.rownames=keep.rownames)
        }, error=function(e)
            stop(.errorMsg("colData must be coercible to a data.frame!"))
        )
    } else {
        colData <- copy(colData)
    }

    if (!is(rowData, "data.table")) {
        tryCatch({
            rowData <- data.table(rowData, keep.rownames=keep.rownames) },
        error=function(e)
            stop(.errorMsg("rowData must be coerceible to a data.frame!"))
        )
    } else {
        rowData <- copy(rowData)
    }

    isDT <- is.items(assays, FUN=is.data.table)
    isDF <- is.items(assays, FUN=is.data.frame) & !isDT
    if (!all(isDT)) {
        tryCatch({
            for (i in which(isDF))
                assays[[i]] <- data.table(assays[[i]], keep.rownames)
        }, error = function(e, assays) {
            message(e)
            types <- lapply(assays, typeof)
            stop(.errorMsg(
                '\nList items are types: ',
                types, '\nPlease ensure all items in the assays list are ',
                'coerceable to a data.frame!'), collapse=', ')
        })
    }
    assays <- copy(assays)

    ## FIXME:: Move all validity checks to top of the function to prevent wasted
    ## computation or into class validity method

    # capture row internal metadata
    if (is.numeric(rowIDs) || is.logical(rowIDs))
        rowIDs <- colnames(rowData)[rowIDs]
    if (!all(rowIDs %in% colnames(rowData)))
        stop(.errorMsg('\nRow IDs not in rowData: ',
            setdiff(rowIDs, colnames(rowData)), collapse=', '))

    # Create the row and column keys for LongTable internal mappings
    if (!('rowKey' %in% colnames(rowData)))
        rowData[, c('rowKey') := .GRP, keyby=c(rowIDs)]
    if (!('colKey' %in% colnames(colData)))
        colData[, c('colKey') := .GRP, keyby=c(colIDs)]

    # initialize the internals object to store private metadata for a LongTable
    internals <- setNames(vector("list", length=6),
        c("rowIDs", "rowMeta", "colIDs", "colMeta", "assayKeys", "assayIndex"))
    internals$rowIDs <- rowIDs
    internals$rowMeta <- setdiff(colnames(rowData[, -'rowKey']), rowIDs)

    # capture column internal metadata
    if (is.numeric(colIDs) || is.logical(colIDs))
        colIDs <- colnames(colData)[colIDs]
    if (!all(colIDs %in% colnames(colData)))
        stop(.errorMsg('\nColumn IDs not in colData: ',
            setdiff(colIDs, colnames(colData)), collapse=', '),
            call.=FALSE)
    internals$colIDs <- colIDs
    internals$colMeta <- setdiff(colnames(colData[, -'colKey']), colIDs)

    # -- capture assays internal metadata
    # sort such that rowIDs are first, then colIDs; ensures reindex returns
    # the same order as construtor
    for (i in seq_along(assayIDs)) {
        rids <- intersect(rowIDs, assayIDs[[i]])
        cids <- intersect(colIDs, assayIDs[[i]])
        assayIDs[[i]] <- c(rids, cids)
    }
    internals$assayKeys <- assayIDs

    # ensure names of assays and assayIDs match
    hasMatchingAssayNames <- names(assays) == names(assayIDs)
    if (!all(hasMatchingAssayNames)) stop(.errorMsg(
        "Mismatched names between assays and assayIDs for:\n\t",
        paste0(names(assays)[!hasMatchingAssayNames], collapse=", ")),
        call.=FALSE)
    # set keys for join with metadata
    for (nm in names(assays)) {
        setkeyv(assays[[nm]], assayIDs[[nm]])
        .nm <- paste0(".", nm)
        assays[[nm]][, (.nm) := .I]
    }
    
    # build the index mapping assay rows to rowKey and colKey
    cat(.infoMsg("Building assay index...\n", time=TRUE))
    assayIndex <- expand.grid(rowKey=rowData$rowKey, colKey=colData$colKey)
    setDT(assayIndex)
    # setkeyv(assayIndex, c("rowKey", "colKey"))
    
    cat(.infoMsg("Joining rowData to assayIndex...\n", time=TRUE))
    setkeyv(rowData, "rowKey")
    setkeyv(assayIndex, "rowKey")
    # assayIndex <- assayIndex[
    #     rowData[, c(rowIDs, "rowKey"), with=FALSE], ,
    #     on="rowKey", allow.cartesian=FALSE
    # ]
    rd <- rowData[, c(rowIDs, "rowKey"), with=FALSE]
    assayIndex <- merge(
        assayIndex, rd,
        by="rowKey", all.x=TRUE, allow.cartesian=FALSE
    )

    # print if rowKey in rowData is not unique
    if(nrow(rowData) != uniqueN(rowData$rowKey)) {
        cat(.warnMsg("rowData rowKey is not unique!"))
        show(assayIndex)
        show(rowData)
    }
    rm(rd)
    gc()
    cat(.infoMsg("Joining colData to assayIndex...\n", time=TRUE))
    setkeyv(colData, "colKey")
    # assayIndex <- assayIndex[
    #     colData[, c(colIDs, "colKey"), with=FALSE], ,
    #     on="colKey", allow.cartesian=FALSE
    # ]
    cd <- colData[, c(colIDs, "colKey"), with=FALSE]
    assayIndex <- merge(
        assayIndex, cd,
        by="colKey", all.x=TRUE, allow.cartesian=FALSE
    )
    rm(cd)
    gc()
    cat(.infoMsg("Joining assays to assayIndex...\n", time=TRUE))

    # Set the key variables for the assayIndex using rowIDs and colIDs
    setkeyv(assayIndex, c(rowIDs, colIDs))


    for (nm in names(assays)) {
        .nm <- paste0(".", nm)
        assayIndex[assays[[nm]], (.nm) := get(.nm)]
    }
    gc()
    assayIndex[, (c(rowIDs, colIDs)) := NULL]
    assayIndex <- assayIndex[
        which(rowAnys(!is.na(assayIndex[, paste0(".", names(assays)), with=FALSE]))),
    ]
    gc()
    cat(.infoMsg("Setting assayIndex key...\n", time=TRUE))
    setkeyv(assayIndex, paste0(".", names(assays)))
    internals$assayIndex <- assayIndex


    # make internals immutable to prevent users from modifying structural metadata
    internals <- immutable(internals)

    gc()
    cat(.infoMsg("Building LongTable...\n", time=TRUE))
    # Drop extra assay columns and key by the assay key in the assay index
    for (i in seq_along(assays)) {
        assays[[i]][, (assayIDs[[i]]) := NULL]
        setkeyv(assays[[i]], paste0(".", names(assays)[i]))
    }

    # Reorder columns to match the keys, this prevents issues in unit tests
    # caused by different column orders
    setkeyv(rowData, "rowKey")
    setkeyv(colData, "colKey")
    setcolorder(rowData, unlist(internals[c("rowIDs", "rowMeta")]))
    setcolorder(colData, unlist(internals[c('colIDs', 'colMeta')]))

    ## Assemble  the pseudo row and column names for the LongTable
    .pasteColons <- function(...) paste(..., collapse=':')
    rowData[, `:=`(.rownames=mapply(.pasteColons, transpose(.SD))),
        .SDcols=rowIDs]
    colData[, `:=`(.colnames=mapply(.pasteColons, transpose(.SD))),
        .SDcols=colIDs]
    return(CoreGx:::.LongTable(rowData=rowData, colData=colData, assays=assays,
        metadata=metadata, .intern=internals))
}

#' Function to combine two LongTables into a single LongTable
#' @param x A `LongTable` object
#' @param y A `LongTable` object
#' 

# ---- Class unions for CoreSet slots
#' A class union to allow multiple types in a CoreSet slot
#'
#' @include LongTable-class.R
setClassUnion('list_OR_LongTable', c('list', 'LongTable'))

# #' Ensure that all rowID and colID keys are valid
# #'
# #' @param rowData A `data.table` containing row level annotations.
# #' @param colData A `data.table` containing column level annotations for a
# #'   `LongTable`.
# #' @param assays A `list` of `data.table`s, one for each assay in an
# #'   `LongTable`.
# #'
# #' @keywords internal
### FIXME:: Finish this and implement class validity methods for LongTable!
#.verifyKeyIntegrity <- function(rowData, colData, assays) {
#    if (!('rowKey' %in% colnames(rowData)) || !is.numeric(rowData$rowID))
#        message(blue('The rowKey column is missing from rowData! Please try
#            rebuilding the LongTable object with the constructor.'))
#    if (!('colKey' %in% colnames(colData)) || !is.numeric(colData$colID))
#        stop()
#}

# ---- LongTable Class Methods

#' Helper function to print slot information
#' @param slotName `character` The name of the slot to print.
#' @param slotData `data.table` The data to print.
#' 
#' @keywords internal
printSlot <- function(slotName, slotData) {
    slotCols <- ncol(slotData)
    slotString <- paste0(slotName, '(', slotCols, '): ')
    slotColnames <- colnames(slotData)
    slotNamesString <-
        if (length(slotColnames) > 6) {
            paste0(.collapse(head(slotColnames, 3)), ' ... ', .collapse(tail(slotColnames, 3)))
        } else {
            .collapse(slotColnames)
        }
    cat("  ", yellow$bold(slotString) %+% green(slotNamesString), '\n')
}

#' Show method for the LongTable class
#'
#' @examples
#' show(merckLongTable)
#'
#' @param object A `LongTable` object to print the results for.
#'
#' @return `invisible` Prints to console.
#'
#' @importFrom crayon %+% yellow red green blue cyan magenta
#' @import data.table
#' @export
setMethod('show', signature(object='LongTable'), function(object) {

    ## FIXME:: Function too long. Can I refactor to a helper that prints each slot?

    # ---- class descriptions
    cat(yellow$bold$italic(paste0("<", class(object)[1], ">"), '\n'))
    cat("  ", yellow$bold('dim: ', .collapse(dim(object)), '\n'))

    # --- assays slot
    assayLength <- length(assayNames(object))
    assaysString <- paste0('assays(', assayLength, '): ')
    assayNames <- assayNames(object)
    assayNamesString <- .collapse(assayNames(object))
    if (nchar(assayNamesString) > options("width")) {
        assayNamesString <- paste0(strwrap(assayNamesString), collapse="\n  ")
    }
    cat("  ", yellow$bold(assaysString) %+% red(assayNamesString), '\n')

    # --- rownames
    rows <- nrow(rowData(object))
    rowsString <- paste0('rownames(', rows, '): ')
    rowNames <- rownames(object)
    rownamesString <-
        if (length(rowNames) > 6) {
            paste0(.collapse(head(rowNames, 2)), ' ... ', .collapse(tail(rowNames, 2)))
        } else {
            .collapse(rowNames)
        }
    cat("  ", yellow$bold(rowsString) %+% green(rownamesString), '\n')

    # ---- rowData slot
    printSlot('rowData', rowData(object))

    # ---- colnames
    cols <- nrow(colData(object))
    colsString <- paste0('colnames(', cols, '): ')
    colnames <- colnames(object)
    colnamesString <-
        if (length(colnames) > 6) {
            paste0(.collapse(head(colnames, 3)), ' ... ', .collapse(tail(colnames, 3)))
        } else {
            .collapse(colnames)
        }
    cat("  ", yellow$bold(colsString) %+% green(colnamesString), '\n')

    # ---- colData slot
    printSlot('colData', colData(object))

    # --- metadata slot
    metadataString <- paste0('metadata(', length(metadata(object)), '): ')
    metadataNames <- names(metadata(object))
    metadataNamesString <-
        if (length(metadataNames) > 6) {
            paste0(.collapse(head(metadataNames, 3), ' ... ', .collapse(tail(metadataNames, 3))))
        } else if (length(metadataNames) >= 1) {
            .collapse(metadataNames)
        } else {
            'none'
        }
    cat("  ", yellow$bold(metadataString) %+% green(metadataNamesString), '\n')
})


# ==== LongTable Accessor Methods

#' Get the id column names for the rowData slot of a LongTable
#'
#' @examples
#' rowIDs(merckLongTable)
#'
#' @param object A `LongTable` to get the rowData id columns for.
#' @param data `logical` Should the rowData for the id columns be returned
#' instead of the column names? Default is FALSE.
#' @param key `logical` Should the key column also be returned?
#'
#' @return A `character` vector of rowData column names if data is FALSE,
#' otherwise a `data.table` with the data from the rowData id columns.
#'
#' @rdname LongTable-class
#' @family LongTable-class
#' @family LongTable-accessors
#'
#' @import data.table
#' @export
setMethod('rowIDs', signature(object='LongTable'),
        function(object, data=FALSE, key=FALSE) {
    cols <- mutable(getIntern(object, 'rowIDs'))
    if (key) cols <- c(cols, 'rowKey')
    if (data) rowData(object, key=key)[, ..cols] else cols
})

#' Get the id column names for the rowData slot of a LongTable
#'
#' @examples
#' rowMeta(merckLongTable)
#'
#' @describeIn LongTable Get the names of the non-id columns from rowData.
#'
#' @param object A `LongTable` to get the rowData metadata columns for.
#' @param data `logical` Should the rowData for the metadata columns be returned
#' instead of the column names? Default is FALSE.
#' @param key `logical` Should the key column also be returned? Default is FALSE
#'
#' @return A `character` vector of rowData column names if data is FALSE,
#' otherwise a `data.table` with the data from the rowData metadta columns.
#'
#' @import data.table
#' @export
setMethod('rowMeta', signature(object='LongTable'),
        function(object, data=FALSE, key=FALSE) {
    cols <- mutable(getIntern(object, 'rowMeta'))
    cols <- cols[!grepl('^\\.', cols)]
    if (key) cols <- c(cols, 'rowKey')
    if (data) rowData(object, key=key)[, ..cols] else cols
})

#' Get the id column names for the colData slot of a LongTable
#'
#' @examples
#' colIDs(merckLongTable)
#'
#' @describeIn LongTable Get the names of the columns in colData required to
#' uniquely identify each row.
#'
#' @param object A `LongTable` to get the colData id columns for.
#' @param data `logical` Should the colData for the id columns be returned
#' instead of the column names? Default is FALSE.
#' @param key `logical` Should the key column also be returned? Default is FALSE.
#'
#' @return A `character` vector of colData column names if data is FALSE,
#' otherwise a `data.table` with the data from the colData id columns.
#'
#' @import data.table
#' @export
setMethod('colIDs', signature(object='LongTable'),
        function(object, data=FALSE, key=FALSE) {

    cols <- mutable(getIntern(object, 'colIDs'))
    if (key) cols <- c(cols, 'colKey')
    if (data) colData(object, key=TRUE)[, ..cols] else cols

})

#' Get the id column names for the colData slot of a LongTable
#'
#' @examples
#' colMeta(merckLongTable)
#'
#' @describeIn LongTable Get the names of the non-id columns in the colData
#'   `data.table`.
#'
#' @param object A `LongTable` to get the colData metadata columns for.
#' @param data `logical` Should the colData for the metadata columns be returned
#'   instead of the column names? Default is FALSE.
#' @param key `logical` Should the key column also be returned?
#'
#' @return A `character` vector of colData column names if data is FALSE,
#'   otherwise a `data.table` with the data from the colData metadta columns.
#'
#' @import data.table
#' @export
setMethod('colMeta', signature(object='LongTable'),
    function(object, data=FALSE, key=FALSE) {

    cols <- mutable(getIntern(object, 'colMeta'))
    cols <- cols[!grepl('^\\.', cols)]
    if (key) cols <- c(cols, 'colKey')
    if (data) colData(object, key=TRUE)[, ..cols] else cols
})



#' Retrieve the unique identifier columns used for primary keys in rowData and
#'    colData.
#'
#' @describeIn LongTable Get the names of all id columns.
#'
#' @examples
#' idCols(merckLongTable)
#'
#' @param object `LongTable`
#'
#' @return `character` A character vector containing the unique rowIDs and
#'   colIDs in a LongTable object.
#'
#' @export
setMethod('idCols', signature('LongTable'),
    function(object) {
    return(unique(c(rowIDs(object), colIDs(object))))
})

#' Retrieve a copy of the assayIndex from the `@.intern` slot.
#'
#' @describeIn LongTable Get the assayIndex item from the objects internal metadata.
#'
#' @param `x` A `LongTable` or inheriting class.
#'
#' @return A `mutable` copy of the "assayIndex" for `x`
#'
#' @examples
#' assayIndex(nci_TRE_small)
#'
#' @aliases assayIndex,LongTable-method
#' @export
setMethod("assayIndex", signature("LongTable"), function(x) {
    mutable(getIntern(x, "assayIndex"))
})

#' Retrieve a copy of the assayKeys from the `@.intern` slot.
#'
#' @describeIn LongTable Get the assayKeys item from the objects internal metadata.
#'
#' @param `x` A `LongTable` or inheriting class.
#' @param `i` An optional valid assay name or index in `x`.
#'
#' @return A `mutable` copy of the "assyKeys" for `x`
#'
#' @examples
#' assayKeys(nci_TRE_small)
#' assayKeys(nci_TRE_small, "sensitivity")
#' assayKeys(nci_TRE_small, 1)
#'
#' @aliases assayKeys,LongTable-method
#' @export
setMethod("assayKeys", signature("LongTable"), function(x, i) {
    keys <- mutable(getIntern(x, "assayKeys"))
    # error handling occurs in `[[`
    if (!missing(i)) keys[[i]] else keys
})


#' Retrieve the value columns for the assays in a LongTable
#'
#' @examples
#' assayCols(merckLongTable)
#'
#' @describeIn LongTable Get a list of column names for each assay in the object.
#'
#' @param object `LongTable`
#' @param i Optional parameter specifying the `character` name or `integer`
#' index of the assay to get the column names for. If missing, returns a
#' list of value column names for all the assays.
#'
#' @return A `list` of `character` vectors containing the value column names for
#' each assay if i is missing, otherwise a `character` vector of value column
#' names for the selected assay.
#'
#' @import data.table
#' @export
setMethod('assayCols', signature(object='LongTable'),
        function(object, i) {
    if (!missing(i)) {
        stopifnot(is.numeric(i) || is.character(i))
        stopifnot(length(i) == 1)
        stopifnot(i %in% assayNames(object) ||
            i %in% seq_along(assayNames(object)))
    }
    keys <- assayKeys(object)
    assayColnames <- Map(setdiff,
        x=lapply(assays(object, raw=TRUE), FUN=colnames),
        y=as.list(paste0(".", assayNames(object)))
    )
    assayCols <- Map(c, keys, assayColnames)
    if (!missing(i)) assayCols[[i]] else assayCols
})

library(testthat)
library(CoreGx)
library(data.table)

data(nci_TRE_small)
data(merckLongTable)

lt <- merckLongTable
tre <- nci_TRE_small

# == LongTable constructor

testthat::test_that("`LongTable` is coercible to TRE", {
    tre <- as(lt, "TreatmentResponseExperiment")
    testthat::expect_s4_class(tre, "TreatmentResponseExperiment")
})

testthat::test_that("`LongTable` constructor method works with valid inputs", {
    ## Extract required parameters to create an TRE object
    parameters <- formalArgs(LongTable)
    parameters <- parameters[
        !(parameters %in% c("metadata", "keep.rownames"))
    ]
    row_data <- rowData(tre)
    row_ids <- rowIDs(tre)
    col_data <- colData(tre)
    col_ids <- colIDs(tre)
    assays_ <- assays(tre)
    assay_ids <- replicate(3, idCols(tre), simplify = FALSE)
    names(assay_ids) <- assayNames(tre)
    ## regex lookaheads to check for ALL missing parameters
    regex <- paste0(sprintf("(?=.*%s)", parameters), collapse = "")
    regex <- paste0("(?s)^", regex) ## handle line breaks in error messages
    ## Line 87: Report all missing parameters in error message
    testthat::expect_error({ ntre <- LongTable() },
        regexp = regex, perl = TRUE
    )
    ## Check for wrong input rowData class (those not coercible to data.frame)
    ## FIX-ME:: We might need extra check for rowData, colData, assays: even NULL is coercible to data.table
    #testthat::expect_error({
    #    ntre <- LongTable(rowData  = NULL,
    #                      rowIDs   = row_ids,
    #                      colData  = col_data,
    #                      colIDs   = col_ids,
    #                      assays   = assays_,
    #                      assayIDs = array_ids)
    #},
    #    regexp = ".*rowData must be coerceible to a data\\.frame"
    #)
    ## Question: should we handle the case where assays' IDs are mislabeled?
    ## Question: should we check for unequal length of names(assays) and names(assayIDs)? (refer to line 171)
    ## Line 172
    testthat::expect_error({
        ntre <- LongTable(rowData = row_data[, -row_ids[1:2], with = FALSE],
                          rowIDs = row_ids,
                          colData = col_data,
                          colIDs = col_ids,
                          assays = assays_,
                          assayIDs = array_ids)
    },
        regexp = paste0(".*Row IDs not in rowData: ",
                        row_ids[1:2],
                        collapse = ",")
    )
    ## Question: should we handle the case where assays' IDs are mislabeled?
    ## Question: should we check for unequal length of names(assays) and names(assayIDs)? (refer to line 171)
    ## Line 172
    testthat::expect_error({
        names(assay_ids)[1] <- "not sensitivity"
        ntre <- LongTable(rowData = row_data,
                          rowIDs = row_ids,
                          colData = col_data,
                          colIDs = col_ids,
                          assays = assays_,
                          assayIDs = assay_ids)
    },
        regexp = paste0(".*Mismatched names between ",
                        "assays and assayIDs for\\:\n\t",
                        paste0(names(assays_)[
                                    names(assays_) != names(assay_ids)
                               ],
                               collapse = ", "),
                        ".*", collapse = "")
    )
})

# == assayCols

testthat::test_that("`assayCols,LongTable-method` retrieves specified assay's column names",{
    testthat::expect_error({ assayCols(tre, i = 1:2) })
    testthat::expect_error({ assayCols(tre, i = (length(assayNames(tre))) + 1) })
    testthat::expect_error({ assayCols(tre, i = paste0(assayNames(tre), collapse = "")) })
})

#' @include CoreSet-class.R CoreSet-accessors.R
#' @importFrom BiocGenerics match %in%
NULL

.local_class <- 'CoreSet'
.local_data <- 'clevelandSmall_cSet'

#### CoreGx dynamic documentation
####
#### Warning: for dynamic docs to work, you must set
#### Roxygen: list(markdown = TRUE, r6=FALSE)
#### in the DESCRPTION file!


# ===================================
# Utility Method Documentation Object
# -----------------------------------


#' @noRd
.docs_CoreSet_utils <- function(...) .parseToRoxygen(
    "
    @title Utility methods for a `{class_}` object.

    @description
    Documentation for utility methods for a `{class_}` object, such as
    set operations like subset and intersect. See @details for information
    on different types of methods and their implementations.

    @param x A `{class_}` object.
    @param samples `character()` vector of sample names. Must be valid rownames
    from `sampleInfo(x)`.
    @param treatments `character()` vector of treatment names. Must be valid
    rownames from `treatmentInfo(x)`. This method does not work with
    `CoreSet` objects yet.
    @param features `character()` vector of feature names. Must be valid feature
    names for a given `mDataType`
    @param mDataTypes `character()` One or more molecular data types to
        to subset features by. Must be valid rownames for the selected
        SummarizedExperiment mDataTypes.

    @return See details.
    ",
    ...
)


#' @name CoreSet-utils
#' @eval .docs_CoreSet_utils(class_=.local_class)
#' @eval .parseToRoxygen("@examples data({data_})", data_=.local_data)
NULL


# ======================================
# Subset Methods
# --------------------------------------


## ===================
## ---- subsetBySample
## -------------------


#' @export
setGeneric('subsetBySample', function(x, samples, ...)
    standardGeneric('subsetBySample'))

#' @noRd
.docs_CoreSet_subsetBySample <- function(...) .parseToRoxygen(
    "
    @details

    ## subset methods
    __subsetBySample__: Subset a `{class_}` object by sample identifier.
    - value: a `{class_}` object containing only `samples`.

    @examples

    ## subset methods

    ### subsetBySample
    samples <- sampleInfo({data_})$sampleid[seq_len(10)]
    {data_}_sub <- subsetBySample({data_}, samples)

    @md
    @aliases subsetBySample subsetBySample,CoreSet-method
    @exportMethod subsetBySample
    ",
    ...
)

#' @rdname CoreSet-utils
#' @eval .docs_CoreSet_subsetBySample(class_=.local_class, data_=.local_data)
setMethod('subsetBySample', signature('CoreSet'), function(x, samples) {

    funContext <- .S4MethodContext('subsetBySample', 'CoreSet')

    sampleNames <- rownames(sampleInfo(x))
    if (!all(samples %in% sampleNames)) {
        .warning(funContext, 'Samples missing from ', class(x)[1], ': ',
            setdiff(samples, sampleNames), '! Please ensure all specified
            samples are valid rownames of sampleInfo(x). Proceeding with
            the valid samples only.')
        samples <- union(samples, sampleNames)
    }

    # -- molecularProfiles slot
    molecSlot <- molecularProfilesSlot(x)
    molecularProfilesSlot(x) <-
        .subsetMolecularProfilesBySample(molecSlot, samples)

    # -- sensitivity slot
    sensSlot <- treatmentResponse(x)
    treatmentResponse(x) <- .subsetSensitivityBySample(sensSlot, samples)

    # -- perturbatiion slot
    ##TODO:: do we still need this?
    
    # -- curation slot
    sampleCuration <- curation(x)$sample
    curation(x)$sample <- sampleCuration[rownames(sampleCuration) %in% samples, ]

    # -- sample slot
    sampleInf <- sampleInfo(x)
    sampleInfo(x) <- sampleInf[rownames(sampleInf) %in% samples, ]

    # -- check object is still valid and return
    tryCatch(checkCsetStructure(x), error = function(e) {})

    return(x)
})

.subsetMolecularProfilesBySample <- function(slotData, samples) {
    funContext <- .funContext(':::.subsetMolecularProfilesBySample')
    if (is(slotData, 'MultiAssayExperiment')) {
        hasSamples <- colData(slotData)$sampleid %in% samples
        if (!all(hasSamples)) .warning(funContext, 'Some specified samples are
            not present in `molecularProfilesSlot(x)`')
        molecProfs <- slotData[, hasSamples]
    } else {
        SEcolData <- lapply(slotData, colData)
        SEsamples <- lapply(SEcolData, FUN=`[[`, i='sampleid')
        hasSEsamples <- lapply(SEsamples, FUN=`%in%`, samples)
        molecProfs <- mapply(`[`, x=slotData, j=hasSEsamples)
    }
    return(molecProfs)
}

.subsetSensitivityBySample <- function(slotData, samples) {
    funContext <- .funContext(':::.subsetSensitivityBySample')
    if (is(slotData, 'LongTable')) {
        slotData <- slotData[, samples]
    } else {
        keepSamples <- slotData$info$sampleid %in% samples
        slotData$profiles <- slotData$profiles[keepSamples, ]
        slotData$raw <- slotData$raw[keepSamples, , ]
        slotData$n <- slotData$n[rownames(slotData$n) %in% samples, ]
        slotData$info <- slotData$info[keepSamples, ]
    }
    return(slotData)
}

## ======================
## ---- subsetByTreatment
## ----------------------

#' @export
setGeneric('subsetByTreatment', function(x, treatments, ...)
    standardGeneric('subsetByTreatment'))

#' @noRd
.docs_CoreSet_subsetByTreatment <- function(...) .parseToRoxygen(
    "
    @details

    ## subset methods
    __subsetByTreatment__: Subset a `{class_}` object by treatment identifier.
    - value: a `{class_}` object containing only `treatments`.

    @examples

    ## subset methods

    ### subsetByTreatment
    #treatments <- {treatment_}Info({data_})${treatment_}id[seq_len(10)]
    #{data_}_sub <- subsetByTreatment({data_}, treatments)

    @md
    @aliases subsetByTreatment subsetByTreatment,{class_}-method
    @exportMethod subsetByTreatment
    ",
    ...
)

#' @rdname CoreSet-utils
#' @eval CoreGx:::.docs_CoreSet_subsetByTreatment(class_=.local_class,
#' data_=.local_data, treatment_='treatment')
setMethod('subsetByTreatment', signature('CoreSet'),
        function(x, treatments) {
    funContext <- .S4MethodContext('subsetByTreatment', 'PharmacoSet')
    treatmentType <- switch(class(x)[1],
        'PharmacoSet'='drug',
        'ToxicoSet'='drug',
        'RadioSet'='radiation',
        'CoreSet'=return(data.frame())
    )
    treatmentNames <- rownames(treatmentInfo(x))
    if (!all(treatments %in% treatmentNames)) {
        .warning(funContext, 'Treatments missing from ', class(x)[1], ': ',
            setdiff(treatments, treatmentNames), '! Please ensure all specified
            treatments are valid rownames of treatmentInfo(x).
            Proceeding with the valid treatments only.')
        treatments <- union(treatments, treatmentNames)
    }
    # -- sensitivity slot
    sensSlot <- treatmentResponse(x)
    treatmentResponse(x) <- .subsetSensitivityByTreatment(sensSlot, treatments,
        treatmentType=treatmentType)

    # -- perturbation slot
    ## TODO: do we still need this?

    # -- curation slot
    treatmentCuration <- curation(x)[[treatmentType]]
    curation(x)[[treatmentType]] <- treatmentCuration[
        rownames(treatmentCuration) %in% treatments, ]

    # -- treatment slot
    treatmentInf <- treatmentInfo(x)
    treatmentInfo(x) <- treatmentInf[rownames(treatmentInf) %in% treatments, ]

    # -- molecularProfiles
    # deal with potential loss of samples when subsetting by treatment
    keepSamples <- sampleNames(x)
    molecSlot <- molecularProfilesSlot(x)
    molecularProfilesSlot(x) <- .subsetMolecularProfilesBySample(molecSlot,
        keepSamples)

    # -- check object is still valid and return
    tryCatch(checkCsetStructure(x), error = function(e) {})

    return(x)
})

.subsetSensitivityByTreatment <- function(slotData, treatments,
        treatmentType) {
    funContext <- .funContext(':::.subsetSensitivityByTreatment')
    treatmentId <- if (treatmentType == 'radiation')
        paste0(treatmentType, '.type') else paste0(treatmentType, 'id')
    if (is(slotData, 'LongTable')) {
        slotData <- slotData[treatments, ]
    } else {
        keepTreatments <- slotData$info[[treatmentId]] %in% treatments
        slotData$profiles <- slotData$profiles[keepTreatments, ]
        slotData$raw <- slotData$raw[keepTreatments, , ]
        slotData$info <- slotData$info[keepTreatments, ]
        slotData$n <- slotData$n[, colnames(slotData$n) %in% treatments]
    }
    return(slotData)
}


## ====================
## ---- subsetByFeature
## --------------------


#' @export
setGeneric('subsetByFeature', function(x, features, ...)
    standardGeneric('subsetByFeature'))

#' @noRd
.docs_CoreSet_subsetByFeature <- function(...) .parseToRoxygen(
    "
    @details

    ## subset methods
    __subsetByFeature__: Subset a `{class_}` object by molecular feature
        identifier.
    - value: a `{class_}` object containing only `features`.

    @examples

    ## subset methods

    ### subsetByFeature
    features <- fNames({data_}, 'rna')[seq_len(5)]
    {data_}_sub <- subsetByFeature({data_}, features, 'rna')

    @md
    @aliases subsetByFeature subsetByFeature,{class_}-method
    @importFrom MultiAssayExperiment MultiAssayExperiment
    @exportMethod subsetByFeature
    ",
    ...
)

#' @rdname CoreSet-utils
#' @eval .docs_CoreSet_subsetByFeature(class_=.local_class, data_=.local_data)
setMethod('subsetByFeature', signature(x='CoreSet'),
        function(x, features, mDataTypes) {
    slotData <- molecularProfilesSlot(x)
    MAE <- if (!is(slotData, 'MultiAssayExperiment'))
        MultiAssayExperiment(slotData) else slotData
    if (missing(mDataTypes)) mDataTypes <- names(MAE)
    suppressMessages({
        suppressWarnings({
            MAE_sub <- MAE[, , mDataTypes]
        })
    })
    keepFeatures <- rownames(MAE_sub) %in% features
    subsetMAE <- MAE[keepFeatures, drop=TRUE]
    newSlotData <- if (is(slotData, 'MultiAssayExperiment')) subsetMAE else
        as.list(experiments(subsetMAE))
    molecularProfilesSlot(x) <- newSlotData
    ## TODO:: What if this drops samples from the PSet?
    return(x)
})

# ==== LongTable Class

#' Get the column names from a `LongTable` object.
#'
#' @examples
#' head(colnames(merckLongTable))
#'
#' @describeIn LongTable Retrieve the pseudo-colnames of a LongTable object,
#'   these are constructed by pasting together the colIDs(longTable) and
#'   can be used in the subset method for regex based queries.
#'
#' @param x A `LongTable` object to get the column names from
#'
#' @return `character` Vector of column names.
#'
#' @export
setMethod('colnames', signature(x='LongTable'), function(x) {
    return(x@colData$.colnames)
})

#' Get the row names from a `LongTable` object.
#'
#' @examples
#' head(rownames(merckLongTable))
#'
#' @describeIn LongTable Retrieve the pseudo-rownames of a LongTable object,
#'   these are constructed by pasting together the rowIDs(longTable) and
#'   can be used in the subset method for regex based queries.
#'
#' @param x A `LongTable` object to get the row names from
#'
#' @return `character` Vector of row names.
#'
#' @export
setMethod('rownames', signature(x='LongTable'), function(x) {
    return(x@rowData$.rownames)
})

#' Getter for the dimnames of a `LongTable` object
#'
#' @examples
#' lapply(dimnames(merckLongTable), head)
#'
#' @describeIn LongTable Get the pseudo-dimnames for a LongTable object. See
#'   colnames and rownames for more information.
#'
#' @param x The `LongTable` object to retrieve the dimnames for.
#'
#' @return `list` List with two character vectors, one for row and one for
#'     column names.
#'
#' @importMethodsFrom Biobase dimnames
#' @export
setMethod('dimnames', signature(x='LongTable'), function(x) {
    return(list(x@rowData$.rownames, x@colData$.colnames))
})

# ==== CoreSet

#' Summarize across replicates for a sensitivity dose-response experiment
#'
#' @examples
#' print("Generics shouldn't need examples?")
#'
#' @param object An `S4` object to summarize sensitivity profiles for.
#' @param ... Allow definition of new arguments to this generic
#'
#' @return Depends on the implemented method
#'
#' @export
setGeneric("summarizeSensitivityProfiles",
    function(object, ...) standardGeneric("summarizeSensitivityProfiles"))


#' Summarize molecular profile data such that there is a single entry for each
#'   sample line/treatment combination
#'
#' @examples
#' print("Generics shouldn't need examples?")
#'
#' @param object An `S4` object to summarize the molecular profiles for.
#' @param ... Allow definition of new arguments to this generic
#'
#' @return Depends on the implemented method
#'
#' @export
setGeneric("summarizeMolecularProfiles",
    function(object, ...) standardGeneric("summarizeMolecularProfiles"))


#' Get the annotations for a `Signature` class object, as returned by
#'   `drugSensitivitysig` or `radSensitivtySig` functions available in
#'   `PharmacoGx` and `RadioGx`, respectively.
#'
#' @examples
#' print("Generics shouldn't need examples?")
#'
#' @param object A `Signature` class object
#' @param ... Allow definition of new arguments to this generic
#'
#' @return NULL Prints the signature annotations to console
#'
#' @export
setGeneric("showSigAnnot",
    function(object, ...) standardGeneric("showSigAnnot"))


#' Generic function to get the annotations for a treatment response experiment
#'   from an S4 class
#'
#' @examples
#' print("Generics shouldn't need examples?")
#'
#' @param object An `S4` object to get treatment response experiment
#'    annotations from.
#' @param ... Allow new arguments to be defined for this generic.
#'
#' @return Depends on the implemented method
#'
#' @export
setGeneric("sensitivityInfo",
    function(object, ...) standardGeneric("sensitivityInfo"))


#' sensitivityInfo<- Generic Method
#'
#' Generic function to get the annotations for a treatment response experiment
#'   from an S4 class.
#'
#' @examples
#' print("Generics shouldn't need examples?")
#'
#' @param object An `S4` object to set treatment response experiment
#'    annotations for.
#' @param ... Allow new arguments to be defined for this generic.
#' @param value The new treatment response experiment annotations.
#'
#' @return Depends on the implemented method
#'
#' @export
setGeneric("sensitivityInfo<-",
    function(object, ..., value) standardGeneric("sensitivityInfo<-"))


#' sensitivityRaw Generic Method
#'
#' Generic function to get the raw data array for a treatment response experiment
#'   from an S4 class.
#'
#' @examples
#' print("Generics shouldn't need examples?")
#'
#' @param object An `S4` object to extract the raw sensitivity experiment
#'     data from.
#' @param ... `pairlist`  Allow new parameters to be defined for this generic.
#'
#' @return Depends on the implemented method
#'
#' @export
setGeneric("sensitivityRaw",
    function(object, ...) standardGeneric("sensitivityRaw"))

#' sensitivityRaw<- Generic
#'
#' Generic function to set the raw data array for a treatment response experiment
#'   in an S4 class.
#'
#' @param object An `S4` object to extract the raw sensitivity data from.
#' @param ... `pairlist` Allow new parameters to be defined for this generic.
#' @param value An object containing dose and viability metrics to update
#'   the object with.
#'
#' @return Depends on the implemented method
#'
#' @export
setGeneric("sensitivityRaw<-",
    function(object, ..., value) standardGeneric("sensitivityRaw<-"))

#' sensitivityProfiles Generic
#'
#' A generic for sensitivityProfiles getter method
#'
#' @examples
#' print("Generics shouldn't need examples?")
#'
#' @param object The `S4` object to retrieve sensitivity profile summaries
#'   from.
#' @param ... `pairlist` Allow defining new arguments for this generic.
#'
#' @return Depends on the implemented method
#'
#' @export
setGeneric("sensitivityProfiles", function(object, ...) standardGeneric("sensitivityProfiles"))

#' sensitivityProfiles<- Generic
#'
#' A generic for the sensitivityProfiles replacement method
#'
#' @param object An `S4` object to update the sensitivity profile summaries
#'    for.
#' @param ... Fallthrough arguments for defining new methods
#' @param value An object with the new sensitivity profiles. If a
#'   matrix object is passed in, converted to data.frame before assignment
#'
#' @return Updated \code{CoreSet}
#'
#' @export
setGeneric("sensitivityProfiles<-",
    function(object, ..., value) standardGeneric("sensitivityProfiles<-"))

#' sensitivityMeasures Generic
#'
#' Get the names of the sensitivity summary metrics available in an S4
#'   object.
#'
#' @examples
#' sensitivityMeasures(clevelandSmall_cSet)
#'
#' @param object An `S4` object to retrieve the names of sensitivty summary
#'    measurements for.
#' @param ... Fallthrough arguements for defining new methods
#'
#' @return Depends on the implemented method
#'
#' @export
setGeneric("sensitivityMeasures",
    function(object, ...) standardGeneric("sensitivityMeasures"))

#' sensitivityMeasures<- Generic
#'
#' Set the names of the sensitivity summary metrics available in an S4
#'   object.
#'
#' @examples
#' print("Generics shouldn't need examples?")
#'
#' @param object An `S4` object to update.
#' @param ... Allow new methods to be defined for this generic.
#' @param value A set of names for sensitivity measures to use to
#'   update the object with.
#'
#' @return Depends on the implemented method
#'
#' @export
setGeneric('sensitivityMeasures<-',
    function(object, ..., value) standardGeneric('sensitivityMeasures<-'))

#' sensitivitySlotToLongTable Generic
#'
#' Convert the sensitivity slot in an object inheriting from a CoreSet from a
#'   list to a LongTable.
#'
#' @examples
#' print("Generics shouldn't need examples?")
#'
#' @param object `CoreSet` Object inheriting from CoreSet.
#' @param ... Allow new arguments to be defined on this generic.
#'
#' @return A `LongTable` object containing the data in the sensitivity slot.
#'
#' @export
setGeneric('sensitivitySlotToLongTable',
    function(object, ...) standardGeneric('sensitivitySlotToLongTable'))

# ==== LongTable Class

#' Generic method for resetting indexing in an S4 object
#'
#' This method allows integer indexes used to maintain referential integrity
#'   internal to an S4 object to be reset. This is useful particularly after
#'   subsetting an object, as certain indexes may no longer be present in the
#'   object data. Reindexing removes gaps integer indexes and ensures that the
#'   smallest contiguous integer values are used in an objects indexes.
#'
#' @examples
#' print("Generics shouldn't need examples?")
#'
#' @param object `S4` An object to redo indexing for
#' @param ... `pairlist` Allow definition of new parameters to this generic.
#'
#' @return Depends on the implemented method
#'
#' @export
setGeneric('reindex', function(object, ...) standardGeneric('reindex'))

#' Build a LongTable object
#'
#' @examples
#' print("Generics shouldn't need examples?")
#'
#' @param from What to build the LongTable from?
#' @param ... `pairlist` Allow definition of new parameters for
#'     implementations of this generic.
#'
#' @return Depends on the implemented method
#'
#' @export
setGeneric('buildLongTable',
    function(from, ...) standardGeneric('buildLongTable'))


#' Perform aggregation over an S4 object, but return an object of the same
#' class.
#'
#' @examples
#' print("Generics shouldn't need examples?")
#'
#' @param x An `S4` object to endomorphically aggregate over.
#' @param ... `pairlist` Allow definition of new parameters for
#'     implementations of this generic.
#'
#' @return An object with the same class as `x`.
#'
#' @export
setGeneric("endoaggregate", function(x, ...) standardGeneric("endoaggregate"))

#' Retrieve a set of assayKeys
#'
#' @examples
#' print("Generics shouldn't need examples?")
#'
#' @param x An `S4` object.
#' @param ... `pairlist` Allow definition of new parameters for
#'     implementations of this generic.
#'
#' @return An object representing the "assayKeys" of an `S4` object.
#'
#' @export
setGeneric("assayKeys", function(x, ...) standardGeneric("assayKeys"))


#' Retrieve and assayIndex
#'
#' @examples
#' print("Generics shouldn't need examples?")
#'
#' @param x An `S4` object.
#' @param ... `pairlist` Allow definition of new parameters for
#'     implementations of this generic.
#'
#' @return An object representing the "assayIndex" of an `S4` object.
#'
#' @export
setGeneric("assayIndex", function(x, ...) standardGeneric("assayIndex"))

# ===== Other Generics


#' Retrieve the specified item from object internal metadata.
#'
#' Internal slot for storing metadata relevant to the internal operation of an
#'     S4 object.
#'
#' Warning: This method is intended for developer use and can be ignored by
#'   users.
#'
#' @examples
#' print("Generics shouldn't need examples?")
#'
#' @param object `S4` An object with an @.itern slot containing an environment.
#' @param x `character` One or more symbol names to retrieve from the
#'    object@.intern environment.
#' @param ... Allow new parmeters to be defined for this generic.
#'
#' @return Depends on the implemented method
#'
#' @export getIntern
setGeneric('getIntern',
    function(object, x, ...) standardGeneric('getIntern'))


#' Set the internal structural metadata for an S4 class
#'
#' @param object An R object to update internal structural metadata for.
#' @param value An `immutable_list` object, being a class union between `list`
#'   and `immutable` S3 classes.
#'
#' @examples
#' print("Generics shouldn't need examples?")
#'
#' @return Updates the object and returns invisibly.
#'
#' @keywords internal
setGeneric("getIntern<-",
    function(object, ..., value) standardGeneric("getIntern<-"))


#' Generic to access the row identifiers from
#'
#' @examples
#' print("Generics shouldn't need examples?")
#'
#' @param object `S4` An object to get row id columns from.
#' @param ... Allow new arguments to this generic.
#'
#' @return Depends on the implemented method.
#'
#' @export
setGeneric('rowIDs', function(object, ...) standardGeneric('rowIDs'))


#' Generic to access the row identifiers from
#'
#' @examples
#' print("Generics shouldn't need examples?")
#'
#' @param object `S4` An object to get row metadata columns from.
#' @param ... Allow new arguments to this generic.
#'
#' @return Depends on the implemented method.
#'
#' @export
setGeneric('rowMeta', function(object, ...) standardGeneric('rowMeta'))


#' Generic to access the row identifiers for an object.
#'
#' @examples
#' print("Generics shouldn't need examples?")
#'
#' @param object `S4` An object to get column id columns from.
#' @param ... ALlow new arguments to this generic
#'
#' @return Depends on the implemented method.
#'
#' @export
setGeneric('colIDs', function(object, ...) standardGeneric('colIDs'))


#' Generic to access the column identifiers for a rectangular object.
#'
#' @examples
#' print("Generics shouldn't need examples?")
#'
#' @param object `S4` An object to get column metadata columns from.
#' @param ... ALlow new arguments to this generic
#'
#' @return Depends on impemented method.
#'
#' @export
setGeneric('colMeta', function(object, ...) standardGeneric('colMeta'))


#' Generic to access the assay columns of a rectangular object.
#'
#' @examples
#' print("Generics shouldn't need examples?")
#'
#' @param object `S4` An object to get assay ids from.
#' @param ... Allow new arguments to this generic.
#'
#' @return Depends on the implemented method.
#'
#' @export
setGeneric('assayCols',
    function(object, ...) standardGeneric('assayCols'))


#' Generic to access the unique id columns in an S4 object used to
#'
#' @examples
#' print("Generics shouldn't need examples?")
#'
#' @param object An `S4` object to get id columns from.
#' @param ... Allow new arguments to this generic.
#'
#' @return Depends on the implemented method
#'
#' @export
setGeneric('idCols',
    function(object, ...) standardGeneric('idCols'))

##' Generic to access the build configuration for an S4 object.
##'
##' @param object `S4` The object to retireve the configuration from.
##'
##'
#setGeneric('getConfig', function(object, ...) standardGeneric(''))

## adaptive Matthews correlation coefficient for binary classification
#' Calculate an Adaptive Matthews Correlation Coefficient
#' 
#' This function calculates an Adaptive Matthews Correlation Coefficient (AMCC) 
#' for two vectors of values of the same length. It assumes the entries in the 
#' two vectors are paired. The Adaptive Matthews Correlation Coefficient for two
#' vectors of values is defined as the Maximum Matthews Coefficient over all 
#' possible binary splits of the ranks of the two vectors. In this way, it 
#' calculates the best possible agreement of a binary classifier on the two 
#' vectors of data. If the AMCC is low, then it is impossible to find any binary 
#' classification of the two vectors with a high degree of concordance.   
#' 
#' @examples 
#' x <- c(1,2,3,4,5,6,7)
#' y <- c(1,3,5,4,2,7,6)
#' amcc(x,y, min.cat=2)
#' 
#' @param x,y Two paired vectors of values. Could be replicates of observations 
#'   for the same experiments for example.  
#' @param step.prct Instead of testing all possible splits of the data, it is 
#'   possible to test steps of a percentage size of the total number of ranks in 
#'   x/y. If this variable is 0, function defaults to testing all possible 
#'   splits.
#' @param min.cat The minimum number of members per category. Classifications 
#'   with less members fitting into both categories will not be considered. 
#' @param nperm The number of perumatation to use for estimating significance. 
#'   If 0, then no p-value is calculated. 
#' @param nthread Number of threads to parallize over. Both the AMCC calculation 
#'   and the permutation testing is done in parallel. 
#' @param ... Additional arguments
#' 
#' @return Returns a list with two elements. $amcc contains the highest 'mcc' 
#'   value over all the splits, the p value, as well as the rank at which the 
#'   split was done.
#'
#' @importFrom BiocParallel bplapply
#' @importFrom stats quantile
#'
#' @export
## FIXME:: We need a more descriptive name for this function
amcc <- function(x, y, step.prct = 0, min.cat = 3, nperm = 1000, nthread = 1, ...) {
    # PARAMETER CHANGE WARNING
    if (!missing(...)) {
        if ("setseed" %in% names(...)) {
            warning("The setseed parameter has been removed in this release to conform
              to Bioconductor coding standards. Please call set.seed in your
              script before running this function.")
        }
    }
    
    if (!min.cat > 1) {
        
        stop("Min.cat should be at least 2")
        
    }
    ccix <- complete.cases(x, y)
    if (sum(ccix) >= (2 * min.cat)) {
        x2 <- rank(-x[ccix], ties.method = "first")
        y2 <- rank(-y[ccix], ties.method = "first")
        ## compute mcc for each rank
        iix <- seq_len(min(max(x2), max(y2)) - 1)
        if (step.prct > 0) {
            iix <- round(quantile(iix, probs = seq(0, 1, by = step.prct)))
        }
        splitix <- parallel::splitIndices(nx = length(iix), ncl = nthread)
        splitix <- splitix[vapply(splitix, length, FUN.VALUE = numeric(1)) > 0]
        mcres <- bplapply(splitix, function(x, iix, x2, y2) {
            res <- t(vapply(iix[x], function(x, x2, y2) {
                x3 <- factor(ifelse(x2 <= x, "1", "0"))
                y3 <- factor(ifelse(y2 <= x, "1", "0"))
                res <- mcc(x = x3, y = y3, nperm = 0, nthread = 1)
                return(res)
            }, x2 = x2, y2 = y2, FUN.VALUE = list(1, 1)))
            ## TODO:: Why is return value a list of doubles when the class of res is matrix in debug ?
            return(res)
        }, iix = iix, x2 = x2, y2 = y2)
        mm <- do.call(rbind, mcres)
        mode(mm) <- "numeric"
        ## remove extreme indices
        rmix <- c(seq_len(min.cat - 1), (nrow(mm) - min.cat + 2):nrow(mm))
        mccix <- max(which(mm[-rmix, "estimate", drop = FALSE] == max(mm[-rmix, "estimate", drop = FALSE], na.rm = TRUE))) + (min.cat - 1)
        ## compute significance only for the AMCC
        x3 <- factor(ifelse(x2 <= mccix, "1", "0"))
        y3 <- factor(ifelse(y2 <= mccix, "1", "0"))
        if (nperm > 0) {
            mm[mccix, "p.value"] <- mcc(x = x3, y = y3, nperm = nperm, nthread = nthread)[["p.value"]]
            ## bonferronni correction mm[mccix, 'p'] <- mm[mccix, 'p'] * length(x3)
        }
        if (!is.na(mm[mccix, "p.value"]) && mm[mccix, "p.value"] > 1) {
            mm[mccix, "p.value"] <- 1
        }
        res <- c(mcc = mm[mccix, ], n1 = mccix, n2 = nrow(mm) - mccix, n = nrow(mm))
    } else {
        res <- c(mcc = NA, p = NA, n1 = 0, n2 = 0, n = sum(ccix))
        mm <- NA
    }
    names(res) <- c("mcc", "p", "n1", "n2", "n")
    return(list(amcc = res, mcc = mm))
}

#' Generic for preprocessing complex data before being used in the constructor
#'   of an `S4` container object.
#'
#' This method is intended to abstract away complex constructor arguments
#'   and data preprocessing steps needed to transform raw data, such as that
#'   produced in a treatment-response or next-gen sequencing experiment, and
#'   automate building of the appropriate `S4` container object. This is
#'   is intended to allow mapping between different experimental designs,
#'   in the form of an `S4` configuration object, and various S4 class
#'   containers in the Bioconductor community and beyond.
#'
#' @param mapper An `S4` object abstracting arguments to an `S4` class
#'   constructor into a well documented `Mapper` object.
#' @param ... Allow new arguments to be defined for this generic.
#'
#' @return An `S4` object for which the class corresponds to the type of
#'   the build configuration object passed to this method.
#'
#' @md
#' @export
setGeneric('metaConstruct', function(mapper, ...) standardGeneric('metaConstruct'))

#' @rdname metaConstruct
#' @title metaConstruct
#'
#' @param mapper An `LongTableDataMapper` object abstracting arguments to an
#'  the `LongTable` constructor.
#'
#' @return A `LongTable` object, as specified in the mapper.
#'
#' @examples
#' data(exampleDataMapper)
#' rowDataMap(exampleDataMapper) <- list(c('treatmentid'), c())
#' colDataMap(exampleDataMapper) <- list(c('sampleid'), c())
#' assayMap(exampleDataMapper) <- list(sensitivity=list(c("treatmentid", "sampleid"), c('viability')))
#' metadataMap(exampleDataMapper) <- list(experiment_metadata=c('metadata'))
#' longTable <- metaConstruct(exampleDataMapper)
#' longTable
#'
#' @md
#' @importFrom data.table key
#' @export
setMethod('metaConstruct', signature(mapper='LongTableDataMapper'),
        function(mapper) {
    .metaConstruct(mapper)
})

#' @rdname metaConstruct
#' @title metaConstruct
#'
#' @param mapper An `TREDataMapper` object abstracting arguments to an
#'  the `TreatmentResponseExperiment` constructor.
#'
#' @return A `TreatmentResponseExperiment` object, as specified in the mapper.
#'
#' @examples
#' data(exampleDataMapper)
#' exampleDataMapper <- as(exampleDataMapper, "TREDataMapper")
#' rowDataMap(exampleDataMapper) <- list(c('treatmentid'), c())
#' colDataMap(exampleDataMapper) <- list(c('sampleid'), c())
#' assayMap(exampleDataMapper) <- list(sensitivity=list(c("treatmentid", "sampleid"), c('viability')))
#' metadataMap(exampleDataMapper) <- list(experiment_metadata=c('metadata'))
#' tre <- metaConstruct(exampleDataMapper)
#' tre
#'
#' @md
#' @importFrom data.table key
#' @export
setMethod('metaConstruct', signature(mapper='TREDataMapper'),
        function(mapper) {
    .metaConstruct(mapper, constructor=CoreGx::TreatmentResponseExperiment)
})

#' @keywords internal
#' @noRd
.metaConstruct <- function(mapper, constructor=CoreGx::LongTable) {
    funContext <- paste0('[', .S4MethodContext('metaConstruct', class(mapper)[1]))

    # subset the rawdata slot to build out each component of LongTable
    rowDataDT <- rowData(mapper)
    rowIDs <- rowDataMap(mapper)[[1]]
    rid_names <- names(rowIDs)
    has_rid_names <- !is.null(rid_names) & rid_names != ""
    rowIDs[has_rid_names] <- rid_names[has_rid_names]
    colDataDT <- colData(mapper)
    colIDs <- colDataMap(mapper)[[1]]
    cid_names <- names(colIDs)
    has_cid_names <- !is.null(cid_names) & cid_names != ""
    colIDs[has_cid_names] <- cid_names[has_cid_names]
    assayDtL <- assays(mapper)
    assayIDs <- lapply(assayMap(mapper), `[[`, i=1)
    for (i in seq_along(assayIDs)) {
        aid_names <- names(assayIDs[[i]])
        notEmptyNames <- !is.null(aid_names) & aid_names != ""
        assayIDs[[i]][notEmptyNames] <- aid_names[notEmptyNames]
    }

    # subset the metadata columns out of raw data and add any additional metadata
    metadataL <- lapply(metadataMap(mapper),
        function(j, x) as.list(unique(x[, j, with=FALSE])), x=rawdata(mapper))
    metadataL <- c(metadataL, metadata(mapper))

    ## FIXME:: Handle TREDataMapper class after updating constructor
    object <- constructor(
        rowData=rowDataDT, rowIDs=rowIDs,
        colData=colDataDT, colIDs=colIDs,
        assays=assayDtL, assayIDs=assayIDs,
        metadata=metadataL)

    return(object)
}

# ==== LongTable

#' Get the dimensions of a `LongTable` object.
#'
#' @examples
#' dim(merckLongTable)
#'
#' @describeIn LongTable Get the number of row annotations by the number of
#'   column annotations from a LongTable object. Please note that row x columns
#'   does not necessarily equal the number of rows in an assay, since it is
#'   not required for each assay to have every row or column present.
#'
#' @examples
#' dim(merckLongTable)
#'
#' @param x A `LongTable` object to retrieve dimensions for.
#'
#' @return `numeric` Vector of object dimensions.
#'
#' @importMethodsFrom SummarizedExperiment dim
#' @export
setMethod('dim', signature(x='LongTable'), function(x) {
    return(c(nrow(rowData(x)), nrow(colData(x))))
})

## - http://en.wikipedia.org/wiki/Pearson_product-moment_correlation_coefficient#Calculating_a_weighted_correlation
## http://www.mathworks.com/matlabcentral/fileexchange/20846-weighted-correlation-matrix F. Pozzi, T. Di Matteo, T. Aste, 'Exponential
## smoothing weighted correlations', The European Physical Journal B, Vol. 85, No 6, 2012. DOI: 10.1140/epjb/e2012-20697-x TODO:: Give
## this function a more descriptive name
#' GWC Score
#' 
#' Calculate the gwc score between two vectors, using either a weighted spearman 
#'  or pearson correlation
#'
#' @examples
#' data(clevelandSmall_cSet)
#' x <- molecularProfiles(clevelandSmall_cSet,'rna')[,1]
#' y <- molecularProfiles(clevelandSmall_cSet,'rna')[,2]
#' x_p <- rep(0.05, times=length(x))
#' y_p <- rep(0.05, times=length(y))
#' names(x_p) <- names(x)
#' names(y_p) <- names(y)
#' gwc(x,x_p,y,y_p, nperm=100)
#'
#'@param x1 \code{numeric} vector of effect sizes (e.g., fold change or t statitsics) for the first experiment
#'@param p1 \code{numeric} vector of p-values for each corresponding effect size for the first experiment
#'@param x2 \code{numeric} effect size (e.g., fold change or t statitsics) for the second experiment
#'@param p2 \code{numeric} vector of p-values for each corresponding effect size for the second experiment
#'@param method.cor \code{character} string identifying if a \code{pearson} or
#'\code{spearman} correlation should be used
#'@param nperm \code{numeric} how many permutations should be done to determine
#'@param truncate.p \code{numeric} Truncation value for extremely low p-values
#'@param ... Other passed down to internal functions
#'
#'@return \code{numeric} a vector of two values, the correlation and associated p-value.
#'@export
gwc <- function(x1, p1, x2, p2, method.cor = c("pearson", "spearman"), nperm = 10000, truncate.p = 1e-16, ...) {
    method.cor <- match.arg(method.cor)
    ## intersection between x and y
    ii <- .intersectList(names(x1), names(p1), names(x2), names(p2))
    if (length(ii) < 10) {
        stop("Less than 10 probes/genes in common between x and y")
    }
    x1 <- x1[ii]
    p1 <- p1[ii]
    x2 <- x2[ii]
    p2 <- p2[ii]
    ## truncate extremely low p-values
    p1[!is.na(p1) & p1 < truncate.p] <- truncate.p
    p2[!is.na(p2) & p2 < truncate.p] <- truncate.p
    ## scaled weights
    p1 <- -log10(p1)
    p1 <- p1/sum(p1, na.rm = TRUE)
    p2 <- -log10(p2)
    p2 <- p2/sum(p2, na.rm = TRUE)
    w <- p1 + p2
    ## compute genome-wide connectivity score
    res <- .corWeighted(x = x1, y = x2, w = w, method = method.cor, nperm = nperm, ...)
    return(res)
}

#### Compute a weighted correlation coefficient inspired from package boot - TODO:: Write function documentation?
#' @importFrom stats cov.wt complete.cases
.corWeighted <- function(x, y, w, method = c("pearson", "spearman"), alternative = c("two.sided", "greater", "less"), nperm = 0, nthread = 1, 
    na.rm = FALSE) {
    
    ###################### 
    wcor <- function(d, w, na.rm = TRUE) {
        CovM <- cov.wt(d, wt = w)[["cov"]]
        res <- CovM[1, 2]/sqrt(CovM[1, 1] * CovM[2, 2])
        return(res)
    }
    
    ###################### 
    
    if (missing(w)) {
        w <- rep(1, length(x))/length(x)
    }
    if (length(x) != length(y) || length(x) != length(w)) {
        stop("x, y, and w must have the same length")
    }
    method <- match.arg(method)
    if (method == "spearman") {
        x <- rank(x)
        y <- rank(y)
    }
    alternative <- match.arg(alternative)
    
    res <- c(rho = NA, p = NA)
    
    ## remove missing values
    ccix <- complete.cases(x, y, w)
    if (!all(ccix) && !na.rm) {
        warning("Missing values are present")
    }
    if (sum(ccix) < 3) {
        return(res)
    }
    x <- x[ccix]
    y <- y[ccix]
    w <- w[ccix]
    
    wc <- wcor(d = cbind(x, y), w = w)
    res["rho"] <- wc
    if (nperm > 1) {
        splitix <- parallel::splitIndices(nx = nperm, ncl = nthread)
        if (!is.list(splitix)) {
            splitix <- list(splitix)
        }
        splitix <- splitix[vapply(splitix, length, FUN.VALUE = numeric(1)) > 0]
        mcres <- BiocParallel::bplapply(splitix, function(x, xx, yy, ww) {
            pres <- vapply(x, function(x, xx, yy, ww) {
                ## permute the data and the weights
                d2 <- cbind(xx[sample(seq_len(length(xx)))], yy[sample(seq_len(length(yy)))])
                w2 <- ww[sample(seq_len(length(ww)))]
                return(wcor(d = d2, w = w2))
            }, xx = xx, yy = yy, ww = ww, FUN.VALUE = double(1))
            return(pres)
        }, xx = x, yy = y, ww = w)
        perms <- do.call(c, mcres)
        
        switch(alternative, two.sided = {
            if (res["rho"] < 0) {
                p <- sum(perms <= res, na.rm = TRUE)
            } else {
                p <- sum(perms >= res, na.rm = TRUE)
            }
            if (p == 0) {
                p <- 1/(nperm + 1)
            } else {
                p <- p/nperm
            }
            p <- p * 2
        }, greater = {
            p <- sum(perms >= res, na.rm = TRUE)
            if (p == 0) {
                p <- 1/(nperm + 1)
            } else {
                p <- p/nperm
            }
        }, less = {
            p <- sum(perms <= res, na.rm = TRUE)
            if (p == 0) {
                p <- 1/(nperm + 1)
            } else {
                p <- p/nperm
            }
        })
        res["p"] <- p
    }
    return(res)
}

#' @title lapply
#' lapply method for `MultiAssayExperiment`
#'
#' @param X A `MultiAssayExperiment` object.
#' @param FUN A function to be applied to each `SummarizedExperiment` in a
#' in `X`.
#' @param ... Fall through parameters to `FUN`
#'
#' @return A `MultiAssayExperiment` object, modifed such that
#' `experiments(X) <- endoapply(experiments(X), FUN, ...)`.s
#'
#' @importMethodsFrom BiocGenerics lapply
#' @importFrom S4Vectors endoapply
#' @importFrom MultiAssayExperiment experiments experiments<-
#' @exportMethod lapply
setMethod('lapply', signature(X='MultiAssayExperiment'),
        function(X, FUN, ...) {
    experiments(X) <- endoapply(experiments(X), FUN, ...)
    return(X)
})

library(testthat)
library(CoreGx)
library(data.table)

data(nci_TRE_small)
tre <- nci_TRE_small

## tre = treatmentResponseExperiment
## ntre = New treatmentResponseExperiment

# see https://github.com/bhklab/CoreGx/wiki/CoreGx-Design-Documentation for
# explanation
test_that("`rowData,LongTable-method` orders data correctly", {
    ## TODO::
})

# == @rowData slot

testthat::test_that("`rowData<-` rowData must be updated with data.table or data.frame", {
    ntre <- copy(tre)
    testthat::expect_error({ rowData(ntre) <- NULL }, ## rowData slot
        regexp = ".*Please pass a data.frame or data.table to update.the rowData slot.*"
    )
})


testthat::test_that("`rowData<-` prevents intentionally breaking referential integrity", {
    ntre <- copy(tre)
    rowData_bad <- rowData(ntre)
    rowData_bad <- rbind(rowData_bad, rowData_bad[.N, ])
    testthat::expect_warning({ rowData(ntre) <- rowData_bad },
        regexp = ".*The ID columns are duplicated for rows [0-9]+! These rows will be dropped before assignment."
    )
})

testthat::test_that("`colData<-` prevents intentionally breaking referential integrity", {
    ntre <- copy(tre)
    colData_bad <- colData(ntre)
    colData_bad <- rbind(colData_bad, colData_bad[.N, ])
    testthat::expect_warning({ colData(ntre) <- colData_bad },
        regexp =  ".*The ID columns are duplicated for rows [0-9]+! These rows will be dropped before assignment."
    )
})

# This warning doesn't trigger if we remove another ID column.
# Instead, an error like in the NCI-ALMANAC script will occur. (Too many duplicate rows)

## FIXME:: We should probably just throw an error for these cases! This test
##   doesn't work for the full NCI_TRE object due to cartesian join
# testthat::test_that("`rowData<-` ensures necessary row ID columns present in the replacement rowData", {
#     ntre <- copy(tre)
#     rowData_missingID <- rowData(ntre)
#     rowData_missingID[, (rowIDs(ntre)[4]) := NULL] # remove one ID column
#     testthat::expect_warning({ rowData(ntre) <- rowData_missingID },
#         regexp = ".*The function will attempt to join with existing rowIDs, but this may fail!.*"
#     )
# })

# == @colData slot

testthat::test_that("`colData<-` colData must be updated with data.table or data.frame", {
    ntre <- copy(tre)
    testthat::expect_error({ colData(ntre) <- NULL }, ## colData slot
        regexp = ".*Please pass a data\\.frame or data\\.table.*"
    )
})

# == @assay slot

testthat::test_that("`assay` invalid assay name and index", {
    testthat::expect_error({ assay(tre, c(1, 2)) },
        regexp = ".*Please specifying a single string assay name or integer.*"
    )
    testthat::expect_error({ assay(tre, paste(assayNames(tre), collapse = '')) },
        regexp = ".*There is no assay.*"
    )
})

testthat::test_that("`assay<-,LongTable-method` prevents invalid assay slot assignment", {
    ntre <- copy(tre)
    testthat::expect_error({
        assay(ntre, i = "sensitivity", withDimnames = FALSE) <- c(1, 2, 3)
    },
    regexp = ".*Only a\ndata.frame or data.table can be assiged to the assay slot!.*"
    )
})

testthat::test_that("`assay,LongTable-method` and `assays,LongTable-method` return equivalent data", {
    assay_list <- lapply(seq_along(assayNames(tre)), FUN=assay,
        x=tre, withDimnames=TRUE, summarize=FALSE)
    assays_ <- assays(tre)
    for (i in seq_along(assay_list)) {
        testthat::expect_true(all.equal(assay_list[[i]], assays_[[i]]))
    }
})

testthat::test_that("`assay<-,LongTable-method` assignment does not corrupt data relationships", {
    ntre <- copy(tre)
    for (nm in assayNames(tre)) {
        ntre[[nm]] <- ntre[[nm]]
        testthat::expect_true(all.equal(ntre[[nm]], tre[[nm]]))
        testthat::expect_true(all.equal(assays(ntre, raw=TRUE)[[nm]], assays(tre, raw=TRUE)[[nm]]))
    }
    testthat::expect_true(all.equal(getIntern(ntre)$assayIndex, getIntern(tre)$assayIndex))
})

testthat::test_that("`assay<-,LongTable-method` allows non-id column updates", {
    ntre <- copy(tre)
    assay_ <- ntre[["sensitivity"]]
    assay_[, viability := rnorm(.N)]
    ntre[["sensitivity"]] <- assay_
    testthat::expect_true(all.equal(ntre[["sensitivity"]], assay_))
    testthat::expect_false(isTRUE(all.equal(ntre[["sensitivity"]], tre[["sensitivity"]])))
})

testthat::test_that("`assay<-LongTable-method` prevents id column updates", {
    ntre <- copy(tre)
    assay_ <- ntre[["sensitivity"]]
#    assay_[, treatment1dose := rnorm(.N)]
#    testthat::expect_error({ ntre[["sensitivity"]] <- assay_ },
#        regexp=".*Identifier columns cannot be modified via assay assignment!.*"
#    )
    testthat::expect_true(all.equal(ntre$sensitivity, tre$sensitivity))
})

testthat::test_that("`assay<-LongTable-method` allows simple summary assignments", {
    ntre <- copy(tre)
    sens <- ntre$sensitivity
    sens_sum <- sens[,
        .(
            mean_treatment1dose=mean(treatment1dose, na.rm=TRUE),
            mean_treatment2dose=mean(treatment2dose, na.rm=TRUE),
            mean_viability=mean(viability, na.rm=TRUE)
        ),
        by=.(treatment1id, treatment2id, sampleid)
    ]
    testthat::expect_silent(ntre$sens_sum <- sens_sum)
    # ensure that the returned assay matches the assigned assay when
    #   summarize=TRUE (the default)
    sens_sum_accessed <- ntre$sens_sum[, colnames(sens_sum), with=FALSE]
    setkeyv(sens_sum, key(sens_sum_accessed))
    testthat::expect_true(all.equal(
        sens_sum,
        sens_sum_accessed,
    ))
    # test that summarzie=FALSE attaches all original data
    testthat::expect_true(all.equal(
        rowIDs(tre, data=TRUE),
        unique(assay(ntre, "sens_sum", summarize=FALSE)[, rowIDs(ntre), with=FALSE]),
        check.attributes=FALSE
    ))
    testthat::expect_true(all.equal(
        colIDs(tre, data=TRUE),
        unique(assay(ntre, "sens_sum", summarize=FALSE)[order(sampleid), colIDs(ntre), with=FALSE]),
        check.attributes=FALSE
    ))
})

testthat::test_that("`assay<-,LongTable-method` summary assignment doesn't break referential integrity", {
    ntre <- copy(tre)
    sens <- ntre$sensitivity
    sens_sum <- sens[,
        .(
            mean_treatment1dose=mean(treatment1dose, na.rm=TRUE),
            mean_treatment2dose=mean(treatment2dose, na.rm=TRUE),
            mean_viability=mean(viability, na.rm=TRUE)
        ),
        by=.(treatment1id, treatment2id, sampleid)
    ]
    testthat::expect_silent(ntre$sens_sum <- sens_sum)
    testthat::expect_true(all.equal(rowData(tre), rowData(ntre)))
    testthat::expect_true(all.equal(colData(tre), colData(ntre)))
    non_summary_assays <- setdiff(assayNames(ntre), "sens_sum")
    for (aname in non_summary_assays) {
        testthat::expect_true(all.equal(
            tre[[aname]],
            ntre[[aname]]
        ))
    }
})

testthat::test_that("`assay<-,LongTable-method` prevents modified row ID in new assay from breaking referential integrity", {
    ntre <- copy(tre)
    sens <- ntre$sensitivity
    sens_sum <- sens[,
        .(
            mean_treatment1dose=mean(treatment1dose, na.rm=TRUE),
            mean_treatment2dose=mean(treatment2dose, na.rm=TRUE),
            mean_viability=mean(viability, na.rm=TRUE)
        ),
        by=.(treatment1id, treatment2id, sampleid)
    ]
    set(sens_sum,
        i     = which(sens_sum[["treatment1id"]] == sens_sum[1, treatment1id]),
        j     = "treatment1id",           # rowID to modify
        value = sens_sum[1, sampleid]) # Replace a treatment id with a sample id
    testthat::expect_error({ ntre$sens_sum <- sens_sum },
        regexp = paste(
            ".*One or more rowIDs\\(x\\) columns have been modified.",
            "Identifier columns cannot be modified via assay assignment!.*"
            )
    )
})

testthat::test_that("`assay<-,LongTable-method` prevents modified column ID in new assay from breaking referential integrity", {
    ntre <- copy(tre)
    sens <- ntre$sensitivity
    sens_sum <- sens[,
        .(
            mean_treatment1dose=mean(treatment1dose, na.rm=TRUE),
            mean_treatment2dose=mean(treatment2dose, na.rm=TRUE),
            mean_viability=mean(viability, na.rm=TRUE)
        ),
        by=.(treatment1id, treatment2id, sampleid)
    ]
    set(sens_sum,
        i = which(sens_sum[["sampleid"]] == sens_sum[1, sampleid]),
        j = "sampleid",             # column ID to modify
        value = sens_sum[1, treatment2id]) # Replace a sample ID with a treatment ID
    testthat::expect_error({ ntre$sens_sum <- sens_sum },
        regexp = paste(
            ".*One or more colIDs\\(x\\) column have been modified.",
            "Identifier columns cannot be modified via assay assignment!.*"
            )
    )
})

# == @metadata slot

testthat::test_that("`metadata<-` invalid metadata slot assignment", {
    ntre <- copy(tre)
    testthat::expect_error({ metadata(ntre) <- NULL },
        regexp = ".*The `metadata` slot must be a list!.*"
    )
})

---
title: "The TreatmentResponseExperiment Class"
author:
- name: Petr Smirnov
  affiliation: &pm Princess Margaret Cancer Centre, University Health Network, Toronto Canada
  email: petr.smirnov@uhnresearch.ca
- name: Christopher Eeles
  affiliation: *pm
  email: christopher.eeles@uhnresearch.ca
package: CoreGx
output:
  BiocStyle::html_document
vignette: |
  %\VignetteIndexEntry{The TreatmentResponseExperiment Class}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
library(formatR)
knitr::opts_chunk$set(echo=FALSE)
```

```{r dependencies, include=FALSE}
library(CoreGx)
library(data.table)
```

# Why Do We Need A New Class?

The current implementation for the `@treatmentResponse` slot in a
`PharmacoSet` has some limitations.

Firstly, it does not natively support dose-response experiments with
multiple drugs and/or cancer cell lines. As a result we have not been
able to include this data into a `PharmacoSet` thus far.

Secondly, drug combination data has the potential to scale to high dimensionality.
As a result we need an object that is highly performant to ensure
computations on such data can be completed in a timely manner.

To resolve these issues, we designed and implement the
`TreamtentResponseExperiment` (or `TRE` for short)!

# Design Philosophy

The current use case is supporting drug combinations experiments in
`r BiocStyle::Biocpkg("PharmacoGx")`, but we wanted to create something flexible
enough to fit other use cases. As such, we have used the generic term
'treatment' to refer to any experimental intervention one can conduct on a
set of samples. In context of `r BiocStyle::Biocpkg("PharmacoGx")`, a
treatment represents application of one or more anti-cancer compounds to a
cancer cell-line. The resulting viability for this cell-line
after treatment is the response metric. We hope that the implementation of
our class is general enough to support other use cases. For example, the
`TreatmentResponseExperiment` class is also being adopted for radiation
dose-response experiments in cancer cell-lines in
`r BiocStyle::Biocpkg("RadioGx")` as well as for investigating compound
toxicity in healthy human and rat cell-lines in
`r BiocStyle::Biocpkg("ToxicoGx")`.

Our design takes the aspects of the `SummarizedExperiment` and
`MultiAssayExperiment` classes and implements them using the `data.table`
package, which provides an R API to a rich set of tools for scalable,
high performance data processing implemented in C.

# Anatomy of a TreatmentResponseExperiment

## Class Diagram

```{r class_diagram, fig.wide=TRUE, caption='LongTable Class Diagram'}
knitr::include_graphics('TreatmentResponseExperimentClassDiagram.drawio.png')
```

We have borrowed directly from the `SummarizedExperiment` class
for the `rowData`, `colData`, `metadata` and `assays` slot names.
We also implemented the `SummarizedExperiment` accessor methods for the
`TreatmentResponseExperiment`. Therefore the interface should be familiar to
users of common Bioconductor packages.

## Object Structure and Cardinality

There are, however, some important differences which make this object more
flexible when dealing with high dimensional data.

```{r tre_structure, fig.wide=TRUE, caption='TreatmentResponseExperiment Structure'}
knitr::include_graphics('TreatmentResponseExperimentStructure.drawio.png')
```

Unlike a `SummarizedExperiment`, there are three distinct
subgroups of columns in `rowData` and `colData`.

The first are the `rowKey` and `colKey` which are implemented internally to
map between each assay observation and its associated treatments or samples
(rows or columns); these will not be returned by the accessors by default.
The second are the `rowIDs` and `colIDs`, these hold all of the information
necessary to uniquely identify a row or column and are used to generate the
`rowKey` and `colKey`. Finally, there are the `rowMeta` and `colMeta` columns,
which store any additional data about treatments or samples not required to
uniquely identify a row in either table.

Within the `TreatmentResponseExperiment`, an `assayIndex` is stored in the
`@.intern` slot which maps between unique combinations of `rowKey` and `colKey`
and the experimental observations in each assay. This relationship is maintained
using a separate primary key for each assay, which can map to one or more `rowKey`
and `colKey` combination. For assays containing raw experimental observations,
generally each assay row will map to one and only one combination of `rowKey`
and `colKey`. However, for metrics computed over experimental observations, It
may be desirable to summarized over some of the `rowID` and/or `colID` columns.
In this case, the relationship between the summarized rows and the metadata
stored in the `rowData` and `colData` slots are retained in the `assayIndex`,
allowing

Also worth noting is the cardinality between `rowData` and `colData` for a given
assay within the assays list. As indicated by the lower connection between these
tables and an assay, for each row or column key there may be zero or more rows in
the assay table. Conversely for each row in the assay there may be zero or one key
in `colData` or `rowData`. When combined, the `rowKey` and `colKey` for a given
row in an assay become a composite key which uniquely identify an observation.

# Constructing a TreatmentResponseExperiment

To deal with the complex kinds of experimental designs which can be stored
in a `LongTable`, we have engineered a new object to help document and validate
the way data is mapped from raw data files, as a single large `data.frame` or
`data.table`, to the various slots of a `TreatmentResponseExperiment` object.

## The DataMapper Class

The `DataMapper` is an abstract class, which means in cannot be instatiated.
Its purpose is to provide a description of the concept of a DataMapper and
define a basic interface for any classes inheriting from it. A DataMapper is
simply a way to map columns from some raw data file to the slots of an S4 class.
It is similar to a schema in SQL in that it defines the valid parts of an
object (analogously a SQL table), but differs in that no types are specified or
enforced at this time.

This object is not important for general users, but may be useful for other
developers who want to map from some raw data to some `S4` class. In this case,
any derived data mapper should inherit from the `DataMapper` abstract class.
Only one slot is defined by default, a `list` or `List` in the `@rawdata` slot.
An accessor method, `rawdata(DataMapper)`, is defined to assign and retrieve
the raw data from your mapper object.

## The TREDataMapper Class

The `TREDataMapper` class is the first concrete sub-class of a
`DataMapper`. It is the object which defines how to go from a single
`data.frame` or `data.table` of raw experimental data to a properly formatted
and valid `TreatmentResponseExperiment` object. This is accomplished by defining
various mappings, which let the the user decide which columns from `rawdata`
should go into which slots of the object. Each slot mapping is implemented as a
list of character vectors specifying the column names from `rawdata` to assign
to each slot.

Additionally, a helper method has been included, `guessMapping`, that will
try to determine which columns of a `TreatmentResponseExperiment`s `rawdata`
should be assigned to which slots, and therefore which maps.

To get started making a `TreatmentResponseExperiment` lets have a look at some
rawdata which is a subset of the data from Oneil *et al.*, 2016. The full set
of rawdata is available for exploration and download from
[SynergxDB.ca](https://www.synergxdb.ca/), a free and open source web-app and
database of publicly available drug combination sensitivity experiments which we
created and released (Seo *et al.*, 2019).

The data was generated as part of the commercial activities of the
pharmaceutical company Merck, and is thus named according.

```{r head_data, echo=TRUE}
filePath <- system.file('extdata', 'merckLongTable.csv', package='CoreGx',
  mustWork=TRUE)
merckDT <- fread(filePath, na.strings=c('NULL'))
colnames(merckDT)
```

```{r sample_data, fig.width=80}
knitr::kable(head(merckDT)[, 1:5])
```

```{r sample_data2, fig.width=80}
knitr::kable(head(merckDT)[, 5:ncol(merckDT)])
```
We can see that all the data related to the treatment response experiment is
contained within this table.

To get an idea of where in a `TreatmentResponseExperiment` this data should go, lets come up
with some guesses for mappings.

```{r guess_mapping, echo=TRUE}
# Our guesses of how we may identify rows, columns and assays
groups <- list(
  justDrugs=c('drug1id', 'drug2id'),
  drugsAndDoses=c('drug1id', 'drug2id', 'drug1dose', 'drug2dose'),
  justCells=c('cellid'),
  cellsAndBatches=c('cellid', 'batchid'),
  assays1=c('drug1id', 'drug2id', 'cellid'),
  assays2=c('drug1id', 'drug2id', 'drug1dose', 'drug2dose', 'cellid', 'batchid')
)

# Decide if we want to subset out mapped columns after each group
subsets <- c(FALSE, TRUE, FALSE, TRUE, FALSE, TRUE)

# First we put our data in the `TRE`
TREdataMapper <- TREDataMapper(rawdata=merckDT)

# Then we can test our hypotheses, subset=FALSE means we don't remove mapped
#   columns after each group is mapped
guess <- guessMapping(TREdataMapper, groups=groups, subset=subsets)
guess
```

Since we want our `TreatmentResponseExperiment` to have drugs as rows and
samples as columns, we see that both `justDrug` and `drugsAndDoses` yield the
same result. So we do not yet prefer one over the other. Looking at `justCells`
and `cellsAndBatches`, we see one column maps to each of them and therefore still
have no preference. For `assay1` however, we see that no columns mapped, while
`assay2` maps many of raw data columns.

Since assays will be subset based on the `rowKey` and `colKey`, we know that
the rowIDs must be `drugsAndDose` and the the colIDs must be `cellsAndBatches`.
Therefore, to uniquely identify an observation in any given assay we need
all of these columns. We can use this information to assign maps to our
`TREDataMapper`.

```{r mappings_to_datamapper, echo=TRUE}
rowDataMap(TREdataMapper) <- guess$drugsAndDose
colDataMap(TREdataMapper) <- guess$cellsAndBatches
```

Looking at our mapped columns for `assay2`, we must decide if we want these
to go into more than one assay. If we do, we should name each item of our
`assayMap` for the `TREDataMapper` and specify it in a list of
`character` vectors, one for each assay. Since viability is the raw experimental
measurement and the final two columns are summaries of it, we will assign them
to two assays:sensitivity and profiles.

```{r split_assays, echo=TRUE}
assays <- list(
  sensitivity=list(
    guess$assays2[[1]],
    guess$assays2[[2]][seq_len(4)]
  ),
  profiles=list(
    guess$assays2[[1]],
    guess$assays2[[2]][c(5, 6)]
  )
)
assays
```

```{r assign_assaymap, echo=TRUE}
assayMap(TREdataMapper) <- assays
```

## metaConstruct Method

The metaConstruct method accepts a `DataMapper` object as its only argument,
and uses the information in that `DataMapper` to preprocess all `rawdata` and
map them to the appropriate slots of an `S4` object. In our case, we are mapping
from the merckDT `data.table` to a `TreatmentResponseExperiment`.

At minimum, a `TREDataMapper` must specify the `rowDataMap`, `colDataMap`,
and `assayMap`. Additional maps are available, see `?TREDataMapper-class`
and `?TREDataMapper-accessors` for more details.

After configuration, creating the object is very straight forward.

```{r echo=TRUE}
tre <- metaConstruct(TREdataMapper)
```

# TreatmentResponseExperiment Object

As mentioned previously, a `TreatmentResponseExperiment` has both list and
table like behaviours. For table like operations, a given
`TreatmentResponseExperiment` can be thought of as a `rowKey` by
`colKey` rectangular object.

To support `data.frame` like sub-setting for this object, the constructor makes
pseudo row and column names, which are the ID columns for each row of
`rowData` or `colData` pasted together with a ':'. The ordering of these
columns is preserved in the pseudo-dim names, so be sure to arrange them
as desirged before creating the `TreatmentResponseExperiment`.

## Row and Column Names

```{r rownames, echo=TRUE}
head(rownames(tre))
```

We see that the rownames for the Merck `TreatmentResponseExperiment` are the cell-line name
pasted to the batch id.

```{r colnames, echo=TRUE}
head(colnames(tre))
```

For the column names, a similar pattern is followed by combining the colID
columns in the form 'drug1:drug2:drug1dose:drug2dose'.

## `data.frame` Subsetting

We can subset a `TreatmentResponseExperiment` using the same row and column name syntax as
with a `data.frame` or `matrix`.

```{r subset_dataframe_character, echo=TRUE}
row <- rownames(tre)[1]
columns <- colnames(tre)[1:2]
tre[row, columns]
```

### Regex Queries

However, unlike a `data.frame` or `matrix` this subsetting also accepts partial
row and column names as well as regex queries.

```{r rowdata_coldata, echo=TRUE}
head(rowData(tre), 3)
head(colData(tre), 3)
```

For example, if we want to get all instance where '5-FU' is the drug:

```{r simple_regex, echo=TRUE}
tre['5-FU', ]
```

This has matched all colnames where 5-FU was in either drug1 or drug2. If we
only want to match drug1, we have several options:

```{r column_specific_regex, echo=TRUE}
all.equal(tre['5-FU:*:*:*', ], tre['^5-FU',  ])
```

As a technicaly note, '\*' is replaced with '.\*' internally for regex queries.
This was implemented to mimic the linux shell style patten matching that most
command-line users are familiar with.

## `data.table` Subsetting

In addition to regex queries, a `TreatmentResponseExperiment` object supports
arbitrarily complex subset queries using the `data.table` API. To access this
API, you will need to use the `.` function, which allows you to pass raw R
expressions to be evaluated inside the `i` and `j` arguments for
`dataTable[i, j]`.

For example if we want to subset to rows where the cell line is VCAP and columns
where drug1 is Temozolomide and drug2 is either Lapatinib or Bortezomib:

```{r , echo=TRUE}
tre[
    # row query
    .(drug1id == 'Temozolomide' & drug2id %in% c('Lapatinib', 'Bortezomib')),
    .(cellid == 'CAOV3') # column query
]
```

We can also invert matches or subset on other columns in `rowData` or `colData`:

```{r echo=TRUE}
subTRE <- tre[
  .(drug1id == 'Temozolomide' & drug2id != 'Lapatinib'),
  .(batchid != 2)
]
```

To show that this works as expected:

```{r echo=TRUE}
print(paste0('drug2id: ', paste0(unique(rowData(subTRE)$drug2id),
    collapse=', ')))
print(paste0('batchid: ', paste0(unique(colData(subTRE)$batchid),
    collapse=', ')))
```

# Accessor Methods

## rowData

```{r echo=TRUE}
head(rowData(tre), 3)
```

```{r echo=TRUE}
head(rowData(tre, key=TRUE), 3)
```

## colData

```{r echo=TRUE}
head(colData(tre), 3)
```

```{r echo=TRUE}
head(colData(tre, key=TRUE), 3)
```

## assays

```{r echo=TRUE}
assays <- assays(tre)
assays[[1]]
```


```{r echo=TRUE}
assays[[2]]
```

```{r echo=TRUE}
assays <- assays(tre, withDimnames=TRUE)
colnames(assays[[1]])
```

```{r echo=TRUE}
assays <- assays(tre, withDimnames=TRUE, metadata=TRUE)
colnames(assays[[2]])
```


```{r echo=TRUE}
assayNames(tre)
```

Using these names we can access specific assays within a
`TreatmentResponseExperiment`.

## assay

```{r echo=TRUE}
colnames(assay(tre, 'sensitivity'))
assay(tre, 'sensitivity')
```

```{r echo=TRUE}
colnames(assay(tre, 'sensitivity', withDimnames=TRUE))
assay(tre, 'sensitivity', withDimnames=TRUE)
```

# References

1. O'Neil J, Benita Y, Feldman I, Chenard M, Roberts B, Liu Y, Li J, Kral A,
Lejnine S, Loboda A, Arthur W, Cristescu R, Haines BB, Winter C, Zhang T,
Bloecher A, Shumway SD. An Unbiased Oncology Compound Screen to Identify Novel
Combination Strategies. Mol Cancer Ther. 2016 Jun;15(6):1155-62.
doi: 10.1158/1535-7163.MCT-15-0843. Epub 2016 Mar 16. PMID: 26983881.

2. Heewon Seo, Denis Tkachuk, Chantal Ho, Anthony Mammoliti, Aria Rezaie,
Seyed Ali Madani Tonekaboni, Benjamin Haibe-Kains, SYNERGxDB: an integrative
pharmacogenomic portal to identify synergistic drug combinations for precision
oncology, Nucleic Acids Research, Volume 48, Issue W1, 02 July 2020, Pages
W494–W501, https://doi.org/10.1093/nar/gkaa421

# sessionInfo

```{r sessionInfo}
sessionInfo()
```

---
title: "CoreGx: Class and Function Abstractions for PharmacoGx, RadioGx and ToxicoGx"

author:
- name: Petr Smirnov
- name: Ian Smith
- name: Christopher Eeles
- name: Benjamin Haibe-Kains

output: BiocStyle::html_document

vignette: |
  %\VignetteEngine{knitr::knitr}
  %\VignetteIndexEntry{CoreGx: Class and Function Abstractions}
---
# CoreGx

```{r setup, include=FALSE}
library(knitr)
knitr::opts_chunk$set(echo = TRUE)
```

This package provides a foundation for the PharmacoGx, RadioGx and ToxicoGx
packages. It is not intended for standalone use, only as a dependency for the
aforementioned software. Its existence allows abstracting generic definitions,
method definitions and class structures common to all three of the Gx suite
packages.

## Importing and Using CoreGx

Load the pacakge:

```{r coregx_load, hide=TRUE, message=FALSE}
library(CoreGx)
library(Biobase)
library(SummarizedExperiment)
```

## The CoreSet Class

The `CoreSet` class is intended as a general purpose data structure for storing
multiomic treatment response data. Extensions of this class have been
customized for their respective fields of study. For example, the `PharmacoSet`
class inherits from the `CoreSet` and is specialized for storing and analyzing
drug sensitivity and perturbation experiments on cancer cell lines together with
associated multiomic data for each treated sample. The RadioSet class serves
a role similar to the PharmacoSet with radiation instead of drug treatments.
Finally, the ToxicoSet class is used to store toxicity data for healthy
human and rat hepatocytes along with the associated multiomic profile for each
treatment.

```{r CoreSet_class}
getClass("CoreSet")
```

The `annotation` slot holds the CoreSet name, the original constructor call, and
a range of metadata about the R session in which the constructor was called.
This allows easy comparison of `CoreSet` versions across time and ensures the
code used to generate a `CoreSet` is well-documented and reproducible.

The `molecularProfiles` slot contains a list of `SummarizedExperiment` objects
for each multi-omic molecular datatype available for a given experiment. Within
the `SummarizedExperiments` are feature and sample annotations for each data
type. We are currently in the process of adopting the `MultiAssayExperiment`
class instead of a `list` for storing molecular profile `SummarizedExperiment`s.
However, the `list` version of the `molecularProfiles` slot is still supported
for backwards compatability.

The `sample` slot contains a `data.frame` with annotations for samples used in
the molecularProfiles or sensitivity slot. It should at minimum have the
standardized column 'sampleid', with a unique identifier for each sample in the
`CoreSet`.

The `treatment` slot contains a `data.frame` of metadata for treatments applied
to samples in the `molecularProfiles` or `treatmentResponse` slot. It should at
minimum have the standarized column 'treatmentid', containing a unique
identifier for each treatment in the `CoreSet`.

The `datasetType` slot contains a character vector indicating the experiment
type the `CoreSet` contains. This slot is soft deprecated and may be removed
in future updates.

The `treatmentResponse` slot contains a list of raw, curated and meta data for
treatment-response experiments. We are currently in the process of adopting
our new S4-class, the `TreamtentResponseExperiment` to store treatment-response
data within a `CoreSet` and inheriting classes. However, the old `list` format
for sensitivity experiments will continue to be support for backwards
compatability.

The `perturbation` slot contains a list of raw, curated and meta data for
perturbation experiments. This slot is soft-deprecated and may be removed
in the future. The reason is that treatment perturbation experiments can be
efficiently stored in the `colData` slot of their respective
`SummarizedExperiment` objects and thus no longer require their own space within
a `CoreSet`.

The `curation` slot contains a list of manually curated identifiers
such as standardized cell-line, tissue and treatment names. Inclusion of such
identifiers ensures a consistent nomenclature is used across all datasets
curated into the classes inheriting from the `CoreSet`, enabling results from
such datasets to be easily compared to validate results from published studies or
combine them for use in larger meta-analyses. The slot contains a list of
`data.frame`s, one for each entity, and should at minimum include a mapping from
curated identifiers used throughout the object to those used in the original
dataset publication.

The `CoreSet` class provides a set of standardized accessor methods which
simplify curation, annotation, and retrieval of data associated with a specfic
treatment response experiment. All accessors are implemented as generics to
allow new methods to be defined on classes inheriting from the `CoreSet`.

```{r CoreSet_accessors}
methods(class="CoreSet")
```

We have provided a sample `CoreSet` (cSet) in this package. In the below code
we load the example cSet and demonstrate a few of the accessor methods.

```{r the_cSet_object}
data(clevelandSmall_cSet)
clevelandSmall_cSet
```

Access a specific molecular profiles:

```{r cSet_accessor_demo}
mProf <- molecularProfiles(clevelandSmall_cSet, "rna")
mProf[seq_len(5), seq_len(5)]
```

Access cell-line metadata:

```{r cSet_accessor_demo1}
cInfo <- sampleInfo(clevelandSmall_cSet)
cInfo[seq_len(5), seq_len(5)]
```

Access treatment-response data:

```{r cSet_accessor_demo2}
sensProf <- sensitivityProfiles(clevelandSmall_cSet)
sensProf[seq_len(5), seq_len(5)]
```

For more information about the accessor methods available for the `CoreSet`
class please see the `class?CoreSet` help page.

## Extending the CoreSet Class

Given that the CoreSet class is intended for extension, we will show some
examples of how to define a new class based on it and implement new methods
for the generics provided for the CoreSet class.

Here we will define a new class, the `DemoSet`, with an additional slot, the
`demoSlot`. We will then view the available methods for this class as well as
define new S4 methods on it.

```{r defining_a_new_class}
DemoSet <- setClass("DemoSet",
                    representation(demoSlot="character"),
                    contains="CoreSet")
getClass("DemoSet")
```

Here we can see the class extending `CoreSet` has all of the same slots as the
original `CoreSet`, plus the new slot we defined: `demoSlot`.

We can see which methods are available for this new class.

```{r inherit_methods}
methods(class="DemoSet")
```

We see that all the accessors defined for the `CoreSet` are also defined for the
inheriting `DemoSet`. These methods all assume the inherit slots have the same
structure as the `CoreSet`. If this is not true, for example, if molecularProfiles
holds `ExpressionSet`s instead of `SummarizedExperiment`s, we can redefine
existing methods as follows:

```{r defining_methods_for_a_new_class}
clevelandSmall_dSet <- DemoSet(clevelandSmall_cSet)
class(clevelandSmall_dSet@molecularProfiles[['rna']])

expressionSets <- lapply(molecularProfilesSlot(clevelandSmall_dSet), FUN=as,
  'ExpressionSet')
molecularProfilesSlot(clevelandSmall_dSet) <- expressionSets

# Now this will error
tryCatch({molecularProfiles(clevelandSmall_dSet, 'rna')},
         error=function(e)
             print(paste("Error: ", e$message)))
```

Since we changed the data in the `molecularProfiles` slot of the `DemoSet`,
the original method from `CoreGx` no longer works. Thus we get an error when
trying to access that slot. To fix this we will need to set a new S4 method
for the molecularProfiles generic function defined in `CoreGx`.

```{r redefine_the_method}
setMethod(molecularProfiles,
          signature("DemoSet"),
          function(object, mDataType) {
            pData(object@molecularProfiles[[mDataType]])
          })
```

This new method is now called whenever we use the `molecularProfiles` method
on a `DemoSet`. Since the new method uses `ExpressionSet` accessor methods
instead of `SummarizedExperiment` accessor methods, we now expect to be able
to access the data in our modified slot.

```{r testing_new_method}
# Now we test our new method
mProf <- molecularProfiles(clevelandSmall_dSet, 'rna')
head(mProf)[seq_len(5), seq_len(5)]
```

We can see our new method works! In order to finish updating the methods
for our new class, we would have to redefine all the methods which access the
modified slot.

However, additional work needs to be done to define accessors for the new
`demoSlot`. Since no generics are available in CoreGx to access this slot,
we need to first define a generic, then implement methods which dispatch on
the 'DemoSet' class to retrieve data in the slot.

```{r defining_setter_methods}
# Define generic for setter method
setGeneric('demoSlot<-', function(object, value) standardGeneric('demoSlot<-'))

# Define a setter method
setReplaceMethod('demoSlot',
                 signature(object='DemoSet', value="character"),
                 function(object, value) {
                   object@demoSlot <- value
                   return(object)
                 })

# Lets add something to our demoSlot
demoSlot(clevelandSmall_dSet) <- c("This", "is", "the", "demoSlot")
```

```{r defining_getter_methods}
# Define generic for getter method
setGeneric('demoSlot', function(object, ...) standardGeneric("demoSlot"))

# Define a getter method
setMethod("demoSlot",
          signature("DemoSet"),
          function(object) {
            paste(object@demoSlot, collapse=" ")
          })

# Test our getter method
demoSlot(clevelandSmall_dSet)
```

Now you should have all the knowledge you need to extend the CoreSet class
for use in other treatment-response experiments!

For more information about this package and the possibility of collaborating on
its extension please contact benjamin.haibe.kains@utoronto.ca.

# sessionInfo

```{r sessionInfo}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deprecated.R
\name{buildLongTable,character-method}
\alias{buildLongTable,character-method}
\title{LongTable build method from character}
\usage{
\S4method{buildLongTable}{character}(from, rowDataCols, colDataCols, assayCols)
}
\arguments{
\item{from}{\code{character} Path to the .csv file containing the data and
metadata from which to build the \code{LongTable}.}

\item{rowDataCols}{\code{list} List with two \code{character} vectors, the first
specifying one or more columns to be used as cell identifiers (e.g.,
cell-line name columns) and the second containing any additional metadata
columns related to the cell identifiers.}

\item{colDataCols}{\code{list} List with two \code{character} vectors, the first
specifying one or more columns to be used as column identifiers (e.g.,
drug name columns) and the second containing any additional metadata
columns related to the column identifiers.}

\item{assayCols}{\code{list} A named list of character vectors specifying how to
parse assay columns into a list of \code{data.table}s. Each list data.table
will be named for the name of corresponding list item and contain the columns
specified in the character vector of column names in each list item.}
}
\value{
A \code{LongTable} object containing one or more assays, indexed by
rowID and colID.
}
\description{
LongTable Create a LongTable object from a single .csv file
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/allGenerics.R
\name{assayIndex}
\alias{assayIndex}
\title{Retrieve and assayIndex}
\usage{
assayIndex(x, ...)
}
\arguments{
\item{x}{An \code{S4} object.}

\item{...}{\code{pairlist} Allow definition of new parameters for
implementations of this generic.}
}
\value{
An object representing the "assayIndex" of an \code{S4} object.
}
\description{
Retrieve and assayIndex
}
\examples{
print("Generics shouldn't need examples?")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CoreSet-utils.R
\name{CoreSet-utils}
\alias{CoreSet-utils}
\alias{subsetBySample,CoreSet-method}
\alias{subsetBySample}
\alias{subsetByTreatment,CoreSet-method}
\alias{subsetByTreatment}
\alias{subsetByFeature,CoreSet-method}
\alias{subsetByFeature}
\title{Utility methods for a \code{CoreSet} object.}
\usage{
\S4method{subsetBySample}{CoreSet}(x, samples)

\S4method{subsetByTreatment}{CoreSet}(x, treatments)

\S4method{subsetByFeature}{CoreSet}(x, features, mDataTypes)
}
\arguments{
\item{x}{A \code{CoreSet} object.}

\item{samples}{\code{character()} vector of sample names. Must be valid rownames
from \code{sampleInfo(x)}.}

\item{treatments}{\code{character()} vector of treatment names. Must be valid
rownames from \code{treatmentInfo(x)}. This method does not work with
\code{CoreSet} objects yet.}

\item{features}{\code{character()} vector of feature names. Must be valid feature
names for a given \code{mDataType}}

\item{mDataTypes}{\code{character()} One or more molecular data types to
to subset features by. Must be valid rownames for the selected
SummarizedExperiment mDataTypes.}
}
\value{
See details.
}
\description{
Documentation for utility methods for a \code{CoreSet} object, such as
set operations like subset and intersect. See @details for information
on different types of methods and their implementations.
}
\details{
\subsection{subset methods}{

\strong{subsetBySample}: Subset a \code{CoreSet} object by sample identifier.
\itemize{
\item value: a \code{CoreSet} object containing only \code{samples}.
}
}

\subsection{subset methods}{

\strong{subsetByTreatment}: Subset a \code{CoreSet} object by treatment identifier.
\itemize{
\item value: a \code{CoreSet} object containing only \code{treatments}.
}
}

\subsection{subset methods}{

\strong{subsetByFeature}: Subset a \code{CoreSet} object by molecular feature
identifier.
\itemize{
\item value: a \code{CoreSet} object containing only \code{features}.
}
}
}
\examples{
data(clevelandSmall_cSet)

## subset methods

### subsetBySample
samples <- sampleInfo(clevelandSmall_cSet)$sampleid[seq_len(10)]
clevelandSmall_cSet_sub <- subsetBySample(clevelandSmall_cSet, samples)


## subset methods

### subsetByTreatment
#treatments <- treatmentInfo(clevelandSmall_cSet)$treatmentid[seq_len(10)]
#clevelandSmall_cSet_sub <- subsetByTreatment(clevelandSmall_cSet, treatments)


## subset methods

### subsetByFeature
features <- fNames(clevelandSmall_cSet, 'rna')[seq_len(5)]
clevelandSmall_cSet_sub <- subsetByFeature(clevelandSmall_cSet, features, 'rna')

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CoreSet-class.R
\name{CoreSet}
\alias{CoreSet}
\title{CoreSet constructor}
\usage{
CoreSet(
  name,
  molecularProfiles = list(),
  sample = data.frame(),
  sensitivityInfo = data.frame(),
  sensitivityRaw = array(dim = c(0, 0, 0)),
  sensitivityProfiles = matrix(),
  sensitivityN = matrix(nrow = 0, ncol = 0),
  perturbationN = array(NA, dim = c(0, 0, 0)),
  curationSample = data.frame(),
  curationTissue = data.frame(),
  curationTreatment = data.frame(),
  treatment = data.frame(),
  datasetType = c("sensitivity", "perturbation", "both"),
  verify = TRUE,
  ...
)
}
\arguments{
\item{name}{A \code{character} string detailing the name of the dataset}

\item{molecularProfiles}{A \code{list} of SummarizedExperiment objects containing
molecular profiles for each molecular data type.}

\item{sample}{A \code{data.frame} containing the annotations for all the sample
profiled in the data set, across all data types. Must contain the mandatory
\code{sampleid} column which uniquely identifies each sample in the object.}

\item{sensitivityInfo}{A \code{data.frame} containing the information for the
sensitivity experiments. Must contain a 'sampleid' column with unique
identifiers to each sample, matching the \code{sample} object and a 'treatmentid'
columns with unique indenifiers for each treatment, matching the \code{treatment}
object.}

\item{sensitivityRaw}{A 3 Dimensional \code{array} contaning the raw drug
dose response data for the sensitivity experiments}

\item{sensitivityProfiles}{\code{data.frame} containing drug sensitivity profile
statistics such as IC50 and AUC}

\item{sensitivityN, perturbationN}{A \code{data.frame} summarizing the
available sensitivity/perturbation data}

\item{curationSample, curationTissue, curationTreatment}{A \code{data.frame} mapping
the names for samples, tissues and treatments used in the data set to
universal identifiers used between different CoreSet objects}

\item{treatment}{A \code{data.frame} containing annotations for all treatments
profiled in the dataset. Must contain the mandatory \code{treatmentid} column
which uniquely identifies each treatment in the object.}

\item{datasetType}{A \code{character(1)} string of 'sensitivity',
'preturbation', or 'both' detailing what type of data can be found in the
\code{CoreSet}, for proper processing of the data}

\item{verify}{\code{logical(1)}Should the function verify the CoreSet and
print out any errors it finds after construction?}

\item{...}{Catch and parse any renamed constructor arguments.}
}
\value{
An object of class \code{CoreSet}
}
\description{
A constructor that simplifies the process of creating CoreSets, as well
as creates empty objects for data not provided to the constructor. Only
objects returned by this constructor are expected to work with the CoreSet
methods.
}
\details{
\subsection{\strong{WARNING}:}{

Parameters to this function have been renamed!
\itemize{
\item cell is now sample
\item drug is now treatment
}
}
}
\examples{
data(clevelandSmall_cSet)
clevelandSmall_cSet

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/allGenerics.R
\name{idCols}
\alias{idCols}
\title{Generic to access the unique id columns in an S4 object used to}
\usage{
idCols(object, ...)
}
\arguments{
\item{object}{An \code{S4} object to get id columns from.}

\item{...}{Allow new arguments to this generic.}
}
\value{
Depends on the implemented method
}
\description{
Generic to access the unique id columns in an S4 object used to
}
\examples{
print("Generics shouldn't need examples?")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TreatmentResponseExperiment-class.R
\name{TreatmentResponseExperiment}
\alias{TreatmentResponseExperiment}
\title{TreatmentResponseExperiment constructor method}
\usage{
TreatmentResponseExperiment(
  rowData,
  rowIDs,
  colData,
  colIDs,
  assays,
  assayIDs,
  metadata = list(),
  keep.rownames = FALSE
)
}
\arguments{
\item{rowData}{\code{data.table}, \code{data.frame}, \code{matrix} A table like object
coercible to a \code{data.table} containing the a unique \code{rowID} column which
is used to key assays, as well as additional row metadata to subset on.}

\item{rowIDs}{\code{character}, \code{integer} A vector specifying
the names or integer indexes of the row data identifier columns. These
columns will be pasted together to make up the rownames of the
\code{TreatmentResponseExperiment} object.}

\item{colData}{\code{data.table}, \code{data.frame}, \code{matrix} A table like object
coercible to a \code{data.table} containing the a unique \code{colID} column which
is used to key assays, as well as additional column metadata to subset on.}

\item{colIDs}{\code{character}, \code{integer} A vector specifying
the names or integer indexes of the column data identifier columns. These
columns will be pasted together to make up the colnames of the
\code{TreatmentResponseExperiment} object.}

\item{assays}{A \code{list} containing one or more objects coercible to a
\code{data.table}, and keyed by rowIDs and colIDs corresponding to the rowID and
colID columns in colData and rowData.}

\item{assayIDs}{\code{list} A list of \code{character} vectors specifying the columns
needed to uniquely identify each row in an \code{assay}. Names must match the
\code{assays} list.}

\item{metadata}{A \code{list} of metadata associated with the
\code{TreatmentResponseExperiment} object being constructed}

\item{keep.rownames}{\code{logical}, \code{character}
Logical: whether rownames should be added as a column if coercing to a
\code{data.table}, default is FALSE. If TRUE, rownames are added to the column
'rn'.
Character: specify a custom column name to store the rownames in.}
}
\value{
A \code{TreatmentResponseExperiment} object containing the data for a
treatment response experiment configured according to the rowIDs and
colIDs arguments.
}
\description{
Builds a \code{TreatmentResponseExperiment} object from rectangular
objects. The \code{rowData} argument should contain row level metadata, while
the \code{colData} argument should contain column level metadata, for the
experimental assays
in the \code{assays} list. The \code{rowIDs} and \code{colIDs} lists are used to configure
the internal keys mapping rows or columns to rows in the assays. Each list
should contain at minimum one character vector, specifying which columns
in \code{rowData} or \code{colData} are required to uniquely identify each row. An
optional second character vector can be included, specifying any metadata
columns for either dimension. These should contain information about each
row but NOT be required to uniquely identify a row in the \code{colData} or
\code{rowData} objects. Additional metadata can be attached to a
\code{TreatmentResponseExperiment} by passing a list to the metadata argument.
}
\details{
For now this class is simply a wrapper around a \code{LongTable} class. In the
future we plan to refactor CoreGx such that the \code{LongTable} class is in a
separate pacakge. We can then specialize the implementation of
\code{TreatmentResponseExperiment} to better capture the biomedical nature of
this object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deprecated.R
\name{buildLongTable,data.frame-method}
\alias{buildLongTable,data.frame-method}
\title{LongTable build method}
\usage{
\S4method{buildLongTable}{data.frame}(from, rowDataCols, colDataCols, assayCols)
}
\arguments{
\item{from}{\code{character} Path to the .csv file containing the data and
metadata from which to build the \code{LongTable}.}

\item{rowDataCols}{\code{list} List with two \code{character} vectors, the first
specifying one or more columns to be used as cell identifiers (e.g.,
cell-line name columns) and the second containing any additional metadata
columns related to the cell identifiers. If you wish to rename any of
these columns, assign the new names to their respective character vectors.}

\item{colDataCols}{\code{list} List with two \code{character} vectors, the first
specifying one or more columns to be used as column identifiers (e.g.,
drug name columns) and the second containing any additional metadata
columns related to the column identifiers. If you wish to rename any of
these columns, assign the new names to their respective character vectors.}

\item{assayCols}{\code{list} A named list of character vectors specifying how to
parse assay columns into a list of \code{data.table}s. Each list data.table
will be named for the name of corresponding list item and contain the columns
specified in the character vector of column names in each list item. If
there are no names for assayCols, the assays will be numbered by instead.}
}
\value{
A \code{LongTable} object containing one or more assays, indexed by
rowID and colID.
}
\description{
Create a LongTable object from a single data.table or
data.frame object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/matthewCor.R
\name{mcc}
\alias{mcc}
\title{Compute a Mathews Correlation Coefficient}
\usage{
mcc(
  x,
  y,
  nperm = 1000,
  nthread = 1,
  alternative = c("two.sided", "less", "greater"),
  ...
)
}
\arguments{
\item{x, y}{\code{factor} of the same length with the same number of levels}

\item{nperm}{\code{numeric} number of permutations for significance
estimation. If 0, no permutation testing is done}

\item{nthread}{\code{numeric} can parallelize permutation texting using
BiocParallels bplapply}

\item{alternative}{indicates the alternative hypothesis and must be one of
‘"two.sided"’, ‘"greater"’ or ‘"less"’.  You can specify just
the initial letter.  ‘"greater"’ corresponds to positive
association, ‘"less"’ to negative association.}

\item{...}{\code{list} Additional arguments}
}
\value{
A list with the MCC as the $estimate, and p value as $p.value
}
\description{
The function computes a Matthews correlation coefficient for two factors
provided to the function. It assumes each factor is a factor of class labels,
and the enteries are paired in order of the vectors.
}
\details{
Please note: we recommend you call set.seed() before using this function to
ensure the reproducibility of your results. Write down the seed number or
save it in a script if you intend to use the results in a publication.
}
\examples{
x <- factor(c(1,2,1,2,3,1))
y <- factor(c(2,1,1,1,2,2))
mcc(x,y)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LongTable-utils.R
\name{$,LongTable-method}
\alias{$,LongTable-method}
\title{Select an assay from a LongTable object}
\usage{
\S4method{$}{LongTable}(x, name)
}
\arguments{
\item{x}{A \code{LongTable} object to retrieve an assay from}

\item{name}{\code{character} The name of the assay to get.}
}
\value{
\code{data.frame} The assay object.
}
\description{
Select an assay from a LongTable object
}
\examples{
merckLongTable$sensitivity

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/datasets.R
\docType{data}
\name{merckLongTable}
\alias{merckLongTable}
\title{Merck Drug Combination Data LongTable}
\format{
LongTable object
}
\usage{
data(merckLongTable)
}
\description{
This is a LongTable object created from some drug combination data provided
to our lab by Merck.
}
\references{
TODO:: Include a reference
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\docType{methods}
\name{buildComboProfiles,LongTable-method}
\alias{buildComboProfiles,LongTable-method}
\title{Build an assay table with selected assay profiles for drug combinations}
\usage{
\S4method{buildComboProfiles}{LongTable}(object, profiles)
}
\arguments{
\item{object}{\code{LongTable} or inheriting class containing curated drug combination data.}

\item{profiles}{\code{character} a vector of profile names, i.e., column names of assays.}
}
\value{
A \code{data.table} containing fields
\code{treatment1id}, \code{treatment1dose}, \code{treatment2id}, \code{treatment2dose}, \code{sampleid},
which are used as keys to keep track of profiles,
along with columns of selected profiles from their assays.
Each \verb{*_1} is the monothearpy profile of treatment 1 in the combination,
and the same rule applies to treatment 2.
}
\description{
Build an assay table with selected assay profiles for drug combinations
}
\examples{
\dontrun{
combo_profile_1 <- buildComboProfiles(tre, c("auc", "SCORE"))
combo_profile_2 <- buildComboProfiles(tre, c("HS", "EC50", "E_inf", "ZIP"))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/connectivityScore.R
\name{connectivityScore}
\alias{connectivityScore}
\title{Function computing connectivity scores between two signatures}
\usage{
connectivityScore(
  x,
  y,
  method = c("fgsea", "gwc"),
  nperm = 10000,
  nthread = 1,
  gwc.method = c("spearman", "pearson"),
  ...
)
}
\arguments{
\item{x}{A \code{matrix} with the first gene signature. In the case of GSEA the vector of
values per gene for GSEA in which we are looking for an enrichment. In the
case of gwc, this should be a matrix, with the per gene responses in the
first column, and the significance values in the second.}

\item{y}{A \code{matrix} with the second signature. In the case of GSEA, this is the
vector of up and down regulated genes we are looking for in our signature,
with the direction being determined from the sign. In the case of gwc, this
should be a matrix of identical size to x, once again with the per gene
responses in the first column, and their significance in the second.}

\item{method}{\code{character} string identifying which method to use, out of 'fgsea' and 'gwc'}

\item{nperm}{\code{numeric}, how many permutations should be done to determine
significance through permutation testing? The minimum is 100, default is
1e4.}

\item{nthread}{\code{numeric}, how many cores to run parallel processing on.}

\item{gwc.method}{\code{character}, should gwc use a weighted spearman or pearson
statistic?}

\item{...}{Additional arguments passed down to gsea and gwc functions}
}
\value{
\code{numeric} a numeric vector with the score and the p-value associated
with it
}
\description{
A function for finding the connectivity between two signatures, using either
the GSEA method based on the KS statistic, or the gwc method based on a
weighted spearman statistic. The GSEA analysis is implemented in the piano package.
}
\examples{
xValue <- c(1,5,23,4,8,9,2,19,11,12,13)
xSig <- c(0.01, 0.001, .97, 0.01,0.01,0.28,0.7,0.01,0.01,0.01,0.01)
yValue <- c(1,5,10,4,8,19,22,19,11,12,13)
ySig <- c(0.01, 0.001, .97,0.01, 0.01,0.78,0.9,0.01,0.01,0.01,0.01)
xx <- cbind(xValue, xSig)
yy <- cbind(yValue, ySig)
rownames(xx) <- rownames(yy) <- c('1','2','3','4','5','6','7','8','9','10','11')
data.cor <- connectivityScore(xx,yy,method='gwc', gwc.method='spearman', nperm=300)

}
\references{
F. Pozzi, T. Di Matteo, T. Aste, 'Exponential smoothing weighted
correlations', The European Physical Journal B, Vol. 85, No 6, 2012. DOI:
10.1140/epjb/e2012-20697-x

Varemo, L., Nielsen, J. and Nookaew, I. (2013) Enriching the gene set
analysis of genome-wide data by incorporating directionality of gene
expression and combining statistical hypotheses and methods. Nucleic
Acids Research. 41 (8), 4378-4391. doi: 10.1093/nar/gkt111
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-iteration.R
\name{lapply,MultiAssayExperiment-method}
\alias{lapply,MultiAssayExperiment-method}
\title{lapply
lapply method for \code{MultiAssayExperiment}}
\usage{
\S4method{lapply}{MultiAssayExperiment}(X, FUN, ...)
}
\arguments{
\item{X}{A \code{MultiAssayExperiment} object.}

\item{FUN}{A function to be applied to each \code{SummarizedExperiment} in a
in \code{X}.}

\item{...}{Fall through parameters to \code{FUN}}
}
\value{
A \code{MultiAssayExperiment} object, modifed such that
\code{experiments(X) <- endoapply(experiments(X), FUN, ...)}.s
}
\description{
lapply
lapply method for \code{MultiAssayExperiment}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deprecated.R
\name{CoreGx-deprecated}
\alias{CoreGx-deprecated}
\alias{CoreGx-defunct}
\title{List of \code{deprecated} or \code{defunct} methods in the \code{CoreGx} R package.}
\description{
List of \code{deprecated} or \code{defunct} methods in the \code{CoreGx} R package.
}
\details{
\subsection{deprecated}{

\code{CoreSet}: The \code{CoreSet} constructor is being updated to have a new API. This
API is currently available via the \code{CoreSet2} constructor. In Bioconductor
3.16, the old constructor will be renamed \code{CoreSet2} and the new constructor
will be renamed \code{CoreSet}.
}

\subsection{defunct}{

\code{buildLongTable}: This function no longer works as building a \code{LongTable} or
\code{TreatmentResponseExperiment} now uses a \code{DataMapper} and the \code{metaConstruct}
method. See \code{vignette("LongTable")} for a detailed description of how
to create a LongTable object.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gwc.R
\name{gwc}
\alias{gwc}
\title{GWC Score}
\usage{
gwc(
  x1,
  p1,
  x2,
  p2,
  method.cor = c("pearson", "spearman"),
  nperm = 10000,
  truncate.p = 1e-16,
  ...
)
}
\arguments{
\item{x1}{\code{numeric} vector of effect sizes (e.g., fold change or t statitsics) for the first experiment}

\item{p1}{\code{numeric} vector of p-values for each corresponding effect size for the first experiment}

\item{x2}{\code{numeric} effect size (e.g., fold change or t statitsics) for the second experiment}

\item{p2}{\code{numeric} vector of p-values for each corresponding effect size for the second experiment}

\item{method.cor}{\code{character} string identifying if a \code{pearson} or
\code{spearman} correlation should be used}

\item{nperm}{\code{numeric} how many permutations should be done to determine}

\item{truncate.p}{\code{numeric} Truncation value for extremely low p-values}

\item{...}{Other passed down to internal functions}
}
\value{
\code{numeric} a vector of two values, the correlation and associated p-value.
}
\description{
Calculate the gwc score between two vectors, using either a weighted spearman
or pearson correlation
}
\examples{
data(clevelandSmall_cSet)
x <- molecularProfiles(clevelandSmall_cSet,'rna')[,1]
y <- molecularProfiles(clevelandSmall_cSet,'rna')[,2]
x_p <- rep(0.05, times=length(x))
y_p <- rep(0.05, times=length(y))
names(x_p) <- names(x)
names(y_p) <- names(y)
gwc(x,x_p,y,y_p, nperm=100)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods-coerce.R
\name{as.long.table}
\alias{as.long.table}
\title{Coerce from data.table to LongTable}
\usage{
as.long.table(x)
}
\arguments{
\item{x}{A \code{data.frame} with the 'longTableDataMapper' attribute, containing
three lists named assayCols, rowDataCols and colDataCols. This attribute
is automatically created when coercing from a LongTable to a data.table.}
}
\value{
\code{LongTable} object configured with the longTableDataMapper
}
\description{
Coerce a data.table with the proper configuration attributes
back to a LongTable
}
\examples{
dataTable <- as(merckLongTable, 'data.table')
print(attr(dataTable, 'longTableDataMapper')) # Method doesn't work without this
as.long.table(dataTable)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/datasets.R
\docType{data}
\name{exampleDataMapper}
\alias{exampleDataMapper}
\title{Example LongTableDataMapper}
\format{
LongTableDataMapper object
}
\usage{
data(exampleDataMapper)
}
\description{
A dummy LongTableDataMapper object to be used in package examples.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{is.items}
\alias{is.items}
\title{Get the types of all items in a list}
\usage{
is.items(list, ..., FUN = is)
}
\arguments{
\item{list}{A \code{list} to get the types from}

\item{...}{\code{pairlist} Additional arguments to FUN}

\item{FUN}{\code{function} or \code{character} Either a function, or the name
of a function which returns a single logical value. The default function
uses \code{is}, specify the desired type in \code{...}. You can also use other
type checking functions such as is.character, is.numeric, or is.data.frame.}
}
\value{
\code{logical} A vector indicating if the list item is the specified
type.
}
\description{
Get the types of all items in a list
}
\examples{
list <- list(c(1,2,3), c('a','b','c'))
is.items(list, 'character')

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/immutable-class.R
\name{mutable}
\alias{mutable}
\title{Remove the "immutable" S3-class from an R object, allowing it to be
modified normally again.}
\usage{
mutable(object)
}
\arguments{
\item{object}{An R object inheriting from the "immutable" class.}
}
\value{
The \code{object} with the "immutable" class stripped from it.
}
\description{
Remove the "immutable" S3-class from an R object, allowing it to be
modified normally again.
}
\examples{
immut_list <- immutable(list())
mutable(immut_list)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/immutable-class.R
\name{c.immutable}
\alias{c.immutable}
\title{Intercept concatenation for "immutable" class objects to return another "immutable" class object.}
\usage{
\method{c}{immutable}(x, ...)
}
\arguments{
\item{x}{An R object inheriting from the "immutable" S3-clas}

\item{...}{Objects to concatenate to \code{x}.}
}
\value{
x with one or more values appended to it.
}
\description{
Ensures that \code{c} and \code{append} to an "immutable" class object return an
immutable class object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/updateObject-methods.R
\name{updateObject,CoreSet-method}
\alias{updateObject,CoreSet-method}
\title{Update the \code{CoreSet} class after changes in it struture or API}
\usage{
\S4method{updateObject}{CoreSet}(object, verify = FALSE, verbose = FALSE)
}
\arguments{
\item{object}{A \code{CoreSet} object to update the class structure for.}

\item{verify}{A \code{logical(1)} indicating is \code{validObject} should be called
after updating the object. Defaults to \code{TRUE}, only set \code{FALSE} for debugging.}

\item{verbose}{TRUE or FALSE, indicating whether information about the update
should be reported}
}
\value{
\code{CoreSet} with update class structure.
}
\description{
Update the \code{CoreSet} class after changes in it struture or API
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/allGenerics.R
\name{endoaggregate}
\alias{endoaggregate}
\title{Perform aggregation over an S4 object, but return an object of the same
class.}
\usage{
endoaggregate(x, ...)
}
\arguments{
\item{x}{An \code{S4} object to endomorphically aggregate over.}

\item{...}{\code{pairlist} Allow definition of new parameters for
implementations of this generic.}
}
\value{
An object with the same class as \code{x}.
}
\description{
Perform aggregation over an S4 object, but return an object of the same
class.
}
\examples{
print("Generics shouldn't need examples?")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/allGenerics.R
\name{sensitivityRaw<-}
\alias{sensitivityRaw<-}
\title{sensitivityRaw<- Generic}
\usage{
sensitivityRaw(object, ...) <- value
}
\arguments{
\item{object}{An \code{S4} object to extract the raw sensitivity data from.}

\item{...}{\code{pairlist} Allow new parameters to be defined for this generic.}

\item{value}{An object containing dose and viability metrics to update
the object with.}
}
\value{
Depends on the implemented method
}
\description{
Generic function to set the raw data array for a treatment response experiment
in an S4 class.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/allGenerics.R
\name{colMeta}
\alias{colMeta}
\title{Generic to access the column identifiers for a rectangular object.}
\usage{
colMeta(object, ...)
}
\arguments{
\item{object}{\code{S4} An object to get column metadata columns from.}

\item{...}{ALlow new arguments to this generic}
}
\value{
Depends on impemented method.
}
\description{
Generic to access the column identifiers for a rectangular object.
}
\examples{
print("Generics shouldn't need examples?")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-optimization.R
\name{make_optim_function}
\alias{make_optim_function}
\title{Takes a non-primitive R function and refactors it to be compatible with
optimization via \code{stats::optim}.}
\usage{
make_optim_function(fn, ...)
}
\arguments{
\item{fn}{\code{function} A non-primitive function}

\item{...}{Arguments to \code{fn} to fix for before building the
function to be optimized. Useful for reducing the number of free parameters
in an optimization if there are insufficient degrees of freedom.}
}
\description{
Takes a non-primitive R function and refactors it to be compatible with optimization via \code{stats::optim}.
}
\seealso{
\code{\link{drop_fn_params}}, \code{\link{collect_fn_params}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CoreSet-class.R
\name{show,CoreSet-method}
\alias{show,CoreSet-method}
\title{Show a CoreSet}
\usage{
\S4method{show}{CoreSet}(object)
}
\arguments{
\item{object}{\code{CoreSet} object to show via \code{cat}.}
}
\value{
Prints the CoreSet object to the output stream, and returns
invisible NULL.
}
\description{
Show a CoreSet
}
\examples{
show(clevelandSmall_cSet)

}
\seealso{
\code{\link{cat}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adaptiveMatthewCor.R
\name{amcc}
\alias{amcc}
\title{Calculate an Adaptive Matthews Correlation Coefficient}
\usage{
amcc(x, y, step.prct = 0, min.cat = 3, nperm = 1000, nthread = 1, ...)
}
\arguments{
\item{x, y}{Two paired vectors of values. Could be replicates of observations
for the same experiments for example.}

\item{step.prct}{Instead of testing all possible splits of the data, it is
possible to test steps of a percentage size of the total number of ranks in
x/y. If this variable is 0, function defaults to testing all possible
splits.}

\item{min.cat}{The minimum number of members per category. Classifications
with less members fitting into both categories will not be considered.}

\item{nperm}{The number of perumatation to use for estimating significance.
If 0, then no p-value is calculated.}

\item{nthread}{Number of threads to parallize over. Both the AMCC calculation
and the permutation testing is done in parallel.}

\item{...}{Additional arguments}
}
\value{
Returns a list with two elements. $amcc contains the highest 'mcc'
value over all the splits, the p value, as well as the rank at which the
split was done.
}
\description{
This function calculates an Adaptive Matthews Correlation Coefficient (AMCC)
for two vectors of values of the same length. It assumes the entries in the
two vectors are paired. The Adaptive Matthews Correlation Coefficient for two
vectors of values is defined as the Maximum Matthews Coefficient over all
possible binary splits of the ranks of the two vectors. In this way, it
calculates the best possible agreement of a binary classifier on the two
vectors of data. If the AMCC is low, then it is impossible to find any binary
classification of the two vectors with a high degree of concordance.
}
\examples{
x <- c(1,2,3,4,5,6,7)
y <- c(1,3,5,4,2,7,6)
amcc(x,y, min.cat=2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DataMapper-class.R
\docType{class}
\name{DataMapper-class}
\alias{DataMapper-class}
\alias{.DataMapper}
\title{An S4 Class For Mapping from Raw Experimental Data to a Specific S4 Object}
\description{
This object will be used as a way to abstract away data preprocessing.
}
\section{Slots}{

\itemize{
\item rawdata: A list-like object containing one or more pieces of raw data
that will be processed and mapped to the slots of an \code{S4} object.
\item metadata: A \code{List} of object level metadata.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/allGenerics.R
\name{sensitivityProfiles<-}
\alias{sensitivityProfiles<-}
\title{sensitivityProfiles<- Generic}
\usage{
sensitivityProfiles(object, ...) <- value
}
\arguments{
\item{object}{An \code{S4} object to update the sensitivity profile summaries
for.}

\item{...}{Fallthrough arguments for defining new methods}

\item{value}{An object with the new sensitivity profiles. If a
matrix object is passed in, converted to data.frame before assignment}
}
\value{
Updated \code{CoreSet}
}
\description{
A generic for the sensitivityProfiles replacement method
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LongTableDataMapper-accessors.R
\name{assay,LongTableDataMapper,ANY-method}
\alias{assay,LongTableDataMapper,ANY-method}
\title{Extract the data for an assay from a \code{LongTableDataMapper}}
\usage{
\S4method{assay}{LongTableDataMapper,ANY}(x, i, withDimnames = TRUE)
}
\arguments{
\item{x}{\code{LongTableDataMapper} The object to retrive assay data form according
to the \code{assayMap} slot.}

\item{i}{\code{character(1)} Name of an assay in the \code{assayMap} slot of \code{x}.}

\item{withDimnames}{\code{logical(1)} For compatibility with
\code{SummarizedExperiment::assay} generic. Not used.}
}
\value{
\code{data.table} Data for the specified assay extracted from the
\code{rawdata} slot of \code{x}.
}
\description{
Extract the data for an assay from a \code{LongTableDataMapper}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/allGenerics.R
\name{getIntern}
\alias{getIntern}
\title{Retrieve the specified item from object internal metadata.}
\usage{
getIntern(object, x, ...)
}
\arguments{
\item{object}{\code{S4} An object with an @.itern slot containing an environment.}

\item{x}{\code{character} One or more symbol names to retrieve from the
object@.intern environment.}

\item{...}{Allow new parmeters to be defined for this generic.}
}
\value{
Depends on the implemented method
}
\description{
Internal slot for storing metadata relevant to the internal operation of an
S4 object.
}
\details{
Warning: This method is intended for developer use and can be ignored by
users.
}
\examples{
print("Generics shouldn't need examples?")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-optimization.R
\name{.fitCurve2}
\alias{.fitCurve2}
\title{Curve fitting via \code{stats::optim} L-BFGS-B with fall-back grid/pattern search
if convergence is not achieved.}
\usage{
.fitCurve2(
  par,
  x,
  y,
  fn,
  loss,
  lower = -Inf,
  upper = Inf,
  precision = 1e-04,
  density = c(2, 10, 5),
  step = 0.5/density,
  ...,
  loss_args = list(),
  span = 1,
  optim_only = FALSE,
  control = list(factr = 1e-08, ndeps = rep(1e-04, times = length(par)), trace = 0)
)
}
\arguments{
\item{par}{\code{numeric} Vector of intial guesses for the parameters. For each
index \code{i} of \code{par}, par[i] must be within the range (\verb{lower\[i\]},
\verb{upper\[i\]}). If only a single \code{upper} or \code{lower} value is present,
that range is used for all parameters in \code{par}.}

\item{x}{\code{numeric} Values to evaluate \code{fn} for.}

\item{y}{\code{numeric} Target output values to optimze \code{fn} against.}

\item{fn}{\code{function} A function to optimize. Any \code{fn} arguments passed via
\code{...} will be treated as constant and removed from the optimization. It
is assumed that the first argument is the x value to optimize over and
any subsequent arguments are free parameters to be optimized. Transformed
to be optim compatible via \code{make_optim_function} is the first arguement
isn't already \code{par}.}

\item{loss}{\code{character(1)} or \code{function} Either the name of one of the bundled
loss functions (see details) or a custom loss function to compute for
the output of \code{fn} over \code{x}.}

\item{lower}{\code{numeric(1)} Lower bound for parameters. Parallel to \code{par}.}

\item{upper}{\code{numeric(1)} Upper bound for paramteres. Parallel to \code{par}.}

\item{precision}{\code{numeric} smallest step size used in pattern search, once
step size drops below this value, the search terminates.}

\item{density}{\code{numeric} how many points in the dimension of each parameter
should be evaluated (density of the grid)}

\item{step}{initial step size for pattern search.}

\item{...}{\code{pairlist} Fall through arguments to \code{fn}.}

\item{loss_args}{\code{list} Additional argument to the \code{loss} function.
These get passed to losss via \code{do.call} analagously to using \code{...}.}

\item{span}{\code{numeric} Can be safely kept at 1, multiplicative ratio for
initial step size in pattern search. Must be larger than precision.}

\item{optim_only}{\code{logical(1)} Should the fall back methods when optim fails
be skipped? Default is \code{FALSE}.}

\item{control}{\code{list} List of control parameters to pass to \code{optim}. See
\code{?optim} for details.}
}
\value{
\code{numeric} Vector of optimal parameters for \code{fn} fit against \code{y}
on the values of \code{x}.
}
\description{
Function to fit curve via stats::optim
}
\details{
TODO
}
\examples{
\dontrun{
  # Four parameter hill curve equation
  hillEqn <- function(x, Emin, Emax, EC50, lambda) {
      (Emin + Emax * (x / EC50)^lambda) / (1 + (x / EC50)^lambda)
  }
  # Make some dummy data
  doses <- rev(1000 / (2^(1:20)))
  lambda <- 1
  Emin <- 1
  Emax <- 0.1
  EC50 <- median(doses)
  response <- hillEqn(doses, Emin=Emin, lambda=lambda, Emax=Emax, EC50=EC50)
  nresponse <- response + rnorm(length(response), sd=sd(response)*0.1) # add noise
  # 3-parameter optimization
  3par <- .fitCurve2(
      par=c(Emax, EC50, lambda),
      x=doses,
      y=nresponse,
      fn=hillEqn,
      Emin=Emin, # set this as constant in the function being optimized (via ...)
      loss=.normal_loss,
      loss_args=list(trunc=FALSE, n=1, scale=0.07),
      upper=c(1, max(doses), 6),
      lower=c(0, min(doses), 0)
  )
  # 2-parameter optimization
  2par <- .fitCurve2(
      par=c(Emax, EC50),
      x=doses,
      y=nresponse,
      fn=hillEqn,
      Emin=Emin, # set this as constant in the function being optimized (via ...)
      lambda=1,
      loss=.normal_loss,
      loss_args=list(trunc=FALSE, n=1, scale=0.07),
      upper=c(1, max(doses)),
      lower=c(0, min(doses))
  )
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/allGenerics.R
\name{summarizeSensitivityProfiles}
\alias{summarizeSensitivityProfiles}
\title{Summarize across replicates for a sensitivity dose-response experiment}
\usage{
summarizeSensitivityProfiles(object, ...)
}
\arguments{
\item{object}{An \code{S4} object to summarize sensitivity profiles for.}

\item{...}{Allow definition of new arguments to this generic}
}
\value{
Depends on the implemented method
}
\description{
Summarize across replicates for a sensitivity dose-response experiment
}
\examples{
print("Generics shouldn't need examples?")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LongTable-utils.R
\name{reindex,LongTable-method}
\alias{reindex,LongTable-method}
\title{Redo indexing for a LongTable object to remove any gaps in integer indexes}
\usage{
\S4method{reindex}{LongTable}(object)
}
\arguments{
\item{object}{The \code{LongTable} object to recalcualte indexes (rowKey and
colKey values) for.}
}
\value{
A copy of the \code{LongTable} with all keys as the smallest set of
contiguous integers possible given the current data.
}
\description{
After subsetting a LongTable, it is possible that values of rowKey or colKey
could no longer be present in the object. As a result there the indexes
will no longer be contiguous integers. This method will calcualte a new
set of rowKey and colKey values such that integer indexes are the smallest
set of contiguous integers possible for the data.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/callingWaterfall.R
\name{.distancePointSegment}
\alias{.distancePointSegment}
\title{Calculate shortest distance between point and line segment}
\usage{
.distancePointSegment(x, y, x1, y1, x2, y2)
}
\arguments{
\item{x}{x-coordinate of point}

\item{y}{y-coordinate of point}

\item{x1}{x-coordinate of one endpoint of the line segment}

\item{y1}{y-coordinate of line segment endpoint with x-coordinate x1}

\item{x2}{x-coordinate of other endpoint of line segment}

\item{y2}{y-coordinate of line segment endpoint with x-coordinate x2}
}
\value{
\code{numeric} The shortest distance between a point and a line
segment
}
\description{
This function calculates the shortest distance between a point
and a line segment in 2D space.
}
\examples{
.distancePointSegment(0, 0, -1, 1, 1, -1)

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/callingWaterfall.R
\name{.distancePointLine}
\alias{.distancePointLine}
\title{Calculate shortest distance between point and line}
\usage{
.distancePointLine(x, y, a = 1, b = 1, c = 0)
}
\arguments{
\item{x}{x-coordinate of point}

\item{y}{y-coordinate of point}

\item{a}{\code{numeric(1)} The coefficient in line equation a * x + b * y + c = 0.
Defaults to 1.}

\item{b}{\code{numeric(1)} The coefficient in line equation a * x + b * y + c = 0.
Defaults to 1.}

\item{c}{\code{numeric(1)} The intercept in line equation a * x + b * y + c = 0.
Defaults to 0.}
}
\value{
\code{numeric} The shortest distance between a point and a line.
}
\description{
This function calculates the shortest distance between a point
and a line in 2D space.
}
\examples{
.distancePointLine(0, 0, 1, -1, 1)

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TREDataMapper-class.R
\name{assays,TREDataMapper-method}
\alias{assays,TREDataMapper-method}
\title{Extract the data for all assays from a \code{TREDataMapper}}
\usage{
\S4method{assays}{TREDataMapper}(x, withDimnames = TRUE)
}
\arguments{
\item{x}{\code{TREDataMapper} The object to retrive assay data form according
to the \code{assayMap} slot.}

\item{withDimnames}{\code{logical(1)} For compatibility with
\code{SummarizedExperiment::assay} generic. Not used.}
}
\value{
\code{list} Data for all assays extracted from the
\code{rawdata} slot of \code{x} as a \code{list} of \code{data.tables}, where the \code{keys} for
each table are their \code{id_columns}.
}
\description{
Extract the data for all assays from a \code{TREDataMapper}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LongTable-accessors.R
\name{getIntern<-,LongTable,immutable_list-method}
\alias{getIntern<-,LongTable,immutable_list-method}
\title{Set the .intern slot of a LongTable}
\usage{
\S4method{getIntern}{LongTable,immutable_list}(object) <- value
}
\arguments{
\item{object}{\code{LongTable}}

\item{value}{An \code{immutable_list} object, being a class union between \code{list}
and \code{immutable} S3 classes.}
}
\value{
Updates the object and returns invisibly.
}
\description{
Set the .intern slot of a LongTable
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/allGenerics.R
\name{sensitivityInfo<-}
\alias{sensitivityInfo<-}
\title{sensitivityInfo<- Generic Method}
\usage{
sensitivityInfo(object, ...) <- value
}
\arguments{
\item{object}{An \code{S4} object to set treatment response experiment
annotations for.}

\item{...}{Allow new arguments to be defined for this generic.}

\item{value}{The new treatment response experiment annotations.}
}
\value{
Depends on the implemented method
}
\description{
Generic function to get the annotations for a treatment response experiment
from an S4 class.
}
\examples{
print("Generics shouldn't need examples?")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LongTable-accessors.R
\name{metadata,LongTable-method}
\alias{metadata,LongTable-method}
\title{Getter method for the metadata slot of a \code{LongTable} object}
\usage{
\S4method{metadata}{LongTable}(x)
}
\arguments{
\item{x}{The \code{LongTable} object from which to retrieve the metadata list.}
}
\value{
\code{list} The contents of the \code{metadata} slot of the \code{LongTable}
object.
}
\description{
Getter method for the metadata slot of a \code{LongTable} object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CoreSet-class.R
\name{CoreSet2}
\alias{CoreSet2}
\title{Make a CoreSet with the updated class structure}
\usage{
CoreSet2(
  name = "emptySet",
  treatment = data.frame(),
  sample = data.frame(),
  molecularProfiles = MultiAssayExperiment(),
  treatmentResponse = LongTable(),
  datasetType = "sensitivity",
  perturbation = list(n = array(dim = 3), info = "No perturbation data!"),
  curation = list(sample = data.frame(), treatment = data.frame())
)
}
\arguments{
\item{name}{A \code{character(1)} vector with the \code{CoreSet} objects name.}

\item{treatment}{A \code{data.frame} with treatment level metadata.}

\item{sample}{A \code{data.frame} with sample level metadata for the union
of samples in \code{treatmentResponse} and \code{molecularProfiles}.}

\item{molecularProfiles}{A \code{MultiAssayExperiment} containing one
\code{SummarizedExperiment} object for each molecular data type.}

\item{treatmentResponse}{A \code{LongTable} or \code{LongTableDataMapper} object
containing all treatment response data associated with the \code{CoreSet}
object.}

\item{datasetType}{A deprecated slot in a \code{CoreSet} object included
for backwards compatibility. This may be removed in future releases.}

\item{perturbation}{A deprecated slot in a \code{CoreSet} object included
for backwards compatibility. This may be removed in future releases.}

\item{curation}{A \code{list(2)} object with two items named \code{treatment} and \code{sample} with
mappings from publication identifiers to standardized identifiers for
both annotations, respectively.}
}
\value{
A \code{CoreSet} object storing standardized and curated treatment
response and multiomic profile data associated with a given publication.
}
\description{
New implementation of the CoreSet constructor to support MAE and TRE. This
constructor will be swapped with the original \code{CoreSet} constructor as
part of an overhaul of the CoreSet class structure.
}
\examples{
data(clevelandSmall_cSet)
clevelandSmall_cSet

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/allGenerics.R
\name{sensitivityMeasures<-}
\alias{sensitivityMeasures<-}
\title{sensitivityMeasures<- Generic}
\usage{
sensitivityMeasures(object, ...) <- value
}
\arguments{
\item{object}{An \code{S4} object to update.}

\item{...}{Allow new methods to be defined for this generic.}

\item{value}{A set of names for sensitivity measures to use to
update the object with.}
}
\value{
Depends on the implemented method
}
\description{
Set the names of the sensitivity summary metrics available in an S4
object.
}
\examples{
print("Generics shouldn't need examples?")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TREDataMapper-class.R
\name{TREDataMapper-accessors}
\alias{TREDataMapper-accessors}
\alias{rawdata,TREDataMapper-method}
\alias{rawdata<-,TREDataMapper,list-method}
\alias{rowDataMap,TREDataMapper-method}
\alias{rowDataMap}
\alias{rowDataMap<-,TREDataMapper,list_OR_List-method}
\alias{rowDataMap<-,TREDataMapper,list-method}
\alias{rowDataMap<-TREDataMapper,List-method}
\alias{rowDataMap<-}
\alias{colDataMap,TREDataMapper-method}
\alias{colDataMap}
\alias{colDataMap<-,TREDataMapper,list_OR_List-method}
\alias{colDataMap<-TREDataMapper,List-method}
\alias{colDataMap<-}
\alias{assayMap,TREDataMapper-method}
\alias{assayMap,TREDataMapper,list-method}
\alias{assayMap,TREDataMapper,List-method}
\alias{assayMap}
\alias{assayMap<-,TREDataMapper,list_OR_List-method}
\alias{assayMap<-,TREDataMapper,list-method}
\alias{assayMap<-,TREDataMapper,List-methhod}
\alias{assayMap<-}
\alias{metadataMap,TREDataMapper-method}
\alias{metadataMap}
\alias{metadataMap<-,TREDataMapper,list_OR_List-method}
\alias{metadataMap<-,TREDataMapper-method}
\alias{metadataMap<-}
\title{Accessing and modifying data in a \code{TREDataMapper} object.}
\usage{
\S4method{rawdata}{TREDataMapper,list}(object) <- value

\S4method{rowDataMap}{TREDataMapper}(object)

\S4method{rowDataMap}{TREDataMapper,list_OR_List}(object) <- value

\S4method{colDataMap}{TREDataMapper}(object)

\S4method{colDataMap}{TREDataMapper,list_OR_List}(object) <- value

\S4method{assayMap}{TREDataMapper}(object)

\S4method{assayMap}{TREDataMapper,list_OR_List}(object) <- value

\S4method{metadataMap}{TREDataMapper}(object)

\S4method{metadataMap}{TREDataMapper,list_OR_List}(object) <- value
}
\arguments{
\item{object}{A \code{TREDataMapper} object to get or set data from.}

\item{value}{See details.}
}
\value{
Accessors: See details

Setters: An update \code{TREDataMapper} object, returned invisibly.
}
\description{
Documentation for the various setters and getters which allow manipulation
of data in the slots of a \code{TREDataMapper} object.
}
\details{
\strong{rawdata}: Get the raw data slot from a \code{TREDataMapper} object. Returns
a list-like containing one or more raw data inputs to the
\code{TREDataMapper} object.

\strong{rawdata}: Set the raw data slot from a \code{TREDataMapper} object.
\strong{value}: The \code{list}-like object to set for the rawdata slot. Note: this
currently only supports \code{data.frame} or \code{data.table} objects.

\strong{rowDataMap}: \code{list} of two \code{character} vectors, the first are the
columns required to uniquely identify each row of a \code{TREDataMapper} and the
second any additional row-level metadata. If the character vectors
have names, the resulting columns are automatically renamed to the
item name of the specified column.

\strong{rowDataMap<-}: Update the \verb{@rowDataMap} slot of a \code{TREDataMapper} object,
returning an invisible NULL. Arguments:
\itemize{
\item value: A \code{list} or \code{List} where the first item is the names of the
identifier columns -- columns needed to uniquely identify each row in
rowData -- and the second item is the metadata associated with those
the identifier columns, but not required to uniquely identify rows in
the object rowData.
}

\strong{colDataMap}: \code{list} of two \code{character} vectors, the first are the
columns required to uniquely identify each row of a \code{TREDataMapper} and the
second any additional col-level metadata. If the character vectors
have names, the resulting columns are automatically renamed to the
item name of the specified column.

\strong{colDataMap<-}: Update the \verb{@colDataMap} slot of a \code{TREDataMapper} object,
returning an invisible NULL. Arguments:
\itemize{
\item value: A \code{list} or \code{List} where the first item is the names of the
identifier columns -- columns needed to uniquely identify each row in
colData -- and the second item is the metadata associated with those
the identifier columns, but not required to uniquely identify rows in
the object rowData.
}

\strong{assayMap}:  A \code{list} of character vectors. The name of each list item
will be the assay in a \code{LongTableDataMapper} object that the columns in the
\code{character} vector will be assigned to. Column renaming occurs automatically
when the character vectors have names (from the value to the name).

\strong{assayMap<-}: Updates the \verb{@assayMap} slot of a \code{TREDataMapper} object,
returning an invisible NULL. Arguments:
\itemize{
\item value:  A \code{list} of character vectors, where the name of each list
item is the name of an assay and the values of each character vector
specify the columns mapping to the assay in the \code{S4} object the
\code{TREDataMapper} constructs.
}

\strong{metadataMap}:  A \code{list} of \code{character} vectors. Each item is an element
of the constructed objects \verb{@metadata} slot.

\strong{metadataMap<-}: Updates \code{TREDataMapper} object in-place, then returns an
\code{invisible(NULL)}. Arguments:
\itemize{
\item value:  A \code{list} of \code{character} vectors. The name of each list item
is the name of the item in the \verb{@metadata} slot of the \code{TREDataMapper} object
created when \code{metaConstruct} is called on the \code{DataMapper}, and a
character vector specifies the columns of \verb{@rawdata} to assign to each item.
}
}
\examples{
rowDataMap(exampleDataMapper)

rowDataMap(exampleDataMapper) <- list(c('treatmentid'), c())

colDataMap(exampleDataMapper)

colDataMap(exampleDataMapper) <- list(c('sampleid'), c())

assayMap(exampleDataMapper)

assayMap(exampleDataMapper) <- list(sensitivity=c(viability1='viability'))

metadataMap(exampleDataMapper)

metadataMap(exampleDataMapper) <- list(object_metadata=c('metadata'))

}
\seealso{
Other DataMapper-accessors: 
\code{\link{DataMapper-accessors}},
\code{\link{LongTableDataMapper-accessors}}
}
\concept{DataMapper-accessors}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods-guessMapping.R
\name{checkColumnCardinality}
\alias{checkColumnCardinality}
\alias{cardinality}
\title{Search a data.frame for 1:\code{cardinality} relationships between a group
of columns (your identifiers) and all other columns.}
\usage{
checkColumnCardinality(df, group, cardinality = 1, ...)
}
\arguments{
\item{df}{A \code{data.frame} to search for 1:\code{cardinality} mappings with
the columns in \code{group}.}

\item{group}{A \code{character} vector of one or more column names to
check the cardinality of other columns against.}

\item{cardinality}{The cardinality of to search for (i.e., 1:\code{cardinality})
relationships with the combination of columns in group. Defaults to 1
(i.e., 1:1 mappings).}

\item{...}{Fall through arguments to data.table::\code{[}. For developer use.
One use case is setting verbose=TRUE to diagnose slow data.table
operations.}
}
\value{
A \code{character} vector with the names of the columns with
cardinality of 1:\code{cardinality} with the columns listed in \code{group}.
}
\description{
Search a data.frame for 1:\code{cardinality} relationships between a group
of columns (your identifiers) and all other columns.
}
\examples{
df <- rawdata(exampleDataMapper)
checkColumnCardinality(df, group='treatmentid')

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TreatmentResponseExperiment-class.R
\docType{class}
\name{TreatmentResponseExperiment-class}
\alias{TreatmentResponseExperiment-class}
\alias{.TreatmentResponseExperiment}
\title{TreatmentResponseExperiment class definition}
\value{
\code{TreatmentResponseExperiment} object containing the assay data from
a treatment response experiment
}
\description{
Define a private constructor method to be used to build a
\code{TreatmentResponseExperiment} object.
}
\section{Slots}{

\describe{
\item{\code{rowData}}{See Slots section.}

\item{\code{colData}}{See Slots section.}

\item{\code{assays}}{See Slots section.}

\item{\code{metadata}}{See Slots section.}

\item{\code{.intern}}{See Slots section.}
}}

\section{Slots}{

\itemize{
\item \emph{rowData}: A \code{data.table} containing the metadata associated with the
row dimension of a \code{TreatmentResponseExperiment}.
\item \emph{colData}: A \code{data.table} containing the metadata associated with the
column dimension of a \code{TreatmentResponseExperiment}.
\item \emph{assays}: A \code{list} of \code{data.table}s, one for each assay in a
\code{TreatmentResponseExperiment}.
\item \emph{metadata}: An optional \code{list} of additional metadata for a
\code{TreatmentResponseExperiment} which doesn't map to one of the dimensions.
\item \emph{.intern}: An \code{environment} that holds internal structural metadata
about a \code{TreatmentResponseExperiment} object, such as which columns are
required to key the object. An environment has been used to allow locking
items, which can prevent accidental modification of a property required
for the class to work.
}
}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods-metaConstruct.R
\name{metaConstruct}
\alias{metaConstruct}
\alias{metaConstruct,LongTableDataMapper-method}
\alias{metaConstruct,TREDataMapper-method}
\title{Generic for preprocessing complex data before being used in the constructor
of an \code{S4} container object.}
\usage{
metaConstruct(mapper, ...)

\S4method{metaConstruct}{LongTableDataMapper}(mapper)

\S4method{metaConstruct}{TREDataMapper}(mapper)
}
\arguments{
\item{mapper}{An \code{TREDataMapper} object abstracting arguments to an
the \code{TreatmentResponseExperiment} constructor.}

\item{...}{Allow new arguments to be defined for this generic.}
}
\value{
An \code{S4} object for which the class corresponds to the type of
the build configuration object passed to this method.

A \code{LongTable} object, as specified in the mapper.

A \code{TreatmentResponseExperiment} object, as specified in the mapper.
}
\description{
This method is intended to abstract away complex constructor arguments
and data preprocessing steps needed to transform raw data, such as that
produced in a treatment-response or next-gen sequencing experiment, and
automate building of the appropriate \code{S4} container object. This is
is intended to allow mapping between different experimental designs,
in the form of an \code{S4} configuration object, and various S4 class
containers in the Bioconductor community and beyond.
}
\examples{
data(exampleDataMapper)
rowDataMap(exampleDataMapper) <- list(c('treatmentid'), c())
colDataMap(exampleDataMapper) <- list(c('sampleid'), c())
assayMap(exampleDataMapper) <- list(sensitivity=list(c("treatmentid", "sampleid"), c('viability')))
metadataMap(exampleDataMapper) <- list(experiment_metadata=c('metadata'))
longTable <- metaConstruct(exampleDataMapper)
longTable

data(exampleDataMapper)
exampleDataMapper <- as(exampleDataMapper, "TREDataMapper")
rowDataMap(exampleDataMapper) <- list(c('treatmentid'), c())
colDataMap(exampleDataMapper) <- list(c('sampleid'), c())
assayMap(exampleDataMapper) <- list(sensitivity=list(c("treatmentid", "sampleid"), c('viability')))
metadataMap(exampleDataMapper) <- list(experiment_metadata=c('metadata'))
tre <- metaConstruct(exampleDataMapper)
tre

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/allGenerics.R
\name{sensitivityMeasures}
\alias{sensitivityMeasures}
\title{sensitivityMeasures Generic}
\usage{
sensitivityMeasures(object, ...)
}
\arguments{
\item{object}{An \code{S4} object to retrieve the names of sensitivty summary
measurements for.}

\item{...}{Fallthrough arguements for defining new methods}
}
\value{
Depends on the implemented method
}
\description{
Get the names of the sensitivity summary metrics available in an S4
object.
}
\examples{
sensitivityMeasures(clevelandSmall_cSet)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CoreSet-class.R
\docType{class}
\name{CoreSet-class}
\alias{CoreSet-class}
\alias{.CoreSet}
\title{CoreSet - A generic data container for molecular profiles and
treatment response data}
\description{
CoreSet - A generic data container for molecular profiles and
treatment response data
}
\details{
The CoreSet (cSet) class was developed as a superclass for pSets in the
PharmacoGx and RadioGx packages to contain the data generated in screens
of cancer sample lines for their genetic profile and sensitivities to therapy
(Pharmacological or Radiation). This class is meant to be a superclass which
is contained within the PharmacoSet (pSet) and RadioSet (rSet) objects
exported by PharmacoGx and RadioGx. The format of the data is similar for
both pSets and rSets, allowing much of the code to be abstracted into
the CoreSet super-class. However, the models involved with quantifying
sampleular response to Pharmacological and Radiation therapy are widely
different, and extension of the cSet class allows the
packages to apply the correct model for the given data.
}
\section{Slots}{

\describe{
\item{\code{annotation}}{See Slots section.}

\item{\code{molecularProfiles}}{See Slots section.}

\item{\code{sample}}{See Slots section.}

\item{\code{treatment}}{See Slots section.}

\item{\code{treatmentResponse}}{See Slots section.}

\item{\code{perturbation}}{See Slots section.}

\item{\code{curation}}{See Slots section.}

\item{\code{datasetType}}{See Slots section.}
}}

\section{Slots}{

\itemize{
\item annotation: A \code{list} of annotation data about the \code{CoreSet},
including the \verb{$name} and the session information for how the object
was created, detailing the exact versions of R and all the packages used.
\item molecularProfiles: A \code{list} or \code{MultiAssayExperiment} containing
\code{CoreSet} object.
\item sample: A \code{data.frame} containg the annotations for all the samples
profiled in the data set, across all molecular data types and
treatment response experiments.
\item treatment: A \code{data.frame} containing the annotations for all treatments
in the dataset, including the mandatory 'treatmentid' column to uniquely
identify each treatment.
\item treatmentResponse: A \code{list} or \code{LongTable} containing all the data for the
treatment response experiment, including \verb{$info}, a \code{data.frame}
containing the experimental info, \verb{$raw} a 3D \code{array} containing raw data,
\verb{$profiles}, a \code{data.frame} containing sensitivity profiles
statistics, and \verb{$n}, a \code{data.frame} detailing the number of
experiments for each sample-drug/radiationInfo pair
\item perturbation: \code{list} containing \verb{$n}, a \code{data.frame}
summarizing the available perturbation data. This slot is currently
being deprecated.
\item curation: \code{list} containing mappings for \code{treatment},
\code{sample} and \code{tissue} names used in the data set to universal
identifiers used between different \code{CoreSet} objects.
\item datasetType: \code{character} string of 'sensitivity',
'perturbation', or both detailing what type of data can be found in the
\code{CoreSet}, for proper processing of the data
}
}

\seealso{
\code{\link{CoreSet-accessors}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods-guessMapping.R
\name{guessMapping,LongTableDataMapper-method}
\alias{guessMapping,LongTableDataMapper-method}
\title{Guess which columns in raw experiment data map to which dimensions.}
\usage{
\S4method{guessMapping}{LongTableDataMapper}(object, groups, subset, data = FALSE)
}
\arguments{
\item{object}{A \code{LongTableDataMapper} object.}

\item{groups}{A \code{list} containing one or more vector of column names
to group-by. The function uses these to determine 1:1 mappings between
the combination of columns in each vector and unique values in the raw
data columns.}

\item{subset}{A \code{logical} vector indicating whether to to subset out mapped
columns after each grouping. Must be a single \code{TRUE} or \code{FALSE} or have
the same length as groups, indicating whether to subset out mapped columns
after each grouping. This will prevent mapping a column to two different
groups.}

\item{data}{A \code{logical} vector indicating whether you would like the data
for mapped columns to be returned instead of their column names. Defaults
to \code{FALSE} for easy use assigning mapped columns to a \code{DataMapper} object.}
}
\value{
A \code{list}, where each item is named for the associated \code{groups} item
the guess is for. The character vector in each item are columns which are
uniquely identified by the identifiers from that group.
}
\description{
Checks for columns which are uniquely identified by a group of identifiers.
This should be used to help identify the columns required to uniquely
identify the rows, columns, assays and metadata of a \code{DataMapper} class
object.
}
\details{
Any unmapped columns will be added to the end of the returned \code{list} in an
item called unmapped.

The function automatically guesses metadata by checking if any columns have
only a single value. This is returned as an additional item in the list.
}
\examples{
guessMapping(exampleDataMapper, groups=list(rows='treatmentid', cols='sampleid'),
subset=FALSE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/immutable-class.R
\name{assignment-immutable}
\alias{assignment-immutable}
\alias{subset<-.immutable}
\alias{subset<-.immutable,}
\alias{[<-.immutable,}
\alias{[[<-.immutable,}
\alias{$<-.immubtale,}
\alias{colnames<-.immutable,}
\alias{rownames<-.immutable,}
\alias{dimnames<-.immutable,}
\alias{names<-.immutable}
\alias{[<-}
\alias{[<-.immutable}
\alias{[[<-}
\alias{[[<-.immutable}
\alias{$<-}
\alias{$<-.immutable}
\alias{names<-}
\alias{dimnames<-}
\alias{dimnames<-.immutable}
\alias{colnames<-}
\alias{colnames<-.immutable}
\alias{rownames<-}
\alias{rownames<-.immutable}
\title{Intercept assignment operations for "immutable" S3 objects.}
\usage{
\\method{subset}{immutable}(object, ...) <- value

\method{[}{immutable}(object, ...) <- value

\method{[[}{immutable}(object, ...) <- value

\method{$}{immutable}(object, ...) <- value

\method{names}{immutable}(x) <- value

\method{dimnames}{immutable}(x) <- value

\\method{colnames}{immutable}(x) <- value

\\method{rownames}{immutable}(x) <- value
}
\arguments{
\item{object, x}{An R object inherting from the "immutable" S3-class.}

\item{...}{Catch subset arguments for various dimensions.}

\item{value}{Not used.}
}
\value{
None, throws an error.
}
\description{
Prevents modification of objects labelled with the "immutable" S3-class by
intercepting assignment during S3-method dispatch and returning an error.
}
\examples{
immutable_df <- immutable(data.frame(a=1:5, b=letters[1:5]))
# return immutable data.frame
immutable_df[1:4, ]
# return immutable vector
immutable_df$a

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/allGenerics.R
\name{sensitivityInfo}
\alias{sensitivityInfo}
\title{Generic function to get the annotations for a treatment response experiment
from an S4 class}
\usage{
sensitivityInfo(object, ...)
}
\arguments{
\item{object}{An \code{S4} object to get treatment response experiment
annotations from.}

\item{...}{Allow new arguments to be defined for this generic.}
}
\value{
Depends on the implemented method
}
\description{
Generic function to get the annotations for a treatment response experiment
from an S4 class
}
\examples{
print("Generics shouldn't need examples?")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CoreSet-class.R
\name{updateSampleId}
\alias{updateSampleId}
\title{Update the sample ids in a cSet object}
\usage{
updateSampleId(object, new.ids = vector("character"))
}
\arguments{
\item{object}{The object for which the sample ids will be updated}

\item{new.ids}{The new ids to assign to the object}
}
\value{
\code{CoreSet} The modified CoreSet object
}
\description{
Update the sample ids in a cSet object
}
\examples{
updateSampleId(clevelandSmall_cSet, sampleNames(clevelandSmall_cSet))

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/allGenerics.R
\name{sensitivitySlotToLongTable}
\alias{sensitivitySlotToLongTable}
\title{sensitivitySlotToLongTable Generic}
\usage{
sensitivitySlotToLongTable(object, ...)
}
\arguments{
\item{object}{\code{CoreSet} Object inheriting from CoreSet.}

\item{...}{Allow new arguments to be defined on this generic.}
}
\value{
A \code{LongTable} object containing the data in the sensitivity slot.
}
\description{
Convert the sensitivity slot in an object inheriting from a CoreSet from a
list to a LongTable.
}
\examples{
print("Generics shouldn't need examples?")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/allGenerics.R
\name{sensitivityProfiles}
\alias{sensitivityProfiles}
\title{sensitivityProfiles Generic}
\usage{
sensitivityProfiles(object, ...)
}
\arguments{
\item{object}{The \code{S4} object to retrieve sensitivity profile summaries
from.}

\item{...}{\code{pairlist} Allow defining new arguments for this generic.}
}
\value{
Depends on the implemented method
}
\description{
A generic for sensitivityProfiles getter method
}
\examples{
print("Generics shouldn't need examples?")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/allGenerics.R
\name{rowIDs}
\alias{rowIDs}
\title{Generic to access the row identifiers from}
\usage{
rowIDs(object, ...)
}
\arguments{
\item{object}{\code{S4} An object to get row id columns from.}

\item{...}{Allow new arguments to this generic.}
}
\value{
Depends on the implemented method.
}
\description{
Generic to access the row identifiers from
}
\examples{
print("Generics shouldn't need examples?")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/allGenerics.R
\name{showSigAnnot}
\alias{showSigAnnot}
\title{Get the annotations for a \code{Signature} class object, as returned by
\code{drugSensitivitysig} or \code{radSensitivtySig} functions available in
\code{PharmacoGx} and \code{RadioGx}, respectively.}
\usage{
showSigAnnot(object, ...)
}
\arguments{
\item{object}{A \code{Signature} class object}

\item{...}{Allow definition of new arguments to this generic}
}
\value{
NULL Prints the signature annotations to console
}
\description{
Get the annotations for a \code{Signature} class object, as returned by
\code{drugSensitivitysig} or \code{radSensitivtySig} functions available in
\code{PharmacoGx} and \code{RadioGx}, respectively.
}
\examples{
print("Generics shouldn't need examples?")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods-coerce.R
\name{.longTableToSummarizedExperiment}
\alias{.longTableToSummarizedExperiment}
\title{Convert LongTable to gDR Style SummarizedExperiment}
\usage{
.longTableToSummarizedExperiment(LT, assay_names)
}
\arguments{
\item{LT}{\code{LongTable} to convert to gDR \code{SummarizedExperiment} format.}

\item{assay_names}{\code{character()} Names to rename the assays to. These
are assumed to be in the same order as \code{assayNames(LT)}.}
}
\value{
\code{SummarizedExperiment} object with all assay from \code{LT} as
\code{BumpyMatrix}es.
}
\description{
Convert LongTable to gDR Style SummarizedExperiment
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aggregate-methods.R
\name{aggregate2}
\alias{aggregate2}
\title{Functional API for data.table aggregation which allows capture of associated
aggregate calls so they can be recomputed later.}
\usage{
aggregate2(
  x,
  by,
  ...,
  nthread = 1,
  progress = interactive(),
  BPPARAM = NULL,
  enlist = TRUE,
  moreArgs = list()
)
}
\arguments{
\item{x}{\code{data.table}}

\item{by}{\code{character} One or more valid column names in \code{x} to compute
groups using.}

\item{...}{\code{call} One or more aggregations to compute for each group by in x.
If you name aggregation calls, that will be the column name of the value
in the resulting \code{data.table} otherwise a default name will be parsed from
the function name and its first argument, which is assumed to be the name
of the column being aggregated over.}

\item{nthread}{\code{numeric(1)} Number of threads to use for split-apply-combine
parallelization. Uses \code{BiocParllel::bplapply} if nthread > 1 or you pass in
\code{BPPARAM}. Does not modify data.table threads, so be sure to use
setDTthreads for reasonable nested parallelism. See details for performance
considerations.}

\item{progress}{\code{logical(1)} Display a progress bar for parallelized
computations? Only works if \verb{bpprogressbar<-} is defined for the current
BiocParallel back-end.}

\item{BPPARAM}{\code{BiocParallelParam} object. Use to customized the
the parallization back-end of bplapply. Note, nthread over-rides any
settings from BPPARAM as long as \verb{bpworkers<-} is defined for that class.}

\item{enlist}{\code{logical(1)} Default is \code{TRUE}. Set to \code{FALSE} to evaluate
the first call in \code{...} within \code{data.table} groups. See details for more
information.}

\item{moreArgs}{\code{list()} A named list where each item is an argument one of
the calls in \code{...} which is not a column in the table being aggregated. Use
to further parameterize you calls. Please note that these are not added
to your aggregate calls unless you specify the names in the call.}
}
\value{
\code{data.table} of aggregation results.
}
\description{
Functional API for data.table aggregation which allows capture of associated
aggregate calls so they can be recomputed later.
}
\details{
\subsection{Use of Non-Standard Evaluation}{

Arguments in \code{...} are substituted and wrapped in a list, which is passed
through to the j argument of \verb{[.data.table} internally. The function currently
tries to build informative column names for unnamed arguments in \code{...} by
appending the name of each function call with the name of its first argument,
which is assumed to be the column name being aggregated over. If an argument
to \code{...} is named, that will be the column name of its value in the resulting
\code{data.table}.
}

\subsection{Enlisting}{

The primary use case for \code{enlist=FALSE} is to allow computation of dependent
aggregations, where the output from a previous aggregation is required in a
subsequent one. For this case, wrap your call in \verb{\{} and assign intermediate
results to variables, returning the final results as a list where each list
item will become a column in the final table with the corresponding name.
Name inference is disabled for this case, since it is assumed you will name
the returned list items appropriately.
A major advantage over multiple calls to \code{aggregate} is that
the overhead of parallelization is paid only once even for complex multi-step
computations like fitting a model, capturing its paramters, and making
predictions using it. It also allows capturing arbitrarily complex calls
which can be recomputed later using the
\verb{update,TreatmentResponseExperiment-method}
A potential disadvantage is increased RAM usage per
thread due to storing intermediate values in variables, as well as any
memory allocation overhead associate therewith.
}
}
\seealso{
\verb{data.table::[.data.table}, \code{BiocParallel::bplapply}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/allGenerics.R
\name{buildLongTable}
\alias{buildLongTable}
\title{Build a LongTable object}
\usage{
buildLongTable(from, ...)
}
\arguments{
\item{from}{What to build the LongTable from?}

\item{...}{\code{pairlist} Allow definition of new parameters for
implementations of this generic.}
}
\value{
Depends on the implemented method
}
\description{
Build a LongTable object
}
\examples{
print("Generics shouldn't need examples?")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LongTableDataMapper-class.R
\docType{class}
\name{LongTableDataMapper-class}
\alias{LongTableDataMapper-class}
\alias{.LongTableDataMapper}
\alias{show,LongTableDataMapper-method}
\title{A Class for Mapping Between Raw Data and an \code{LongTable} Object}
\usage{
\S4method{show}{LongTableDataMapper}(object)
}
\arguments{
\item{object}{A \code{LongTableDataMapper} to display in the console.}
}
\value{
\code{invisible} Prints to console.
}
\description{
A Class for Mapping Between Raw Data and an \code{LongTable} Object
}
\section{Functions}{
\itemize{
\item \code{show(LongTableDataMapper)}: Show method for LongTableDataMapper.
Determines how the object is displayed in the console.

}}
\section{Slots}{

\describe{
\item{\code{rawdata}}{See Slots section.}

\item{\code{rowDataMap}}{See Slots section.}

\item{\code{colDataMap}}{See Slots section.}

\item{\code{assayMap}}{See Slots section.}

\item{\code{metadataMap}}{See Slots section.}
}}

\section{Slots}{

\itemize{
\item rowDataMap: A list-like object containing two \code{character} vectors.
The first is column names in \code{rawdata} needed to uniquely identify each row,
the second is additional columns which map to rows, but are not required to
uniquely identify them. Rows should be drugs.
\item colDataMap: A list-like object containing two \code{character} vectors.
The first is column names in \code{rawdata} needed to uniquely identify each
column, the second is additional columns which map to rows, but are not
required to uniquely identify them. Columns should be samples.
\item assayMap A list-like where each item is a \code{list} with two elements
specifying an assay, the first being the identifier columns in \code{rawdata}
needed to uniquely identify each row an assay, and the second a list of
\code{rawdata} columns to be mapped to that assay. The names of \code{assayMap}
will be the names of the assays in the \code{LongTable} that is created when
calling \code{metaConstruct} on this \code{DataMapper} object.
\item metadataMap: A list-like where each item is a \code{character} vector of
\code{rawdata} column names to assign to the \verb{@metadata} of the \code{LongTable},
where the name of that assay is the name of the list item. If names are
omitted, assays will be numbered by their index in the list.
}


\itemize{
\item rawdata: A list-like object containing one or more pieces of raw data
that will be processed and mapped to the slots of an \code{S4} object.
\item metadata: A \code{List} of object level metadata.
}
}

\examples{
show(exampleDataMapper)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LongTableDataMapper-class.R
\name{LongTableDataMapper}
\alias{LongTableDataMapper}
\title{Constructor for the \code{LongTableDataMapper} class, which maps from one or
more raw experimental data files to the slots of a \code{LongTable} object.}
\usage{
LongTableDataMapper(
  rawdata = data.frame(),
  rowDataMap = list(character(), character()),
  colDataMap = list(character(), character()),
  assayMap = list(list(character(), character())),
  metadataMap = list(character())
)
}
\arguments{
\item{rawdata}{A \code{data.frame} of raw data from a treatment response
experiment. This will be coerced to a \code{data.table} internally. We recommend
using joins to aggregate your raw data if it is not present in a single file.}

\item{rowDataMap}{A list-like object containing two \code{character} vectors.
The first is column names in \code{rawdata} needed to uniquely identify each row,
the second is additional columns which map to rows, but are not required to
uniquely identify them. Rows should be drugs.}

\item{colDataMap}{A list-like object containing two \code{character} vectors.
The first is column names in \code{rawdata} needed to uniquely identify each
column, the second is additional columns which map to rows, but are not
required to uniquely identify them. Columns should be samples.}

\item{assayMap}{A list-like where each item is a \code{list} with two \code{character}
vectors defining an assay, the first containing the identifier columns in
\code{rawdata} needed to uniquely identify each row an assay, and the second the
\code{rawdata} columns to be mapped to that assay. The names of \code{assayMap}
will be the names of the assays in the \code{LongTable} that is created when
calling \code{metaConstruct} on this \code{DataMapper} object. If the character vectors
have names, the value columns will be renamed accordingly.}

\item{metadataMap}{A list-like where each item is a \code{character} vector of
\code{rawdata} column names to assign to the \verb{@metadata} of the \code{LongTable},
where the name of that assay is the name of the list item. If names are
omitted, assays will be numbered by their index in the list.}
}
\value{
A \code{LongTable} object, with columns mapped to it's slots according
to the various maps in the \code{LongTableDataMapper} object.
}
\description{
Constructor for the \code{LongTableDataMapper} class, which maps from one or
more raw experimental data files to the slots of a \code{LongTable} object.
}
\details{
The \code{guessMapping} method can be used to test hypotheses about the
cardinality of one or more sets of identifier columns. This is helpful
to determine the id columns for \code{rowDataMap} and \code{colDataMap}, as well
as identify columns mapping to \code{assays} or \code{metadata}.

To attach metadata not associated with \code{rawdata}, please use the \code{metadata}
assignment method on your \code{LongTableDataMapper}. This metadata will be
merged with any metadata from \code{metadataMap} and added to the \code{LongTable}
which this object ultimately constructs.
}
\examples{
data(exampleDataMapper)
exampleDataMapper

}
\seealso{
\code{\link{guessMapping}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/allGenerics.R
\name{getIntern<-}
\alias{getIntern<-}
\title{Set the internal structural metadata for an S4 class}
\usage{
getIntern(object, ...) <- value
}
\arguments{
\item{object}{An R object to update internal structural metadata for.}

\item{value}{An \code{immutable_list} object, being a class union between \code{list}
and \code{immutable} S3 classes.}
}
\value{
Updates the object and returns invisibly.
}
\description{
Set the internal structural metadata for an S4 class
}
\examples{
print("Generics shouldn't need examples?")

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DataMapper-class.R
\name{DataMapper-accessors}
\alias{DataMapper-accessors}
\alias{rawdata,DataMapper-method}
\alias{rawdata}
\alias{rawdata<-,DataMapper,ANY-method}
\alias{rawdata<-}
\title{Accessing and modifying data in a \code{DataMapper} object.}
\usage{
\S4method{rawdata}{DataMapper}(object)

\S4method{rawdata}{DataMapper,ANY}(object) <- value
}
\arguments{
\item{object}{A \code{DataMapper} object to get or set data from.}

\item{value}{A \code{list}-like object to assign to the rawdata slot. Should be
a \code{data.frame} or \code{data.table} with the current implementation.}
}
\value{
Accessors: See details

Setters: An update \code{DataMapper} object, returned invisibly.
}
\description{
Documentation for the various setters and getters which allow manipulation
of data in the slots of a \code{DataMapper} object.
}
\details{
\strong{rawdata}: Get the raw data slot from a \code{DataMapper} object. Returns
a list-like containing one or more raw data inputs to the
\code{DataMapper} object.

\strong{rawdata}: Set the raw data slot from a \code{DataMapper} object.
\strong{value}: The \code{list}-like object to set for the rawdata slot. Note: this
currently only supports \code{data.frame} or \code{data.table} objects.
}
\seealso{
Other DataMapper-accessors: 
\code{\link{LongTableDataMapper-accessors}},
\code{\link{TREDataMapper-accessors}}
}
\concept{DataMapper-accessors}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deprecated.R
\name{buildLongTable,list-method}
\alias{buildLongTable,list-method}
\title{LongTable build method from list}
\usage{
\S4method{buildLongTable}{list}(from, rowDataCols, colDataCols, assayCols)
}
\arguments{
\item{from}{\code{list} A list containing any combination of character file paths,
data.tables and data.frames which will be used to construct the LongTable.}

\item{rowDataCols}{\code{list} List with two \code{character} vectors, the first
specifying one or more columns to be used as cell identifiers (e.g.,
cell-line name columns) and the second containing any additional metadata
columns related to the cell identifiers.}

\item{colDataCols}{\code{list} List with two \code{character} vectors, the first
specifying one or more columns to be used as column identifiers (e.g.,
drug name columns) and the second containing any additional metadata
columns related to the column identifiers.}

\item{assayCols}{\code{list} A named list of character vectors specifying how to
parse assay columns into a list of \code{data.table}s. Each list data.table
will be named for the name of corresponding list item and contain the columns
specified in the character vector of column names in each list item.}
}
\value{
A \code{LongTable} object constructed with the data in \code{from}.
}
\description{
Create a LongTable object from a list containing file paths,
data.frames and data.tables.
}
\examples{
\dontrun{
assayList <- assays(merckLongTable, withDimnames=TRUE)
rowDataCols <- list(rowIDs(merckLongTable), rowMeta(merckLongTable))
colDataCols <- list(colIDs(merckLongTable), colMeta(merckLongTable))
assayCols <- assayCols(merckLongTable)
longTable <- buildLongTable(from=assayList, rowDataCols, colDataCols, assayCols)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{.symSetDiffList}
\alias{.symSetDiffList}
\title{Utility to find the symmetric set difference of a list of two or more
vectors or lists}
\usage{
.symSetDiffList(...)
}
\arguments{
\item{...}{A list of or any number of vector like objects of the same mode,
which could also be operated on by the native R set operations}
}
\value{
A vector like object of the same mode as the first argument,
containing only the symmetric set difference
}
\description{
The function finds the symmetric set differnces between all the arguments,
defined as Union(args)-Intersection(args)
}
\examples{
list1 <- list('a', 'b', 'c')
list2 <- list('a', 'c')
list3 <- list('a', 'c', 'd')
listAll <- .symSetDiffList(list1, list2, list3)
listAll

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{.intersectList}
\alias{.intersectList}
\title{Intersect A List of More Than Two Vectors}
\usage{
.intersectList(...)
}
\arguments{
\item{...}{A list of or any number of vector like objects of the same mode,
which could also be operated on by the native R set operations}
}
\value{
A vector like object of the same mode as the first argument,
containing only the intersection common to all arguments to the function
}
\description{
Utility to find the intersection between a list of more than two vectors or
lists This function extends the native intersect function to work on two
or more arguments.
}
\examples{
list1 <- list('a', 'b', 'c')
list2 <- list('a', 'c')
list3 <- list('a', 'c', 'd')
listAll <- .intersectList(list1, list2, list3)
listAll

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mergeAssays-method.R
\name{mergeAssays,LongTable-method}
\alias{mergeAssays,LongTable-method}
\title{Endomorphically merge assays within a \code{LongTable} or inheriting class}
\usage{
\S4method{mergeAssays}{LongTable}(object, x, y, target = x, ..., metadata = FALSE)
}
\arguments{
\item{object}{A \code{LongTable} or inheriting class.}

\item{x}{\code{character(1)} A valid assay name in \code{object}.}

\item{y}{\code{character(1)} A valid assay name in \code{object}.}

\item{target}{\code{character(1)} Name of the assay to assign the result to.
Can be a new or existing assay. Defaults to \code{x}.}

\item{...}{Fallthrough arguments to merge.data.table to specify the join
type. Use this to specify which columns to merge on. If excluded, defaults
to by=assayKeys(objecty, y).}

\item{metadata}{\code{logical} A logical vector indicating whether to attach
metadata to either assay before the merge occurs. If only one value is
passed that value is used for both assays. Defaults to \code{FALSE}.}
}
\value{
A copy of \code{object} with assays \code{x} and \code{y} merged and assigned to
\code{target}.
}
\description{
Endomorphically merge assays within a \code{LongTable} or inheriting class
}
\seealso{
\code{\link{merge.data.table}}
}
\author{
Christopher Eeles
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LongTable-class.R
\name{LongTable}
\alias{LongTable}
\title{LongTable constructor method}
\usage{
LongTable(
  rowData,
  rowIDs,
  colData,
  colIDs,
  assays,
  assayIDs,
  metadata = list(),
  keep.rownames = FALSE
)
}
\arguments{
\item{rowData}{\code{data.frame} A rectangular object coercible to a \code{data.table}.}

\item{rowIDs}{\code{character} A vector of \code{rowData} column names needed to
uniquely identify each row in a \code{LongTable}.}

\item{colData}{\code{data.frame} A rectangular object coercible to a \code{data.table.}}

\item{colIDs}{\code{chacter} A vector of \code{colData} column names needed to uniquely
identify each column in a \code{LongTable}.}

\item{assays}{\code{list} A list of rectangular objects, each coercible to
a \code{data.table}. Must be named and item names must match the \code{assayIDs}
list.}

\item{assayIDs}{\code{list} A list of \code{character} vectors specifying the columns
needed to uniquely identify each row in an \code{assay}. Names must match the
\code{assays} list.}

\item{metadata}{\code{list} A list of one or more metadata items associated with
a LongTable experiment.}

\item{keep.rownames}{\code{logical(1)} or \code{character(1)} Should rownames be
retained when coercing to \code{data.table} inside the constructor. Default
is FALSE. If TRUE, adds a 'rn' column to each rectangular object that
gets coerced from \code{data.frame} to \code{data.table}. If a string, that becomes
the name of the rownames column.}
}
\value{
A \code{LongTable} object containing the data for a treatment response
experiment and configured according to the rowIDs and colIDs arguments.
}
\description{
LongTable constructor method
}
\examples{
"See vignette('The LongTable Class', package='CoreGx')"

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-optimization.R
\name{drop_fn_params}
\alias{drop_fn_params}
\title{Drop parameters from a function and replace them with constants
inside the function body.}
\usage{
drop_fn_params(fn, args)
}
\arguments{
\item{fn}{\code{function} A non-primitive function to remove parameters from
(via \code{base::formals(fn)}).}

\item{args}{\code{list} A list where names are the function arguments (parameters)
to remove and the values are the appopriate value to replace the parameter
with in the function body.}
}
\value{
\code{function} A new non-primitize function with the parameters named in
\code{args} deleted and their values fixed with the values from \code{args} in the
function body.
}
\description{
Drop parameters from a function and replace them with constants
inside the function body.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LongTable-class.R, R/LongTable-accessors.R,
%   R/LongTable-utils.R, R/methods-dim.R, R/methods-dimnames.R
\docType{class}
\name{LongTable-class}
\alias{LongTable-class}
\alias{.LongTable}
\alias{rowIDs,LongTable-method}
\alias{rowMeta,LongTable-method}
\alias{colIDs,LongTable-method}
\alias{colMeta,LongTable-method}
\alias{idCols,LongTable-method}
\alias{assayIndex,LongTable-method}
\alias{assayKeys,LongTable-method}
\alias{assayCols,LongTable-method}
\alias{getIntern,LongTable,character-method}
\alias{getIntern,LongTable,missing-method}
\alias{rowData,LongTable-method}
\alias{rowData<-,LongTable-method}
\alias{colData,LongTable-method}
\alias{colData<-,LongTable,ANY-method}
\alias{assays,LongTable-method}
\alias{assays<-,LongTable,list-method}
\alias{assay,LongTable,ANY-method}
\alias{assay<-,LongTable,ANY-method}
\alias{assayNames,LongTable-method}
\alias{names,LongTable-method}
\alias{names}
\alias{[[,LongTable,ANY,ANY-method}
\alias{dim,LongTable-method}
\alias{colnames,LongTable-method}
\alias{rownames,LongTable-method}
\alias{dimnames,LongTable-method}
\title{LongTable class definition}
\usage{
\S4method{rowIDs}{LongTable}(object, data = FALSE, key = FALSE)

\S4method{rowMeta}{LongTable}(object, data = FALSE, key = FALSE)

\S4method{colIDs}{LongTable}(object, data = FALSE, key = FALSE)

\S4method{colMeta}{LongTable}(object, data = FALSE, key = FALSE)

\S4method{idCols}{LongTable}(object)

\S4method{assayIndex}{LongTable}(x)

\S4method{assayKeys}{LongTable}(x, i)

\S4method{assayCols}{LongTable}(object, i)

\S4method{getIntern}{LongTable,character}(object, x)

\S4method{getIntern}{LongTable,missing}(object, x)

\S4method{rowData}{LongTable}(x, key = FALSE, use.names = FALSE, ...)

\S4method{rowData}{LongTable}(x, ...) <- value

\S4method{colData}{LongTable}(x, key = FALSE, dimnames = FALSE, ...)

\S4method{colData}{LongTable,ANY}(x, ...) <- value

\S4method{assays}{LongTable}(
  x,
  withDimnames = TRUE,
  metadata = withDimnames,
  key = !withDimnames,
  ...
)

\S4method{assays}{LongTable,list}(x, withDimnames = TRUE, ...) <- value

\S4method{assay}{LongTable,ANY}(
  x,
  i,
  withDimnames = TRUE,
  summarize = withDimnames,
  metadata = !summarize,
  key = !(summarize || withDimnames),
  ...
)

\S4method{assay}{LongTable,ANY}(x, i) <- value

\S4method{assayNames}{LongTable}(x)

\S4method{[[}{LongTable,ANY,ANY}(x, i)

\S4method{dim}{LongTable}(x)

\S4method{colnames}{LongTable}(x)

\S4method{rownames}{LongTable}(x)

\S4method{dimnames}{LongTable}(x)
}
\arguments{
\item{object}{\code{LongTable}}

\item{data}{\code{logical} Should the colData for the metadata columns be returned
instead of the column names? Default is FALSE.}

\item{key}{\code{logical} Should the key columns also be returned? Defaults to
!withDimnames. This is incompatible with \code{summarize=TRUE}, which will
drop the key columns regardless of the value of this argument.}

\item{x}{The \code{LongTable} object to retrieve the dimnames for.}

\item{i}{\code{character(1)} name or \code{integer} index of the desired assay.}

\item{use.names}{\code{logical} This parameter is just here to stop matching
the positional argument to use.names from the rowData generic. It
doesn't do anything at this time and can be ignored.}

\item{...}{For developer use only! Pass raw=TRUE to return the slot for
modification by reference.}

\item{value}{A \code{data.frame} or \code{data.table} to update the assay data
with. This must at minumum contain the row and column data identifier
columns to allow correctly mapping the assay keys. We recommend modifying
the results returned by assay(longTable, 'assayName', withDimnames=TRUE).
For convenience, both the \code{[[} and \code{$} LongTable accessors return an assay
with the dimnames.}

\item{withDimnames}{\code{logical(1)} Should the dimension names be returned
joined to the assay. This retrieves both the row and column identifiers
and returns them joined to the assay. For}

\item{metadata}{\code{logical(1)} Should all of the metadata also be joined to
the assay. This is useful when modifying assays as the resulting list
has all the information needed to recreated the LongTable object. Defaults
to \code{withDimnames}.}

\item{summarize}{\code{logical(1)} If the assays is a summary where some of
\code{idCols(x)} are not in \code{assayKeys(x, i)}, then those missing columns
are dropped. Defaults to \code{FALSE}. When \code{metadata} is \code{TRUE}, only
metadata columns with 1:1 cardinality with the assay keys for \code{i}.}

\item{`x`}{A \code{LongTable} or inheriting class.}

\item{`i`}{An optional valid assay name or index in \code{x}.}
}
\value{
\code{LongTable} object containing the assay data from a treatment
response experiment

A \code{character} vector of rowData column names if data is FALSE,
otherwise a \code{data.table} with the data from the rowData id columns.

A \code{character} vector of rowData column names if data is FALSE,
otherwise a \code{data.table} with the data from the rowData metadta columns.

A \code{character} vector of colData column names if data is FALSE,
otherwise a \code{data.table} with the data from the colData id columns.

A \code{character} vector of colData column names if data is FALSE,
otherwise a \code{data.table} with the data from the colData metadta columns.

\code{character} A character vector containing the unique rowIDs and
colIDs in a LongTable object.

A \code{mutable} copy of the "assayIndex" for \code{x}

A \code{mutable} copy of the "assyKeys" for \code{x}

A \code{list} of \code{character} vectors containing the value column names for
each assay if i is missing, otherwise a \code{character} vector of value column
names for the selected assay.

\code{immutable} value of x if length(x) == 1 else named list of values
for all symbols in x.

An \code{immutable} list.

A \code{data.table} containing rowID, row identifiers, and row metadata.

A copy of the \code{LongTable} object with the \code{rowData}
slot updated.

A \code{data.table} containing row identifiers and metadata.

A copy of the \code{LongTable} object with the \code{colData}
slot updated.

A \code{list} of \code{data.table} objects, one per assay in the object.

A copy of the \code{LongTable} with the assays modified.

\code{LongTable} With updated assays slot.

\code{character} Names of the assays contained in the \code{LongTable}.

\code{numeric} Vector of object dimensions.

\code{character} Vector of column names.

\code{character} Vector of row names.

\code{list} List with two character vectors, one for row and one for
column names.
}
\description{
Define a private constructor method to be used to build a
\code{LongTable} object.

This is used as an alternative to R attributes for storing structural
metadata of an S4 objects.

Add or replace an assay in a LongTable by name. Currently
this function only works when the assay has all columns in row and column
data tables (i.e., when assays is retured withDimnames=TRUE).

Select an assay from within a LongTable object.
}
\section{Methods (by generic)}{
\itemize{
\item \code{rowMeta(LongTable)}: Get the names of the non-id columns from rowData.

\item \code{colIDs(LongTable)}: Get the names of the columns in colData required to
uniquely identify each row.

\item \code{colMeta(LongTable)}: Get the names of the non-id columns in the colData
\code{data.table}.

\item \code{idCols(LongTable)}: Get the names of all id columns.

\item \code{assayIndex(LongTable)}: Get the assayIndex item from the objects internal metadata.

\item \code{assayKeys(LongTable)}: Get the assayKeys item from the objects internal metadata.

\item \code{assayCols(LongTable)}: Get a list of column names for each assay in the object.

\item \code{getIntern(object = LongTable, x = character)}: Access structural metadata present within a
LongTable object. This is mostly for developmer use.

\item \code{getIntern(object = LongTable, x = missing)}: Access all structural metadata present within a
LongTable object. This is primarily for developmer use.

\item \code{rowData(LongTable)}: Get the row level annotations for a \code{LongTable} object.

\item \code{rowData(LongTable) <- value}: Update the row annotations for a \code{LongTable} object.
Currently requires that all columns in rowIDs(longTable) be present in
value.

\item \code{colData(LongTable)}: Get the column level annotations for a LongTable
object.

\item \code{colData(x = LongTable) <- value}: Update the colData of a LongTable object. Currently
requires that all of the colIDs(longTable) be in the value object.

\item \code{assays(LongTable)}: Get a list containing all the assays in a \code{LongTable}.

\item \code{assays(x = LongTable) <- value}: Update the assays in a LongTable object. The rowIDs
and colIDs must be present in all assays to allow successfully remapping
the keys. We recommend modifying the list returned by
assays(longTable, withDimnames=TRUE) and the reassigning to the
\code{LongTable}.

\item \code{assay(x = LongTable, i = ANY)}: Retrieve an assay \code{data.table} object from the
\code{assays} slot of a \code{LongTable} object.

\item \code{assay(x = LongTable, i = ANY) <- value}: 

\item \code{assayNames(LongTable)}: Return the names of the assays contained in a
\code{LongTable}

\item \code{x[[i}: Get an assay from a LongTable object. This method
returns the row and column annotations by default to make assignment
and aggregate operations easiers.

\item \code{dim(LongTable)}: Get the number of row annotations by the number of
column annotations from a LongTable object. Please note that row x columns
does not necessarily equal the number of rows in an assay, since it is
not required for each assay to have every row or column present.

\item \code{colnames(LongTable)}: Retrieve the pseudo-colnames of a LongTable object,
these are constructed by pasting together the colIDs(longTable) and
can be used in the subset method for regex based queries.

\item \code{rownames(LongTable)}: Retrieve the pseudo-rownames of a LongTable object,
these are constructed by pasting together the rowIDs(longTable) and
can be used in the subset method for regex based queries.

\item \code{dimnames(LongTable)}: Get the pseudo-dimnames for a LongTable object. See
colnames and rownames for more information.

}}
\section{Slots}{

\describe{
\item{\code{rowData}}{See Slots section.}

\item{\code{colData}}{See Slots section.}

\item{\code{assays}}{See Slots section.}

\item{\code{metadata}}{See Slots section.}

\item{\code{.intern}}{See Slots section.}
}}

\section{Slots}{

\itemize{
\item \emph{rowData}: A \code{data.table} containing the metadata associated with the
row dimension of a \code{LongTable}.
\item \emph{colData}: A \code{data.table} containing the metadata associated with the
column dimension of a \code{LongTable}.
\item \emph{assays}: A \code{list} of \code{data.table}s, one for each assay in a
\code{LongTable}.
\item \emph{metadata}: An optional \code{list} of additional metadata for a \code{LongTable}
which doesn't map to one of the dimensions.
\item \emph{.intern}: An \code{immutable} \code{list} that holds internal structural metadata
about a LongTable object, such as which columns are required to key
the object.
}
}

\examples{
rowIDs(merckLongTable)

rowMeta(merckLongTable)

colIDs(merckLongTable)

colMeta(merckLongTable)

idCols(merckLongTable)

assayIndex(nci_TRE_small)

assayKeys(nci_TRE_small)
assayKeys(nci_TRE_small, "sensitivity")
assayKeys(nci_TRE_small, 1)

assayCols(merckLongTable)

getIntern(merckLongTable, 'rowIDs')
getIntern(merckLongTable, c('colIDs', 'colMeta'))

getIntern(merckLongTable)

rowData(merckLongTable)

rowData(merckLongTable) <- rowData(merckLongTable)

colData(merckLongTable)

# Get the keys as well, mostly for internal use
colData(merckLongTable, key=TRUE)

colData(merckLongTable) <- colData(merckLongTable)

assays(merckLongTable)

assays(merckLongTable) <- assays(merckLongTable, withDimnames=TRUE)

# Default annotations, just the key columns
assay(merckLongTable, 'sensitivity')
assay(merckLongTable, 1)

# With identifiers joined
assay(merckLongTable, 'sensitivity', withDimnames=TRUE)

# With identifiers and metadata
assay(merckLongTable, 'profiles', withDimnames=TRUE, metadata=TRUE)

assay(merckLongTable, 'sensitivity') <-
     assay(merckLongTable, 'sensitivity', withDimnames=TRUE)
assay(merckLongTable, 'sensitivity') <- merckLongTable$sensitivity

assayNames(merckLongTable)
names(merckLongTable)

merckLongTable[['sensitivity']]

dim(merckLongTable)

dim(merckLongTable)

head(colnames(merckLongTable))

head(rownames(merckLongTable))

lapply(dimnames(merckLongTable), head)

}
\concept{LongTable-accessors}
\concept{LongTable-class}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/allGenerics.R
\name{summarizeMolecularProfiles}
\alias{summarizeMolecularProfiles}
\title{Summarize molecular profile data such that there is a single entry for each
sample line/treatment combination}
\usage{
summarizeMolecularProfiles(object, ...)
}
\arguments{
\item{object}{An \code{S4} object to summarize the molecular profiles for.}

\item{...}{Allow definition of new arguments to this generic}
}
\value{
Depends on the implemented method
}
\description{
Summarize molecular profile data such that there is a single entry for each
sample line/treatment combination
}
\examples{
print("Generics shouldn't need examples?")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/datasets.R
\docType{data}
\name{nci_TRE_small}
\alias{nci_TRE_small}
\title{NCI-ALMANAC Drug Combination Data TreatmentResponseExperiment Subset}
\format{
TreatmentResponseExperiment object
}
\usage{
data(nci_TRE_small)
}
\description{
This is a \code{TreatmentResponseExperiment} object containing a subset of
NCI-ALMANAC drug combination screening data,
with 2347 unique treatment combinations on 10 cancer cell lines selected.
}
\references{
Susan L. Holbeck, Richard Camalier, James A. Crowell, Jeevan Prasaad Govindharajulu, Melinda Hollingshead, Lawrence W. Anderson, Eric Polley, Larry Rubinstein, Apurva Srivastava, Deborah Wilsker, Jerry M. Collins, James H. Doroshow; The National Cancer Institute ALMANAC: A Comprehensive Screening Resource for the Detection of Anticancer Drug Pairs with Enhanced Therapeutic Activity. Cancer Res 1 July 2017; 77 (13): 3564–3576. https://doi.org/10.1158/0008-5472.CAN-17-0489
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LongTable-accessors.R
\name{metadata<-,LongTable-method}
\alias{metadata<-,LongTable-method}
\title{Setter method for the metadata slot of a \code{LongTable} object}
\usage{
\S4method{metadata}{LongTable}(x) <- value
}
\arguments{
\item{x}{\code{LongTable} The LongTable to update}

\item{value}{\code{list} A list of new metadata associated with a \code{LongTable}
object.}
}
\value{
\code{LongTable} A copy of the \code{LongTable} object with the \code{value} in
the metadata slot.
}
\description{
Setter method for the metadata slot of a \code{LongTable} object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-optimization.R
\name{collect_fn_params}
\alias{collect_fn_params}
\title{Collects all function arguments other than the first into a single list
parameter.}
\usage{
collect_fn_params(fn)
}
\arguments{
\item{fn}{\code{function} A non-primitive function to refactor such that the first
argument becomes the second argument and all other parameters must be
passed as a vector to the first argument of the new function via the \code{par}
parameter.}
}
\value{
\code{function} A new non-primitive function where the first argument is
\code{par}, which takes a vector of parameters being optimized, and the
second argument is the old first argument to \code{fn} (usually \code{x} since this
is the independent variable to optimize the function over).
}
\description{
Useful for converting a regular function into a function amenable to
optimization via \code{stats::optim}, which requires all free parameters be
passed as a single vector \code{par}.
}
\details{
Takes a function of the form f(x, ...), where ... is any number of additional
function parameters (bot not literal \code{...}!) and parses it to a function of
the form f(par, x) where \code{par} is a vector of values for ... in
the same order as the arguments appear in \code{fn}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cosinePerm.R
\name{cosinePerm}
\alias{cosinePerm}
\title{Cosine Permutations}
\usage{
cosinePerm(
  x,
  y,
  nperm = 1000,
  alternative = c("two.sided", "less", "greater"),
  include.perm = FALSE,
  nthread = 1,
  ...
)
}
\arguments{
\item{x}{\code{factor} is the factors for the first variable}

\item{y}{\code{factor} is the factors for the second variable}

\item{nperm}{\code{integer} is the number of permutations to compute the null
distribution of MCC estimates}

\item{alternative}{\code{string} indicates the alternative hypothesis and
must be one of ‘'two.sided'’, ‘'greater'’ or ‘'less'’.  You can specify
just the initial letter.  ‘'greater'’ corresponds to positive association,
‘'less'’ to negative association. Options are 'two.sided', 'less', or
'greater'}

\item{include.perm}{\code{boolean} indicates whether the estimates for the
null distribution should be returned. Default set to 'FALSE'}

\item{nthread}{\code{integer} is the number of threads to be used to perform
the permutations in parallel}

\item{...}{A \code{list} of fallthrough parameters}
}
\value{
A \code{list} estimate of the cosine similarity, p-value and
estimates after random permutations (null distribution) in include.perm is
set to 'TRUE'
}
\description{
Computes the cosine similarity and significance using permutation test. This
function uses random numbers, to ensure reproducibility please call
\code{set.seed()} before running the function.
}
\examples{
x <- factor(c(1,2,1,2,1))
y <- factor(c(2,2,1,1,1))
cosinePerm(x, y)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TreatmentResponseExperiment-class.R,
%   R/methods-coerce.R
\name{as}
\alias{as}
\title{Coerce a \code{LongTable} to a \code{TreatmentResponseExperiment}}
\arguments{
\item{from}{A \code{LongTableDataMapper} to coerce.}
}
\value{
The data in \code{object}, as the child-class
\code{TreatmentResponseExperiment}.

A \code{data.table} with the data from a LongTable.

\code{data.table} containing the data from the LongTable, with the
`longTableDataMapper' attribute containg the metadata needed to reverse
the coercing operation.

\code{LongTable} object configured with the longTableDataMapper

\code{data.table} with long format of data in \code{from}

\code{data.frame} with long format of data in \code{from}.

\code{SummarizedExperiment} with each assay as a \code{BumpyMatrix}

A \code{TREDataMapper} object.
}
\description{
Coerce a LongTable into a \code{data.table}.

Currently only supports coercing to data.table or data.frame

Coerce a data.table with the proper configuration attributes
back to a LongTable
}
\examples{
data(clevelandSmall_cSet)
TRE <- as(treatmentResponse(clevelandSmall_cSet),
    "TreatmentResponseExperiment")
TRE

as(merckLongTable, 'data.table')

dataTable <- as(merckLongTable, 'data.table')
print(attr(dataTable, 'longTableDataMapper')) # Method doesn't work without this
as(dataTable, 'LongTable')

SE <- molecularProfilesSlot(clevelandSmall_cSet)[[1]]
as(SE, 'data.table')

SE <- molecularProfilesSlot(clevelandSmall_cSet)[[1]]
as(SE, 'data.frame')

}
\seealso{
\code{\link{TreatmentResponseExperiment}}

\code{\link[BumpyMatrix:BumpyMatrix]{BumpyMatrix::BumpyMatrix}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{.unionList}
\alias{.unionList}
\title{Utility to find the union between a list of more than two vectors or
lists}
\usage{
.unionList(...)
}
\arguments{
\item{...}{A list of or any number of vector like objects of the same mode,
which could also be operated on by the native R set operations}
}
\value{
A vector like object of the same mode as the first argument,
containing all the elements of all arguments passed to the function
}
\description{
This function extends the native union function to work on two or more
arguments.
}
\examples{
list1 <- list('a', 'b')
list2 <- list('a', 'c')
list3 <- list('c', 'd')
listAll <- .unionList(list1, list2, list3)
listAll

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CoreSet-accessors.R
\name{CoreSet-accessors}
\alias{CoreSet-accessors}
\alias{annotation,CoreSet-method}
\alias{annotation}
\alias{annotation<-,CoreSet,list-method}
\alias{annotation<-}
\alias{dateCreated,CoreSet-method}
\alias{dateCreated}
\alias{dateCreated<-,CoreSet,character-method}
\alias{dateCreated<-,CoreSet-method}
\alias{dateCreated<-}
\alias{name,CoreSet-method}
\alias{name}
\alias{name<-,CoreSet-method}
\alias{name<-,CoreSet,character-method}
\alias{name<-}
\alias{sampleInfo,CoreSet-method}
\alias{sampleInfo}
\alias{cellInfo,CoreSet-method}
\alias{cellInfo}
\alias{sampleInfo<-,CoreSet,data.frame-method}
\alias{sampleInfo<-}
\alias{cellInfo<-,CoreSet,data.frame-method}
\alias{cellInfo<-}
\alias{sampleNames,CoreSet-method}
\alias{sampleName,CoreSet-method}
\alias{sampleNames}
\alias{cellName,CoreSet-method}
\alias{cellNames}
\alias{sampleNames<-,CoreSet,character-method}
\alias{sampleNames<-,CoreSet,list-method}
\alias{sampleNames<-}
\alias{cellNames<-,CoreSet,list-method}
\alias{cellNames<-}
\alias{treatmentInfo,CoreSet-method}
\alias{treatmentInfo}
\alias{treatmentInfo<-,CoreSet,data.frame-method}
\alias{treatmentInfo<-}
\alias{treatmentNames,CoreSet-method}
\alias{treatmentNames}
\alias{treatmentNames<-,CoreSet,character-method}
\alias{treatmentNames<-}
\alias{curation,CoreSet-method}
\alias{curation}
\alias{curation<-,CoreSet,list-method}
\alias{curation<-}
\alias{datasetType,CoreSet-method}
\alias{datasetType}
\alias{datasetType<-,CoreSet,character-method}
\alias{datasetType<-}
\alias{molecularProfiles,CoreSet-method}
\alias{molecularProfiles}
\alias{molecularProfiles<-,CoreSet,character,character,matrix-method}
\alias{molecularProfiles<-,CoreSet,character,missing,matrix-method}
\alias{molecularProfiles<-,CoreSet,missing,missing,list-method}
\alias{molecularProfiles<-,CoreSet,missing,missing,MutliAssayExperiment-method}
\alias{molecularProfiles<-}
\alias{molecularProfiles<-,CoreSet,missing,missing,list_OR_MAE-method}
\alias{featureInfo,CoreSet-method}
\alias{featureInfo}
\alias{featureInfo<-,CoreSet,character,data.frame-method}
\alias{featureInfo<-,CoreSet,character,DataFrame-method}
\alias{featureInfo<-}
\alias{phenoInfo,CoreSet,character-method}
\alias{phenoInfo}
\alias{phenoInfo<-,CoreSet,character,data.frame-method}
\alias{phenoInfo<-,CoreSet,character,DataFrame-method}
\alias{phenoInfo<-}
\alias{fNames,CoreSet,character-method}
\alias{fNames}
\alias{fNames<-,CoreSet,character,character-method}
\alias{fNames<-}
\alias{mDataNames,CoreSet-method}
\alias{mDataNames}
\alias{mDataNames<-,CoreSet-method}
\alias{mDataNames<-,CoreSet,ANY-method}
\alias{mDataNames<-}
\alias{molecularProfilesSlot,CoreSet-method}
\alias{moleculerProfilesSlot,CoreSet-method}
\alias{molecularProfilesSlot}
\alias{molecularProfilesSlot<-,CoreSet,list_OR_MAE-method}
\alias{molecularProfilesSlot<-,CoreSet,list-method}
\alias{molecularProfilesSlot<-CoreSet,MultiAssayExperiment-method}
\alias{molecularProfilesSlot<-}
\alias{sensitivityInfo,CoreSet-method}
\alias{sensitivityInfo,CoreSet,missing-method}
\alias{sensitivityInfo,CoreSet,character-method}
\alias{sensitivityInfo<-,CoreSet,data.frame-method}
\alias{sensitivityInfo<-,CoreSet,missing,data.frame-method}
\alias{sensitvityInfo<-,CoreSet,character,data.frame-method}
\alias{sensitivityMeasures,CoreSet-method}
\alias{sensitivityMeasures<-,CoreSet,character-method}
\alias{sensitivityProfiles,CoreSet-method}
\alias{sensitivityProfiles<-,CoreSet,data.frame-method}
\alias{sensitivityRaw,CoreSet-method}
\alias{sensitivityRaw<-,CoreSet,array-method}
\alias{treatmentResponse,CoreSet-method}
\alias{treatmentResponse}
\alias{sensitivitySlot}
\alias{treatmentResponse<-,CoreSet,list_OR_LongTable-method}
\alias{treatmentResponse<-}
\alias{treamentResponse<-,CoreSet,list-method}
\alias{treatmentResponse<-,CoreSet,LongTable-method}
\alias{sensitivitySlot<-}
\alias{sensNumber,CoreSet-method}
\alias{sensNumber}
\alias{sensNumber<-,CoreSet,matrix-method}
\alias{sensNumber<-}
\alias{pertNumber,CoreSet-method}
\alias{pertNumber}
\alias{pertNumber<-,CoreSet,array-method}
\alias{pertNumber<-}
\title{Accessing and modifying information in a \code{CoreSet}}
\usage{
\S4method{annotation}{CoreSet}(object)

\S4method{annotation}{CoreSet,list}(object) <- value

\S4method{dateCreated}{CoreSet}(object)

\S4method{dateCreated}{CoreSet,character}(object) <- value

\S4method{name}{CoreSet}(object)

\S4method{name}{CoreSet}(object) <- value

\S4method{sampleInfo}{CoreSet}(object)

\S4method{sampleInfo}{CoreSet,data.frame}(object) <- value

\S4method{sampleNames}{CoreSet}(object)

\S4method{sampleNames}{CoreSet,character}(object) <- value

\S4method{treatmentInfo}{CoreSet}(object)

\S4method{treatmentInfo}{CoreSet,data.frame}(object) <- value

\S4method{treatmentNames}{CoreSet}(object)

\S4method{treatmentNames}{CoreSet,character}(object) <- value

\S4method{curation}{CoreSet}(object)

\S4method{curation}{CoreSet,list}(object) <- value

\S4method{datasetType}{CoreSet}(object)

\S4method{datasetType}{CoreSet,character}(object) <- value

\S4method{molecularProfiles}{CoreSet}(object, mDataType, assay)

\S4method{molecularProfiles}{CoreSet,character,character,matrix}(object, mDataType, assay) <- value

\S4method{molecularProfiles}{CoreSet,character,missing,matrix}(object, mDataType, assay) <- value

\S4method{molecularProfiles}{CoreSet,missing,missing,list_OR_MAE}(object, mDataType, assay) <- value

\S4method{featureInfo}{CoreSet}(object, mDataType)

\S4method{featureInfo}{CoreSet,character,data.frame}(object, mDataType) <- value

\S4method{phenoInfo}{CoreSet,character}(object, mDataType)

\S4method{phenoInfo}{CoreSet,character,data.frame}(object, mDataType) <- value

\S4method{fNames}{CoreSet,character}(object, mDataType)

\S4method{fNames}{CoreSet,character,character}(object, mDataType) <- value

\S4method{mDataNames}{CoreSet}(object)

\S4method{mDataNames}{CoreSet}(object) <- value

\S4method{molecularProfilesSlot}{CoreSet}(object)

\S4method{molecularProfilesSlot}{CoreSet,list_OR_MAE}(object) <- value

\S4method{sensitivityInfo}{CoreSet}(object, dimension, ...)

\S4method{sensitivityInfo}{CoreSet,data.frame}(object, dimension, ...) <- value

\S4method{sensitivityMeasures}{CoreSet}(object)

\S4method{sensitivityMeasures}{CoreSet,character}(object) <- value

\S4method{sensitivityProfiles}{CoreSet}(object)

\S4method{sensitivityProfiles}{CoreSet,data.frame}(object) <- value

\S4method{sensitivityRaw}{CoreSet}(object)

\S4method{sensitivityRaw}{CoreSet,array}(object) <- value

\S4method{treatmentResponse}{CoreSet}(object)

\S4method{treatmentResponse}{CoreSet,list_OR_LongTable}(object) <- value

\S4method{sensNumber}{CoreSet}(object)

\S4method{sensNumber}{CoreSet,matrix}(object) <- value

\S4method{pertNumber}{CoreSet}(object)

\S4method{pertNumber}{CoreSet,array}(object) <- value
}
\arguments{
\item{object}{A \code{CoreSet} object.}

\item{value}{See details.}

\item{mDataType}{\code{character(1)} The name of a molecular datatype to access
from the \code{molecularProfiles} of a \code{CoreSet} object.}

\item{assay}{\code{character(1)} A valid assay name in the \code{SummarizedExperiment}
of \verb{@molecularProfiles} of a CoreSet object for data type \code{mDataType}.}

\item{dimension}{See details.}

\item{...}{See details.}
}
\value{
Accessors: See details.

Setters: An updated \code{CoreSet} object, returned invisibly.
}
\description{
Documentation for the various setters and getters which allow manipulation
of data in the slots of a \code{CoreSet} object.
}
\details{
\subsection{@annotation}{

\strong{annotation}: A \code{list} of CoreSet annotations with items: 'name',
the name of the object; 'dateCreated', date the object was created; 'sessionInfo',
the \code{sessionInfo()} when the object was created; 'call', the R constructor call;
and 'version', the object version.
}

\strong{annotation<-}: Setter method for the annotation slot. Arguments:
\itemize{
\item value: a \code{list} of annotations to update the CoreSet with.
}

\subsection{@dateCreated}{

\strong{dateCreated}: \code{character(1)} The date the \code{CoreSet} object was
created, as returned by the \code{date()} function.
}

\strong{dateCreated<-}: Update the 'dateCreated' item in the \code{annotation} slot of
a \code{CoreSet} object. Arguments:
\itemize{
\item value: A \code{character(1)} vector, as returned by the \code{date()} function.
}

\strong{name}: \code{character(1)} The name of the \code{CoreSet}, retreived from
the \verb{@annotation} slot.

\strong{name<-}: Update the \verb{@annotation$name} value in a \code{CoreSet}  object.
\itemize{
\item value: \code{character(1)} The name of the \code{CoreSet} object.
}

\strong{cellInfo}: \code{data.frame} Metadata for all sample in a \code{CoreSet} object.

\strong{sampleInfo<-}: assign updated sample annotations to the \code{CoreSet}
object.
Arguments:
\itemize{
\item value: a \code{data.frame} object.
}

\strong{sampleNames}: \code{character} Retrieve the rownames of the \code{data.frame} in
the \code{sample} slot from a CoreSet object.

\strong{sampleNames<-}: assign new rownames to the sampleInfo \code{data.frame} for
a CoreSet object.
Arguments:
\itemize{
\item value: \code{character} vector of rownames for the \code{sampleInfo(object)} \code{data.frame}.
}

\strong{treatmentInfo}: \code{data.frame} Metadata for all treatments in a \code{CoreSet}
object. Arguments:
\itemize{
\item object: \code{CoreSet} An object to retrieve treatment metadata from.
}

\strong{treatmentInfo<-}: \code{CoreSet} object with updated treatment metadata.
object. Arguments:
\itemize{
\item object: \code{CoreSet} An object to set treatment metadata for.
\item value: \code{data.frame} A new table of treatment metadata for \code{object}.
}

\strong{treatmentNames}: \code{character} Names for all treatments in a \code{CoreSet}
object. Arguments:
\itemize{
\item object: \code{CoreSet} An object to retrieve treatment names from.
}

\strong{treatmentNames<-}: \code{CoreSet} Object with updates treatment names.
object. Arguments:
\itemize{
\item object: \code{CoreSet} An object to set treatment names from.
\item value: \code{character} A character vector of updated treatment names.
}

\subsection{@curation}{

\strong{curation}: A \code{list} of curated mappings between identifiers in the
CoreSet object and the original data publication. Contains two \code{data.frame}s, 'sample' with sample ids and
'tissue' with tissue ids.
}

\strong{curation<-}: Update the \code{curation} slot of a CoreSet object. Arugments:
\itemize{
\item value: A \code{list} of \code{data.frame}s, one for each type of curated
identifier. For a \code{CoreSet} object the slot should contain tissue and
sample id \code{data.frame}s.
}

\subsection{datasetType slot}{

\strong{datasetType}: \code{character(1)} The type treatment response in the
\code{sensitivity} slot. Valid values are 'sensitivity', 'perturbation' or 'both'.
}

\strong{datasetType<-}: Update the datasetType slot of a CoreSet object.
Arguments:
\itemize{
\item value: A \code{character(1)} vector with one of 'sensitivity', 'perturbation'
or 'both'
}

\subsection{@molecularProfiles}{

\strong{molecularProfiles}: \code{matrix()} Retrieve an assay in a
\code{SummarizedExperiment} from the \code{molecularProfiles} slot of a \code{CoreSet}
object with the specified \code{mDataType}. Valid \code{mDataType} arguments can be
found with \code{mDataNames(object)}. Exclude \code{mDataType} and \code{assay} to
access the entire slot. Arguments:
\itemize{
\item assay: Optional \code{character(1)} vector specifying an assay in the
\code{SummarizedExperiment} of the \code{molecularProfiles} slot of the
\code{CoreSet} object for the specified \code{mDataType}. If excluded,
defaults to modifying the first assay in the \code{SummarizedExperiment} for
the given \code{mDataType}.
}
}

\strong{molecularProfiles<-}: Update an assay in a \code{SummarizedExperiment} from
the \code{molecularProfiles} slot of a CoreSet object with the specified
\code{mDataType}. Valid \code{mDataType} arguments can be found with
\code{mDataNames(object)}. Omit \code{mDataType} and \code{assay} to update the slot.
\itemize{
\item assay: Optional \code{character(1)} vector specifying an assay in the
\code{SummarizedExperiment} of the \code{molecularProfiles} slot of the
\code{CoreSet} object for the specified \code{mDataType}. If excluded,
defaults to modifying the first assay in the \code{SummarizedExperiment} for
the given \code{mDataType}.
\item value: A \code{matrix} of values to assign to the \code{assay} slot of the
\code{SummarizedExperiment} for the selected \code{mDataType}. The rownames and
column names must match the associated \code{SummarizedExperiment}.
}

\strong{featureInfo}: Retrieve a \code{DataFrame} of feature metadata for the specified
\code{mDataType} from the \code{molecularProfiles} slot of a \code{CoreSet} object. More
specifically, retrieve the \verb{@rowData} slot from the \code{SummarizedExperiment}
from the \verb{@molecularProfiles} of a \code{CoreSet} object with the name
\code{mDataType}.

\strong{featureInfo<-}: Update the \code{featureInfo(object, mDataType)} \code{DataFrame}
with new feature metadata. Arguments:
\itemize{
\item value: A \code{data.frame} or \code{DataFrame} with updated feature metadata for
the specified molecular profile in the \code{molecularProfiles} slot of a
\code{CoreSet} object.
}

\strong{phenoInfo}: Return the \verb{@colData} slot from the \code{SummarizedExperiment} of
\code{mDataType}, containing sample-level metadata, from a \code{CoreSet} object.

\strong{phenoInfo<-}: Update the \verb{@colData} slot of the \code{SummarizedExperiment}
of \code{mDataType} in the \verb{@molecularProfiles} slot of a \code{CoreSet} object.
This updates the sample-level metadata in-place.
\itemize{
\item value: A \code{data.frame} or \code{DataFrame} object where rows are samples
and columns are sample metadata.
}

\strong{fNames}: \code{character()} The features names from the \code{rowData} slot of a
\code{SummarizedExperiment} of \code{mDataType} within a \code{CoreSet} object.

\strong{fNames}: Updates the rownames of the feature metadata (i.e., \code{rowData})
for a \code{SummarizedExperiment} of \code{mDataType} within a \code{CoreSet} object.
\itemize{
\item value: \code{character()} A character vector of new features names for the
\code{rowData} of the \code{SummarizedExperiment} of \code{mDataType} in the
\verb{@molecularProfiles} slot of a \code{CoreSet} object. Must be the same
length as \code{nrow(featureInfo(object, mDataType))},
the number of rows in the feature metadata.
}

\strong{mDataNames}: \code{character} Retrieve the names of the molecular data types
available in the \code{molecularProfiles} slot of a \code{CoreSet} object. These
are the options which can be used in the \code{mDataType} parameter of various
\code{molecularProfiles} slot accessors methods.

\strong{mDataNames}: Update the molecular data type names of the
\code{molecularProfiles} slot of a CoreSet object. Arguments:
\itemize{
\item value: \code{character} vector of molecular datatype names, with length
equal to \code{length(molecularProfilesSlot(object))}.
}

\strong{molecularProfilesSlot}: Return the contents of the \verb{@molecularProfiles}
slot of a \code{CoreSet} object. This will either be a \code{list} or
\code{MultiAssayExperiment} of \code{SummarizedExperiment}s.

\strong{molecularProfilesSlot<-}: Update the contents of the \verb{@molecularProfiles}
slot of a \code{CoreSet} object. Arguemnts:
\itemize{
\item value: A \code{list} or \code{MultiAssayExperiment} of \code{SummarizedExperiment}s. The
\code{list} and \code{assays} should be named for the molecular datatype in each
\code{SummarizedExperiment}.
}

\subsection{@treatmentResponse}{
\subsection{Arguments:}{
\itemize{
\item \code{dimension}: Optional \code{character(1)} One of 'treatment', 'sample' or
'assay' to retrieve \code{rowData}, \code{colData} or the 'assay_metadata' assay from
the \code{CoreSet} \verb{@sensitvity} \code{LongTable} object, respectively. Ignored with
warning if \verb{@treatmentResponse} is not a \code{LongTable} object.
\item \code{...}: Additional arguments to the \code{rowData} or \code{colData}.
\code{LongTable} methods. Only used if the sensitivity slot contains a
\code{LongTable} object instead of a \code{list} and the \code{dimension} argument is
specified.
}
}

\subsection{Methods:}{

\strong{sensitivityInfo}: \code{DataFrame} or \code{data.frame} of sensitivity treatment combo
by sample metadata for the \code{CoreSet} object. When the \code{dimension}
parameter is used, it allows retrieval of the dimension specific metadata
from the \code{LongTable} object in \verb{@treatmentResponse} of a CoreSet object.
}

}

\strong{sensitivityInfo}<-: Update the \verb{@treatmentResponse} slot metadata for a
\code{CoreSet} object. When used without the \code{dimension} argument is behaves
similar to the old CoreSet implementation, where the \verb{@treatmentResponse} slot
contained a list with a \verb{$info} \code{data.frame} item. When the \code{dimension}
arugment is used, more complicated assignments can occur where 'sample'
modifies the \verb{@sensitvity} \code{LongTable} colData, 'treatment' the rowData and
'assay' the 'assay_metadata' assay.
Arguments:
\itemize{
\item value: A \code{data.frame} of treatment response experiment metadata,
documenting experiment level metadata (mapping to treatments and samples). If
the \verb{@treatmentResponse} slot doesn't contain a \code{LongTable} and \code{dimension} is
not specified, you can only modify existing columns as returned by
\code{sensitivityInfo(object)}.
}

\strong{sensitivityMeaures}: Get the 'sensitivityMeasures' available in a \code{CoreSet}
object. Each measure reprents some summary of sample sensitivity to a given
treatment, such as ic50, ec50, AUC, AAC, etc. The results are returned as a
\code{character} vector with all available metrics for the PSet object.

\strong{sensitivityMeaures}: Update the sensitivity meaure in a \code{CoreSet}
object. Thesee values are the column names of the 'profiles' assay and
represent various compued sensitviity metrics such as ic50, ec50, AUC, AAC,
etc.
\itemize{
\item value: A \code{character} vector of new sensitivity measure names, the
then length of the character vector must matcht he number of columns of the
'profiles' assay, excluding metadata and key columns.
}

\strong{sensitivityProfiles}: Return the sensitivity profile summaries from the
sensitivity slot. This data.frame cotanins vaarious sensitivity summary
metrics, such as ic50, amax, EC50, aac, HS, etc as columns, with rows as
treatment by sample experiments.

\strong{sensitivityProfiles<-}: Update the sensitivity profile summaries the
sensitivity slot. Arguments:
-value: A \code{data.frame} the the same number of rows as as returned by
\code{sensitivityProfiles(object)}, but potentially modified columns, such as the
computation of additional summary metrics.

\strong{sensitivityRaw}: Access the raw sensitiity measurents for a CoreSet
object. A 3D \code{array} where rows are experiment_ids, columns are doses
and the third dimension is metric, either 'Dose' for the doses used or
'Viability' for the sample viability at that dose.

\strong{sensitvityRaw<-}: Update the raw dose and viability data in a \code{CoreSet}.
\itemize{
\item value: A 3D \code{array} object where rows are experiment_ids, columns are
replicates and pages are c('Dose', 'Viability'), with the corresponding
dose or viability measurement for that experiment_id and replicate.
}

\strong{sensNumber}: Return a count of viability observations in a \code{CoreSet}
object for each treatment-combo by sample combination.

\strong{sensNumber<-}: Update the 'n' item, which holds a matrix with a count
of treatment by sample-line experiment counts, in the \code{list} in \verb{@treatmentResponse}
slot of a \code{CoreSet} object. Will error when \verb{@sensitviity} contains
a \code{LongTable} object, since the counts are computed on the fly. Arguments:
\itemize{
\item value: A \code{matrix} where rows are samples and columns are treatments, with a
count of the number of experiments for each combination as the values.
}

\strong{pertNumber}: \code{array} Summary of available perturbation experiments
from in a \code{CoreSet} object. Returns a 3D \code{array} with the number of
perturbation experiments per treatment and sample, and data type.

\strong{pertNumber<-}: Update the \verb{@perturbation$n} value in a \code{CoreSet} object,
which stores a summary of the available perturbation experiments. Arguments:
\itemize{
\item value: A new 3D \code{array} with the number of perturbation experiments per
treatment and sample, and data type
}
}
\examples{
data(clevelandSmall_cSet)

## @annotation

annotation(clevelandSmall_cSet)

annotation(clevelandSmall_cSet) <- annotation(clevelandSmall_cSet)

dateCreated(clevelandSmall_cSet)

## dateCreated
dateCreated(clevelandSmall_cSet) <- date()

name(clevelandSmall_cSet)

name(clevelandSmall_cSet) <- 'new_name'

sampleInfo(clevelandSmall_cSet) <- sampleInfo(clevelandSmall_cSet)

sampleNames(clevelandSmall_cSet)

sampleNames(clevelandSmall_cSet) <- sampleNames(clevelandSmall_cSet)

treatmentInfo(clevelandSmall_cSet)

treatmentInfo(clevelandSmall_cSet) <- treatmentInfo(clevelandSmall_cSet)

treatmentNames(clevelandSmall_cSet)

treatmentNames(clevelandSmall_cSet) <- treatmentNames(clevelandSmall_cSet)

## curation
curation(clevelandSmall_cSet)

curation(clevelandSmall_cSet) <- curation(clevelandSmall_cSet)

datasetType(clevelandSmall_cSet)

datasetType(clevelandSmall_cSet) <- 'both'

# No assay specified
molecularProfiles(clevelandSmall_cSet, 'rna') <- molecularProfiles(clevelandSmall_cSet, 'rna')

# Specific assay
molecularProfiles(clevelandSmall_cSet, 'rna', 'exprs') <-
    molecularProfiles(clevelandSmall_cSet, 'rna', 'exprs')

# Replace the whole slot
molecularProfiles(clevelandSmall_cSet) <- molecularProfiles(clevelandSmall_cSet)

featureInfo(clevelandSmall_cSet, 'rna')

featureInfo(clevelandSmall_cSet, 'rna') <- featureInfo(clevelandSmall_cSet, 'rna')

phenoInfo(clevelandSmall_cSet, 'rna')

phenoInfo(clevelandSmall_cSet, 'rna') <- phenoInfo(clevelandSmall_cSet, 'rna')

fNames(clevelandSmall_cSet, 'rna')

fNames(clevelandSmall_cSet, 'rna') <- fNames(clevelandSmall_cSet, 'rna')

mDataNames(clevelandSmall_cSet)

mDataNames(clevelandSmall_cSet) <- mDataNames(clevelandSmall_cSet)

molecularProfilesSlot(clevelandSmall_cSet)

molecularProfilesSlot(clevelandSmall_cSet) <- molecularProfilesSlot(clevelandSmall_cSet)

sensitivityInfo(clevelandSmall_cSet)

sensitivityInfo(clevelandSmall_cSet) <- sensitivityInfo(clevelandSmall_cSet)

sensitivityMeasures(clevelandSmall_cSet) <- sensitivityMeasures(clevelandSmall_cSet)

sensitivityMeasures(clevelandSmall_cSet) <- sensitivityMeasures(clevelandSmall_cSet)

sensitivityProfiles(clevelandSmall_cSet)

sensitivityProfiles(clevelandSmall_cSet) <- sensitivityProfiles(clevelandSmall_cSet)

head(sensitivityRaw(clevelandSmall_cSet))

sensitivityRaw(clevelandSmall_cSet) <- sensitivityRaw(clevelandSmall_cSet)

treatmentResponse(clevelandSmall_cSet)

treatmentResponse(clevelandSmall_cSet) <- treatmentResponse(clevelandSmall_cSet)

sensNumber(clevelandSmall_cSet)

sensNumber(clevelandSmall_cSet) <- sensNumber(clevelandSmall_cSet)

pertNumber(clevelandSmall_cSet)

pertNumber(clevelandSmall_cSet) <- pertNumber(clevelandSmall_cSet)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LongTable-utils.R
\name{[[<-,LongTable,ANY,ANY-method}
\alias{[[<-,LongTable,ANY,ANY-method}
\title{\verb{[[<-} Method for LongTable Class}
\usage{
\S4method{[[}{LongTable,ANY,ANY}(x, i) <- value
}
\arguments{
\item{x}{A \code{LongTable} to update.}

\item{i}{The name of the assay to update, must be in \code{assayNames(object)}.}

\item{value}{A \code{data.frame}}
}
\value{
A \code{LongTable} object with the assay \code{i} updated using \code{value}.
}
\description{
Just a wrapper around assay<- for convenience. See
\verb{?'assay<-,LongTable,character-method'.}
}
\examples{
merckLongTable[['sensitivity']] <- merckLongTable[['sensitivity']]

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/allGenerics.R
\name{colIDs}
\alias{colIDs}
\title{Generic to access the row identifiers for an object.}
\usage{
colIDs(object, ...)
}
\arguments{
\item{object}{\code{S4} An object to get column id columns from.}

\item{...}{ALlow new arguments to this generic}
}
\value{
Depends on the implemented method.
}
\description{
Generic to access the row identifiers for an object.
}
\examples{
print("Generics shouldn't need examples?")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LongTable-utils.R
\name{$<-,LongTable-method}
\alias{$<-,LongTable-method}
\title{Update an assay from a LongTable object}
\usage{
\S4method{$}{LongTable}(x, name) <- value
}
\arguments{
\item{x}{A \code{LongTable} to update an assay for.}

\item{name}{\code{character(1)} The name of the assay to update}

\item{value}{A \code{data.frame} or \code{data.table} to update the assay with.}
}
\value{
Updates the assay \code{name} in \code{x} with \code{value}, returning an invisible
NULL.
}
\description{
Update an assay from a LongTable object
}
\examples{
merckLongTable$sensitivity <- merckLongTable$sensitivity

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CoreSet-class.R
\name{checkCsetStructure}
\alias{checkCsetStructure}
\title{A function to verify the structure of a CoreSet}
\usage{
checkCsetStructure(object, plotDist = FALSE, result.dir = tempdir())
}
\arguments{
\item{object}{A \code{CoreSet} to be verified}

\item{plotDist}{Should the function also plot the distribution of molecular
data?}

\item{result.dir}{The path to the directory for saving the plots as a string.
Defaults to this R sessions \code{tempdir()}.}
}
\value{
Prints out messages whenever describing the errors found in the
structure of the cSet object passed in.
}
\description{
This function checks the structure of a PharamcoSet, ensuring that the
correct annotations are in place and all the required slots are filled so
that matching of samples and drugs can be properly done across different types
of data and with other studies.
}
\examples{
checkCsetStructure(clevelandSmall_cSet)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LongTableDataMapper-accessors.R
\name{rowData,LongTableDataMapper-method}
\alias{rowData,LongTableDataMapper-method}
\title{Convenience method to subset the \code{rowData} out of the \code{rawdata} slot using
the assigned \code{rowDataMap} metadata.}
\usage{
\S4method{rowData}{LongTableDataMapper}(x, key = TRUE)
}
\arguments{
\item{x}{\code{LongTableDataMapper} object with valid data in the \code{rawdata} and
\code{colDataMap} slots.}

\item{key}{\code{logical(1)} Should the table be keyed according to the
\code{id_columns} of the \code{rowDataMap} slot? This will sort the table in memory.
Default is TRUE.}
}
\value{
\code{data.table} The \code{rowData} as specified in the \code{rowDataMap} slot.
}
\description{
Convenience method to subset the \code{rowData} out of the \code{rawdata} slot using
the assigned \code{rowDataMap} metadata.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods-coerce.R
\name{.assayToBumpyMatrix}
\alias{.assayToBumpyMatrix}
\title{Convert a LongTable assay into a BumpyMatrix object}
\usage{
.assayToBumpyMatrix(LT, assay, rows, cols, sparse = TRUE)
}
\arguments{
\item{LT}{\code{LongTable} with assay to convert into \code{BumpyMatrix}}

\item{assay}{\code{character(1)} A valid assay name in \code{LT}, as returned by
\code{assayNames(LT)}.}

\item{rows}{\code{character()} The rownames associated with the assay rowKey}

\item{cols}{\code{character()} The names associated with the assay colKey}

\item{sparse}{\code{logical(1)} Should the \code{BumpyMatrix} be sparse (i.e., is the
assay sparse).}
}
\value{
\code{BumpyMatrix} containing the data from \code{assay}.
}
\description{
Convert a LongTable assay into a BumpyMatrix object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods-guessMapping.R
\name{guessMapping}
\alias{guessMapping}
\title{Generic for Guessing the Mapping Between Some Raw Data and an S4 Object}
\usage{
guessMapping(object, ...)
}
\arguments{
\item{object}{An \code{S4} object containing so raw data to guess data to
object slot mappings for.}

\item{...}{Allow new arguments to be defined for this generic.}
}
\value{
A \code{list} with mapping guesses as items.
}
\description{
Generic for Guessing the Mapping Between Some Raw Data and an S4 Object
}
\examples{
"Generics shouldn't need examples!"

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LongTable-class.R
\name{printSlot}
\alias{printSlot}
\title{Helper function to print slot information}
\usage{
printSlot(slotName, slotData)
}
\arguments{
\item{slotName}{\code{character} The name of the slot to print.}

\item{slotData}{\code{data.table} The data to print.}
}
\description{
Helper function to print slot information
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{buildComboProfiles}
\alias{buildComboProfiles}
\title{Build an assay table with an \code{S4} object.}
\usage{
buildComboProfiles(object, ...)
}
\arguments{
\item{object}{\code{S4} An S4 object a list-like slot containing assays for the
object.}

\item{...}{Allow new arguments to be defined for this generic.}
}
\value{
\code{data.table}.
}
\description{
Build an assay table with an \code{S4} object.
}
\examples{
"This is a generic method!"

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LongTable-utils.R
\name{.}
\alias{.}
\title{Convenience function for converting R code to a call}
\usage{
.(...)
}
\arguments{
\item{...}{\code{pairlist} One or more R expressions to convert to calls.}
}
\value{
\code{call} An R call object containing the quoted expression.
}
\description{
This is used to pass through unevaluated R expressions into subset and
\code{[}, where they will be evaluated in the correct context.
}
\examples{
.(sample_line1 == 'A2058')

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimizeCoreGx.R
\name{optimizeCoreGx}
\alias{optimizeCoreGx}
\title{A helper method to find the best multithreading configuration for your
computer}
\usage{
optimizeCoreGx(sample_data, set = FALSE, report = !set)
}
\arguments{
\item{sample_data}{\code{TreatmentResponseExperiment}}

\item{set}{\code{logical(1)} Should the function modify your R environment
with the predicted optimal settings? This changes the global state of your
R session!}

\item{report}{\code{logical(1)} Should a \code{data.frame} of results be returned
by number of threads and operation be returned? Defaults to \code{!set}.}
}
\value{
If \code{set=TRUE}, modifies \code{data.table} threads via \code{setDTthreads()}, otherwise
displays a message indicating the optimal number of threads.
If \code{report=TRUE}, also returns a \code{data.frame} of the benchmark results.
}
\description{
A helper method to find the best multithreading configuration for your
computer
}
\examples{
\donttest{
  data(merckLongTable)
  optimizeCoreGx(merckLongTable)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{.convertCSetMolecularProfilesToSE}
\alias{.convertCSetMolecularProfilesToSE}
\title{CSet molecularProfiles from ESets to SEs}
\usage{
.convertCSetMolecularProfilesToSE(cSet)
}
\arguments{
\item{cSet}{\code{S4} A CoreSet containing molecular data in ExpressionSets}
}
\value{
\code{S4} A CoreSet containing molecular data in a
SummarizedExperiments
}
\description{
Converts all ExpressionSet objects within the molecularProfiles slot of a
CoreSet to SummarizedExperiments
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-optimization.R
\name{is_optim_compatible}
\alias{is_optim_compatible}
\title{Check whether a function signature is amenable to optimization via \code{stats::optim}.}
\usage{
is_optim_compatible(fn)
}
\arguments{
\item{fn}{\code{function} A non-primitive function.}
}
\value{
\code{logical(1)} \code{TRUE} if the first value of \code{formalArg(fn)} is "par",
otherwise \code{FALSE}.
}
\description{
Functions compatible with \code{optim} have the parameter named \code{par} as their
first formal argument where each value is a respective free parameter to
be optimized.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/immutable-class.R
\name{setOps-immutable}
\alias{setOps-immutable}
\alias{subset.immutable}
\alias{subset,}
\alias{[.immutable,}
\alias{[[.immutable,}
\alias{$.immutable}
\alias{[}
\alias{[.immutable}
\alias{[[}
\alias{[[.immutable}
\alias{$}
\title{Subset an immutable object, returning another immutable object.}
\usage{
subset.immutable(x, ...)

\method{[}{immutable}(x, ...)

\method{[[}{immutable}(x, ...)

\method{$}{immutable}(x, ...)
}
\arguments{
\item{x}{An R object inheriting from the "immutable" S3-class.}

\item{...}{Catch any additional parameters. Lets objects with arbitrary
dimensions be made immutable.}
}
\value{
An immutable subset of \code{x}.
}
\description{
Subset an immutable object, returning another immutable object.
}
\examples{
immut_mat <- immutable(matrix(1:100, 10, 10))
immut_mat[1:5, 1:5]

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/allGenerics.R
\name{reindex}
\alias{reindex}
\title{Generic method for resetting indexing in an S4 object}
\usage{
reindex(object, ...)
}
\arguments{
\item{object}{\code{S4} An object to redo indexing for}

\item{...}{\code{pairlist} Allow definition of new parameters to this generic.}
}
\value{
Depends on the implemented method
}
\description{
This method allows integer indexes used to maintain referential integrity
internal to an S4 object to be reset. This is useful particularly after
subsetting an object, as certain indexes may no longer be present in the
object data. Reindexing removes gaps integer indexes and ensures that the
smallest contiguous integer values are used in an objects indexes.
}
\examples{
print("Generics shouldn't need examples?")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LongTable-utils.R
\name{subset,LongTable-method}
\alias{subset,LongTable-method}
\title{Subset method for a LongTable object.}
\usage{
\S4method{subset}{LongTable}(x, i, j, assays = assayNames(x), reindex = TRUE)
}
\arguments{
\item{x}{\code{LongTable} The object to subset.}

\item{i}{\code{character}, \code{numeric}, \code{logical} or \code{call}
Character: pass in a character vector of rownames for the \code{LongTable} object
or a valid regex query which will be evaluated against the rownames.
Numeric or Logical: vector of indices or a logical vector to subset
the rows of a \code{LongTable}.
Call: Accepts valid query statements to the \code{data.table} i parameter,
this can be used to make complex queries using the \code{data.table} API
for the \code{rowData} data.table.}

\item{j}{\code{character}, \code{numeric}, \code{logical} or \code{call}
Character: pass in a character vector of colnames for the \code{LongTable} object
or a valid regex query which will be evaluated against the colnames.
Numeric or Logical: vector of indices or a logical vector to subset
the columns of a \code{LongTable}.
Call: Accepts valid query statements to the \code{data.table} i parameter,
this can be used to make complex queries using the \code{data.table} API
for the \code{colData} data.table.}

\item{assays}{\code{character}, \code{numeric} or \code{logical} Optional list of assay
names to subset. Can be used to subset the assays list further,
returning only the selected items in the new LongTable.}

\item{reindex}{\code{logical(1)} Should index values be reset such that they
are the smallest possible set of consecutive integers. Modifies the
"rowKey", "colKey", and all assayKey columns. Initial benchmarks indicate
\code{reindex=FALSE} saves ~20\% of both execution time and memory allocation. The
cost of reindexing decreases the smaller your subset gets.}
}
\value{
\code{LongTable} A new \code{LongTable} object subset based on the specified
parameters.
}
\description{
Allows use of the colData and rowData \code{data.table} objects to query based on
rowID and colID, which is then used to subset all assay \code{data.table}s stored
in the \code{assays} slot.
This function is endomorphic, it always returns a LongTable object.
}
\examples{
# Character
subset(merckLongTable, 'ABT-888', 'CAOV3')
# Numeric
subset(merckLongTable, 1, c(1, 2))
# Logical
subset(merckLongTable, , colData(merckLongTable)$sampleid == 'A2058')
# Call
subset(merckLongTable, drug1id == 'Dasatinib' & drug2id != '5-FU',
    sampleid == 'A2058')

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LongTable-class.R
\name{show,LongTable-method}
\alias{show,LongTable-method}
\title{Show method for the LongTable class}
\usage{
\S4method{show}{LongTable}(object)
}
\arguments{
\item{object}{A \code{LongTable} object to print the results for.}
}
\value{
\code{invisible} Prints to console.
}
\description{
Show method for the LongTable class
}
\examples{
show(merckLongTable)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods-drugSensitivitySig.R
\name{drugSensitivitySig}
\alias{drugSensitivitySig}
\title{Compute the correlation between a molecular feature and treatment response}
\usage{
drugSensitivitySig(object, ...)
}
\arguments{
\item{object}{An object inheriting form the \code{CoreGx::CoreSet} class}

\item{...}{Allow definition of new arguments to this generic}
}
\value{
A 3D array of genes x drugs x metric
}
\description{
Compute the correlation between a molecular feature and treatment response
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LongTable-utils.R
\name{[,LongTable,ANY,ANY,ANY-method}
\alias{[,LongTable,ANY,ANY,ANY-method}
\title{[ LongTable Method}
\usage{
\S4method{[}{LongTable,ANY,ANY,ANY}(x, i, j, assays = assayNames(x), ..., drop = FALSE)
}
\arguments{
\item{x}{\code{LongTable} The object to subset.}

\item{i}{\code{character}, \code{numeric}, \code{logical} or \code{call}
Character: pass in a character vector of drug names, which will subset the
object on all row id columns matching the vector. This parameter also
supports valid R regex query strings which will match on the colnames
of \code{x}. For convenience, * is converted to .* automatically. Colon
can be to denote a specific part of the colnames string to query.
Numeric or Logical: these select based on the rowKey from the \code{rowData}
method for the \code{LongTable}.
Call: Accepts valid query statements to the \code{data.table} i parameter as
a call object. We have provided the function .() to conveniently
convert raw R statements into a call for use in this function.}

\item{j}{\code{character}, \code{numeric}, \code{logical} or \code{call}
Character: pass in a character vector of drug names, which will subset the
object on all drug id columns matching the vector. This parameter also
supports regex queries. Colon can be to denote a specific part of the
colnames string to query.
Numeric or Logical: these select based on the rowID from the \code{rowData}
method for the \code{LongTable}.
Call: Accepts valid query statements to the \code{data.table} i parameter as
a call object. We have provided the function .() to conveniently
convert raw R statements into a call for use in this function.}

\item{assays}{\code{character} Names of assays which should be kept in the
\code{LongTable} after subsetting.}

\item{...}{Included to ensure drop can only be set by name.}

\item{drop}{\code{logical} Included for compatibility with the '[' primitive,
it defaults to FALSE and changing it does nothing.}
}
\value{
A \code{LongTable} containing only the data specified in the function
parameters.
}
\description{
Single bracket subsetting for a LongTable object. See subset for more details.
}
\details{
This function is endomorphic, it always returns a LongTable object.
}
\examples{
# Character
merckLongTable['ABT-888', 'CAOV3']
# Numeric
merckLongTable[1, c(1, 2)]
# Logical
merckLongTable[, colData(merckLongTable)$sampleid == 'A2058']
# Call
merckLongTable[
     .(drug1id == 'Dasatinib' & drug2id != '5-FU'),
     .(sampleid == 'A2058'),
 ]

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods-subsetTo.R
\name{subsetTo}
\alias{subsetTo}
\title{Subset a CoreSet object based on various parameters, such as cell lines, molecular features}
\usage{
subsetTo(object, ...)
}
\arguments{
\item{object}{An object inheriting from the \code{CoreGx::CoreSet} class}

\item{...}{Allow definition of new arguments to this generic}
}
\value{
A subsetted version of the original \code{object}
}
\description{
Subset a CoreSet object based on various parameters, such as cell lines, molecular features
}
\examples{
"Generics shouldn't need examples!"

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/allGenerics.R
\name{sensitivityRaw}
\alias{sensitivityRaw}
\title{sensitivityRaw Generic Method}
\usage{
sensitivityRaw(object, ...)
}
\arguments{
\item{object}{An \code{S4} object to extract the raw sensitivity experiment
data from.}

\item{...}{\code{pairlist}  Allow new parameters to be defined for this generic.}
}
\value{
Depends on the implemented method
}
\description{
Generic function to get the raw data array for a treatment response experiment
from an S4 class.
}
\examples{
print("Generics shouldn't need examples?")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/callingWaterfall.R
\name{callingWaterfall}
\alias{callingWaterfall}
\title{Drug sensitivity calling using waterfall plots}
\usage{
callingWaterfall(
  x,
  type = c("IC50", "AUC", "AMAX"),
  intermediate.fold = c(4, 1.2, 1.2),
  cor.min.linear = 0.95,
  name = "Drug",
  plot = FALSE
)
}
\arguments{
\item{x}{What type of object does this take in?}

\item{type}{ic50: IC50 values in micro molar (positive values)
actarea: Activity Area, that is area under the drug activity curve (positive values)
amax: Activity at max concentration (positive values)}

\item{intermediate.fold}{vector of fold changes used to define the intermediate sensitivities for ic50, actarea and amax respectively}

\item{cor.min.linear}{\code{numeric} The minimum linear correlation to
require?}

\item{name}{\code{character} The name of the output to use in plot}

\item{plot}{\code{boolean} Whether to plot the results}
}
\value{
\code{factor} Containing the drug sensitivity status of each
sample.
}
\description{
\enumerate{
\item Sensitivity calls were made using one of IC50, ActArea or Amax
}
}
\details{
\enumerate{
\item Sort log IC50s (or ActArea or Amax) of the samples to generate a
“waterfall distribution”
\item Identify cutoff:
}

3.1 If the waterfall distribution is non-linear (pearson cc to the linear
fit <=0.95), estimate the major inflection point of the log IC50 curve as
the point on the curve with the maximal distance to a line drawn between
the start and end points of the distribution.

3.2 If the waterfall distribution appears linear (pearson cc to the linear
fit > 0.95), then use the median IC50 instead.
\enumerate{
\item Samples within a 4-fold IC50 (or within a 1.2-fold ActArea or 20\% Amax
difference) difference centered around this inflection point are classified
as being “intermediate”,  samples with lower IC50s (or ActArea/Amax
values) than this range are defined as sensitive, and those with IC50s (or
ActArea/Amax) higher than this range are called “insensitive”.
\item Require at least x sensitive and x insensitive samples after applying
these criteria (x=5 in our case).
}
}
\examples{
# Dummy example
1 + 1

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LongTable-accessors.R
\name{LongTable-accessors}
\alias{LongTable-accessors}
\title{Accessing and modifying information in a \code{LongTable}}
\value{
Accessors: See details.

Setters: An updated \code{LongTable} object, returned invisibly.
}
\description{
Documentation for the various setters and getters which allow manipulation
of data in the slots of a \code{LongTable} object.
}
\examples{
data(merckLongTable)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/datasets.R
\docType{data}
\name{clevelandSmall_cSet}
\alias{clevelandSmall_cSet}
\title{Cleaveland_mut RadioSet subsetted and cast as CoreSet}
\format{
CoreSet object
}
\usage{
data(clevelandSmall_cSet)
}
\description{
This dataset is just a dummy object derived from the Cleveland_mut RadioSet
in the RadioGx R package. It's contents should not be interpreted and it
is only present to test the functions in this package and provide
examples
}
\references{
Lamb et al. The Connectivity Map: using gene-expression signatures to connect
small molecules, genes, and disease. Science, 2006.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mergeAssays-method.R
\name{mergeAssays}
\alias{mergeAssays}
\title{Merge assays with an \code{S4} object.}
\usage{
mergeAssays(object, ...)
}
\arguments{
\item{object}{\code{S4} An S4 object a list-like slot containing assays for the
object.}

\item{...}{Allow new arguments to be defined for this generic.}
}
\value{
A modified version of \code{object}.
}
\description{
Merge assays with an \code{S4} object.
}
\examples{
"This is a generic method!"

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/immutable-class.R
\docType{class}
\name{immutable}
\alias{immutable}
\alias{immutable_list-class}
\alias{is.immutable}
\alias{print.immutable}
\alias{show.immutable}
\title{Constructor for "immutable" S3-class property}
\usage{
immutable(object)

is.immutable(object)

\method{print}{immutable}(x, ...)

show.immutable(x)
}
\arguments{
\item{object, x}{Any R object which uses S3 method dispatch}

\item{...}{Fallthrough arguments to \code{print.default}.}
}
\value{
The \code{object} with "immutable" prepended to its class attribute.

\code{logical(1)} Does the object inherit from the "immutable" S3-class?

None, \code{invisible(NULL)}
}
\description{
This method should allow any S3 object in R to become immutable by
intercepting \verb{[<-}, \verb{[[<-}, \verb{$<-} and \code{c} during S3-method dispatch and
returning an error.

Reverse with call to the \code{mutable} function.
}
\details{
The motivation for this class was to create pseudo-private slots in an R
S4 object by preventing mutation of those slots outside of the accessors
written for the class. It should behave as expected for R object which
operate with 'copy-on-modify' semantics, including most base R functions and
S3 objects.

An environment was not suitable for this case due
to the 'copy-by-reference' semantics, such that normal R assignment, which
users assume makes a copy of the object, actually references the same
environment in both the original and copy of the object.

WARNING: This implementation is unable to intercept modifications to a
\code{data.table} via the \verb{set*} group of methods. This is because these methods
are not S3 generics and therefore no mechanism exists for hooking into them
to extend their functionality. In general, this helper class will only work
for objects with an S3 interface.
}
\examples{
immutable_list <- immutable(as.list(1:5))
class(immutable_list)
# errors during assignment operations
tryCatch({ immutable_list$new <- 1 }, error=print)

immutable_list <- immutable(as.list(1:5))
is.immutable(immutable_list)

}
\seealso{
\code{\link{assignment-immutable}}, \code{\link{setOps-immutable}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LongTable-class.R
\docType{class}
\name{list_OR_LongTable-class}
\alias{list_OR_LongTable-class}
\title{A class union to allow multiple types in a CoreSet slot}
\description{
A class union to allow multiple types in a CoreSet slot
}
